; The following line has the only difference between release and debug versions of DXFeed lib
LIBRARY DXFeed_64
VERSION 5.10
EXPORTS
    dxf_create_connection
    dxf_create_connection_auth_basic
    dxf_create_connection_auth_bearer
    dxf_create_connection_auth_custom
    dxf_close_connection
    dxf_create_subscription
    dxf_create_subscription_with_flags
    dxf_create_subscription_timed
    dxf_create_subscription_timed_with_flags
    dxf_close_subscription
    dxf_add_symbol
    dxf_add_symbols
    dxf_add_candle_symbol
    dxf_remove_candle_symbol
    dxf_remove_symbol
    dxf_remove_symbols
    dxf_get_symbols
    dxf_set_symbols
    dxf_clear_symbols
    dxf_attach_event_listener
    dxf_detach_event_listener
    dxf_attach_event_listener_v2
    dxf_detach_event_listener_v2
    dxf_get_subscription_event_types
    dxf_get_last_event
    dxf_get_last_error
    dxf_initialize_logger
    dxf_set_order_source
    dxf_add_order_source
    dxf_create_candle_symbol_attributes
    dxf_delete_candle_symbol_attributes
    dxf_create_snapshot
    dxf_create_order_snapshot
    dxf_create_candle_snapshot
    dxf_close_snapshot
    dxf_attach_snapshot_listener
    dxf_detach_snapshot_listener
    dxf_attach_snapshot_inc_listener
    dxf_detach_snapshot_inc_listener
    dxf_get_snapshot_symbol
    dxf_create_price_level_book
    dxf_close_price_level_book
    dxf_attach_price_level_book_listener
    dxf_detach_price_level_book_listener
    dxf_create_regional_book
    dxf_close_regional_book
    dxf_attach_regional_book_listener
    dxf_detach_regional_book_listener
    dxf_attach_regional_book_listener_v2
    dxf_detach_regional_book_listener_v2
    dxf_write_raw_data
    dxf_get_connection_properties_snapshot
    dxf_free_connection_properties_snapshot
    dxf_get_current_connected_address
    dxf_get_current_connection_status
    dxf_free
    dx_get_event_data_item
