/* Generated by Cython 0.29.16 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "language": "c++",
        "libraries": [
            "ws2_32"
        ],
        "name": "dxfeed.core.DXFeedPy",
        "sources": [
            "dxfeed/core/DXFeedPy.pyx"
        ]
    },
    "module_name": "dxfeed.core.DXFeedPy"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_16"
#define CYTHON_HEX_VERSION 0x001D10F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__dxfeed__core__DXFeedPy
#define __PYX_HAVE_API__dxfeed__core__DXFeedPy
/* Early includes */
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <deque>
#include "DXTypes.h"
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include "RecordData.h"
#include "EventData.h"
#include "DXFeed.h"
#include "DXErrorCodes.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "dxfeed\\core\\DXFeedPy.pyx",
  "stringsource",
  "type.pxd",
  "dxfeed\\core\\listeners\\listener.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper;
struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass;
struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass;
struct __pyx_opt_args_6dxfeed_4core_8DXFeedPy_process_last_error;

/* "dxfeed/core/DXFeedPy.pyx":20
 * from dxfeed.core.pxd_include.EventData cimport *
 * 
 * cpdef int process_last_error(verbose: bool=True):             # <<<<<<<<<<<<<<
 *     """
 *     Function retrieves last error
 */
struct __pyx_opt_args_6dxfeed_4core_8DXFeedPy_process_last_error {
  int __pyx_n;
  PyObject *verbose;
};

/* "dxfeed/core/listeners/listener.pxd":5
 * 
 * 
 * cdef class FuncWrapper:             # <<<<<<<<<<<<<<
 * 	cdef dxf_event_listener_t func
 * 	@staticmethod
 */
struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper {
  PyObject_HEAD
  struct __pyx_vtabstruct_6dxfeed_4core_9listeners_8listener_FuncWrapper *__pyx_vtab;
  dxf_event_listener_t func;
};


/* "dxfeed/core/DXFeedPy.pyx":51
 * 
 * 
 * cdef class ConnectionClass:             # <<<<<<<<<<<<<<
 *     """
 *     Data structure that contains connection
 */
struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass {
  PyObject_HEAD
  struct __pyx_vtabstruct_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_vtab;
  dxf_connection_t connection;
  std::deque<dxf_subscription_t *>  sub_ptr_list;
  int subs_order;
};


/* "dxfeed/core/DXFeedPy.pyx":77
 * 
 * 
 * cdef class SubscriptionClass:             # <<<<<<<<<<<<<<
 *     """
 *     Data structure that contains subscription and related fields
 */
struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass {
  PyObject_HEAD
  dxf_connection_t connection;
  dxf_subscription_t subscription;
  int subscription_order;
  std::deque<dxf_subscription_t *>  *con_sub_list_ptr;
  dxf_event_listener_t listener;
  PyObject *event_type_str;
  PyObject *columns;
  PyObject *data;
  void *u_data;
};



/* "dxfeed/core/listeners/listener.pxd":5
 * 
 * 
 * cdef class FuncWrapper:             # <<<<<<<<<<<<<<
 * 	cdef dxf_event_listener_t func
 * 	@staticmethod
 */

struct __pyx_vtabstruct_6dxfeed_4core_9listeners_8listener_FuncWrapper {
  struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *(*make_from_ptr)(dxf_event_listener_t);
};
static struct __pyx_vtabstruct_6dxfeed_4core_9listeners_8listener_FuncWrapper *__pyx_vtabptr_6dxfeed_4core_9listeners_8listener_FuncWrapper;


/* "dxfeed/core/DXFeedPy.pyx":51
 * 
 * 
 * cdef class ConnectionClass:             # <<<<<<<<<<<<<<
 *     """
 *     Data structure that contains connection
 */

struct __pyx_vtabstruct_6dxfeed_4core_8DXFeedPy_ConnectionClass {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *(*make_new_subscription)(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_vtabptr_6dxfeed_4core_8DXFeedPy_ConnectionClass;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* BuildPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char);

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_str(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_str(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* IncludeStringH.proto */
#include <string.h>

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* decode_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_bytes(
         PyObject* string, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    return __Pyx_decode_c_bytes(
        PyBytes_AS_STRING(string), PyBytes_GET_SIZE(string),
        start, stop, encoding, errors, decode_func);
}

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* None.proto */
#include <new>

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_dxf_connection_status_t(dxf_connection_status_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE dxf_long_t __Pyx_PyInt_As_dxf_long_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CStringEquals.proto */
static CYTHON_INLINE int __Pyx_StrEq(const char *, const char *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_f_6dxfeed_4core_8DXFeedPy_15ConnectionClass_make_new_subscription(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_self, PyObject *__pyx_v_data_len, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from 'libcpp.deque' */

/* Module declarations from 'libc.stddef' */

/* Module declarations from 'dxfeed.core.pxd_include.DXTypes' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'dxfeed.core.utils.helpers' */
static PyObject *(*__pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t)(dxf_const_string_t); /*proto*/
static dxf_const_string_t (*__pyx_f_6dxfeed_4core_5utils_7helpers_dxf_const_string_t_from_unicode)(PyObject *); /*proto*/

/* Module declarations from 'dxfeed.core.pxd_include.RecordData' */

/* Module declarations from 'dxfeed.core.pxd_include.EventData' */

/* Module declarations from 'dxfeed.core.pxd_include.DXFeed' */

/* Module declarations from 'dxfeed.core.pxd_include.DXErrorCodes' */

/* Module declarations from 'dxfeed.core.listeners.listener' */
static PyTypeObject *__pyx_ptype_6dxfeed_4core_9listeners_8listener_FuncWrapper = 0;
static void (*__pyx_f_6dxfeed_4core_9listeners_8listener_trade_default_listener)(int, dxf_const_string_t, dxf_event_data_t const *, int, void *); /*proto*/
static void (*__pyx_f_6dxfeed_4core_9listeners_8listener_quote_default_listener)(int, dxf_const_string_t, dxf_event_data_t const *, int, void *); /*proto*/
static void (*__pyx_f_6dxfeed_4core_9listeners_8listener_summary_default_listener)(int, dxf_const_string_t, dxf_event_data_t const *, int, void *); /*proto*/
static void (*__pyx_f_6dxfeed_4core_9listeners_8listener_profile_default_listener)(int, dxf_const_string_t, dxf_event_data_t const *, int, void *); /*proto*/
static void (*__pyx_f_6dxfeed_4core_9listeners_8listener_time_and_sale_default_listener)(int, dxf_const_string_t, dxf_event_data_t const *, int, void *); /*proto*/
static void (*__pyx_f_6dxfeed_4core_9listeners_8listener_candle_default_listener)(int, dxf_const_string_t, dxf_event_data_t const *, int, void *); /*proto*/
static void (*__pyx_f_6dxfeed_4core_9listeners_8listener_order_default_listener)(int, dxf_const_string_t, dxf_event_data_t const *, int, void *); /*proto*/
static void (*__pyx_f_6dxfeed_4core_9listeners_8listener_greeks_default_listener)(int, dxf_const_string_t, dxf_event_data_t const *, int, void *); /*proto*/
static void (*__pyx_f_6dxfeed_4core_9listeners_8listener_theo_price_default_listener)(int, dxf_const_string_t, dxf_event_data_t const *, int, void *); /*proto*/
static void (*__pyx_f_6dxfeed_4core_9listeners_8listener_underlying_default_listener)(int, dxf_const_string_t, dxf_event_data_t const *, int, void *); /*proto*/
static void (*__pyx_f_6dxfeed_4core_9listeners_8listener_series_default_listener)(int, dxf_const_string_t, dxf_event_data_t const *, int, void *); /*proto*/
static void (*__pyx_f_6dxfeed_4core_9listeners_8listener_configuration_default_listener)(int, dxf_const_string_t, dxf_event_data_t const *, int, void *); /*proto*/

/* Module declarations from 'dxfeed.core.DXFeedPy' */
static PyTypeObject *__pyx_ptype_6dxfeed_4core_8DXFeedPy_ConnectionClass = 0;
static PyTypeObject *__pyx_ptype_6dxfeed_4core_8DXFeedPy_SubscriptionClass = 0;
static int __pyx_f_6dxfeed_4core_8DXFeedPy_process_last_error(int __pyx_skip_dispatch, struct __pyx_opt_args_6dxfeed_4core_8DXFeedPy_process_last_error *__pyx_optional_args); /*proto*/
#define __Pyx_MODULE_NAME "dxfeed.core.DXFeedPy"
extern int __pyx_module_is_main_dxfeed__core__DXFeedPy;
int __pyx_module_is_main_dxfeed__core__DXFeedPy = 0;

/* Implementation of 'dxfeed.core.DXFeedPy' */
static PyObject *__pyx_builtin_print;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_range;
static const char __pyx_k_i[] = "i";
static const char __pyx_k_cc[] = "cc";
static const char __pyx_k_fw[] = "fw";
static const char __pyx_k_pd[] = "pd";
static const char __pyx_k_sc[] = "sc";
static const char __pyx_k__13[] = "";
static const char __pyx_k__14[] = "*";
static const char __pyx_k_idx[] = "idx";
static const char __pyx_k_lis[] = "lis";
static const char __pyx_k_loc[] = "loc";
static const char __pyx_k_str[] = "str";
static const char __pyx_k_sym[] = "sym";
static const char __pyx_k_Time[] = "Time";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_keep[] = "keep";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_warn[] = "warn";
static const char __pyx_k_M8_ms[] = "<M8[ms]";
static const char __pyx_k_Order[] = "Order";
static const char __pyx_k_Quote[] = "Quote";
static const char __pyx_k_Trade[] = "Trade";
static const char __pyx_k_Union[] = "Union";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_token[] = "token";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_Candle[] = "Candle";
static const char __pyx_k_Greeks[] = "Greeks";
static const char __pyx_k_Series[] = "Series";
static const char __pyx_k_astype[] = "astype";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_maxlen[] = "maxlen";
static const char __pyx_k_pandas[] = "pandas";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_status[] = "status";
static const char __pyx_k_symbol[] = "symbol";
static const char __pyx_k_typing[] = "typing";
static const char __pyx_k_utcnow[] = "utcnow";
static const char __pyx_k_Profile[] = "Profile";
static const char __pyx_k_Summary[] = "Summary";
static const char __pyx_k_address[] = "address";
static const char __pyx_k_columns[] = "columns";
static const char __pyx_k_symbols[] = "symbols";
static const char __pyx_k_verbose[] = "verbose";
static const char __pyx_k_Iterable[] = "Iterable";
static const char __pyx_k_Optional[] = "Optional";
static const char __pyx_k_TradeETH[] = "TradeETH";
static const char __pyx_k_contains[] = "contains";
static const char __pyx_k_data_len[] = "data_len";
static const char __pyx_k_datetime[] = "datetime";
static const char __pyx_k_deque_wl[] = "deque_wl";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_occurred[] = " occurred!";
static const char __pyx_k_safe_get[] = "safe_get";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_strptime[] = "strptime";
static const char __pyx_k_warnings[] = "warnings";
static const char __pyx_k_Connected[] = "Connected";
static const char __pyx_k_DataFrame[] = "DataFrame";
static const char __pyx_k_TheoPrice[] = "TheoPrice";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_file_name[] = "file_name";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_timestamp[] = "timestamp";
static const char __pyx_k_Underlying[] = "Underlying";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_error_code[] = "error code = ";
static const char __pyx_k_et_mapping[] = "et_mapping";
static const char __pyx_k_event_type[] = "event_type";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_return_str[] = "return_str";
static const char __pyx_k_SpreadOrder[] = "SpreadOrder";
static const char __pyx_k_TimeAndSale[] = "TimeAndSale";
static const char __pyx_k_Y_m_d_H_M_S[] = "%Y-%m-%d %H:%M:%S";
static const char __pyx_k_candle_time[] = "candle_time";
static const char __pyx_k_description[] = ", description = ";
static const char __pyx_k_et_type_int[] = "et_type_int";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_error_code_2[] = "error_code";
static const char __pyx_k_event_type_2[] = " event type";
static const char __pyx_k_rewrite_file[] = "rewrite_file";
static const char __pyx_k_symbol_count[] = "symbol_count";
static const char __pyx_k_symbols_list[] = "symbols_list";
static const char __pyx_k_Configuration[] = "Configuration";
static const char __pyx_k_DequeWithLock[] = "DequeWithLock";
static const char __pyx_k_Not_connected[] = "Not connected";
static const char __pyx_k_ORDER_COLUMNS[] = "ORDER_COLUMNS";
static const char __pyx_k_QUOTE_COLUMNS[] = "QUOTE_COLUMNS";
static const char __pyx_k_TRADE_COLUMNS[] = "TRADE_COLUMNS";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_total_seconds[] = "total_seconds";
static const char __pyx_k_CANDLE_COLUMNS[] = "CANDLE_COLUMNS";
static const char __pyx_k_GREEKS_COLUMNS[] = "GREEKS_COLUMNS";
static const char __pyx_k_Login_required[] = "Login required";
static const char __pyx_k_SERIES_COLUMNS[] = "SERIES_COLUMNS";
static const char __pyx_k_status_mapping[] = "status_mapping";
static const char __pyx_k_ConnectionClass[] = "ConnectionClass";
static const char __pyx_k_PROFILE_COLUMNS[] = "PROFILE_COLUMNS";
static const char __pyx_k_SUMMARY_COLUMNS[] = "SUMMARY_COLUMNS";
static const char __pyx_k_dxf_add_symbols[] = "dxf_add_symbols";
static const char __pyx_k_dxf_get_symbols[] = "dxf_get_symbols";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_SubscriptionClass[] = "SubscriptionClass";
static const char __pyx_k_dxf_clear_symbols[] = "dxf_clear_symbols";
static const char __pyx_k_THEO_PRICE_COLUMNS[] = "THEO_PRICE_COLUMNS";
static const char __pyx_k_UNDERLYING_COLUMNS[] = "UNDERLYING_COLUMNS";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_dxf_remove_symbols[] = "dxf_remove_symbols";
static const char __pyx_k_event_type_convert[] = "event_type_convert";
static const char __pyx_k_show_timezone_info[] = "show_timezone_info";
static const char __pyx_k_dxf_attach_listener[] = "dxf_attach_listener";
static const char __pyx_k_dxf_detach_listener[] = "dxf_detach_listener";
static const char __pyx_k_Incorrect_event_type[] = "Incorrect event type!";
static const char __pyx_k_Invalid_subscription[] = "Invalid subscription";
static const char __pyx_k_demo_dxfeed_com_7300[] = "demo.dxfeed.com:7300";
static const char __pyx_k_dxf_close_connection[] = "dxf_close_connection";
static const char __pyx_k_dxfeed_core_DXFeedPy[] = "dxfeed.core.DXFeedPy";
static const char __pyx_k_CONFIGURATION_COLUMNS[] = "CONFIGURATION_COLUMNS";
static const char __pyx_k_TIME_AND_SALE_COLUMNS[] = "TIME_AND_SALE_COLUMNS";
static const char __pyx_k_dxf_create_connection[] = "dxf_create_connection";
static const char __pyx_k_dxf_initialize_logger[] = "dxf_initialize_logger";
static const char __pyx_k_make_new_subscription[] = "make_new_subscription";
static const char __pyx_k_dxf_close_subscription[] = "dxf_close_subscription";
static const char __pyx_k_Connection_is_not_valid[] = "Connection is not valid";
static const char __pyx_k_No_default_listener_for[] = "No default listener for ";
static const char __pyx_k_dxf_create_subscription[] = "dxf_create_subscription";
static const char __pyx_k_Connected_and_authorized[] = "Connected and authorized";
static const char __pyx_k_dxfeed_core_DXFeedPy_pyx[] = "dxfeed\\core\\DXFeedPy.pyx";
static const char __pyx_k_Subscription_is_not_valid[] = "Subscription is not valid";
static const char __pyx_k_dxfeed_core_utils_helpers[] = "dxfeed.core.utils.helpers";
static const char __pyx_k_dxf_attach_custom_listener[] = "dxf_attach_custom_listener";
static const char __pyx_k_dxfeed_core_utils_data_class[] = "dxfeed.core.utils.data_class";
static const char __pyx_k_dxfeed_core_listeners_listener[] = "dxfeed.core.listeners.listener";
static const char __pyx_k_has_type_different_from_string[] = " has type different from string";
static const char __pyx_k_no_error_information_is_stored[] = "no error information is stored";
static const char __pyx_k_Error_occurred_and_successfully[] = "Error occurred and successfully retrieved:\n";
static const char __pyx_k_Inapropriate_date_format_should[] = "Inapropriate date format, should be %Y-%m-%d %H:%M:%S";
static const char __pyx_k_In_underlying_C_API_library_erro[] = "In underlying C-API library error ";
static const char __pyx_k_dxf_create_connection_auth_beare[] = "dxf_create_connection_auth_bearer";
static const char __pyx_k_dxf_get_current_connected_addres[] = "dxf_get_current_connected_address";
static const char __pyx_k_dxf_get_current_connection_statu[] = "dxf_get_current_connection_status";
static const char __pyx_k_dxf_get_subscription_event_types[] = "dxf_get_subscription_event_types";
static const char __pyx_k_self_con_sub_list_ptr_self_conne[] = "self.con_sub_list_ptr,self.connection,self.listener,self.subscription,self.u_data cannot be converted to a Python object for pickling";
static const char __pyx_k_self_connection_self_sub_ptr_lis[] = "self.connection,self.sub_ptr_list cannot be converted to a Python object for pickling";
static PyObject *__pyx_n_s_CANDLE_COLUMNS;
static PyObject *__pyx_n_s_CONFIGURATION_COLUMNS;
static PyObject *__pyx_n_u_Candle;
static PyObject *__pyx_n_u_Configuration;
static PyObject *__pyx_n_u_Connected;
static PyObject *__pyx_kp_u_Connected_and_authorized;
static PyObject *__pyx_n_s_ConnectionClass;
static PyObject *__pyx_kp_u_Connection_is_not_valid;
static PyObject *__pyx_n_s_DataFrame;
static PyObject *__pyx_n_s_DequeWithLock;
static PyObject *__pyx_kp_u_Error_occurred_and_successfully;
static PyObject *__pyx_n_s_GREEKS_COLUMNS;
static PyObject *__pyx_n_u_Greeks;
static PyObject *__pyx_kp_u_In_underlying_C_API_library_erro;
static PyObject *__pyx_kp_u_Inapropriate_date_format_should;
static PyObject *__pyx_kp_u_Incorrect_event_type;
static PyObject *__pyx_kp_u_Invalid_subscription;
static PyObject *__pyx_n_s_Iterable;
static PyObject *__pyx_kp_u_Login_required;
static PyObject *__pyx_kp_u_M8_ms;
static PyObject *__pyx_kp_u_No_default_listener_for;
static PyObject *__pyx_kp_u_Not_connected;
static PyObject *__pyx_n_s_ORDER_COLUMNS;
static PyObject *__pyx_n_s_Optional;
static PyObject *__pyx_n_u_Order;
static PyObject *__pyx_n_s_PROFILE_COLUMNS;
static PyObject *__pyx_n_u_Profile;
static PyObject *__pyx_n_s_QUOTE_COLUMNS;
static PyObject *__pyx_n_u_Quote;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_SERIES_COLUMNS;
static PyObject *__pyx_n_s_SUMMARY_COLUMNS;
static PyObject *__pyx_n_u_Series;
static PyObject *__pyx_n_u_SpreadOrder;
static PyObject *__pyx_n_s_SubscriptionClass;
static PyObject *__pyx_kp_u_Subscription_is_not_valid;
static PyObject *__pyx_n_u_Summary;
static PyObject *__pyx_n_s_THEO_PRICE_COLUMNS;
static PyObject *__pyx_n_s_TIME_AND_SALE_COLUMNS;
static PyObject *__pyx_n_s_TRADE_COLUMNS;
static PyObject *__pyx_n_u_TheoPrice;
static PyObject *__pyx_n_u_Time;
static PyObject *__pyx_n_u_TimeAndSale;
static PyObject *__pyx_n_u_Trade;
static PyObject *__pyx_n_u_TradeETH;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_UNDERLYING_COLUMNS;
static PyObject *__pyx_n_u_Underlying;
static PyObject *__pyx_n_s_Union;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_u_Y_m_d_H_M_S;
static PyObject *__pyx_kp_u__13;
static PyObject *__pyx_n_s__14;
static PyObject *__pyx_n_s_address;
static PyObject *__pyx_n_s_astype;
static PyObject *__pyx_n_s_candle_time;
static PyObject *__pyx_n_s_cc;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_columns;
static PyObject *__pyx_n_s_contains;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_data_len;
static PyObject *__pyx_n_s_datetime;
static PyObject *__pyx_kp_u_demo_dxfeed_com_7300;
static PyObject *__pyx_n_s_deque_wl;
static PyObject *__pyx_kp_u_description;
static PyObject *__pyx_n_s_dxf_add_symbols;
static PyObject *__pyx_n_s_dxf_attach_custom_listener;
static PyObject *__pyx_n_s_dxf_attach_listener;
static PyObject *__pyx_n_s_dxf_clear_symbols;
static PyObject *__pyx_n_s_dxf_close_connection;
static PyObject *__pyx_n_s_dxf_close_subscription;
static PyObject *__pyx_n_s_dxf_create_connection;
static PyObject *__pyx_n_s_dxf_create_connection_auth_beare;
static PyObject *__pyx_n_s_dxf_create_subscription;
static PyObject *__pyx_n_s_dxf_detach_listener;
static PyObject *__pyx_n_s_dxf_get_current_connected_addres;
static PyObject *__pyx_n_s_dxf_get_current_connection_statu;
static PyObject *__pyx_n_s_dxf_get_subscription_event_types;
static PyObject *__pyx_n_s_dxf_get_symbols;
static PyObject *__pyx_n_s_dxf_initialize_logger;
static PyObject *__pyx_n_s_dxf_remove_symbols;
static PyObject *__pyx_n_s_dxfeed_core_DXFeedPy;
static PyObject *__pyx_kp_s_dxfeed_core_DXFeedPy_pyx;
static PyObject *__pyx_n_s_dxfeed_core_listeners_listener;
static PyObject *__pyx_n_s_dxfeed_core_utils_data_class;
static PyObject *__pyx_n_s_dxfeed_core_utils_helpers;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_kp_u_error_code;
static PyObject *__pyx_n_s_error_code_2;
static PyObject *__pyx_n_s_et_mapping;
static PyObject *__pyx_n_s_et_type_int;
static PyObject *__pyx_n_s_event_type;
static PyObject *__pyx_kp_u_event_type_2;
static PyObject *__pyx_n_s_event_type_convert;
static PyObject *__pyx_n_s_file_name;
static PyObject *__pyx_n_s_fw;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_kp_u_has_type_different_from_string;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_idx;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_keep;
static PyObject *__pyx_n_s_lis;
static PyObject *__pyx_n_s_loc;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_make_new_subscription;
static PyObject *__pyx_n_s_maxlen;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_kp_u_no_error_information_is_stored;
static PyObject *__pyx_kp_u_occurred;
static PyObject *__pyx_n_s_pandas;
static PyObject *__pyx_n_s_pd;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_result;
static PyObject *__pyx_n_s_return_str;
static PyObject *__pyx_n_s_rewrite_file;
static PyObject *__pyx_n_s_safe_get;
static PyObject *__pyx_n_s_sc;
static PyObject *__pyx_kp_s_self_con_sub_list_ptr_self_conne;
static PyObject *__pyx_kp_s_self_connection_self_sub_ptr_lis;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_show_timezone_info;
static PyObject *__pyx_n_s_status;
static PyObject *__pyx_n_s_status_mapping;
static PyObject *__pyx_n_s_str;
static PyObject *__pyx_n_s_strptime;
static PyObject *__pyx_n_s_sym;
static PyObject *__pyx_n_s_symbol;
static PyObject *__pyx_n_s_symbol_count;
static PyObject *__pyx_n_s_symbols;
static PyObject *__pyx_n_s_symbols_list;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_timestamp;
static PyObject *__pyx_n_s_token;
static PyObject *__pyx_n_s_total_seconds;
static PyObject *__pyx_n_s_typing;
static PyObject *__pyx_n_s_utcnow;
static PyObject *__pyx_kp_u_utf_8;
static PyObject *__pyx_n_s_verbose;
static PyObject *__pyx_n_s_warn;
static PyObject *__pyx_n_s_warnings;
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_process_last_error(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_verbose); /* proto */
static int __pyx_pf_6dxfeed_4core_8DXFeedPy_15ConnectionClass___init__(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_self); /* proto */
static void __pyx_pf_6dxfeed_4core_8DXFeedPy_15ConnectionClass_2__dealloc__(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_15ConnectionClass_4make_new_subscription(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_self, PyObject *__pyx_v_data_len); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_15ConnectionClass_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_15ConnectionClass_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_6dxfeed_4core_8DXFeedPy_17SubscriptionClass___init__(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_self, PyObject *__pyx_v_data_len); /* proto */
static void __pyx_pf_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_2__dealloc__(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_4get_data(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_6to_dataframe(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_self, PyObject *__pyx_v_keep); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_2dxf_create_connection(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_address); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_4dxf_create_connection_auth_bearer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_address, PyObject *__pyx_v_token); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_6dxf_create_subscription(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_cc, PyObject *__pyx_v_event_type, PyObject *__pyx_v_candle_time, PyObject *__pyx_v_data_len); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_8dxf_add_symbols(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc, PyObject *__pyx_v_symbols); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_10dxf_attach_listener(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_12dxf_attach_custom_listener(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc, struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *__pyx_v_fw, PyObject *__pyx_v_columns, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_14dxf_detach_listener(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_16dxf_close_connection(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_cc); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_18dxf_close_subscription(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_20dxf_get_current_connection_status(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_cc, PyObject *__pyx_v_return_str); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_22dxf_get_current_connected_address(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_cc); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_24dxf_initialize_logger(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_file_name, PyObject *__pyx_v_rewrite_file, PyObject *__pyx_v_show_timezone_info, PyObject *__pyx_v_verbose); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_26dxf_get_subscription_event_types(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc, PyObject *__pyx_v_return_str); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_28dxf_get_symbols(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_30dxf_remove_symbols(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc, PyObject *__pyx_v_symbols); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_32dxf_clear_symbols(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc); /* proto */
static PyObject *__pyx_tp_new_6dxfeed_4core_8DXFeedPy_ConnectionClass(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6dxfeed_4core_8DXFeedPy_SubscriptionClass(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_16;
static PyObject *__pyx_int_32;
static PyObject *__pyx_int_64;
static PyObject *__pyx_int_128;
static PyObject *__pyx_int_256;
static PyObject *__pyx_int_512;
static PyObject *__pyx_int_1000;
static PyObject *__pyx_int_1024;
static PyObject *__pyx_int_1970;
static PyObject *__pyx_int_2048;
static PyObject *__pyx_int_4096;
static PyObject *__pyx_int_5000;
static PyObject *__pyx_int_8192;
static PyObject *__pyx_int_100000;
static PyObject *__pyx_int_neg_16384;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_slice__4;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_codeobj__16;
static PyObject *__pyx_codeobj__18;
static PyObject *__pyx_codeobj__20;
static PyObject *__pyx_codeobj__22;
static PyObject *__pyx_codeobj__24;
static PyObject *__pyx_codeobj__26;
static PyObject *__pyx_codeobj__28;
static PyObject *__pyx_codeobj__30;
static PyObject *__pyx_codeobj__32;
static PyObject *__pyx_codeobj__34;
static PyObject *__pyx_codeobj__36;
static PyObject *__pyx_codeobj__38;
static PyObject *__pyx_codeobj__40;
static PyObject *__pyx_codeobj__42;
static PyObject *__pyx_codeobj__44;
static PyObject *__pyx_codeobj__46;
/* Late includes */

/* "dxfeed/core/DXFeedPy.pyx":20
 * from dxfeed.core.pxd_include.EventData cimport *
 * 
 * cpdef int process_last_error(verbose: bool=True):             # <<<<<<<<<<<<<<
 *     """
 *     Function retrieves last error
 */

static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_1process_last_error(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_6dxfeed_4core_8DXFeedPy_process_last_error(CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_6dxfeed_4core_8DXFeedPy_process_last_error *__pyx_optional_args) {
  PyObject *__pyx_v_verbose = ((PyObject *)Py_True);
  int __pyx_v_error_code;
  dxf_const_string_t __pyx_v_error_descr;
  int __pyx_v_res;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("process_last_error", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_verbose = __pyx_optional_args->verbose;
    }
  }

  /* "dxfeed/core/DXFeedPy.pyx":34
 *         Error code is returned
 *     """
 *     cdef int error_code = dxec.dx_ec_success             # <<<<<<<<<<<<<<
 *     cdef dxf_const_string_t error_descr = NULL
 *     cdef int res
 */
  __pyx_v_error_code = dx_ec_success;

  /* "dxfeed/core/DXFeedPy.pyx":35
 *     """
 *     cdef int error_code = dxec.dx_ec_success
 *     cdef dxf_const_string_t error_descr = NULL             # <<<<<<<<<<<<<<
 *     cdef int res
 * 
 */
  __pyx_v_error_descr = NULL;

  /* "dxfeed/core/DXFeedPy.pyx":38
 *     cdef int res
 * 
 *     res = clib.dxf_get_last_error(&error_code, &error_descr)             # <<<<<<<<<<<<<<
 * 
 *     if res == clib.DXF_SUCCESS:
 */
  __pyx_v_res = dxf_get_last_error((&__pyx_v_error_code), (&__pyx_v_error_descr));

  /* "dxfeed/core/DXFeedPy.pyx":40
 *     res = clib.dxf_get_last_error(&error_code, &error_descr)
 * 
 *     if res == clib.DXF_SUCCESS:             # <<<<<<<<<<<<<<
 *         if error_code == dxec.dx_ec_success and verbose:
 *             print("no error information is stored")
 */
  __pyx_t_1 = ((__pyx_v_res == DXF_SUCCESS) != 0);
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":41
 * 
 *     if res == clib.DXF_SUCCESS:
 *         if error_code == dxec.dx_ec_success and verbose:             # <<<<<<<<<<<<<<
 *             print("no error information is stored")
 * 
 */
    __pyx_t_2 = ((__pyx_v_error_code == dx_ec_success) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_verbose); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 41, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_2;
    __pyx_L5_bool_binop_done:;
    if (__pyx_t_1) {

      /* "dxfeed/core/DXFeedPy.pyx":42
 *     if res == clib.DXF_SUCCESS:
 *         if error_code == dxec.dx_ec_success and verbose:
 *             print("no error information is stored")             # <<<<<<<<<<<<<<
 * 
 *         if verbose:
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 42, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "dxfeed/core/DXFeedPy.pyx":41
 * 
 *     if res == clib.DXF_SUCCESS:
 *         if error_code == dxec.dx_ec_success and verbose:             # <<<<<<<<<<<<<<
 *             print("no error information is stored")
 * 
 */
    }

    /* "dxfeed/core/DXFeedPy.pyx":44
 *             print("no error information is stored")
 * 
 *         if verbose:             # <<<<<<<<<<<<<<
 *             print("Error occurred and successfully retrieved:\n",
 *                   f"error code = {error_code}, description = {unicode_from_dxf_const_string_t(error_descr)}")
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_verbose); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 44, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "dxfeed/core/DXFeedPy.pyx":46
 *         if verbose:
 *             print("Error occurred and successfully retrieved:\n",
 *                   f"error code = {error_code}, description = {unicode_from_dxf_const_string_t(error_descr)}")             # <<<<<<<<<<<<<<
 * 
 *     return error_code
 */
      __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 46, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_5 = 127;
      __Pyx_INCREF(__pyx_kp_u_error_code);
      __pyx_t_4 += 13;
      __Pyx_GIVEREF(__pyx_kp_u_error_code);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_error_code);
      __pyx_t_6 = __Pyx_PyUnicode_From_int(__pyx_v_error_code, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 46, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_kp_u_description);
      __pyx_t_4 += 16;
      __Pyx_GIVEREF(__pyx_kp_u_description);
      PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u_description);
      __pyx_t_6 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t(__pyx_v_error_descr); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 46, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_FormatSimple(__pyx_t_6, __pyx_empty_unicode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 46, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_7) : __pyx_t_5;
      __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyUnicode_Join(__pyx_t_3, 4, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 46, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "dxfeed/core/DXFeedPy.pyx":45
 * 
 *         if verbose:
 *             print("Error occurred and successfully retrieved:\n",             # <<<<<<<<<<<<<<
 *                   f"error code = {error_code}, description = {unicode_from_dxf_const_string_t(error_descr)}")
 * 
 */
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 45, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_kp_u_Error_occurred_and_successfully);
      __Pyx_GIVEREF(__pyx_kp_u_Error_occurred_and_successfully);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Error_occurred_and_successfully);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_3, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 45, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "dxfeed/core/DXFeedPy.pyx":44
 *             print("no error information is stored")
 * 
 *         if verbose:             # <<<<<<<<<<<<<<
 *             print("Error occurred and successfully retrieved:\n",
 *                   f"error code = {error_code}, description = {unicode_from_dxf_const_string_t(error_descr)}")
 */
    }

    /* "dxfeed/core/DXFeedPy.pyx":40
 *     res = clib.dxf_get_last_error(&error_code, &error_descr)
 * 
 *     if res == clib.DXF_SUCCESS:             # <<<<<<<<<<<<<<
 *         if error_code == dxec.dx_ec_success and verbose:
 *             print("no error information is stored")
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":48
 *                   f"error code = {error_code}, description = {unicode_from_dxf_const_string_t(error_descr)}")
 * 
 *     return error_code             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_error_code;
  goto __pyx_L0;

  /* "dxfeed/core/DXFeedPy.pyx":20
 * from dxfeed.core.pxd_include.EventData cimport *
 * 
 * cpdef int process_last_error(verbose: bool=True):             # <<<<<<<<<<<<<<
 *     """
 *     Function retrieves last error
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_WriteUnraisable("dxfeed.core.DXFeedPy.process_last_error", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_1process_last_error(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_process_last_error[] = "\n    Function retrieves last error\n\n    Parameters\n    ----------\n    verbose: bool\n        If True error description is printed\n        \n    Returns\n    -------\n    error_code: int\n        Error code is returned\n    ";
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_1process_last_error(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_verbose = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("process_last_error (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_verbose,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_verbose);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "process_last_error") < 0)) __PYX_ERR(0, 20, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_verbose = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("process_last_error", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 20, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.process_last_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_process_last_error(__pyx_self, __pyx_v_verbose);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_process_last_error(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_verbose) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  struct __pyx_opt_args_6dxfeed_4core_8DXFeedPy_process_last_error __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("process_last_error", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.verbose = __pyx_v_verbose;
  __pyx_t_1 = __pyx_f_6dxfeed_4core_8DXFeedPy_process_last_error(0, &__pyx_t_2); 
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.process_last_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":61
 *     cdef int subs_order
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.subs_order = 0
 * 
 */

/* Python wrapper */
static int __pyx_pw_6dxfeed_4core_8DXFeedPy_15ConnectionClass_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6dxfeed_4core_8DXFeedPy_15ConnectionClass_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 0))) return -1;
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_15ConnectionClass___init__(((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6dxfeed_4core_8DXFeedPy_15ConnectionClass___init__(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "dxfeed/core/DXFeedPy.pyx":62
 * 
 *     def __init__(self):
 *         self.subs_order = 0             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->subs_order = 0;

  /* "dxfeed/core/DXFeedPy.pyx":61
 *     cdef int subs_order
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.subs_order = 0
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":64
 *         self.subs_order = 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         dxf_close_connection(self)
 * 
 */

/* Python wrapper */
static void __pyx_pw_6dxfeed_4core_8DXFeedPy_15ConnectionClass_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6dxfeed_4core_8DXFeedPy_15ConnectionClass_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6dxfeed_4core_8DXFeedPy_15ConnectionClass_2__dealloc__(((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6dxfeed_4core_8DXFeedPy_15ConnectionClass_2__dealloc__(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "dxfeed/core/DXFeedPy.pyx":65
 * 
 *     def __dealloc__(self):
 *         dxf_close_connection(self)             # <<<<<<<<<<<<<<
 * 
 *     cpdef SubscriptionClass make_new_subscription(self, data_len: int):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_dxf_close_connection); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, ((PyObject *)__pyx_v_self)) : __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_self));
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":64
 *         self.subs_order = 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         dxf_close_connection(self)
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("dxfeed.core.DXFeedPy.ConnectionClass.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "dxfeed/core/DXFeedPy.pyx":67
 *         dxf_close_connection(self)
 * 
 *     cpdef SubscriptionClass make_new_subscription(self, data_len: int):             # <<<<<<<<<<<<<<
 *         cdef SubscriptionClass out = SubscriptionClass(data_len)
 *         out.connection = self.connection
 */

static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_15ConnectionClass_5make_new_subscription(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_f_6dxfeed_4core_8DXFeedPy_15ConnectionClass_make_new_subscription(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_self, PyObject *__pyx_v_data_len, int __pyx_skip_dispatch) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_out = 0;
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  dxf_connection_t __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("make_new_subscription", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_make_new_subscription); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_6dxfeed_4core_8DXFeedPy_15ConnectionClass_5make_new_subscription)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_data_len) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_data_len);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_6dxfeed_4core_8DXFeedPy_SubscriptionClass))))) __PYX_ERR(0, 67, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "dxfeed/core/DXFeedPy.pyx":68
 * 
 *     cpdef SubscriptionClass make_new_subscription(self, data_len: int):
 *         cdef SubscriptionClass out = SubscriptionClass(data_len)             # <<<<<<<<<<<<<<
 *         out.connection = self.connection
 *         self.sub_ptr_list.push_back(&out.subscription)  # append pointer to new subscription
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_6dxfeed_4core_8DXFeedPy_SubscriptionClass), __pyx_v_data_len); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_out = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":69
 *     cpdef SubscriptionClass make_new_subscription(self, data_len: int):
 *         cdef SubscriptionClass out = SubscriptionClass(data_len)
 *         out.connection = self.connection             # <<<<<<<<<<<<<<
 *         self.sub_ptr_list.push_back(&out.subscription)  # append pointer to new subscription
 *         out.subscription_order = self.subs_order  # assign each subscription an index
 */
  __pyx_t_5 = __pyx_v_self->connection;
  __pyx_v_out->connection = __pyx_t_5;

  /* "dxfeed/core/DXFeedPy.pyx":70
 *         cdef SubscriptionClass out = SubscriptionClass(data_len)
 *         out.connection = self.connection
 *         self.sub_ptr_list.push_back(&out.subscription)  # append pointer to new subscription             # <<<<<<<<<<<<<<
 *         out.subscription_order = self.subs_order  # assign each subscription an index
 *         self.subs_order += 1
 */
  __pyx_v_self->sub_ptr_list.push_back((&__pyx_v_out->subscription));

  /* "dxfeed/core/DXFeedPy.pyx":71
 *         out.connection = self.connection
 *         self.sub_ptr_list.push_back(&out.subscription)  # append pointer to new subscription
 *         out.subscription_order = self.subs_order  # assign each subscription an index             # <<<<<<<<<<<<<<
 *         self.subs_order += 1
 *         out.con_sub_list_ptr = &self.sub_ptr_list  # reverse pointer to pointers list
 */
  __pyx_t_6 = __pyx_v_self->subs_order;
  __pyx_v_out->subscription_order = __pyx_t_6;

  /* "dxfeed/core/DXFeedPy.pyx":72
 *         self.sub_ptr_list.push_back(&out.subscription)  # append pointer to new subscription
 *         out.subscription_order = self.subs_order  # assign each subscription an index
 *         self.subs_order += 1             # <<<<<<<<<<<<<<
 *         out.con_sub_list_ptr = &self.sub_ptr_list  # reverse pointer to pointers list
 *         return out
 */
  __pyx_v_self->subs_order = (__pyx_v_self->subs_order + 1);

  /* "dxfeed/core/DXFeedPy.pyx":73
 *         out.subscription_order = self.subs_order  # assign each subscription an index
 *         self.subs_order += 1
 *         out.con_sub_list_ptr = &self.sub_ptr_list  # reverse pointer to pointers list             # <<<<<<<<<<<<<<
 *         return out
 * 
 */
  __pyx_v_out->con_sub_list_ptr = (&__pyx_v_self->sub_ptr_list);

  /* "dxfeed/core/DXFeedPy.pyx":74
 *         self.subs_order += 1
 *         out.con_sub_list_ptr = &self.sub_ptr_list  # reverse pointer to pointers list
 *         return out             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_out));
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "dxfeed/core/DXFeedPy.pyx":67
 *         dxf_close_connection(self)
 * 
 *     cpdef SubscriptionClass make_new_subscription(self, data_len: int):             # <<<<<<<<<<<<<<
 *         cdef SubscriptionClass out = SubscriptionClass(data_len)
 *         out.connection = self.connection
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.ConnectionClass.make_new_subscription", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_out);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_15ConnectionClass_5make_new_subscription(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_15ConnectionClass_5make_new_subscription(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_data_len = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("make_new_subscription (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data_len,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data_len)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "make_new_subscription") < 0)) __PYX_ERR(0, 67, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_data_len = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("make_new_subscription", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 67, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.ConnectionClass.make_new_subscription", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_15ConnectionClass_4make_new_subscription(((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *)__pyx_v_self), __pyx_v_data_len);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_15ConnectionClass_4make_new_subscription(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_self, PyObject *__pyx_v_data_len) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("make_new_subscription", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_6dxfeed_4core_8DXFeedPy_15ConnectionClass_make_new_subscription(__pyx_v_self, __pyx_v_data_len, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.ConnectionClass.make_new_subscription", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.connection,self.sub_ptr_list cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_15ConnectionClass_7__reduce_cython__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_15ConnectionClass_7__reduce_cython__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return NULL;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_15ConnectionClass_6__reduce_cython__(((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_15ConnectionClass_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.connection,self.sub_ptr_list cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.connection,self.sub_ptr_list cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.connection,self.sub_ptr_list cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.ConnectionClass.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.connection,self.sub_ptr_list cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.connection,self.sub_ptr_list cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_15ConnectionClass_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_15ConnectionClass_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.ConnectionClass.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_15ConnectionClass_8__setstate_cython__(((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_15ConnectionClass_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.connection,self.sub_ptr_list cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.connection,self.sub_ptr_list cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.connection,self.sub_ptr_list cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.connection,self.sub_ptr_list cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.ConnectionClass.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":91
 *     cdef void *u_data
 * 
 *     def __init__(self, data_len: int):             # <<<<<<<<<<<<<<
 *         """
 *         Parameters
 */

/* Python wrapper */
static int __pyx_pw_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_17SubscriptionClass___init__[] = "\n        Parameters\n        ----------\n        data_len: int\n            Sets maximum amount of events, that are kept in Subscription class\n        ";
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_6dxfeed_4core_8DXFeedPy_17SubscriptionClass___init__;
#endif
static int __pyx_pw_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_data_len = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data_len,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data_len)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 91, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_data_len = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 91, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.SubscriptionClass.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_17SubscriptionClass___init__(((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)__pyx_v_self), __pyx_v_data_len);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6dxfeed_4core_8DXFeedPy_17SubscriptionClass___init__(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_self, PyObject *__pyx_v_data_len) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "dxfeed/core/DXFeedPy.pyx":98
 *             Sets maximum amount of events, that are kept in Subscription class
 *         """
 *         self.subscription = NULL             # <<<<<<<<<<<<<<
 *         self.columns = list()
 *         if data_len > 0:
 */
  __pyx_v_self->subscription = NULL;

  /* "dxfeed/core/DXFeedPy.pyx":99
 *         """
 *         self.subscription = NULL
 *         self.columns = list()             # <<<<<<<<<<<<<<
 *         if data_len > 0:
 *             self.data = deque_wl(maxlen=data_len)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->columns);
  __Pyx_DECREF(__pyx_v_self->columns);
  __pyx_v_self->columns = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":100
 *         self.subscription = NULL
 *         self.columns = list()
 *         if data_len > 0:             # <<<<<<<<<<<<<<
 *             self.data = deque_wl(maxlen=data_len)
 *         else:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_data_len, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 100, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "dxfeed/core/DXFeedPy.pyx":101
 *         self.columns = list()
 *         if data_len > 0:
 *             self.data = deque_wl(maxlen=data_len)             # <<<<<<<<<<<<<<
 *         else:
 *             self.data = deque_wl()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_deque_wl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_maxlen, __pyx_v_data_len) < 0) __PYX_ERR(0, 101, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GIVEREF(__pyx_t_4);
    __Pyx_GOTREF(__pyx_v_self->data);
    __Pyx_DECREF(__pyx_v_self->data);
    __pyx_v_self->data = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":100
 *         self.subscription = NULL
 *         self.columns = list()
 *         if data_len > 0:             # <<<<<<<<<<<<<<
 *             self.data = deque_wl(maxlen=data_len)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "dxfeed/core/DXFeedPy.pyx":103
 *             self.data = deque_wl(maxlen=data_len)
 *         else:
 *             self.data = deque_wl()             # <<<<<<<<<<<<<<
 *         self.u_data = <void *> self.data
 *         self.listener = NULL
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_deque_wl); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GIVEREF(__pyx_t_4);
    __Pyx_GOTREF(__pyx_v_self->data);
    __Pyx_DECREF(__pyx_v_self->data);
    __pyx_v_self->data = __pyx_t_4;
    __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "dxfeed/core/DXFeedPy.pyx":104
 *         else:
 *             self.data = deque_wl()
 *         self.u_data = <void *> self.data             # <<<<<<<<<<<<<<
 *         self.listener = NULL
 * 
 */
  __pyx_v_self->u_data = ((void *)__pyx_v_self->data);

  /* "dxfeed/core/DXFeedPy.pyx":105
 *             self.data = deque_wl()
 *         self.u_data = <void *> self.data
 *         self.listener = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->listener = NULL;

  /* "dxfeed/core/DXFeedPy.pyx":91
 *     cdef void *u_data
 * 
 *     def __init__(self, data_len: int):             # <<<<<<<<<<<<<<
 *         """
 *         Parameters
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.SubscriptionClass.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":107
 *         self.listener = NULL
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.subscription:  # if connection is not closed
 *             clib.dxf_close_subscription(self.subscription)
 */

/* Python wrapper */
static void __pyx_pw_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_2__dealloc__(((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_2__dealloc__(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "dxfeed/core/DXFeedPy.pyx":108
 * 
 *     def __dealloc__(self):
 *         if self.subscription:  # if connection is not closed             # <<<<<<<<<<<<<<
 *             clib.dxf_close_subscription(self.subscription)
 *             self.con_sub_list_ptr[0][self.subscription_order] = NULL
 */
  __pyx_t_1 = (__pyx_v_self->subscription != 0);
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":109
 *     def __dealloc__(self):
 *         if self.subscription:  # if connection is not closed
 *             clib.dxf_close_subscription(self.subscription)             # <<<<<<<<<<<<<<
 *             self.con_sub_list_ptr[0][self.subscription_order] = NULL
 * 
 */
    (void)(dxf_close_subscription(__pyx_v_self->subscription));

    /* "dxfeed/core/DXFeedPy.pyx":110
 *         if self.subscription:  # if connection is not closed
 *             clib.dxf_close_subscription(self.subscription)
 *             self.con_sub_list_ptr[0][self.subscription_order] = NULL             # <<<<<<<<<<<<<<
 * 
 *     def get_data(self):
 */
    ((__pyx_v_self->con_sub_list_ptr[0])[__pyx_v_self->subscription_order]) = NULL;

    /* "dxfeed/core/DXFeedPy.pyx":108
 * 
 *     def __dealloc__(self):
 *         if self.subscription:  # if connection is not closed             # <<<<<<<<<<<<<<
 *             clib.dxf_close_subscription(self.subscription)
 *             self.con_sub_list_ptr[0][self.subscription_order] = NULL
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":107
 *         self.listener = NULL
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.subscription:  # if connection is not closed
 *             clib.dxf_close_subscription(self.subscription)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "dxfeed/core/DXFeedPy.pyx":112
 *             self.con_sub_list_ptr[0][self.subscription_order] = NULL
 * 
 *     def get_data(self):             # <<<<<<<<<<<<<<
 *         """
 *         Method returns list with data, specified in event listener and returned data will be removed from object buffer
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_5get_data(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_4get_data[] = "\n        Method returns list with data, specified in event listener and returned data will be removed from object buffer\n\n        Returns\n        -------\n        list\n            List with data\n        ";
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_5get_data(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_data (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_data", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return NULL;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_data", 0))) return NULL;
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_4get_data(((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_4get_data(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("get_data", 0);

  /* "dxfeed/core/DXFeedPy.pyx":121
 *             List with data
 *         """
 *         return self.data.safe_get()             # <<<<<<<<<<<<<<
 * 
 *     def to_dataframe(self, keep: bool=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data, __pyx_n_s_safe_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "dxfeed/core/DXFeedPy.pyx":112
 *             self.con_sub_list_ptr[0][self.subscription_order] = NULL
 * 
 *     def get_data(self):             # <<<<<<<<<<<<<<
 *         """
 *         Method returns list with data, specified in event listener and returned data will be removed from object buffer
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.SubscriptionClass.get_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":123
 *         return self.data.safe_get()
 * 
 *     def to_dataframe(self, keep: bool=True):             # <<<<<<<<<<<<<<
 *         """
 *         Method converts data to the Pandas DataFrame
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_7to_dataframe(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_6to_dataframe[] = "\n        Method converts data to the Pandas DataFrame\n\n        Parameters\n        ----------\n        keep: bool\n            When True copies data to dataframe, otherwise pops. Default True\n\n        Returns\n        -------\n        df: pandas DataFrame\n        ";
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_7to_dataframe(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_keep = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_dataframe (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_keep,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_keep);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "to_dataframe") < 0)) __PYX_ERR(0, 123, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_keep = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("to_dataframe", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 123, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.SubscriptionClass.to_dataframe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_6to_dataframe(((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)__pyx_v_self), __pyx_v_keep);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_6to_dataframe(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_self, PyObject *__pyx_v_keep) {
  PyObject *__pyx_v_df_data = NULL;
  PyObject *__pyx_v_df = NULL;
  PyObject *__pyx_v_time_columns = NULL;
  PyObject *__pyx_v_column = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("to_dataframe", 0);

  /* "dxfeed/core/DXFeedPy.pyx":136
 *         df: pandas DataFrame
 *         """
 *         if keep:             # <<<<<<<<<<<<<<
 *             df_data = self.data.copy()
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_keep); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 136, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":137
 *         """
 *         if keep:
 *             df_data = self.data.copy()             # <<<<<<<<<<<<<<
 *         else:
 *             df_data = self.data.safe_get()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data, __pyx_n_s_copy); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_df_data = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":136
 *         df: pandas DataFrame
 *         """
 *         if keep:             # <<<<<<<<<<<<<<
 *             df_data = self.data.copy()
 *         else:
 */
    goto __pyx_L3;
  }

  /* "dxfeed/core/DXFeedPy.pyx":139
 *             df_data = self.data.copy()
 *         else:
 *             df_data = self.data.safe_get()             # <<<<<<<<<<<<<<
 * 
 *         df = pd.DataFrame(df_data, columns=self.columns)
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->data, __pyx_n_s_safe_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_df_data = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "dxfeed/core/DXFeedPy.pyx":141
 *             df_data = self.data.safe_get()
 * 
 *         df = pd.DataFrame(df_data, columns=self.columns)             # <<<<<<<<<<<<<<
 *         time_columns = df.columns[df.columns.str.contains('Time')]
 *         for column in time_columns:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pd); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_DataFrame); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_df_data);
  __Pyx_GIVEREF(__pyx_v_df_data);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_df_data);
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_columns, __pyx_v_self->columns) < 0) __PYX_ERR(0, 141, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_df = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":142
 * 
 *         df = pd.DataFrame(df_data, columns=self.columns)
 *         time_columns = df.columns[df.columns.str.contains('Time')]             # <<<<<<<<<<<<<<
 *         for column in time_columns:
 *             df.loc[:, column] = df.loc[:, column].astype('<M8[ms]')
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_df, __pyx_n_s_columns); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_df, __pyx_n_s_columns); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_str); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_contains); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_n_u_Time) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_n_u_Time);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_time_columns = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":143
 *         df = pd.DataFrame(df_data, columns=self.columns)
 *         time_columns = df.columns[df.columns.str.contains('Time')]
 *         for column in time_columns:             # <<<<<<<<<<<<<<
 *             df.loc[:, column] = df.loc[:, column].astype('<M8[ms]')
 *         return df
 */
  if (likely(PyList_CheckExact(__pyx_v_time_columns)) || PyTuple_CheckExact(__pyx_v_time_columns)) {
    __pyx_t_2 = __pyx_v_time_columns; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_time_columns); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 143, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 143, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 143, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 143, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 143, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_7(__pyx_t_2);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 143, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_column, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":144
 *         time_columns = df.columns[df.columns.str.contains('Time')]
 *         for column in time_columns:
 *             df.loc[:, column] = df.loc[:, column].astype('<M8[ms]')             # <<<<<<<<<<<<<<
 *         return df
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_df, __pyx_n_s_loc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_slice__4);
    __Pyx_GIVEREF(__pyx_slice__4);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_slice__4);
    __Pyx_INCREF(__pyx_v_column);
    __Pyx_GIVEREF(__pyx_v_column);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_column);
    __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_astype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_8, __pyx_kp_u_M8_ms) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u_M8_ms);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_df, __pyx_n_s_loc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_slice__4);
    __Pyx_GIVEREF(__pyx_slice__4);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_slice__4);
    __Pyx_INCREF(__pyx_v_column);
    __Pyx_GIVEREF(__pyx_v_column);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_v_column);
    if (unlikely(PyObject_SetItem(__pyx_t_3, __pyx_t_8, __pyx_t_4) < 0)) __PYX_ERR(0, 144, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":143
 *         df = pd.DataFrame(df_data, columns=self.columns)
 *         time_columns = df.columns[df.columns.str.contains('Time')]
 *         for column in time_columns:             # <<<<<<<<<<<<<<
 *             df.loc[:, column] = df.loc[:, column].astype('<M8[ms]')
 *         return df
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":145
 *         for column in time_columns:
 *             df.loc[:, column] = df.loc[:, column].astype('<M8[ms]')
 *         return df             # <<<<<<<<<<<<<<
 * 
 * def dxf_create_connection(address: Union[str, unicode, bytes] = 'demo.dxfeed.com:7300'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_df);
  __pyx_r = __pyx_v_df;
  goto __pyx_L0;

  /* "dxfeed/core/DXFeedPy.pyx":123
 *         return self.data.safe_get()
 * 
 *     def to_dataframe(self, keep: bool=True):             # <<<<<<<<<<<<<<
 *         """
 *         Method converts data to the Pandas DataFrame
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.SubscriptionClass.to_dataframe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_df_data);
  __Pyx_XDECREF(__pyx_v_df);
  __Pyx_XDECREF(__pyx_v_time_columns);
  __Pyx_XDECREF(__pyx_v_column);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.con_sub_list_ptr,self.connection,self.listener,self.subscription,self.u_data cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_9__reduce_cython__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_9__reduce_cython__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return NULL;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_8__reduce_cython__(((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.con_sub_list_ptr,self.connection,self.listener,self.subscription,self.u_data cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.con_sub_list_ptr,self.connection,self.listener,self.subscription,self.u_data cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.con_sub_list_ptr,self.connection,self.listener,self.subscription,self.u_data cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.SubscriptionClass.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.con_sub_list_ptr,self.connection,self.listener,self.subscription,self.u_data cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.con_sub_list_ptr,self.connection,self.listener,self.subscription,self.u_data cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.SubscriptionClass.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_10__setstate_cython__(((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.con_sub_list_ptr,self.connection,self.listener,self.subscription,self.u_data cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.con_sub_list_ptr,self.connection,self.listener,self.subscription,self.u_data cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.con_sub_list_ptr,self.connection,self.listener,self.subscription,self.u_data cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.con_sub_list_ptr,self.connection,self.listener,self.subscription,self.u_data cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.SubscriptionClass.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":147
 *         return df
 * 
 * def dxf_create_connection(address: Union[str, unicode, bytes] = 'demo.dxfeed.com:7300'):             # <<<<<<<<<<<<<<
 *     """
 *     Function creates connection to dxfeed given url address
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_3dxf_create_connection(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_2dxf_create_connection[] = "\n    Function creates connection to dxfeed given url address\n\n    Parameters\n    ----------\n    address: str\n        dxfeed url address\n\n    Returns\n    -------\n    cc: ConnectionClass\n        Cython ConnectionClass with information about connection\n    ";
static PyMethodDef __pyx_mdef_6dxfeed_4core_8DXFeedPy_3dxf_create_connection = {"dxf_create_connection", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_3dxf_create_connection, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_2dxf_create_connection};
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_3dxf_create_connection(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_address = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dxf_create_connection (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_address,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)__pyx_kp_u_demo_dxfeed_com_7300);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_address);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dxf_create_connection") < 0)) __PYX_ERR(0, 147, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_address = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dxf_create_connection", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 147, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_create_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_2dxf_create_connection(__pyx_self, __pyx_v_address);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_2dxf_create_connection(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_address) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_cc = NULL;
  int __pyx_v_error_code;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char const *__pyx_t_4;
  int __pyx_t_5;
  struct __pyx_opt_args_6dxfeed_4core_8DXFeedPy_process_last_error __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_UCS4 __pyx_t_9;
  __Pyx_RefNannySetupContext("dxf_create_connection", 0);
  __Pyx_INCREF(__pyx_v_address);

  /* "dxfeed/core/DXFeedPy.pyx":161
 *         Cython ConnectionClass with information about connection
 *     """
 *     cc = ConnectionClass()             # <<<<<<<<<<<<<<
 *     address = address.encode('utf-8')
 *     clib.dxf_create_connection(address, NULL, NULL, NULL, NULL, NULL, &cc.connection)
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6dxfeed_4core_8DXFeedPy_ConnectionClass)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cc = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":162
 *     """
 *     cc = ConnectionClass()
 *     address = address.encode('utf-8')             # <<<<<<<<<<<<<<
 *     clib.dxf_create_connection(address, NULL, NULL, NULL, NULL, NULL, &cc.connection)
 *     error_code = process_last_error(verbose=False)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_address, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_address, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":163
 *     cc = ConnectionClass()
 *     address = address.encode('utf-8')
 *     clib.dxf_create_connection(address, NULL, NULL, NULL, NULL, NULL, &cc.connection)             # <<<<<<<<<<<<<<
 *     error_code = process_last_error(verbose=False)
 *     if error_code:
 */
  __pyx_t_4 = __Pyx_PyObject_AsString(__pyx_v_address); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 163, __pyx_L1_error)
  (void)(dxf_create_connection(__pyx_t_4, NULL, NULL, NULL, NULL, NULL, (&__pyx_v_cc->connection)));

  /* "dxfeed/core/DXFeedPy.pyx":164
 *     address = address.encode('utf-8')
 *     clib.dxf_create_connection(address, NULL, NULL, NULL, NULL, NULL, &cc.connection)
 *     error_code = process_last_error(verbose=False)             # <<<<<<<<<<<<<<
 *     if error_code:
 *         raise RuntimeError(f"In underlying C-API library error {error_code} occurred!")
 */
  __pyx_t_6.__pyx_n = 1;
  __pyx_t_6.verbose = Py_False;
  __pyx_t_5 = __pyx_f_6dxfeed_4core_8DXFeedPy_process_last_error(0, &__pyx_t_6); 
  __pyx_v_error_code = __pyx_t_5;

  /* "dxfeed/core/DXFeedPy.pyx":165
 *     clib.dxf_create_connection(address, NULL, NULL, NULL, NULL, NULL, &cc.connection)
 *     error_code = process_last_error(verbose=False)
 *     if error_code:             # <<<<<<<<<<<<<<
 *         raise RuntimeError(f"In underlying C-API library error {error_code} occurred!")
 *     return cc
 */
  __pyx_t_7 = (__pyx_v_error_code != 0);
  if (unlikely(__pyx_t_7)) {

    /* "dxfeed/core/DXFeedPy.pyx":166
 *     error_code = process_last_error(verbose=False)
 *     if error_code:
 *         raise RuntimeError(f"In underlying C-API library error {error_code} occurred!")             # <<<<<<<<<<<<<<
 *     return cc
 * 
 */
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = 0;
    __pyx_t_9 = 127;
    __Pyx_INCREF(__pyx_kp_u_In_underlying_C_API_library_erro);
    __pyx_t_8 += 34;
    __Pyx_GIVEREF(__pyx_kp_u_In_underlying_C_API_library_erro);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_In_underlying_C_API_library_erro);
    __pyx_t_2 = __Pyx_PyUnicode_From_int(__pyx_v_error_code, 0, ' ', 'd'); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_occurred);
    __pyx_t_8 += 10;
    __Pyx_GIVEREF(__pyx_kp_u_occurred);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_occurred);
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 166, __pyx_L1_error)

    /* "dxfeed/core/DXFeedPy.pyx":165
 *     clib.dxf_create_connection(address, NULL, NULL, NULL, NULL, NULL, &cc.connection)
 *     error_code = process_last_error(verbose=False)
 *     if error_code:             # <<<<<<<<<<<<<<
 *         raise RuntimeError(f"In underlying C-API library error {error_code} occurred!")
 *     return cc
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":167
 *     if error_code:
 *         raise RuntimeError(f"In underlying C-API library error {error_code} occurred!")
 *     return cc             # <<<<<<<<<<<<<<
 * 
 * def dxf_create_connection_auth_bearer(address: Union[str, unicode, bytes],
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_cc));
  __pyx_r = ((PyObject *)__pyx_v_cc);
  goto __pyx_L0;

  /* "dxfeed/core/DXFeedPy.pyx":147
 *         return df
 * 
 * def dxf_create_connection(address: Union[str, unicode, bytes] = 'demo.dxfeed.com:7300'):             # <<<<<<<<<<<<<<
 *     """
 *     Function creates connection to dxfeed given url address
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_create_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cc);
  __Pyx_XDECREF(__pyx_v_address);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":169
 *     return cc
 * 
 * def dxf_create_connection_auth_bearer(address: Union[str, unicode, bytes],             # <<<<<<<<<<<<<<
 *                                      token: Union[str, unicode, bytes]):
 *     """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_5dxf_create_connection_auth_bearer(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_4dxf_create_connection_auth_bearer[] = "\n    Function creates connection to dxfeed given url address and token\n\n    Parameters\n    ----------\n    address: str\n        dxfeed url address\n    token: str\n        dxfeed token\n\n    Returns\n    -------\n    cc: ConnectionClass\n        Cython ConnectionClass with information about connection\n    ";
static PyMethodDef __pyx_mdef_6dxfeed_4core_8DXFeedPy_5dxf_create_connection_auth_bearer = {"dxf_create_connection_auth_bearer", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_5dxf_create_connection_auth_bearer, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_4dxf_create_connection_auth_bearer};
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_5dxf_create_connection_auth_bearer(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_address = 0;
  PyObject *__pyx_v_token = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dxf_create_connection_auth_bearer (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_address,&__pyx_n_s_token,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_address)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_token)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dxf_create_connection_auth_bearer", 1, 2, 2, 1); __PYX_ERR(0, 169, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dxf_create_connection_auth_bearer") < 0)) __PYX_ERR(0, 169, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_address = values[0];
    __pyx_v_token = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dxf_create_connection_auth_bearer", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 169, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_create_connection_auth_bearer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_4dxf_create_connection_auth_bearer(__pyx_self, __pyx_v_address, __pyx_v_token);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_4dxf_create_connection_auth_bearer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_address, PyObject *__pyx_v_token) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_cc = NULL;
  int __pyx_v_error_code;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char const *__pyx_t_4;
  char const *__pyx_t_5;
  int __pyx_t_6;
  struct __pyx_opt_args_6dxfeed_4core_8DXFeedPy_process_last_error __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_UCS4 __pyx_t_10;
  __Pyx_RefNannySetupContext("dxf_create_connection_auth_bearer", 0);
  __Pyx_INCREF(__pyx_v_address);
  __Pyx_INCREF(__pyx_v_token);

  /* "dxfeed/core/DXFeedPy.pyx":186
 *         Cython ConnectionClass with information about connection
 *     """
 *     cc = ConnectionClass()             # <<<<<<<<<<<<<<
 *     address = address.encode('utf-8')
 *     token = token.encode('utf-8')
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6dxfeed_4core_8DXFeedPy_ConnectionClass)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cc = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":187
 *     """
 *     cc = ConnectionClass()
 *     address = address.encode('utf-8')             # <<<<<<<<<<<<<<
 *     token = token.encode('utf-8')
 *     clib.dxf_create_connection_auth_bearer(address, token,
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_address, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_address, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":188
 *     cc = ConnectionClass()
 *     address = address.encode('utf-8')
 *     token = token.encode('utf-8')             # <<<<<<<<<<<<<<
 *     clib.dxf_create_connection_auth_bearer(address, token,
 *                                           NULL, NULL, NULL, NULL, NULL, &cc.connection)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_token, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_token, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":189
 *     address = address.encode('utf-8')
 *     token = token.encode('utf-8')
 *     clib.dxf_create_connection_auth_bearer(address, token,             # <<<<<<<<<<<<<<
 *                                           NULL, NULL, NULL, NULL, NULL, &cc.connection)
 *     error_code = process_last_error(verbose=False)
 */
  __pyx_t_4 = __Pyx_PyObject_AsString(__pyx_v_address); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 189, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_AsString(__pyx_v_token); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 189, __pyx_L1_error)

  /* "dxfeed/core/DXFeedPy.pyx":190
 *     token = token.encode('utf-8')
 *     clib.dxf_create_connection_auth_bearer(address, token,
 *                                           NULL, NULL, NULL, NULL, NULL, &cc.connection)             # <<<<<<<<<<<<<<
 *     error_code = process_last_error(verbose=False)
 *     if error_code:
 */
  (void)(dxf_create_connection_auth_bearer(__pyx_t_4, __pyx_t_5, NULL, NULL, NULL, NULL, NULL, (&__pyx_v_cc->connection)));

  /* "dxfeed/core/DXFeedPy.pyx":191
 *     clib.dxf_create_connection_auth_bearer(address, token,
 *                                           NULL, NULL, NULL, NULL, NULL, &cc.connection)
 *     error_code = process_last_error(verbose=False)             # <<<<<<<<<<<<<<
 *     if error_code:
 *         raise RuntimeError(f"In underlying C-API library error {error_code} occurred!")
 */
  __pyx_t_7.__pyx_n = 1;
  __pyx_t_7.verbose = Py_False;
  __pyx_t_6 = __pyx_f_6dxfeed_4core_8DXFeedPy_process_last_error(0, &__pyx_t_7); 
  __pyx_v_error_code = __pyx_t_6;

  /* "dxfeed/core/DXFeedPy.pyx":192
 *                                           NULL, NULL, NULL, NULL, NULL, &cc.connection)
 *     error_code = process_last_error(verbose=False)
 *     if error_code:             # <<<<<<<<<<<<<<
 *         raise RuntimeError(f"In underlying C-API library error {error_code} occurred!")
 *     return cc
 */
  __pyx_t_8 = (__pyx_v_error_code != 0);
  if (unlikely(__pyx_t_8)) {

    /* "dxfeed/core/DXFeedPy.pyx":193
 *     error_code = process_last_error(verbose=False)
 *     if error_code:
 *         raise RuntimeError(f"In underlying C-API library error {error_code} occurred!")             # <<<<<<<<<<<<<<
 *     return cc
 * 
 */
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = 0;
    __pyx_t_10 = 127;
    __Pyx_INCREF(__pyx_kp_u_In_underlying_C_API_library_erro);
    __pyx_t_9 += 34;
    __Pyx_GIVEREF(__pyx_kp_u_In_underlying_C_API_library_erro);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_In_underlying_C_API_library_erro);
    __pyx_t_2 = __Pyx_PyUnicode_From_int(__pyx_v_error_code, 0, ' ', 'd'); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_occurred);
    __pyx_t_9 += 10;
    __Pyx_GIVEREF(__pyx_kp_u_occurred);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_occurred);
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 193, __pyx_L1_error)

    /* "dxfeed/core/DXFeedPy.pyx":192
 *                                           NULL, NULL, NULL, NULL, NULL, &cc.connection)
 *     error_code = process_last_error(verbose=False)
 *     if error_code:             # <<<<<<<<<<<<<<
 *         raise RuntimeError(f"In underlying C-API library error {error_code} occurred!")
 *     return cc
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":194
 *     if error_code:
 *         raise RuntimeError(f"In underlying C-API library error {error_code} occurred!")
 *     return cc             # <<<<<<<<<<<<<<
 * 
 * def dxf_create_subscription(ConnectionClass cc, event_type: str, candle_time: Optional[str] = None,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_cc));
  __pyx_r = ((PyObject *)__pyx_v_cc);
  goto __pyx_L0;

  /* "dxfeed/core/DXFeedPy.pyx":169
 *     return cc
 * 
 * def dxf_create_connection_auth_bearer(address: Union[str, unicode, bytes],             # <<<<<<<<<<<<<<
 *                                      token: Union[str, unicode, bytes]):
 *     """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_create_connection_auth_bearer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cc);
  __Pyx_XDECREF(__pyx_v_address);
  __Pyx_XDECREF(__pyx_v_token);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":196
 *     return cc
 * 
 * def dxf_create_subscription(ConnectionClass cc, event_type: str, candle_time: Optional[str] = None,             # <<<<<<<<<<<<<<
 *                             data_len: int = 100000):
 *     """
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_7dxf_create_subscription(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_6dxf_create_subscription[] = "\n    Function creates subscription and writes all relevant information to SubscriptionClass\n\n    Parameters\n    ----------\n    cc: ConnectionClass\n        Variable with connection information\n    event_type: str\n        Event types: 'Trade', 'Quote', 'Summary', 'Profile', 'Order', 'TimeAndSale', 'Candle', 'TradeETH',\n        'SpreadOrder', 'Greeks', 'TheoPrice', 'Underlying', 'Series', 'Configuration' or ''\n    candle_time: str\n        String of %Y-%m-%d %H:%M:%S datetime format for retrieving candles. By default set to now\n    data_len: int\n        Sets maximum amount of events, that are kept in Subscription class\n\n    Returns\n    -------\n    sc: SubscriptionClass\n        Cython SubscriptionClass with information about subscription\n    ";
static PyMethodDef __pyx_mdef_6dxfeed_4core_8DXFeedPy_7dxf_create_subscription = {"dxf_create_subscription", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_7dxf_create_subscription, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_6dxf_create_subscription};
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_7dxf_create_subscription(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_cc = 0;
  PyObject *__pyx_v_event_type = 0;
  PyObject *__pyx_v_candle_time = 0;
  PyObject *__pyx_v_data_len = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dxf_create_subscription (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cc,&__pyx_n_s_event_type,&__pyx_n_s_candle_time,&__pyx_n_s_data_len,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject *)Py_None);
    values[3] = ((PyObject *)__pyx_int_100000);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_event_type)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dxf_create_subscription", 0, 2, 4, 1); __PYX_ERR(0, 196, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_candle_time);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data_len);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dxf_create_subscription") < 0)) __PYX_ERR(0, 196, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_cc = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *)values[0]);
    __pyx_v_event_type = ((PyObject*)values[1]);
    __pyx_v_candle_time = values[2];
    __pyx_v_data_len = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dxf_create_subscription", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 196, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_create_subscription", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_cc), __pyx_ptype_6dxfeed_4core_8DXFeedPy_ConnectionClass, 1, "cc", 0))) __PYX_ERR(0, 196, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_event_type), (&PyUnicode_Type), 1, "event_type", 1))) __PYX_ERR(0, 196, __pyx_L1_error)
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_6dxf_create_subscription(__pyx_self, __pyx_v_cc, __pyx_v_event_type, __pyx_v_candle_time, __pyx_v_data_len);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_6dxf_create_subscription(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_cc, PyObject *__pyx_v_event_type, PyObject *__pyx_v_candle_time, PyObject *__pyx_v_data_len) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc = NULL;
  PyObject *__pyx_v_et_type_int = NULL;
  PyObject *__pyx_v_timestamp = NULL;
  int __pyx_v_error_code;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  dxf_long_t __pyx_t_14;
  struct __pyx_opt_args_6dxfeed_4core_8DXFeedPy_process_last_error __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_UCS4 __pyx_t_17;
  __Pyx_RefNannySetupContext("dxf_create_subscription", 0);
  __Pyx_INCREF(__pyx_v_candle_time);

  /* "dxfeed/core/DXFeedPy.pyx":218
 *         Cython SubscriptionClass with information about subscription
 *     """
 *     if not cc.connection:             # <<<<<<<<<<<<<<
 *         raise ValueError('Connection is not valid')
 *     if event_type not in ['Trade', 'Quote', 'Summary', 'Profile', 'Order', 'TimeAndSale', 'Candle', 'TradeETH',
 */
  __pyx_t_1 = ((!(__pyx_v_cc->connection != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "dxfeed/core/DXFeedPy.pyx":219
 *     """
 *     if not cc.connection:
 *         raise ValueError('Connection is not valid')             # <<<<<<<<<<<<<<
 *     if event_type not in ['Trade', 'Quote', 'Summary', 'Profile', 'Order', 'TimeAndSale', 'Candle', 'TradeETH',
 *                           'SpreadOrder', 'Greeks', 'TheoPrice', 'Underlying', 'Series', 'Configuration', ]:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 219, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 219, __pyx_L1_error)

    /* "dxfeed/core/DXFeedPy.pyx":218
 *         Cython SubscriptionClass with information about subscription
 *     """
 *     if not cc.connection:             # <<<<<<<<<<<<<<
 *         raise ValueError('Connection is not valid')
 *     if event_type not in ['Trade', 'Quote', 'Summary', 'Profile', 'Order', 'TimeAndSale', 'Candle', 'TradeETH',
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":220
 *     if not cc.connection:
 *         raise ValueError('Connection is not valid')
 *     if event_type not in ['Trade', 'Quote', 'Summary', 'Profile', 'Order', 'TimeAndSale', 'Candle', 'TradeETH',             # <<<<<<<<<<<<<<
 *                           'SpreadOrder', 'Greeks', 'TheoPrice', 'Underlying', 'Series', 'Configuration', ]:
 *         raise ValueError('Incorrect event type!')
 */
  __Pyx_INCREF(__pyx_v_event_type);
  __pyx_t_3 = __pyx_v_event_type;
  __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_Trade, Py_NE)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_Quote, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_Summary, Py_NE)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_Profile, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_Order, Py_NE)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_TimeAndSale, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_Candle, Py_NE)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_TradeETH, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_SpreadOrder, Py_NE)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_Greeks, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_TheoPrice, Py_NE)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_Underlying, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_Series, Py_NE)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_Configuration, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_5 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L5_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (unlikely(__pyx_t_4)) {

    /* "dxfeed/core/DXFeedPy.pyx":222
 *     if event_type not in ['Trade', 'Quote', 'Summary', 'Profile', 'Order', 'TimeAndSale', 'Candle', 'TradeETH',
 *                           'SpreadOrder', 'Greeks', 'TheoPrice', 'Underlying', 'Series', 'Configuration', ]:
 *         raise ValueError('Incorrect event type!')             # <<<<<<<<<<<<<<
 * 
 *     sc = cc.make_new_subscription(data_len=data_len)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 222, __pyx_L1_error)

    /* "dxfeed/core/DXFeedPy.pyx":220
 *     if not cc.connection:
 *         raise ValueError('Connection is not valid')
 *     if event_type not in ['Trade', 'Quote', 'Summary', 'Profile', 'Order', 'TimeAndSale', 'Candle', 'TradeETH',             # <<<<<<<<<<<<<<
 *                           'SpreadOrder', 'Greeks', 'TheoPrice', 'Underlying', 'Series', 'Configuration', ]:
 *         raise ValueError('Incorrect event type!')
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":224
 *         raise ValueError('Incorrect event type!')
 * 
 *     sc = cc.make_new_subscription(data_len=data_len)             # <<<<<<<<<<<<<<
 *     sc.event_type_str = event_type
 *     et_type_int = event_type_convert(event_type)
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_6dxfeed_4core_8DXFeedPy_ConnectionClass *)__pyx_v_cc->__pyx_vtab)->make_new_subscription(__pyx_v_cc, __pyx_v_data_len, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_sc = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":225
 * 
 *     sc = cc.make_new_subscription(data_len=data_len)
 *     sc.event_type_str = event_type             # <<<<<<<<<<<<<<
 *     et_type_int = event_type_convert(event_type)
 * 
 */
  __Pyx_INCREF(__pyx_v_event_type);
  __Pyx_GIVEREF(__pyx_v_event_type);
  __Pyx_GOTREF(__pyx_v_sc->event_type_str);
  __Pyx_DECREF(__pyx_v_sc->event_type_str);
  __pyx_v_sc->event_type_str = __pyx_v_event_type;

  /* "dxfeed/core/DXFeedPy.pyx":226
 *     sc = cc.make_new_subscription(data_len=data_len)
 *     sc.event_type_str = event_type
 *     et_type_int = event_type_convert(event_type)             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_event_type_convert); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_v_event_type) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_event_type);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_et_type_int = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":228
 *     et_type_int = event_type_convert(event_type)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         candle_time = datetime.strptime(candle_time, '%Y-%m-%d %H:%M:%S') if candle_time else datetime.utcnow()
 *         timestamp = int((candle_time - datetime(1970, 1, 1)).total_seconds()) * 1000 - 5000
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_10);
    /*try:*/ {

      /* "dxfeed/core/DXFeedPy.pyx":229
 * 
 *     try:
 *         candle_time = datetime.strptime(candle_time, '%Y-%m-%d %H:%M:%S') if candle_time else datetime.utcnow()             # <<<<<<<<<<<<<<
 *         timestamp = int((candle_time - datetime(1970, 1, 1)).total_seconds()) * 1000 - 5000
 *     except ValueError:
 */
      __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_candle_time); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 229, __pyx_L19_error)
      if (__pyx_t_4) {
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_datetime); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 229, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_strptime); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 229, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        __pyx_t_12 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
            __pyx_t_12 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_11)) {
          PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_candle_time, __pyx_kp_u_Y_m_d_H_M_S};
          __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 229, __pyx_L19_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_6);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
          PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_candle_time, __pyx_kp_u_Y_m_d_H_M_S};
          __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 229, __pyx_L19_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_6);
        } else
        #endif
        {
          __pyx_t_13 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 229, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_13);
          if (__pyx_t_7) {
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_7); __pyx_t_7 = NULL;
          }
          __Pyx_INCREF(__pyx_v_candle_time);
          __Pyx_GIVEREF(__pyx_v_candle_time);
          PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_12, __pyx_v_candle_time);
          __Pyx_INCREF(__pyx_kp_u_Y_m_d_H_M_S);
          __Pyx_GIVEREF(__pyx_kp_u_Y_m_d_H_M_S);
          PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_12, __pyx_kp_u_Y_m_d_H_M_S);
          __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_13, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 229, __pyx_L19_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_2 = __pyx_t_6;
        __pyx_t_6 = 0;
      } else {
        __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_datetime); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 229, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_utcnow); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 229, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
          }
        }
        __pyx_t_6 = (__pyx_t_11) ? __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_t_11) : __Pyx_PyObject_CallNoArg(__pyx_t_13);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 229, __pyx_L19_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_2 = __pyx_t_6;
        __pyx_t_6 = 0;
      }
      __Pyx_DECREF_SET(__pyx_v_candle_time, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "dxfeed/core/DXFeedPy.pyx":230
 *     try:
 *         candle_time = datetime.strptime(candle_time, '%Y-%m-%d %H:%M:%S') if candle_time else datetime.utcnow()
 *         timestamp = int((candle_time - datetime(1970, 1, 1)).total_seconds()) * 1000 - 5000             # <<<<<<<<<<<<<<
 *     except ValueError:
 *         raise Exception("Inapropriate date format, should be %Y-%m-%d %H:%M:%S")
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_datetime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 230, __pyx_L19_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 230, __pyx_L19_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyNumber_Subtract(__pyx_v_candle_time, __pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 230, __pyx_L19_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_total_seconds); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 230, __pyx_L19_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_13);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_13, function);
        }
      }
      __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_13);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 230, __pyx_L19_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_13 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 230, __pyx_L19_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Multiply(__pyx_t_13, __pyx_int_1000); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 230, __pyx_L19_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_13 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_5000, 0x1388, 0, 0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 230, __pyx_L19_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_timestamp = __pyx_t_13;
      __pyx_t_13 = 0;

      /* "dxfeed/core/DXFeedPy.pyx":228
 *     et_type_int = event_type_convert(event_type)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         candle_time = datetime.strptime(candle_time, '%Y-%m-%d %H:%M:%S') if candle_time else datetime.utcnow()
 *         timestamp = int((candle_time - datetime(1970, 1, 1)).total_seconds()) * 1000 - 5000
 */
    }
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    goto __pyx_L24_try_end;
    __pyx_L19_error:;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":231
 *         candle_time = datetime.strptime(candle_time, '%Y-%m-%d %H:%M:%S') if candle_time else datetime.utcnow()
 *         timestamp = int((candle_time - datetime(1970, 1, 1)).total_seconds()) * 1000 - 5000
 *     except ValueError:             # <<<<<<<<<<<<<<
 *         raise Exception("Inapropriate date format, should be %Y-%m-%d %H:%M:%S")
 * 
 */
    __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
    if (__pyx_t_12) {
      __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_create_subscription", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_13, &__pyx_t_2, &__pyx_t_6) < 0) __PYX_ERR(0, 231, __pyx_L21_except_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_6);

      /* "dxfeed/core/DXFeedPy.pyx":232
 *         timestamp = int((candle_time - datetime(1970, 1, 1)).total_seconds()) * 1000 - 5000
 *     except ValueError:
 *         raise Exception("Inapropriate date format, should be %Y-%m-%d %H:%M:%S")             # <<<<<<<<<<<<<<
 * 
 *     if event_type == 'Candle':
 */
      __pyx_t_11 = __Pyx_PyObject_Call(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 232, __pyx_L21_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_Raise(__pyx_t_11, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __PYX_ERR(0, 232, __pyx_L21_except_error)
    }
    goto __pyx_L21_except_error;
    __pyx_L21_except_error:;

    /* "dxfeed/core/DXFeedPy.pyx":228
 *     et_type_int = event_type_convert(event_type)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         candle_time = datetime.strptime(candle_time, '%Y-%m-%d %H:%M:%S') if candle_time else datetime.utcnow()
 *         timestamp = int((candle_time - datetime(1970, 1, 1)).total_seconds()) * 1000 - 5000
 */
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
    goto __pyx_L1_error;
    __pyx_L24_try_end:;
  }

  /* "dxfeed/core/DXFeedPy.pyx":234
 *         raise Exception("Inapropriate date format, should be %Y-%m-%d %H:%M:%S")
 * 
 *     if event_type == 'Candle':             # <<<<<<<<<<<<<<
 *         clib.dxf_create_subscription_timed(sc.connection, et_type_int, timestamp, &sc.subscription)
 *     else:
 */
  __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_v_event_type, __pyx_n_u_Candle, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 234, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_4 != 0);
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":235
 * 
 *     if event_type == 'Candle':
 *         clib.dxf_create_subscription_timed(sc.connection, et_type_int, timestamp, &sc.subscription)             # <<<<<<<<<<<<<<
 *     else:
 *         clib.dxf_create_subscription(sc.connection, et_type_int, &sc.subscription)
 */
    __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_v_et_type_int); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 235, __pyx_L1_error)
    __pyx_t_14 = __Pyx_PyInt_As_dxf_long_t(__pyx_v_timestamp); if (unlikely((__pyx_t_14 == ((dxf_long_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 235, __pyx_L1_error)
    (void)(dxf_create_subscription_timed(__pyx_v_sc->connection, __pyx_t_12, __pyx_t_14, (&__pyx_v_sc->subscription)));

    /* "dxfeed/core/DXFeedPy.pyx":234
 *         raise Exception("Inapropriate date format, should be %Y-%m-%d %H:%M:%S")
 * 
 *     if event_type == 'Candle':             # <<<<<<<<<<<<<<
 *         clib.dxf_create_subscription_timed(sc.connection, et_type_int, timestamp, &sc.subscription)
 *     else:
 */
    goto __pyx_L27;
  }

  /* "dxfeed/core/DXFeedPy.pyx":237
 *         clib.dxf_create_subscription_timed(sc.connection, et_type_int, timestamp, &sc.subscription)
 *     else:
 *         clib.dxf_create_subscription(sc.connection, et_type_int, &sc.subscription)             # <<<<<<<<<<<<<<
 * 
 *     error_code = process_last_error(verbose=False)
 */
  /*else*/ {
    __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_v_et_type_int); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 237, __pyx_L1_error)
    (void)(dxf_create_subscription(__pyx_v_sc->connection, __pyx_t_12, (&__pyx_v_sc->subscription)));
  }
  __pyx_L27:;

  /* "dxfeed/core/DXFeedPy.pyx":239
 *         clib.dxf_create_subscription(sc.connection, et_type_int, &sc.subscription)
 * 
 *     error_code = process_last_error(verbose=False)             # <<<<<<<<<<<<<<
 *     if error_code:
 *         raise RuntimeError(f"In underlying C-API library error {error_code} occurred!")
 */
  __pyx_t_15.__pyx_n = 1;
  __pyx_t_15.verbose = Py_False;
  __pyx_t_12 = __pyx_f_6dxfeed_4core_8DXFeedPy_process_last_error(0, &__pyx_t_15); 
  __pyx_v_error_code = __pyx_t_12;

  /* "dxfeed/core/DXFeedPy.pyx":240
 * 
 *     error_code = process_last_error(verbose=False)
 *     if error_code:             # <<<<<<<<<<<<<<
 *         raise RuntimeError(f"In underlying C-API library error {error_code} occurred!")
 *     return sc
 */
  __pyx_t_1 = (__pyx_v_error_code != 0);
  if (unlikely(__pyx_t_1)) {

    /* "dxfeed/core/DXFeedPy.pyx":241
 *     error_code = process_last_error(verbose=False)
 *     if error_code:
 *         raise RuntimeError(f"In underlying C-API library error {error_code} occurred!")             # <<<<<<<<<<<<<<
 *     return sc
 * 
 */
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_16 = 0;
    __pyx_t_17 = 127;
    __Pyx_INCREF(__pyx_kp_u_In_underlying_C_API_library_erro);
    __pyx_t_16 += 34;
    __Pyx_GIVEREF(__pyx_kp_u_In_underlying_C_API_library_erro);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_kp_u_In_underlying_C_API_library_erro);
    __pyx_t_2 = __Pyx_PyUnicode_From_int(__pyx_v_error_code, 0, ' ', 'd'); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_16 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_occurred);
    __pyx_t_16 += 10;
    __Pyx_GIVEREF(__pyx_kp_u_occurred);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_kp_u_occurred);
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_6, 3, __pyx_t_16, __pyx_t_17); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 241, __pyx_L1_error)

    /* "dxfeed/core/DXFeedPy.pyx":240
 * 
 *     error_code = process_last_error(verbose=False)
 *     if error_code:             # <<<<<<<<<<<<<<
 *         raise RuntimeError(f"In underlying C-API library error {error_code} occurred!")
 *     return sc
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":242
 *     if error_code:
 *         raise RuntimeError(f"In underlying C-API library error {error_code} occurred!")
 *     return sc             # <<<<<<<<<<<<<<
 * 
 * def dxf_add_symbols(SubscriptionClass sc, symbols: Iterable[str]):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_sc));
  __pyx_r = ((PyObject *)__pyx_v_sc);
  goto __pyx_L0;

  /* "dxfeed/core/DXFeedPy.pyx":196
 *     return cc
 * 
 * def dxf_create_subscription(ConnectionClass cc, event_type: str, candle_time: Optional[str] = None,             # <<<<<<<<<<<<<<
 *                             data_len: int = 100000):
 *     """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_create_subscription", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_sc);
  __Pyx_XDECREF(__pyx_v_et_type_int);
  __Pyx_XDECREF(__pyx_v_timestamp);
  __Pyx_XDECREF(__pyx_v_candle_time);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":244
 *     return sc
 * 
 * def dxf_add_symbols(SubscriptionClass sc, symbols: Iterable[str]):             # <<<<<<<<<<<<<<
 *     """
 *     Adds symbols to subscription
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_9dxf_add_symbols(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_8dxf_add_symbols[] = "\n    Adds symbols to subscription\n\n    Parameters\n    ----------\n    sc: SubscriptionClass\n        SubscriptionClass with information about subscription\n    symbols: list\n        List of symbols to add\n    ";
static PyMethodDef __pyx_mdef_6dxfeed_4core_8DXFeedPy_9dxf_add_symbols = {"dxf_add_symbols", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_9dxf_add_symbols, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_8dxf_add_symbols};
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_9dxf_add_symbols(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc = 0;
  PyObject *__pyx_v_symbols = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dxf_add_symbols (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sc,&__pyx_n_s_symbols,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_symbols)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dxf_add_symbols", 1, 2, 2, 1); __PYX_ERR(0, 244, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dxf_add_symbols") < 0)) __PYX_ERR(0, 244, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_sc = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)values[0]);
    __pyx_v_symbols = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dxf_add_symbols", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 244, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_add_symbols", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sc), __pyx_ptype_6dxfeed_4core_8DXFeedPy_SubscriptionClass, 1, "sc", 0))) __PYX_ERR(0, 244, __pyx_L1_error)
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_8dxf_add_symbols(__pyx_self, __pyx_v_sc, __pyx_v_symbols);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_8dxf_add_symbols(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc, PyObject *__pyx_v_symbols) {
  CYTHON_UNUSED PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_sym = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("dxf_add_symbols", 0);

  /* "dxfeed/core/DXFeedPy.pyx":255
 *         List of symbols to add
 *     """
 *     if not sc.subscription:             # <<<<<<<<<<<<<<
 *         raise ValueError('Subscription is not valid')
 *     for idx, sym in enumerate(symbols):
 */
  __pyx_t_1 = ((!(__pyx_v_sc->subscription != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "dxfeed/core/DXFeedPy.pyx":256
 *     """
 *     if not sc.subscription:
 *         raise ValueError('Subscription is not valid')             # <<<<<<<<<<<<<<
 *     for idx, sym in enumerate(symbols):
 *         if not isinstance(sym, str):
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 256, __pyx_L1_error)

    /* "dxfeed/core/DXFeedPy.pyx":255
 *         List of symbols to add
 *     """
 *     if not sc.subscription:             # <<<<<<<<<<<<<<
 *         raise ValueError('Subscription is not valid')
 *     for idx, sym in enumerate(symbols):
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":257
 *     if not sc.subscription:
 *         raise ValueError('Subscription is not valid')
 *     for idx, sym in enumerate(symbols):             # <<<<<<<<<<<<<<
 *         if not isinstance(sym, str):
 *             warn(f'{sym} has type different from string')
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_2 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_symbols)) || PyTuple_CheckExact(__pyx_v_symbols)) {
    __pyx_t_3 = __pyx_v_symbols; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_symbols); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 257, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 257, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 257, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 257, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 257, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      }
    } else {
      __pyx_t_6 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_6)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 257, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_6);
    }
    __Pyx_XDECREF_SET(__pyx_v_sym, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_2);
    __pyx_t_6 = __Pyx_PyInt_AddObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":258
 *         raise ValueError('Subscription is not valid')
 *     for idx, sym in enumerate(symbols):
 *         if not isinstance(sym, str):             # <<<<<<<<<<<<<<
 *             warn(f'{sym} has type different from string')
 *             continue
 */
    __pyx_t_1 = PyUnicode_Check(__pyx_v_sym); 
    __pyx_t_7 = ((!(__pyx_t_1 != 0)) != 0);
    if (__pyx_t_7) {

      /* "dxfeed/core/DXFeedPy.pyx":259
 *     for idx, sym in enumerate(symbols):
 *         if not isinstance(sym, str):
 *             warn(f'{sym} has type different from string')             # <<<<<<<<<<<<<<
 *             continue
 *         if not clib.dxf_add_symbol(sc.subscription, dxf_const_string_t_from_unicode(sym)):
 */
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_warn); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 259, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyObject_FormatSimple(__pyx_v_sym, __pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 259, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyUnicode_Concat(__pyx_t_9, __pyx_kp_u_has_type_different_from_string); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 259, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_6 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_10);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 259, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "dxfeed/core/DXFeedPy.pyx":260
 *         if not isinstance(sym, str):
 *             warn(f'{sym} has type different from string')
 *             continue             # <<<<<<<<<<<<<<
 *         if not clib.dxf_add_symbol(sc.subscription, dxf_const_string_t_from_unicode(sym)):
 *             process_last_error()
 */
      goto __pyx_L4_continue;

      /* "dxfeed/core/DXFeedPy.pyx":258
 *         raise ValueError('Subscription is not valid')
 *     for idx, sym in enumerate(symbols):
 *         if not isinstance(sym, str):             # <<<<<<<<<<<<<<
 *             warn(f'{sym} has type different from string')
 *             continue
 */
    }

    /* "dxfeed/core/DXFeedPy.pyx":261
 *             warn(f'{sym} has type different from string')
 *             continue
 *         if not clib.dxf_add_symbol(sc.subscription, dxf_const_string_t_from_unicode(sym)):             # <<<<<<<<<<<<<<
 *             process_last_error()
 * 
 */
    __pyx_t_7 = ((!(dxf_add_symbol(__pyx_v_sc->subscription, __pyx_f_6dxfeed_4core_5utils_7helpers_dxf_const_string_t_from_unicode(__pyx_v_sym)) != 0)) != 0);
    if (__pyx_t_7) {

      /* "dxfeed/core/DXFeedPy.pyx":262
 *             continue
 *         if not clib.dxf_add_symbol(sc.subscription, dxf_const_string_t_from_unicode(sym)):
 *             process_last_error()             # <<<<<<<<<<<<<<
 * 
 * def dxf_attach_listener(SubscriptionClass sc):
 */
      (void)(__pyx_f_6dxfeed_4core_8DXFeedPy_process_last_error(0, NULL));

      /* "dxfeed/core/DXFeedPy.pyx":261
 *             warn(f'{sym} has type different from string')
 *             continue
 *         if not clib.dxf_add_symbol(sc.subscription, dxf_const_string_t_from_unicode(sym)):             # <<<<<<<<<<<<<<
 *             process_last_error()
 * 
 */
    }

    /* "dxfeed/core/DXFeedPy.pyx":257
 *     if not sc.subscription:
 *         raise ValueError('Subscription is not valid')
 *     for idx, sym in enumerate(symbols):             # <<<<<<<<<<<<<<
 *         if not isinstance(sym, str):
 *             warn(f'{sym} has type different from string')
 */
    __pyx_L4_continue:;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":244
 *     return sc
 * 
 * def dxf_add_symbols(SubscriptionClass sc, symbols: Iterable[str]):             # <<<<<<<<<<<<<<
 *     """
 *     Adds symbols to subscription
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_add_symbols", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_sym);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":264
 *             process_last_error()
 * 
 * def dxf_attach_listener(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Function attaches default listener according to subscription type
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_11dxf_attach_listener(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_10dxf_attach_listener[] = "\n    Function attaches default listener according to subscription type\n\n    Parameters\n    ----------\n    sc: SubscriptionClass\n        SubscriptionClass with information about subscription\n    ";
static PyMethodDef __pyx_mdef_6dxfeed_4core_8DXFeedPy_11dxf_attach_listener = {"dxf_attach_listener", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_11dxf_attach_listener, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_10dxf_attach_listener};
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_11dxf_attach_listener(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dxf_attach_listener (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sc,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dxf_attach_listener") < 0)) __PYX_ERR(0, 264, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_sc = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dxf_attach_listener", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 264, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_attach_listener", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sc), __pyx_ptype_6dxfeed_4core_8DXFeedPy_SubscriptionClass, 1, "sc", 0))) __PYX_ERR(0, 264, __pyx_L1_error)
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_10dxf_attach_listener(__pyx_self, __pyx_v_sc);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_10dxf_attach_listener(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  __Pyx_RefNannySetupContext("dxf_attach_listener", 0);

  /* "dxfeed/core/DXFeedPy.pyx":273
 *         SubscriptionClass with information about subscription
 *     """
 *     if not sc.subscription:             # <<<<<<<<<<<<<<
 *         raise ValueError('Subscription is not valid')
 *     if sc.event_type_str == 'Trade':
 */
  __pyx_t_1 = ((!(__pyx_v_sc->subscription != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "dxfeed/core/DXFeedPy.pyx":274
 *     """
 *     if not sc.subscription:
 *         raise ValueError('Subscription is not valid')             # <<<<<<<<<<<<<<
 *     if sc.event_type_str == 'Trade':
 *         sc.columns = lis.TRADE_COLUMNS
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 274, __pyx_L1_error)

    /* "dxfeed/core/DXFeedPy.pyx":273
 *         SubscriptionClass with information about subscription
 *     """
 *     if not sc.subscription:             # <<<<<<<<<<<<<<
 *         raise ValueError('Subscription is not valid')
 *     if sc.event_type_str == 'Trade':
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":275
 *     if not sc.subscription:
 *         raise ValueError('Subscription is not valid')
 *     if sc.event_type_str == 'Trade':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.TRADE_COLUMNS
 *         sc.listener = lis.trade_default_listener
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_sc->event_type_str, __pyx_n_u_Trade, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 275, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":276
 *         raise ValueError('Subscription is not valid')
 *     if sc.event_type_str == 'Trade':
 *         sc.columns = lis.TRADE_COLUMNS             # <<<<<<<<<<<<<<
 *         sc.listener = lis.trade_default_listener
 *     elif sc.event_type_str == 'Quote':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_lis); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_TRADE_COLUMNS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_sc->columns);
    __Pyx_DECREF(__pyx_v_sc->columns);
    __pyx_v_sc->columns = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":277
 *     if sc.event_type_str == 'Trade':
 *         sc.columns = lis.TRADE_COLUMNS
 *         sc.listener = lis.trade_default_listener             # <<<<<<<<<<<<<<
 *     elif sc.event_type_str == 'Quote':
 *         sc.columns = lis.QUOTE_COLUMNS
 */
    __pyx_v_sc->listener = __pyx_f_6dxfeed_4core_9listeners_8listener_trade_default_listener;

    /* "dxfeed/core/DXFeedPy.pyx":275
 *     if not sc.subscription:
 *         raise ValueError('Subscription is not valid')
 *     if sc.event_type_str == 'Trade':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.TRADE_COLUMNS
 *         sc.listener = lis.trade_default_listener
 */
    goto __pyx_L4;
  }

  /* "dxfeed/core/DXFeedPy.pyx":278
 *         sc.columns = lis.TRADE_COLUMNS
 *         sc.listener = lis.trade_default_listener
 *     elif sc.event_type_str == 'Quote':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.QUOTE_COLUMNS
 *         sc.listener = lis.quote_default_listener
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_sc->event_type_str, __pyx_n_u_Quote, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 278, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":279
 *         sc.listener = lis.trade_default_listener
 *     elif sc.event_type_str == 'Quote':
 *         sc.columns = lis.QUOTE_COLUMNS             # <<<<<<<<<<<<<<
 *         sc.listener = lis.quote_default_listener
 *     elif sc.event_type_str == 'Summary':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_lis); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_QUOTE_COLUMNS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_sc->columns);
    __Pyx_DECREF(__pyx_v_sc->columns);
    __pyx_v_sc->columns = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":280
 *     elif sc.event_type_str == 'Quote':
 *         sc.columns = lis.QUOTE_COLUMNS
 *         sc.listener = lis.quote_default_listener             # <<<<<<<<<<<<<<
 *     elif sc.event_type_str == 'Summary':
 *         sc.columns = lis.SUMMARY_COLUMNS
 */
    __pyx_v_sc->listener = __pyx_f_6dxfeed_4core_9listeners_8listener_quote_default_listener;

    /* "dxfeed/core/DXFeedPy.pyx":278
 *         sc.columns = lis.TRADE_COLUMNS
 *         sc.listener = lis.trade_default_listener
 *     elif sc.event_type_str == 'Quote':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.QUOTE_COLUMNS
 *         sc.listener = lis.quote_default_listener
 */
    goto __pyx_L4;
  }

  /* "dxfeed/core/DXFeedPy.pyx":281
 *         sc.columns = lis.QUOTE_COLUMNS
 *         sc.listener = lis.quote_default_listener
 *     elif sc.event_type_str == 'Summary':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.SUMMARY_COLUMNS
 *         sc.listener = lis.summary_default_listener
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_sc->event_type_str, __pyx_n_u_Summary, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 281, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":282
 *         sc.listener = lis.quote_default_listener
 *     elif sc.event_type_str == 'Summary':
 *         sc.columns = lis.SUMMARY_COLUMNS             # <<<<<<<<<<<<<<
 *         sc.listener = lis.summary_default_listener
 *     elif sc.event_type_str == 'Profile':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_lis); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_SUMMARY_COLUMNS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 282, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_sc->columns);
    __Pyx_DECREF(__pyx_v_sc->columns);
    __pyx_v_sc->columns = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":283
 *     elif sc.event_type_str == 'Summary':
 *         sc.columns = lis.SUMMARY_COLUMNS
 *         sc.listener = lis.summary_default_listener             # <<<<<<<<<<<<<<
 *     elif sc.event_type_str == 'Profile':
 *         sc.columns = lis.PROFILE_COLUMNS
 */
    __pyx_v_sc->listener = __pyx_f_6dxfeed_4core_9listeners_8listener_summary_default_listener;

    /* "dxfeed/core/DXFeedPy.pyx":281
 *         sc.columns = lis.QUOTE_COLUMNS
 *         sc.listener = lis.quote_default_listener
 *     elif sc.event_type_str == 'Summary':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.SUMMARY_COLUMNS
 *         sc.listener = lis.summary_default_listener
 */
    goto __pyx_L4;
  }

  /* "dxfeed/core/DXFeedPy.pyx":284
 *         sc.columns = lis.SUMMARY_COLUMNS
 *         sc.listener = lis.summary_default_listener
 *     elif sc.event_type_str == 'Profile':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.PROFILE_COLUMNS
 *         sc.listener = lis.profile_default_listener
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_sc->event_type_str, __pyx_n_u_Profile, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 284, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":285
 *         sc.listener = lis.summary_default_listener
 *     elif sc.event_type_str == 'Profile':
 *         sc.columns = lis.PROFILE_COLUMNS             # <<<<<<<<<<<<<<
 *         sc.listener = lis.profile_default_listener
 *     elif sc.event_type_str == 'TimeAndSale':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_lis); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_PROFILE_COLUMNS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_sc->columns);
    __Pyx_DECREF(__pyx_v_sc->columns);
    __pyx_v_sc->columns = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":286
 *     elif sc.event_type_str == 'Profile':
 *         sc.columns = lis.PROFILE_COLUMNS
 *         sc.listener = lis.profile_default_listener             # <<<<<<<<<<<<<<
 *     elif sc.event_type_str == 'TimeAndSale':
 *         sc.columns = lis.TIME_AND_SALE_COLUMNS
 */
    __pyx_v_sc->listener = __pyx_f_6dxfeed_4core_9listeners_8listener_profile_default_listener;

    /* "dxfeed/core/DXFeedPy.pyx":284
 *         sc.columns = lis.SUMMARY_COLUMNS
 *         sc.listener = lis.summary_default_listener
 *     elif sc.event_type_str == 'Profile':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.PROFILE_COLUMNS
 *         sc.listener = lis.profile_default_listener
 */
    goto __pyx_L4;
  }

  /* "dxfeed/core/DXFeedPy.pyx":287
 *         sc.columns = lis.PROFILE_COLUMNS
 *         sc.listener = lis.profile_default_listener
 *     elif sc.event_type_str == 'TimeAndSale':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.TIME_AND_SALE_COLUMNS
 *         sc.listener = lis.time_and_sale_default_listener
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_sc->event_type_str, __pyx_n_u_TimeAndSale, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 287, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":288
 *         sc.listener = lis.profile_default_listener
 *     elif sc.event_type_str == 'TimeAndSale':
 *         sc.columns = lis.TIME_AND_SALE_COLUMNS             # <<<<<<<<<<<<<<
 *         sc.listener = lis.time_and_sale_default_listener
 *     elif sc.event_type_str == 'Candle':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_lis); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_TIME_AND_SALE_COLUMNS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 288, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_sc->columns);
    __Pyx_DECREF(__pyx_v_sc->columns);
    __pyx_v_sc->columns = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":289
 *     elif sc.event_type_str == 'TimeAndSale':
 *         sc.columns = lis.TIME_AND_SALE_COLUMNS
 *         sc.listener = lis.time_and_sale_default_listener             # <<<<<<<<<<<<<<
 *     elif sc.event_type_str == 'Candle':
 *         sc.columns = lis.CANDLE_COLUMNS
 */
    __pyx_v_sc->listener = __pyx_f_6dxfeed_4core_9listeners_8listener_time_and_sale_default_listener;

    /* "dxfeed/core/DXFeedPy.pyx":287
 *         sc.columns = lis.PROFILE_COLUMNS
 *         sc.listener = lis.profile_default_listener
 *     elif sc.event_type_str == 'TimeAndSale':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.TIME_AND_SALE_COLUMNS
 *         sc.listener = lis.time_and_sale_default_listener
 */
    goto __pyx_L4;
  }

  /* "dxfeed/core/DXFeedPy.pyx":290
 *         sc.columns = lis.TIME_AND_SALE_COLUMNS
 *         sc.listener = lis.time_and_sale_default_listener
 *     elif sc.event_type_str == 'Candle':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.CANDLE_COLUMNS
 *         sc.listener = lis.candle_default_listener
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_sc->event_type_str, __pyx_n_u_Candle, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 290, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":291
 *         sc.listener = lis.time_and_sale_default_listener
 *     elif sc.event_type_str == 'Candle':
 *         sc.columns = lis.CANDLE_COLUMNS             # <<<<<<<<<<<<<<
 *         sc.listener = lis.candle_default_listener
 *     elif sc.event_type_str == 'Order':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_lis); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 291, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CANDLE_COLUMNS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 291, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 291, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_sc->columns);
    __Pyx_DECREF(__pyx_v_sc->columns);
    __pyx_v_sc->columns = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":292
 *     elif sc.event_type_str == 'Candle':
 *         sc.columns = lis.CANDLE_COLUMNS
 *         sc.listener = lis.candle_default_listener             # <<<<<<<<<<<<<<
 *     elif sc.event_type_str == 'Order':
 *         sc.columns = lis.ORDER_COLUMNS
 */
    __pyx_v_sc->listener = __pyx_f_6dxfeed_4core_9listeners_8listener_candle_default_listener;

    /* "dxfeed/core/DXFeedPy.pyx":290
 *         sc.columns = lis.TIME_AND_SALE_COLUMNS
 *         sc.listener = lis.time_and_sale_default_listener
 *     elif sc.event_type_str == 'Candle':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.CANDLE_COLUMNS
 *         sc.listener = lis.candle_default_listener
 */
    goto __pyx_L4;
  }

  /* "dxfeed/core/DXFeedPy.pyx":293
 *         sc.columns = lis.CANDLE_COLUMNS
 *         sc.listener = lis.candle_default_listener
 *     elif sc.event_type_str == 'Order':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.ORDER_COLUMNS
 *         sc.listener = lis.order_default_listener
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_sc->event_type_str, __pyx_n_u_Order, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 293, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":294
 *         sc.listener = lis.candle_default_listener
 *     elif sc.event_type_str == 'Order':
 *         sc.columns = lis.ORDER_COLUMNS             # <<<<<<<<<<<<<<
 *         sc.listener = lis.order_default_listener
 *     elif sc.event_type_str == 'TradeETH':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_lis); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ORDER_COLUMNS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_sc->columns);
    __Pyx_DECREF(__pyx_v_sc->columns);
    __pyx_v_sc->columns = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":295
 *     elif sc.event_type_str == 'Order':
 *         sc.columns = lis.ORDER_COLUMNS
 *         sc.listener = lis.order_default_listener             # <<<<<<<<<<<<<<
 *     elif sc.event_type_str == 'TradeETH':
 *         sc.columns = lis.TRADE_COLUMNS
 */
    __pyx_v_sc->listener = __pyx_f_6dxfeed_4core_9listeners_8listener_order_default_listener;

    /* "dxfeed/core/DXFeedPy.pyx":293
 *         sc.columns = lis.CANDLE_COLUMNS
 *         sc.listener = lis.candle_default_listener
 *     elif sc.event_type_str == 'Order':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.ORDER_COLUMNS
 *         sc.listener = lis.order_default_listener
 */
    goto __pyx_L4;
  }

  /* "dxfeed/core/DXFeedPy.pyx":296
 *         sc.columns = lis.ORDER_COLUMNS
 *         sc.listener = lis.order_default_listener
 *     elif sc.event_type_str == 'TradeETH':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.TRADE_COLUMNS
 *         sc.listener = lis.trade_default_listener
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_sc->event_type_str, __pyx_n_u_TradeETH, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 296, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":297
 *         sc.listener = lis.order_default_listener
 *     elif sc.event_type_str == 'TradeETH':
 *         sc.columns = lis.TRADE_COLUMNS             # <<<<<<<<<<<<<<
 *         sc.listener = lis.trade_default_listener
 *     elif sc.event_type_str == 'SpreadOrder':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_lis); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 297, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_TRADE_COLUMNS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 297, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 297, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_sc->columns);
    __Pyx_DECREF(__pyx_v_sc->columns);
    __pyx_v_sc->columns = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":298
 *     elif sc.event_type_str == 'TradeETH':
 *         sc.columns = lis.TRADE_COLUMNS
 *         sc.listener = lis.trade_default_listener             # <<<<<<<<<<<<<<
 *     elif sc.event_type_str == 'SpreadOrder':
 *         sc.columns = lis.ORDER_COLUMNS
 */
    __pyx_v_sc->listener = __pyx_f_6dxfeed_4core_9listeners_8listener_trade_default_listener;

    /* "dxfeed/core/DXFeedPy.pyx":296
 *         sc.columns = lis.ORDER_COLUMNS
 *         sc.listener = lis.order_default_listener
 *     elif sc.event_type_str == 'TradeETH':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.TRADE_COLUMNS
 *         sc.listener = lis.trade_default_listener
 */
    goto __pyx_L4;
  }

  /* "dxfeed/core/DXFeedPy.pyx":299
 *         sc.columns = lis.TRADE_COLUMNS
 *         sc.listener = lis.trade_default_listener
 *     elif sc.event_type_str == 'SpreadOrder':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.ORDER_COLUMNS
 *         sc.listener = lis.order_default_listener
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_sc->event_type_str, __pyx_n_u_SpreadOrder, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 299, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":300
 *         sc.listener = lis.trade_default_listener
 *     elif sc.event_type_str == 'SpreadOrder':
 *         sc.columns = lis.ORDER_COLUMNS             # <<<<<<<<<<<<<<
 *         sc.listener = lis.order_default_listener
 *     elif sc.event_type_str == 'Greeks':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_lis); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ORDER_COLUMNS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 300, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_sc->columns);
    __Pyx_DECREF(__pyx_v_sc->columns);
    __pyx_v_sc->columns = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":301
 *     elif sc.event_type_str == 'SpreadOrder':
 *         sc.columns = lis.ORDER_COLUMNS
 *         sc.listener = lis.order_default_listener             # <<<<<<<<<<<<<<
 *     elif sc.event_type_str == 'Greeks':
 *         sc.columns = lis.GREEKS_COLUMNS
 */
    __pyx_v_sc->listener = __pyx_f_6dxfeed_4core_9listeners_8listener_order_default_listener;

    /* "dxfeed/core/DXFeedPy.pyx":299
 *         sc.columns = lis.TRADE_COLUMNS
 *         sc.listener = lis.trade_default_listener
 *     elif sc.event_type_str == 'SpreadOrder':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.ORDER_COLUMNS
 *         sc.listener = lis.order_default_listener
 */
    goto __pyx_L4;
  }

  /* "dxfeed/core/DXFeedPy.pyx":302
 *         sc.columns = lis.ORDER_COLUMNS
 *         sc.listener = lis.order_default_listener
 *     elif sc.event_type_str == 'Greeks':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.GREEKS_COLUMNS
 *         sc.listener = lis.greeks_default_listener
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_sc->event_type_str, __pyx_n_u_Greeks, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 302, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":303
 *         sc.listener = lis.order_default_listener
 *     elif sc.event_type_str == 'Greeks':
 *         sc.columns = lis.GREEKS_COLUMNS             # <<<<<<<<<<<<<<
 *         sc.listener = lis.greeks_default_listener
 *     elif sc.event_type_str == 'TheoPrice':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_lis); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_GREEKS_COLUMNS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_sc->columns);
    __Pyx_DECREF(__pyx_v_sc->columns);
    __pyx_v_sc->columns = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":304
 *     elif sc.event_type_str == 'Greeks':
 *         sc.columns = lis.GREEKS_COLUMNS
 *         sc.listener = lis.greeks_default_listener             # <<<<<<<<<<<<<<
 *     elif sc.event_type_str == 'TheoPrice':
 *         sc.columns = lis.THEO_PRICE_COLUMNS
 */
    __pyx_v_sc->listener = __pyx_f_6dxfeed_4core_9listeners_8listener_greeks_default_listener;

    /* "dxfeed/core/DXFeedPy.pyx":302
 *         sc.columns = lis.ORDER_COLUMNS
 *         sc.listener = lis.order_default_listener
 *     elif sc.event_type_str == 'Greeks':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.GREEKS_COLUMNS
 *         sc.listener = lis.greeks_default_listener
 */
    goto __pyx_L4;
  }

  /* "dxfeed/core/DXFeedPy.pyx":305
 *         sc.columns = lis.GREEKS_COLUMNS
 *         sc.listener = lis.greeks_default_listener
 *     elif sc.event_type_str == 'TheoPrice':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.THEO_PRICE_COLUMNS
 *         sc.listener = lis.theo_price_default_listener
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_sc->event_type_str, __pyx_n_u_TheoPrice, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 305, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":306
 *         sc.listener = lis.greeks_default_listener
 *     elif sc.event_type_str == 'TheoPrice':
 *         sc.columns = lis.THEO_PRICE_COLUMNS             # <<<<<<<<<<<<<<
 *         sc.listener = lis.theo_price_default_listener
 *     elif sc.event_type_str == 'Underlying':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_lis); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_THEO_PRICE_COLUMNS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_sc->columns);
    __Pyx_DECREF(__pyx_v_sc->columns);
    __pyx_v_sc->columns = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":307
 *     elif sc.event_type_str == 'TheoPrice':
 *         sc.columns = lis.THEO_PRICE_COLUMNS
 *         sc.listener = lis.theo_price_default_listener             # <<<<<<<<<<<<<<
 *     elif sc.event_type_str == 'Underlying':
 *         sc.columns = lis.UNDERLYING_COLUMNS
 */
    __pyx_v_sc->listener = __pyx_f_6dxfeed_4core_9listeners_8listener_theo_price_default_listener;

    /* "dxfeed/core/DXFeedPy.pyx":305
 *         sc.columns = lis.GREEKS_COLUMNS
 *         sc.listener = lis.greeks_default_listener
 *     elif sc.event_type_str == 'TheoPrice':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.THEO_PRICE_COLUMNS
 *         sc.listener = lis.theo_price_default_listener
 */
    goto __pyx_L4;
  }

  /* "dxfeed/core/DXFeedPy.pyx":308
 *         sc.columns = lis.THEO_PRICE_COLUMNS
 *         sc.listener = lis.theo_price_default_listener
 *     elif sc.event_type_str == 'Underlying':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.UNDERLYING_COLUMNS
 *         sc.listener = lis.underlying_default_listener
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_sc->event_type_str, __pyx_n_u_Underlying, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 308, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":309
 *         sc.listener = lis.theo_price_default_listener
 *     elif sc.event_type_str == 'Underlying':
 *         sc.columns = lis.UNDERLYING_COLUMNS             # <<<<<<<<<<<<<<
 *         sc.listener = lis.underlying_default_listener
 *     elif sc.event_type_str == 'Series':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_lis); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_UNDERLYING_COLUMNS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 309, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_sc->columns);
    __Pyx_DECREF(__pyx_v_sc->columns);
    __pyx_v_sc->columns = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":310
 *     elif sc.event_type_str == 'Underlying':
 *         sc.columns = lis.UNDERLYING_COLUMNS
 *         sc.listener = lis.underlying_default_listener             # <<<<<<<<<<<<<<
 *     elif sc.event_type_str == 'Series':
 *         sc.columns = lis.SERIES_COLUMNS
 */
    __pyx_v_sc->listener = __pyx_f_6dxfeed_4core_9listeners_8listener_underlying_default_listener;

    /* "dxfeed/core/DXFeedPy.pyx":308
 *         sc.columns = lis.THEO_PRICE_COLUMNS
 *         sc.listener = lis.theo_price_default_listener
 *     elif sc.event_type_str == 'Underlying':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.UNDERLYING_COLUMNS
 *         sc.listener = lis.underlying_default_listener
 */
    goto __pyx_L4;
  }

  /* "dxfeed/core/DXFeedPy.pyx":311
 *         sc.columns = lis.UNDERLYING_COLUMNS
 *         sc.listener = lis.underlying_default_listener
 *     elif sc.event_type_str == 'Series':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.SERIES_COLUMNS
 *         sc.listener = lis.series_default_listener
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_sc->event_type_str, __pyx_n_u_Series, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 311, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":312
 *         sc.listener = lis.underlying_default_listener
 *     elif sc.event_type_str == 'Series':
 *         sc.columns = lis.SERIES_COLUMNS             # <<<<<<<<<<<<<<
 *         sc.listener = lis.series_default_listener
 *     elif sc.event_type_str == 'Configuration':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_lis); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_SERIES_COLUMNS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 312, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_sc->columns);
    __Pyx_DECREF(__pyx_v_sc->columns);
    __pyx_v_sc->columns = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":313
 *     elif sc.event_type_str == 'Series':
 *         sc.columns = lis.SERIES_COLUMNS
 *         sc.listener = lis.series_default_listener             # <<<<<<<<<<<<<<
 *     elif sc.event_type_str == 'Configuration':
 *         sc.columns = lis.CONFIGURATION_COLUMNS
 */
    __pyx_v_sc->listener = __pyx_f_6dxfeed_4core_9listeners_8listener_series_default_listener;

    /* "dxfeed/core/DXFeedPy.pyx":311
 *         sc.columns = lis.UNDERLYING_COLUMNS
 *         sc.listener = lis.underlying_default_listener
 *     elif sc.event_type_str == 'Series':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.SERIES_COLUMNS
 *         sc.listener = lis.series_default_listener
 */
    goto __pyx_L4;
  }

  /* "dxfeed/core/DXFeedPy.pyx":314
 *         sc.columns = lis.SERIES_COLUMNS
 *         sc.listener = lis.series_default_listener
 *     elif sc.event_type_str == 'Configuration':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.CONFIGURATION_COLUMNS
 *         sc.listener = lis.configuration_default_listener
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_sc->event_type_str, __pyx_n_u_Configuration, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 314, __pyx_L1_error)
  if (likely(__pyx_t_1)) {

    /* "dxfeed/core/DXFeedPy.pyx":315
 *         sc.listener = lis.series_default_listener
 *     elif sc.event_type_str == 'Configuration':
 *         sc.columns = lis.CONFIGURATION_COLUMNS             # <<<<<<<<<<<<<<
 *         sc.listener = lis.configuration_default_listener
 *     else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_lis); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_CONFIGURATION_COLUMNS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 315, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_sc->columns);
    __Pyx_DECREF(__pyx_v_sc->columns);
    __pyx_v_sc->columns = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":316
 *     elif sc.event_type_str == 'Configuration':
 *         sc.columns = lis.CONFIGURATION_COLUMNS
 *         sc.listener = lis.configuration_default_listener             # <<<<<<<<<<<<<<
 *     else:
 *         raise Exception(f'No default listener for {sc.event_type_str} event type')
 */
    __pyx_v_sc->listener = __pyx_f_6dxfeed_4core_9listeners_8listener_configuration_default_listener;

    /* "dxfeed/core/DXFeedPy.pyx":314
 *         sc.columns = lis.SERIES_COLUMNS
 *         sc.listener = lis.series_default_listener
 *     elif sc.event_type_str == 'Configuration':             # <<<<<<<<<<<<<<
 *         sc.columns = lis.CONFIGURATION_COLUMNS
 *         sc.listener = lis.configuration_default_listener
 */
    goto __pyx_L4;
  }

  /* "dxfeed/core/DXFeedPy.pyx":318
 *         sc.listener = lis.configuration_default_listener
 *     else:
 *         raise Exception(f'No default listener for {sc.event_type_str} event type')             # <<<<<<<<<<<<<<
 * 
 *     if not clib.dxf_attach_event_listener(sc.subscription, sc.listener, sc.u_data):
 */
  /*else*/ {
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_No_default_listener_for);
    __pyx_t_4 += 24;
    __Pyx_GIVEREF(__pyx_kp_u_No_default_listener_for);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_No_default_listener_for);
    __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_sc->event_type_str, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_kp_u_event_type_2);
    __pyx_t_4 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_event_type_2);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_event_type_2);
    __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_2, 3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 318, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 318, __pyx_L1_error)
  }
  __pyx_L4:;

  /* "dxfeed/core/DXFeedPy.pyx":320
 *         raise Exception(f'No default listener for {sc.event_type_str} event type')
 * 
 *     if not clib.dxf_attach_event_listener(sc.subscription, sc.listener, sc.u_data):             # <<<<<<<<<<<<<<
 *         process_last_error()
 * 
 */
  __pyx_t_1 = ((!(dxf_attach_event_listener(__pyx_v_sc->subscription, __pyx_v_sc->listener, __pyx_v_sc->u_data) != 0)) != 0);
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":321
 * 
 *     if not clib.dxf_attach_event_listener(sc.subscription, sc.listener, sc.u_data):
 *         process_last_error()             # <<<<<<<<<<<<<<
 * 
 * def dxf_attach_custom_listener(SubscriptionClass sc, lis.FuncWrapper fw, columns: Iterable[str], data: Iterable = None):
 */
    (void)(__pyx_f_6dxfeed_4core_8DXFeedPy_process_last_error(0, NULL));

    /* "dxfeed/core/DXFeedPy.pyx":320
 *         raise Exception(f'No default listener for {sc.event_type_str} event type')
 * 
 *     if not clib.dxf_attach_event_listener(sc.subscription, sc.listener, sc.u_data):             # <<<<<<<<<<<<<<
 *         process_last_error()
 * 
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":264
 *             process_last_error()
 * 
 * def dxf_attach_listener(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Function attaches default listener according to subscription type
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_attach_listener", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":323
 *         process_last_error()
 * 
 * def dxf_attach_custom_listener(SubscriptionClass sc, lis.FuncWrapper fw, columns: Iterable[str], data: Iterable = None):             # <<<<<<<<<<<<<<
 *     """
 *     Attaches custom listener
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_13dxf_attach_custom_listener(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_12dxf_attach_custom_listener[] = "\n    Attaches custom listener\n\n    Parameters\n    ----------\n    sc: SubscriptionClass\n        SubscriptionClass with information about subscription\n    fw: FuncWrapper\n        c function wrapped in FuncWrapper class with Cython\n    columns: list\n        Columns for internal data of SubscriptionClass\n    data: dict\n        Dict with new internal data structure of  SubscriptionClass\n    ";
static PyMethodDef __pyx_mdef_6dxfeed_4core_8DXFeedPy_13dxf_attach_custom_listener = {"dxf_attach_custom_listener", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_13dxf_attach_custom_listener, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_12dxf_attach_custom_listener};
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_13dxf_attach_custom_listener(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc = 0;
  struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *__pyx_v_fw = 0;
  PyObject *__pyx_v_columns = 0;
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dxf_attach_custom_listener (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sc,&__pyx_n_s_fw,&__pyx_n_s_columns,&__pyx_n_s_data,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fw)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dxf_attach_custom_listener", 0, 3, 4, 1); __PYX_ERR(0, 323, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_columns)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dxf_attach_custom_listener", 0, 3, 4, 2); __PYX_ERR(0, 323, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dxf_attach_custom_listener") < 0)) __PYX_ERR(0, 323, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_sc = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)values[0]);
    __pyx_v_fw = ((struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *)values[1]);
    __pyx_v_columns = values[2];
    __pyx_v_data = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dxf_attach_custom_listener", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 323, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_attach_custom_listener", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sc), __pyx_ptype_6dxfeed_4core_8DXFeedPy_SubscriptionClass, 1, "sc", 0))) __PYX_ERR(0, 323, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_fw), __pyx_ptype_6dxfeed_4core_9listeners_8listener_FuncWrapper, 1, "fw", 0))) __PYX_ERR(0, 323, __pyx_L1_error)
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_12dxf_attach_custom_listener(__pyx_self, __pyx_v_sc, __pyx_v_fw, __pyx_v_columns, __pyx_v_data);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_12dxf_attach_custom_listener(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc, struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *__pyx_v_fw, PyObject *__pyx_v_columns, PyObject *__pyx_v_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  dxf_event_listener_t __pyx_t_3;
  __Pyx_RefNannySetupContext("dxf_attach_custom_listener", 0);

  /* "dxfeed/core/DXFeedPy.pyx":338
 *         Dict with new internal data structure of  SubscriptionClass
 *     """
 *     if not sc.subscription:             # <<<<<<<<<<<<<<
 *         raise ValueError('Subscription is not valid')
 *     if data:
 */
  __pyx_t_1 = ((!(__pyx_v_sc->subscription != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "dxfeed/core/DXFeedPy.pyx":339
 *     """
 *     if not sc.subscription:
 *         raise ValueError('Subscription is not valid')             # <<<<<<<<<<<<<<
 *     if data:
 *         sc.data = data
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 339, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 339, __pyx_L1_error)

    /* "dxfeed/core/DXFeedPy.pyx":338
 *         Dict with new internal data structure of  SubscriptionClass
 *     """
 *     if not sc.subscription:             # <<<<<<<<<<<<<<
 *         raise ValueError('Subscription is not valid')
 *     if data:
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":340
 *     if not sc.subscription:
 *         raise ValueError('Subscription is not valid')
 *     if data:             # <<<<<<<<<<<<<<
 *         sc.data = data
 *     sc.columns = columns
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_data); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 340, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":341
 *         raise ValueError('Subscription is not valid')
 *     if data:
 *         sc.data = data             # <<<<<<<<<<<<<<
 *     sc.columns = columns
 *     sc.listener = fw.func
 */
    __Pyx_INCREF(__pyx_v_data);
    __Pyx_GIVEREF(__pyx_v_data);
    __Pyx_GOTREF(__pyx_v_sc->data);
    __Pyx_DECREF(__pyx_v_sc->data);
    __pyx_v_sc->data = __pyx_v_data;

    /* "dxfeed/core/DXFeedPy.pyx":340
 *     if not sc.subscription:
 *         raise ValueError('Subscription is not valid')
 *     if data:             # <<<<<<<<<<<<<<
 *         sc.data = data
 *     sc.columns = columns
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":342
 *     if data:
 *         sc.data = data
 *     sc.columns = columns             # <<<<<<<<<<<<<<
 *     sc.listener = fw.func
 *     if not clib.dxf_attach_event_listener(sc.subscription, sc.listener, sc.u_data):
 */
  if (!(likely(PyList_CheckExact(__pyx_v_columns))||((__pyx_v_columns) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_columns)->tp_name), 0))) __PYX_ERR(0, 342, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_columns;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_sc->columns);
  __Pyx_DECREF(__pyx_v_sc->columns);
  __pyx_v_sc->columns = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":343
 *         sc.data = data
 *     sc.columns = columns
 *     sc.listener = fw.func             # <<<<<<<<<<<<<<
 *     if not clib.dxf_attach_event_listener(sc.subscription, sc.listener, sc.u_data):
 *         process_last_error()
 */
  __pyx_t_3 = __pyx_v_fw->func;
  __pyx_v_sc->listener = __pyx_t_3;

  /* "dxfeed/core/DXFeedPy.pyx":344
 *     sc.columns = columns
 *     sc.listener = fw.func
 *     if not clib.dxf_attach_event_listener(sc.subscription, sc.listener, sc.u_data):             # <<<<<<<<<<<<<<
 *         process_last_error()
 * 
 */
  __pyx_t_1 = ((!(dxf_attach_event_listener(__pyx_v_sc->subscription, __pyx_v_sc->listener, __pyx_v_sc->u_data) != 0)) != 0);
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":345
 *     sc.listener = fw.func
 *     if not clib.dxf_attach_event_listener(sc.subscription, sc.listener, sc.u_data):
 *         process_last_error()             # <<<<<<<<<<<<<<
 * 
 * def dxf_detach_listener(SubscriptionClass sc):
 */
    (void)(__pyx_f_6dxfeed_4core_8DXFeedPy_process_last_error(0, NULL));

    /* "dxfeed/core/DXFeedPy.pyx":344
 *     sc.columns = columns
 *     sc.listener = fw.func
 *     if not clib.dxf_attach_event_listener(sc.subscription, sc.listener, sc.u_data):             # <<<<<<<<<<<<<<
 *         process_last_error()
 * 
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":323
 *         process_last_error()
 * 
 * def dxf_attach_custom_listener(SubscriptionClass sc, lis.FuncWrapper fw, columns: Iterable[str], data: Iterable = None):             # <<<<<<<<<<<<<<
 *     """
 *     Attaches custom listener
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_attach_custom_listener", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":347
 *         process_last_error()
 * 
 * def dxf_detach_listener(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Detaches any listener
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_15dxf_detach_listener(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_14dxf_detach_listener[] = "\n    Detaches any listener\n\n    Parameters\n    ----------\n    sc: SubscriptionClass\n        SubscriptionClass with information about subscription\n    ";
static PyMethodDef __pyx_mdef_6dxfeed_4core_8DXFeedPy_15dxf_detach_listener = {"dxf_detach_listener", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_15dxf_detach_listener, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_14dxf_detach_listener};
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_15dxf_detach_listener(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dxf_detach_listener (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sc,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dxf_detach_listener") < 0)) __PYX_ERR(0, 347, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_sc = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dxf_detach_listener", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 347, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_detach_listener", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sc), __pyx_ptype_6dxfeed_4core_8DXFeedPy_SubscriptionClass, 1, "sc", 0))) __PYX_ERR(0, 347, __pyx_L1_error)
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_14dxf_detach_listener(__pyx_self, __pyx_v_sc);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_14dxf_detach_listener(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("dxf_detach_listener", 0);

  /* "dxfeed/core/DXFeedPy.pyx":356
 *         SubscriptionClass with information about subscription
 *     """
 *     if not sc.subscription:             # <<<<<<<<<<<<<<
 *         raise ValueError('Subscription is not valid')
 *     if not clib.dxf_detach_event_listener(sc.subscription, sc.listener):
 */
  __pyx_t_1 = ((!(__pyx_v_sc->subscription != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "dxfeed/core/DXFeedPy.pyx":357
 *     """
 *     if not sc.subscription:
 *         raise ValueError('Subscription is not valid')             # <<<<<<<<<<<<<<
 *     if not clib.dxf_detach_event_listener(sc.subscription, sc.listener):
 *         process_last_error()
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 357, __pyx_L1_error)

    /* "dxfeed/core/DXFeedPy.pyx":356
 *         SubscriptionClass with information about subscription
 *     """
 *     if not sc.subscription:             # <<<<<<<<<<<<<<
 *         raise ValueError('Subscription is not valid')
 *     if not clib.dxf_detach_event_listener(sc.subscription, sc.listener):
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":358
 *     if not sc.subscription:
 *         raise ValueError('Subscription is not valid')
 *     if not clib.dxf_detach_event_listener(sc.subscription, sc.listener):             # <<<<<<<<<<<<<<
 *         process_last_error()
 * 
 */
  __pyx_t_1 = ((!(dxf_detach_event_listener(__pyx_v_sc->subscription, __pyx_v_sc->listener) != 0)) != 0);
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":359
 *         raise ValueError('Subscription is not valid')
 *     if not clib.dxf_detach_event_listener(sc.subscription, sc.listener):
 *         process_last_error()             # <<<<<<<<<<<<<<
 * 
 * def dxf_close_connection(ConnectionClass cc):
 */
    (void)(__pyx_f_6dxfeed_4core_8DXFeedPy_process_last_error(0, NULL));

    /* "dxfeed/core/DXFeedPy.pyx":358
 *     if not sc.subscription:
 *         raise ValueError('Subscription is not valid')
 *     if not clib.dxf_detach_event_listener(sc.subscription, sc.listener):             # <<<<<<<<<<<<<<
 *         process_last_error()
 * 
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":347
 *         process_last_error()
 * 
 * def dxf_detach_listener(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Detaches any listener
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_detach_listener", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":361
 *         process_last_error()
 * 
 * def dxf_close_connection(ConnectionClass cc):             # <<<<<<<<<<<<<<
 *     """
 *     Closes connection
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_17dxf_close_connection(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_16dxf_close_connection[] = "\n    Closes connection\n\n\n    Parameters\n    ----------\n    cc: ConnectionClass\n        Variable with connection information\n    ";
static PyMethodDef __pyx_mdef_6dxfeed_4core_8DXFeedPy_17dxf_close_connection = {"dxf_close_connection", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_17dxf_close_connection, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_16dxf_close_connection};
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_17dxf_close_connection(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_cc = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dxf_close_connection (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cc,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dxf_close_connection") < 0)) __PYX_ERR(0, 361, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_cc = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dxf_close_connection", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 361, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_close_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_cc), __pyx_ptype_6dxfeed_4core_8DXFeedPy_ConnectionClass, 1, "cc", 0))) __PYX_ERR(0, 361, __pyx_L1_error)
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_16dxf_close_connection(__pyx_self, __pyx_v_cc);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_16dxf_close_connection(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_cc) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("dxf_close_connection", 0);

  /* "dxfeed/core/DXFeedPy.pyx":373
 * 
 *     # close all subscriptions before closing the connection
 *     for i in range(cc.sub_ptr_list.size()):             # <<<<<<<<<<<<<<
 *         if cc.sub_ptr_list[i]:  # subscription should not be closed previously
 *             clib.dxf_close_subscription(cc.sub_ptr_list[i][0])
 */
  __pyx_t_1 = __pyx_v_cc->sub_ptr_list.size();
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "dxfeed/core/DXFeedPy.pyx":374
 *     # close all subscriptions before closing the connection
 *     for i in range(cc.sub_ptr_list.size()):
 *         if cc.sub_ptr_list[i]:  # subscription should not be closed previously             # <<<<<<<<<<<<<<
 *             clib.dxf_close_subscription(cc.sub_ptr_list[i][0])
 *             cc.sub_ptr_list[i][0] = NULL  # mark subscription as closed
 */
    __pyx_t_4 = ((__pyx_v_cc->sub_ptr_list[__pyx_v_i]) != 0);
    if (__pyx_t_4) {

      /* "dxfeed/core/DXFeedPy.pyx":375
 *     for i in range(cc.sub_ptr_list.size()):
 *         if cc.sub_ptr_list[i]:  # subscription should not be closed previously
 *             clib.dxf_close_subscription(cc.sub_ptr_list[i][0])             # <<<<<<<<<<<<<<
 *             cc.sub_ptr_list[i][0] = NULL  # mark subscription as closed
 * 
 */
      (void)(dxf_close_subscription(((__pyx_v_cc->sub_ptr_list[__pyx_v_i])[0])));

      /* "dxfeed/core/DXFeedPy.pyx":376
 *         if cc.sub_ptr_list[i]:  # subscription should not be closed previously
 *             clib.dxf_close_subscription(cc.sub_ptr_list[i][0])
 *             cc.sub_ptr_list[i][0] = NULL  # mark subscription as closed             # <<<<<<<<<<<<<<
 * 
 *     cc.sub_ptr_list.clear()
 */
      ((__pyx_v_cc->sub_ptr_list[__pyx_v_i])[0]) = NULL;

      /* "dxfeed/core/DXFeedPy.pyx":374
 *     # close all subscriptions before closing the connection
 *     for i in range(cc.sub_ptr_list.size()):
 *         if cc.sub_ptr_list[i]:  # subscription should not be closed previously             # <<<<<<<<<<<<<<
 *             clib.dxf_close_subscription(cc.sub_ptr_list[i][0])
 *             cc.sub_ptr_list[i][0] = NULL  # mark subscription as closed
 */
    }
  }

  /* "dxfeed/core/DXFeedPy.pyx":378
 *             cc.sub_ptr_list[i][0] = NULL  # mark subscription as closed
 * 
 *     cc.sub_ptr_list.clear()             # <<<<<<<<<<<<<<
 * 
 *     if cc.connection:
 */
  __pyx_v_cc->sub_ptr_list.clear();

  /* "dxfeed/core/DXFeedPy.pyx":380
 *     cc.sub_ptr_list.clear()
 * 
 *     if cc.connection:             # <<<<<<<<<<<<<<
 *         clib.dxf_close_connection(cc.connection)
 *         cc.connection = NULL
 */
  __pyx_t_4 = (__pyx_v_cc->connection != 0);
  if (__pyx_t_4) {

    /* "dxfeed/core/DXFeedPy.pyx":381
 * 
 *     if cc.connection:
 *         clib.dxf_close_connection(cc.connection)             # <<<<<<<<<<<<<<
 *         cc.connection = NULL
 * 
 */
    (void)(dxf_close_connection(__pyx_v_cc->connection));

    /* "dxfeed/core/DXFeedPy.pyx":382
 *     if cc.connection:
 *         clib.dxf_close_connection(cc.connection)
 *         cc.connection = NULL             # <<<<<<<<<<<<<<
 * 
 * def dxf_close_subscription(SubscriptionClass sc):
 */
    __pyx_v_cc->connection = NULL;

    /* "dxfeed/core/DXFeedPy.pyx":380
 *     cc.sub_ptr_list.clear()
 * 
 *     if cc.connection:             # <<<<<<<<<<<<<<
 *         clib.dxf_close_connection(cc.connection)
 *         cc.connection = NULL
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":361
 *         process_last_error()
 * 
 * def dxf_close_connection(ConnectionClass cc):             # <<<<<<<<<<<<<<
 *     """
 *     Closes connection
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":384
 *         cc.connection = NULL
 * 
 * def dxf_close_subscription(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Closes subscription
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_19dxf_close_subscription(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_18dxf_close_subscription[] = "\n    Closes subscription\n\n    Parameters\n    ----------\n    sc: SubscriptionClass\n        SubscriptionClass with information about subscription\n    ";
static PyMethodDef __pyx_mdef_6dxfeed_4core_8DXFeedPy_19dxf_close_subscription = {"dxf_close_subscription", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_19dxf_close_subscription, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_18dxf_close_subscription};
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_19dxf_close_subscription(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dxf_close_subscription (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sc,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dxf_close_subscription") < 0)) __PYX_ERR(0, 384, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_sc = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dxf_close_subscription", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 384, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_close_subscription", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sc), __pyx_ptype_6dxfeed_4core_8DXFeedPy_SubscriptionClass, 1, "sc", 0))) __PYX_ERR(0, 384, __pyx_L1_error)
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_18dxf_close_subscription(__pyx_self, __pyx_v_sc);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_18dxf_close_subscription(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("dxf_close_subscription", 0);

  /* "dxfeed/core/DXFeedPy.pyx":393
 *         SubscriptionClass with information about subscription
 *     """
 *     if sc.subscription:             # <<<<<<<<<<<<<<
 *         clib.dxf_close_subscription(sc.subscription)
 *         sc.subscription = NULL
 */
  __pyx_t_1 = (__pyx_v_sc->subscription != 0);
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":394
 *     """
 *     if sc.subscription:
 *         clib.dxf_close_subscription(sc.subscription)             # <<<<<<<<<<<<<<
 *         sc.subscription = NULL
 * 
 */
    (void)(dxf_close_subscription(__pyx_v_sc->subscription));

    /* "dxfeed/core/DXFeedPy.pyx":395
 *     if sc.subscription:
 *         clib.dxf_close_subscription(sc.subscription)
 *         sc.subscription = NULL             # <<<<<<<<<<<<<<
 * 
 * def dxf_get_current_connection_status(ConnectionClass cc, return_str: bool=True):
 */
    __pyx_v_sc->subscription = NULL;

    /* "dxfeed/core/DXFeedPy.pyx":393
 *         SubscriptionClass with information about subscription
 *     """
 *     if sc.subscription:             # <<<<<<<<<<<<<<
 *         clib.dxf_close_subscription(sc.subscription)
 *         sc.subscription = NULL
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":384
 *         cc.connection = NULL
 * 
 * def dxf_close_subscription(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Closes subscription
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":397
 *         sc.subscription = NULL
 * 
 * def dxf_get_current_connection_status(ConnectionClass cc, return_str: bool=True):             # <<<<<<<<<<<<<<
 *     """
 *     Returns one of four possible statuses
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_21dxf_get_current_connection_status(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_20dxf_get_current_connection_status[] = "\n    Returns one of four possible statuses\n\n    Parameters\n    ----------\n    cc: ConnectionClass\n        Variable with connection information\n    return_str: bool\n        When True returns connection status in string format, otherwise internal c representation as integer\n\n\n    ";
static PyMethodDef __pyx_mdef_6dxfeed_4core_8DXFeedPy_21dxf_get_current_connection_status = {"dxf_get_current_connection_status", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_21dxf_get_current_connection_status, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_20dxf_get_current_connection_status};
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_21dxf_get_current_connection_status(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_cc = 0;
  PyObject *__pyx_v_return_str = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dxf_get_current_connection_status (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cc,&__pyx_n_s_return_str,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_return_str);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dxf_get_current_connection_status") < 0)) __PYX_ERR(0, 397, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_cc = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *)values[0]);
    __pyx_v_return_str = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dxf_get_current_connection_status", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 397, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_get_current_connection_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_cc), __pyx_ptype_6dxfeed_4core_8DXFeedPy_ConnectionClass, 1, "cc", 0))) __PYX_ERR(0, 397, __pyx_L1_error)
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_20dxf_get_current_connection_status(__pyx_self, __pyx_v_cc, __pyx_v_return_str);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_20dxf_get_current_connection_status(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_cc, PyObject *__pyx_v_return_str) {
  PyObject *__pyx_v_status_mapping = NULL;
  dxf_connection_status_t __pyx_v_status;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("dxf_get_current_connection_status", 0);

  /* "dxfeed/core/DXFeedPy.pyx":411
 *     """
 *     status_mapping = {
 *         0: 'Not connected',             # <<<<<<<<<<<<<<
 *         1: 'Connected',
 *         2: 'Login required',
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_0, __pyx_kp_u_Not_connected) < 0) __PYX_ERR(0, 411, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_1, __pyx_n_u_Connected) < 0) __PYX_ERR(0, 411, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_2, __pyx_kp_u_Login_required) < 0) __PYX_ERR(0, 411, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_3, __pyx_kp_u_Connected_and_authorized) < 0) __PYX_ERR(0, 411, __pyx_L1_error)
  __pyx_v_status_mapping = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":418
 * 
 *     cdef clib.dxf_connection_status_t status
 *     clib.dxf_get_current_connection_status(cc.connection, &status)             # <<<<<<<<<<<<<<
 *     result = status
 *     if return_str:
 */
  (void)(dxf_get_current_connection_status(__pyx_v_cc->connection, (&__pyx_v_status)));

  /* "dxfeed/core/DXFeedPy.pyx":419
 *     cdef clib.dxf_connection_status_t status
 *     clib.dxf_get_current_connection_status(cc.connection, &status)
 *     result = status             # <<<<<<<<<<<<<<
 *     if return_str:
 *         result = status_mapping[status]
 */
  __pyx_t_1 = __Pyx_PyInt_From_dxf_connection_status_t(__pyx_v_status); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":420
 *     clib.dxf_get_current_connection_status(cc.connection, &status)
 *     result = status
 *     if return_str:             # <<<<<<<<<<<<<<
 *         result = status_mapping[status]
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_return_str); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 420, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "dxfeed/core/DXFeedPy.pyx":421
 *     result = status
 *     if return_str:
 *         result = status_mapping[status]             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
    __pyx_t_1 = __Pyx_PyInt_From_dxf_connection_status_t(__pyx_v_status); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_status_mapping, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":420
 *     clib.dxf_get_current_connection_status(cc.connection, &status)
 *     result = status
 *     if return_str:             # <<<<<<<<<<<<<<
 *         result = status_mapping[status]
 * 
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":423
 *         result = status_mapping[status]
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * def dxf_get_current_connected_address(ConnectionClass cc):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "dxfeed/core/DXFeedPy.pyx":397
 *         sc.subscription = NULL
 * 
 * def dxf_get_current_connection_status(ConnectionClass cc, return_str: bool=True):             # <<<<<<<<<<<<<<
 *     """
 *     Returns one of four possible statuses
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_get_current_connection_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_status_mapping);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":425
 *     return result
 * 
 * def dxf_get_current_connected_address(ConnectionClass cc):             # <<<<<<<<<<<<<<
 *     """
 *     Returns current connected address
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_23dxf_get_current_connected_address(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_22dxf_get_current_connected_address[] = "\n    Returns current connected address\n\n    Parameters\n    ----------\n    cc: ConnectionClass\n        Variable with connection information\n\n    Returns\n    -------\n    address: str\n        Current connected address\n    ";
static PyMethodDef __pyx_mdef_6dxfeed_4core_8DXFeedPy_23dxf_get_current_connected_address = {"dxf_get_current_connected_address", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_23dxf_get_current_connected_address, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_22dxf_get_current_connected_address};
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_23dxf_get_current_connected_address(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_cc = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dxf_get_current_connected_address (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cc,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dxf_get_current_connected_address") < 0)) __PYX_ERR(0, 425, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_cc = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dxf_get_current_connected_address", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 425, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_get_current_connected_address", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_cc), __pyx_ptype_6dxfeed_4core_8DXFeedPy_ConnectionClass, 1, "cc", 0))) __PYX_ERR(0, 425, __pyx_L1_error)
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_22dxf_get_current_connected_address(__pyx_self, __pyx_v_cc);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_22dxf_get_current_connected_address(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *__pyx_v_cc) {
  char *__pyx_v_address;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("dxf_get_current_connected_address", 0);

  /* "dxfeed/core/DXFeedPy.pyx":439
 *         Current connected address
 *     """
 *     if not cc.connection:             # <<<<<<<<<<<<<<
 *         raise ValueError('Connection is not valid')
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_cc->connection != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "dxfeed/core/DXFeedPy.pyx":440
 *     """
 *     if not cc.connection:
 *         raise ValueError('Connection is not valid')             # <<<<<<<<<<<<<<
 * 
 *     cdef char * address
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 440, __pyx_L1_error)

    /* "dxfeed/core/DXFeedPy.pyx":439
 *         Current connected address
 *     """
 *     if not cc.connection:             # <<<<<<<<<<<<<<
 *         raise ValueError('Connection is not valid')
 * 
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":443
 * 
 *     cdef char * address
 *     clib.dxf_get_current_connected_address(cc.connection, &address)             # <<<<<<<<<<<<<<
 *     return (<bytes>address).decode('UTF-8')
 * 
 */
  (void)(dxf_get_current_connected_address(__pyx_v_cc->connection, (&__pyx_v_address)));

  /* "dxfeed/core/DXFeedPy.pyx":444
 *     cdef char * address
 *     clib.dxf_get_current_connected_address(cc.connection, &address)
 *     return (<bytes>address).decode('UTF-8')             # <<<<<<<<<<<<<<
 * 
 * def dxf_initialize_logger(file_name: str, rewrite_file: bool, show_timezone_info: bool, verbose: bool):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBytes_FromString(__pyx_v_address); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(__pyx_t_2 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "decode");
    __PYX_ERR(0, 444, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_decode_bytes(((PyObject*)__pyx_t_2), 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "dxfeed/core/DXFeedPy.pyx":425
 *     return result
 * 
 * def dxf_get_current_connected_address(ConnectionClass cc):             # <<<<<<<<<<<<<<
 *     """
 *     Returns current connected address
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_get_current_connected_address", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":446
 *     return (<bytes>address).decode('UTF-8')
 * 
 * def dxf_initialize_logger(file_name: str, rewrite_file: bool, show_timezone_info: bool, verbose: bool):             # <<<<<<<<<<<<<<
 *     """
 *     Initializes the internal logger.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_25dxf_initialize_logger(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_24dxf_initialize_logger[] = "\n    Initializes the internal logger.\n    Various actions and events, including the errors, are being logged throughout the library. They may be stored\n    into the file.\n\n    Parameters\n    ----------\n    file_name: str\n        A full path to the file where the log is to be stored\n    rewrite_file: bool\n        A flag defining the file open mode if it's True then the log file will be rewritten\n    show_timezone_info: bool\n        A flag defining the time display option in the log file if it's True then the time will be displayed\n        with the timezone suffix\n    verbose: bool\n        A flag defining the logging mode if it's True then the verbose logging will be enabled\n\n    ";
static PyMethodDef __pyx_mdef_6dxfeed_4core_8DXFeedPy_25dxf_initialize_logger = {"dxf_initialize_logger", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_25dxf_initialize_logger, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_24dxf_initialize_logger};
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_25dxf_initialize_logger(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_file_name = 0;
  PyObject *__pyx_v_rewrite_file = 0;
  PyObject *__pyx_v_show_timezone_info = 0;
  PyObject *__pyx_v_verbose = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dxf_initialize_logger (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_file_name,&__pyx_n_s_rewrite_file,&__pyx_n_s_show_timezone_info,&__pyx_n_s_verbose,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_file_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rewrite_file)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dxf_initialize_logger", 1, 4, 4, 1); __PYX_ERR(0, 446, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_show_timezone_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dxf_initialize_logger", 1, 4, 4, 2); __PYX_ERR(0, 446, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_verbose)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dxf_initialize_logger", 1, 4, 4, 3); __PYX_ERR(0, 446, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dxf_initialize_logger") < 0)) __PYX_ERR(0, 446, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_file_name = ((PyObject*)values[0]);
    __pyx_v_rewrite_file = values[1];
    __pyx_v_show_timezone_info = values[2];
    __pyx_v_verbose = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dxf_initialize_logger", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 446, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_initialize_logger", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_file_name), (&PyUnicode_Type), 1, "file_name", 1))) __PYX_ERR(0, 446, __pyx_L1_error)
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_24dxf_initialize_logger(__pyx_self, __pyx_v_file_name, __pyx_v_rewrite_file, __pyx_v_show_timezone_info, __pyx_v_verbose);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_24dxf_initialize_logger(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_file_name, PyObject *__pyx_v_rewrite_file, PyObject *__pyx_v_show_timezone_info, PyObject *__pyx_v_verbose) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char const *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("dxf_initialize_logger", 0);

  /* "dxfeed/core/DXFeedPy.pyx":465
 * 
 *     """
 *     clib.dxf_initialize_logger(file_name.encode('UTF-8'), int(rewrite_file), int(show_timezone_info), int(verbose))             # <<<<<<<<<<<<<<
 * 
 * def dxf_get_subscription_event_types(SubscriptionClass sc, return_str: bool=True):
 */
  if (unlikely(__pyx_v_file_name == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 465, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_AsUTF8String(__pyx_v_file_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBytes_AsString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 465, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_v_rewrite_file); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_v_show_timezone_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_v_verbose); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  (void)(dxf_initialize_logger(__pyx_t_2, __pyx_t_4, __pyx_t_5, __pyx_t_6));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":446
 *     return (<bytes>address).decode('UTF-8')
 * 
 * def dxf_initialize_logger(file_name: str, rewrite_file: bool, show_timezone_info: bool, verbose: bool):             # <<<<<<<<<<<<<<
 *     """
 *     Initializes the internal logger.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_initialize_logger", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":467
 *     clib.dxf_initialize_logger(file_name.encode('UTF-8'), int(rewrite_file), int(show_timezone_info), int(verbose))
 * 
 * def dxf_get_subscription_event_types(SubscriptionClass sc, return_str: bool=True):             # <<<<<<<<<<<<<<
 *     """
 *     Gets subscription event type
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_27dxf_get_subscription_event_types(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_26dxf_get_subscription_event_types[] = "\n    Gets subscription event type\n\n    Parameters\n    ----------\n    sc: SubscriptionClass\n        SubscriptionClass with information about subscription\n    return_str: bool\n        When True returns event type in string format, otherwise internal c representation as integer\n\n    Returns\n    -------\n    str or int\n        Subscription type\n    ";
static PyMethodDef __pyx_mdef_6dxfeed_4core_8DXFeedPy_27dxf_get_subscription_event_types = {"dxf_get_subscription_event_types", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_27dxf_get_subscription_event_types, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_26dxf_get_subscription_event_types};
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_27dxf_get_subscription_event_types(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc = 0;
  PyObject *__pyx_v_return_str = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dxf_get_subscription_event_types (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sc,&__pyx_n_s_return_str,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_return_str);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dxf_get_subscription_event_types") < 0)) __PYX_ERR(0, 467, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_sc = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)values[0]);
    __pyx_v_return_str = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dxf_get_subscription_event_types", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 467, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_get_subscription_event_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sc), __pyx_ptype_6dxfeed_4core_8DXFeedPy_SubscriptionClass, 1, "sc", 0))) __PYX_ERR(0, 467, __pyx_L1_error)
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_26dxf_get_subscription_event_types(__pyx_self, __pyx_v_sc, __pyx_v_return_str);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_26dxf_get_subscription_event_types(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc, PyObject *__pyx_v_return_str) {
  int __pyx_v_event_type;
  PyObject *__pyx_v_et_mapping = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("dxf_get_subscription_event_types", 0);

  /* "dxfeed/core/DXFeedPy.pyx":483
 *         Subscription type
 *     """
 *     if not sc.subscription:             # <<<<<<<<<<<<<<
 *         raise ValueError('Invalid subscription')
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_sc->subscription != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "dxfeed/core/DXFeedPy.pyx":484
 *     """
 *     if not sc.subscription:
 *         raise ValueError('Invalid subscription')             # <<<<<<<<<<<<<<
 * 
 *     cdef int event_type
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 484, __pyx_L1_error)

    /* "dxfeed/core/DXFeedPy.pyx":483
 *         Subscription type
 *     """
 *     if not sc.subscription:             # <<<<<<<<<<<<<<
 *         raise ValueError('Invalid subscription')
 * 
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":489
 * 
 *     et_mapping = {
 *         1: 'Trade',             # <<<<<<<<<<<<<<
 *         2: 'Quote',
 *         4: 'Summary',
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_1, __pyx_n_u_Trade) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_2, __pyx_n_u_Quote) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_4, __pyx_n_u_Summary) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_8, __pyx_n_u_Profile) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_16, __pyx_n_u_Order) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_32, __pyx_n_u_TimeAndSale) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_64, __pyx_n_u_Candle) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_128, __pyx_n_u_TradeETH) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_256, __pyx_n_u_SpreadOrder) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_512, __pyx_n_u_Greeks) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_1024, __pyx_n_u_TheoPrice) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_2048, __pyx_n_u_Underlying) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_4096, __pyx_n_u_Series) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_8192, __pyx_n_u_Configuration) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_int_neg_16384, __pyx_kp_u__13) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  __pyx_v_et_mapping = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":506
 *     }
 * 
 *     clib.dxf_get_subscription_event_types (sc.subscription, &event_type)             # <<<<<<<<<<<<<<
 *     result = event_type
 *     if return_str:
 */
  (void)(dxf_get_subscription_event_types(__pyx_v_sc->subscription, (&__pyx_v_event_type)));

  /* "dxfeed/core/DXFeedPy.pyx":507
 * 
 *     clib.dxf_get_subscription_event_types (sc.subscription, &event_type)
 *     result = event_type             # <<<<<<<<<<<<<<
 *     if return_str:
 *         result = et_mapping[event_type]
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_event_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_result = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":508
 *     clib.dxf_get_subscription_event_types (sc.subscription, &event_type)
 *     result = event_type
 *     if return_str:             # <<<<<<<<<<<<<<
 *         result = et_mapping[event_type]
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_return_str); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 508, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "dxfeed/core/DXFeedPy.pyx":509
 *     result = event_type
 *     if return_str:
 *         result = et_mapping[event_type]             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_event_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_et_mapping, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":508
 *     clib.dxf_get_subscription_event_types (sc.subscription, &event_type)
 *     result = event_type
 *     if return_str:             # <<<<<<<<<<<<<<
 *         result = et_mapping[event_type]
 * 
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":511
 *         result = et_mapping[event_type]
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * def dxf_get_symbols(SubscriptionClass sc):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "dxfeed/core/DXFeedPy.pyx":467
 *     clib.dxf_initialize_logger(file_name.encode('UTF-8'), int(rewrite_file), int(show_timezone_info), int(verbose))
 * 
 * def dxf_get_subscription_event_types(SubscriptionClass sc, return_str: bool=True):             # <<<<<<<<<<<<<<
 *     """
 *     Gets subscription event type
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_get_subscription_event_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_et_mapping);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":513
 *     return result
 * 
 * def dxf_get_symbols(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Retrieves the list of symbols currently added to the subscription.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_29dxf_get_symbols(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_28dxf_get_symbols[] = "\n    Retrieves the list of symbols currently added to the subscription.\n\n    Parameters\n    ----------\n    sc: SubscriptionClass\n        SubscriptionClass with information about subscription\n\n    Returns\n    -------\n    list\n        List of unicode strings of subscription symbols\n    ";
static PyMethodDef __pyx_mdef_6dxfeed_4core_8DXFeedPy_29dxf_get_symbols = {"dxf_get_symbols", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_29dxf_get_symbols, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_28dxf_get_symbols};
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_29dxf_get_symbols(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dxf_get_symbols (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sc,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dxf_get_symbols") < 0)) __PYX_ERR(0, 513, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_sc = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dxf_get_symbols", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 513, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_get_symbols", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sc), __pyx_ptype_6dxfeed_4core_8DXFeedPy_SubscriptionClass, 1, "sc", 0))) __PYX_ERR(0, 513, __pyx_L1_error)
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_28dxf_get_symbols(__pyx_self, __pyx_v_sc);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_28dxf_get_symbols(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc) {
  dxf_const_string_t *__pyx_v_symbols;
  PyObject *__pyx_v_symbols_list = NULL;
  int __pyx_v_symbol_count;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("dxf_get_symbols", 0);

  /* "dxfeed/core/DXFeedPy.pyx":527
 *         List of unicode strings of subscription symbols
 *     """
 *     if not sc.subscription:             # <<<<<<<<<<<<<<
 *         raise ValueError('Invalid subscription')
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_sc->subscription != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "dxfeed/core/DXFeedPy.pyx":528
 *     """
 *     if not sc.subscription:
 *         raise ValueError('Invalid subscription')             # <<<<<<<<<<<<<<
 * 
 *     cdef dxf_const_string_t * symbols
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 528, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 528, __pyx_L1_error)

    /* "dxfeed/core/DXFeedPy.pyx":527
 *         List of unicode strings of subscription symbols
 *     """
 *     if not sc.subscription:             # <<<<<<<<<<<<<<
 *         raise ValueError('Invalid subscription')
 * 
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":531
 * 
 *     cdef dxf_const_string_t * symbols
 *     symbols_list = list()             # <<<<<<<<<<<<<<
 *     cdef int symbol_count
 *     cdef int i
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_symbols_list = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":535
 *     cdef int i
 * 
 *     clib.dxf_get_symbols(sc.subscription, &symbols, &symbol_count)             # <<<<<<<<<<<<<<
 *     for i in range(symbol_count):
 *         symbols_list.append(unicode_from_dxf_const_string_t(symbols[i]))
 */
  (void)(dxf_get_symbols(__pyx_v_sc->subscription, (&__pyx_v_symbols), (&__pyx_v_symbol_count)));

  /* "dxfeed/core/DXFeedPy.pyx":536
 * 
 *     clib.dxf_get_symbols(sc.subscription, &symbols, &symbol_count)
 *     for i in range(symbol_count):             # <<<<<<<<<<<<<<
 *         symbols_list.append(unicode_from_dxf_const_string_t(symbols[i]))
 * 
 */
  __pyx_t_3 = __pyx_v_symbol_count;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "dxfeed/core/DXFeedPy.pyx":537
 *     clib.dxf_get_symbols(sc.subscription, &symbols, &symbol_count)
 *     for i in range(symbol_count):
 *         symbols_list.append(unicode_from_dxf_const_string_t(symbols[i]))             # <<<<<<<<<<<<<<
 * 
 *     return symbols_list
 */
    __pyx_t_2 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t((__pyx_v_symbols[__pyx_v_i])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 537, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_symbols_list, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 537, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "dxfeed/core/DXFeedPy.pyx":539
 *         symbols_list.append(unicode_from_dxf_const_string_t(symbols[i]))
 * 
 *     return symbols_list             # <<<<<<<<<<<<<<
 * 
 * def dxf_remove_symbols(SubscriptionClass sc, symbols: Iterable[str]):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_symbols_list);
  __pyx_r = __pyx_v_symbols_list;
  goto __pyx_L0;

  /* "dxfeed/core/DXFeedPy.pyx":513
 *     return result
 * 
 * def dxf_get_symbols(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Retrieves the list of symbols currently added to the subscription.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_get_symbols", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_symbols_list);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":541
 *     return symbols_list
 * 
 * def dxf_remove_symbols(SubscriptionClass sc, symbols: Iterable[str]):             # <<<<<<<<<<<<<<
 *     """
 *     Removes several symbols from the subscription
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_31dxf_remove_symbols(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_30dxf_remove_symbols[] = "\n    Removes several symbols from the subscription\n\n    Parameters\n    ----------\n    sc: SubscriptionClass\n        SubscriptionClass with information about subscription\n    symbols: list\n        List of symbols to remove\n    ";
static PyMethodDef __pyx_mdef_6dxfeed_4core_8DXFeedPy_31dxf_remove_symbols = {"dxf_remove_symbols", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_31dxf_remove_symbols, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_30dxf_remove_symbols};
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_31dxf_remove_symbols(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc = 0;
  PyObject *__pyx_v_symbols = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dxf_remove_symbols (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sc,&__pyx_n_s_symbols,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_symbols)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dxf_remove_symbols", 1, 2, 2, 1); __PYX_ERR(0, 541, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dxf_remove_symbols") < 0)) __PYX_ERR(0, 541, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_sc = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)values[0]);
    __pyx_v_symbols = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dxf_remove_symbols", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 541, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_remove_symbols", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sc), __pyx_ptype_6dxfeed_4core_8DXFeedPy_SubscriptionClass, 1, "sc", 0))) __PYX_ERR(0, 541, __pyx_L1_error)
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_30dxf_remove_symbols(__pyx_self, __pyx_v_sc, __pyx_v_symbols);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_30dxf_remove_symbols(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc, PyObject *__pyx_v_symbols) {
  PyObject *__pyx_v_symbol = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("dxf_remove_symbols", 0);

  /* "dxfeed/core/DXFeedPy.pyx":552
 *         List of symbols to remove
 *     """
 *     if not sc.subscription:             # <<<<<<<<<<<<<<
 *         raise ValueError('Invalid subscription')
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_sc->subscription != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "dxfeed/core/DXFeedPy.pyx":553
 *     """
 *     if not sc.subscription:
 *         raise ValueError('Invalid subscription')             # <<<<<<<<<<<<<<
 * 
 *     for symbol in symbols:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 553, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 553, __pyx_L1_error)

    /* "dxfeed/core/DXFeedPy.pyx":552
 *         List of symbols to remove
 *     """
 *     if not sc.subscription:             # <<<<<<<<<<<<<<
 *         raise ValueError('Invalid subscription')
 * 
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":555
 *         raise ValueError('Invalid subscription')
 * 
 *     for symbol in symbols:             # <<<<<<<<<<<<<<
 *         clib.dxf_remove_symbol(sc.subscription, dxf_const_string_t_from_unicode(symbol))
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_symbols)) || PyTuple_CheckExact(__pyx_v_symbols)) {
    __pyx_t_2 = __pyx_v_symbols; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_symbols); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 555, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 555, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 555, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 555, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 555, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 555, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 555, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_symbol, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "dxfeed/core/DXFeedPy.pyx":556
 * 
 *     for symbol in symbols:
 *         clib.dxf_remove_symbol(sc.subscription, dxf_const_string_t_from_unicode(symbol))             # <<<<<<<<<<<<<<
 * 
 * def dxf_clear_symbols(SubscriptionClass sc):
 */
    (void)(dxf_remove_symbol(__pyx_v_sc->subscription, __pyx_f_6dxfeed_4core_5utils_7helpers_dxf_const_string_t_from_unicode(__pyx_v_symbol)));

    /* "dxfeed/core/DXFeedPy.pyx":555
 *         raise ValueError('Invalid subscription')
 * 
 *     for symbol in symbols:             # <<<<<<<<<<<<<<
 *         clib.dxf_remove_symbol(sc.subscription, dxf_const_string_t_from_unicode(symbol))
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":541
 *     return symbols_list
 * 
 * def dxf_remove_symbols(SubscriptionClass sc, symbols: Iterable[str]):             # <<<<<<<<<<<<<<
 *     """
 *     Removes several symbols from the subscription
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_remove_symbols", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_symbol);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/DXFeedPy.pyx":558
 *         clib.dxf_remove_symbol(sc.subscription, dxf_const_string_t_from_unicode(symbol))
 * 
 * def dxf_clear_symbols(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Removes all symbols from the subscription
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_33dxf_clear_symbols(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6dxfeed_4core_8DXFeedPy_32dxf_clear_symbols[] = "\n    Removes all symbols from the subscription\n\n    Parameters\n    ----------\n    sc: SubscriptionClass\n        SubscriptionClass with information about subscription\n    ";
static PyMethodDef __pyx_mdef_6dxfeed_4core_8DXFeedPy_33dxf_clear_symbols = {"dxf_clear_symbols", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_33dxf_clear_symbols, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_32dxf_clear_symbols};
static PyObject *__pyx_pw_6dxfeed_4core_8DXFeedPy_33dxf_clear_symbols(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dxf_clear_symbols (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sc,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sc)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dxf_clear_symbols") < 0)) __PYX_ERR(0, 558, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_sc = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dxf_clear_symbols", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 558, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_clear_symbols", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sc), __pyx_ptype_6dxfeed_4core_8DXFeedPy_SubscriptionClass, 1, "sc", 0))) __PYX_ERR(0, 558, __pyx_L1_error)
  __pyx_r = __pyx_pf_6dxfeed_4core_8DXFeedPy_32dxf_clear_symbols(__pyx_self, __pyx_v_sc);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_8DXFeedPy_32dxf_clear_symbols(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *__pyx_v_sc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("dxf_clear_symbols", 0);

  /* "dxfeed/core/DXFeedPy.pyx":567
 *         SubscriptionClass with information about subscription
 *     """
 *     if not sc.subscription:             # <<<<<<<<<<<<<<
 *         raise ValueError('Invalid subscription')
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_sc->subscription != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "dxfeed/core/DXFeedPy.pyx":568
 *     """
 *     if not sc.subscription:
 *         raise ValueError('Invalid subscription')             # <<<<<<<<<<<<<<
 * 
 *     clib.dxf_clear_symbols(sc.subscription)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 568, __pyx_L1_error)

    /* "dxfeed/core/DXFeedPy.pyx":567
 *         SubscriptionClass with information about subscription
 *     """
 *     if not sc.subscription:             # <<<<<<<<<<<<<<
 *         raise ValueError('Invalid subscription')
 * 
 */
  }

  /* "dxfeed/core/DXFeedPy.pyx":570
 *         raise ValueError('Invalid subscription')
 * 
 *     clib.dxf_clear_symbols(sc.subscription)             # <<<<<<<<<<<<<<
 */
  (void)(dxf_clear_symbols(__pyx_v_sc->subscription));

  /* "dxfeed/core/DXFeedPy.pyx":558
 *         clib.dxf_remove_symbol(sc.subscription, dxf_const_string_t_from_unicode(symbol))
 * 
 * def dxf_clear_symbols(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Removes all symbols from the subscription
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("dxfeed.core.DXFeedPy.dxf_clear_symbols", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_6dxfeed_4core_8DXFeedPy_ConnectionClass __pyx_vtable_6dxfeed_4core_8DXFeedPy_ConnectionClass;

static PyObject *__pyx_tp_new_6dxfeed_4core_8DXFeedPy_ConnectionClass(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *)o);
  p->__pyx_vtab = __pyx_vtabptr_6dxfeed_4core_8DXFeedPy_ConnectionClass;
  new((void*)&(p->sub_ptr_list)) std::deque<dxf_subscription_t *> ();
  return o;
}

static void __pyx_tp_dealloc_6dxfeed_4core_8DXFeedPy_ConnectionClass(PyObject *o) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *p = (struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_6dxfeed_4core_8DXFeedPy_15ConnectionClass_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  __Pyx_call_destructor(p->sub_ptr_list);
  (*Py_TYPE(o)->tp_free)(o);
}

static PyMethodDef __pyx_methods_6dxfeed_4core_8DXFeedPy_ConnectionClass[] = {
  {"make_new_subscription", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_15ConnectionClass_5make_new_subscription, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_15ConnectionClass_7__reduce_cython__, METH_VARARGS|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_15ConnectionClass_9__setstate_cython__, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6dxfeed_4core_8DXFeedPy_ConnectionClass = {
  PyVarObject_HEAD_INIT(0, 0)
  "dxfeed.core.DXFeedPy.ConnectionClass", /*tp_name*/
  sizeof(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6dxfeed_4core_8DXFeedPy_ConnectionClass, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "\n    Data structure that contains connection\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6dxfeed_4core_8DXFeedPy_ConnectionClass, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6dxfeed_4core_8DXFeedPy_15ConnectionClass_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6dxfeed_4core_8DXFeedPy_ConnectionClass, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_6dxfeed_4core_8DXFeedPy_SubscriptionClass(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)o);
  p->event_type_str = Py_None; Py_INCREF(Py_None);
  p->columns = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->data = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6dxfeed_4core_8DXFeedPy_SubscriptionClass(PyObject *o) {
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *p = (struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->event_type_str);
  Py_CLEAR(p->columns);
  Py_CLEAR(p->data);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6dxfeed_4core_8DXFeedPy_SubscriptionClass(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *p = (struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)o;
  if (p->event_type_str) {
    e = (*v)(p->event_type_str, a); if (e) return e;
  }
  if (p->columns) {
    e = (*v)(p->columns, a); if (e) return e;
  }
  if (p->data) {
    e = (*v)(p->data, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6dxfeed_4core_8DXFeedPy_SubscriptionClass(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *p = (struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *)o;
  tmp = ((PyObject*)p->event_type_str);
  p->event_type_str = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->columns);
  p->columns = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->data);
  p->data = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6dxfeed_4core_8DXFeedPy_SubscriptionClass[] = {
  {"get_data", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_5get_data, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_4get_data},
  {"to_dataframe", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_7to_dataframe, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_6to_dataframe},
  {"__reduce_cython__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_9__reduce_cython__, METH_VARARGS|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_11__setstate_cython__, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6dxfeed_4core_8DXFeedPy_SubscriptionClass = {
  PyVarObject_HEAD_INIT(0, 0)
  "dxfeed.core.DXFeedPy.SubscriptionClass", /*tp_name*/
  sizeof(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6dxfeed_4core_8DXFeedPy_SubscriptionClass, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "\n    Data structure that contains subscription and related fields\n    ", /*tp_doc*/
  __pyx_tp_traverse_6dxfeed_4core_8DXFeedPy_SubscriptionClass, /*tp_traverse*/
  __pyx_tp_clear_6dxfeed_4core_8DXFeedPy_SubscriptionClass, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6dxfeed_4core_8DXFeedPy_SubscriptionClass, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6dxfeed_4core_8DXFeedPy_17SubscriptionClass_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6dxfeed_4core_8DXFeedPy_SubscriptionClass, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"process_last_error", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6dxfeed_4core_8DXFeedPy_1process_last_error, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6dxfeed_4core_8DXFeedPy_process_last_error},
  {0, 0, 0, 0}
};

static int __pyx_import_star_set(PyObject *o, PyObject* py_name, char *name) {
  static const char* internal_type_names[] = {
    "ConnectionClass",
    "ERRORCODE",
    "PyObject",
    "SubscriptionClass",
    "__pyx_ctuple_int",
    "__pyx_ctuple_int_struct",
    "__pyx_ctuple_long__and_long__and_long",
    "__pyx_ctuple_long__and_long__and_long_struct",
    "__pyx_ctuple_size_t",
    "__pyx_ctuple_size_t_struct",
    "__pyx_opt_args_6dxfeed_4core_8DXFeedPy_process_last_error",
    "cppdq",
    "dx_candle_t",
    "dx_configuration_t",
    "dx_event_id_t",
    "dx_event_subscription_param_list_t",
    "dx_event_subscription_param_t",
    "dx_greeks_t",
    "dx_market_maker_t",
    "dx_order_source_array_ptr_t",
    "dx_order_source_array_t",
    "dx_order_t",
    "dx_profile_t",
    "dx_quote_t",
    "dx_record_id_t",
    "dx_record_info_id_t",
    "dx_series_t",
    "dx_spread_order_t",
    "dx_subscription_type_t",
    "dx_suffix_t",
    "dx_summary_t",
    "dx_theo_price_t",
    "dx_time_and_sale_t",
    "dx_trade_eth_t",
    "dx_trade_t",
    "dx_underlying_t",
    "dxf_bool_t",
    "dxf_byte_array_t",
    "dxf_byte_t",
    "dxf_candle_alignment_attribute_t",
    "dxf_candle_attributes_t",
    "dxf_candle_price_attribute_t",
    "dxf_candle_session_attribute_t",
    "dxf_candle_t",
    "dxf_candle_type_period_attribute_t",
    "dxf_char_t",
    "dxf_configuration_t",
    "dxf_connection_status_t",
    "dxf_connection_t",
    "dxf_const_string_t",
    "dxf_dayid_t",
    "dxf_direction_t",
    "dxf_double_t",
    "dxf_event_data_t",
    "dxf_event_flag",
    "dxf_event_flags_t",
    "dxf_event_listener_t",
    "dxf_event_listener_v2_t",
    "dxf_event_params_t",
    "dxf_float_t",
    "dxf_greeks_t",
    "dxf_int_t",
    "dxf_long_t",
    "dxf_order_scope_t",
    "dxf_order_side_t",
    "dxf_order_t",
    "dxf_price_level_book_data_ptr_t",
    "dxf_price_level_book_data_t",
    "dxf_price_level_book_listener_t",
    "dxf_price_level_book_t",
    "dxf_price_level_element_t",
    "dxf_price_type_t",
    "dxf_profile_t",
    "dxf_property_item_t",
    "dxf_quote_t",
    "dxf_regional_book_t",
    "dxf_regional_quote_listener_t",
    "dxf_series_t",
    "dxf_short_sale_restriction_t",
    "dxf_short_t",
    "dxf_snapshot_data_ptr_t",
    "dxf_snapshot_data_t",
    "dxf_snapshot_inc_listener_t",
    "dxf_snapshot_listener_t",
    "dxf_snapshot_t",
    "dxf_string_t",
    "dxf_subscription_t",
    "dxf_summary_t",
    "dxf_theo_price_t",
    "dxf_time_and_sale_t",
    "dxf_time_int_field_t",
    "dxf_tns_type_t",
    "dxf_trade_t",
    "dxf_trading_status_t",
    "dxf_ubyte_t",
    "dxf_uint_t",
    "dxf_ulong_t",
    "dxf_underlying_t",
    "dxf_ushort_t",
    "wchar_t",
    0
  };
  const char** type_name = internal_type_names;
  while (*type_name) {
    if (__Pyx_StrEq(name, *type_name)) {
      PyErr_Format(PyExc_TypeError, "Cannot overwrite C type %s", name);
      goto bad;
    }
    type_name++;
  }
  if (0);
  else {
    if (PyObject_SetAttr(__pyx_m, py_name, o) < 0) goto bad;
  }
  return 0;
  bad:
  return -1;
}

static int
__Pyx_import_all_from(PyObject *locals, PyObject *v)
{
    PyObject *all = PyObject_GetAttrString(v, "__all__");
    PyObject *dict, *name, *value;
    int skip_leading_underscores = 0;
    int pos, err;
    if (all == NULL) {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError))
            return -1;
        PyErr_Clear();
        dict = PyObject_GetAttrString(v, "__dict__");
        if (dict == NULL) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError))
                return -1;
            PyErr_SetString(PyExc_ImportError,
            "from-import-* object has no __dict__ and no __all__");
            return -1;
        }
#if PY_MAJOR_VERSION < 3
        all = PyObject_CallMethod(dict, (char *)"keys", NULL);
#else
        all = PyMapping_Keys(dict);
#endif
        Py_DECREF(dict);
        if (all == NULL)
            return -1;
        skip_leading_underscores = 1;
    }
    for (pos = 0, err = 0; ; pos++) {
        name = PySequence_GetItem(all, pos);
        if (name == NULL) {
            if (!PyErr_ExceptionMatches(PyExc_IndexError))
                err = -1;
            else
                PyErr_Clear();
            break;
        }
        if (skip_leading_underscores &&
#if PY_MAJOR_VERSION < 3
            PyString_Check(name) &&
            PyString_AS_STRING(name)[0] == '_')
#else
            PyUnicode_Check(name) &&
            PyUnicode_AS_UNICODE(name)[0] == '_')
#endif
        {
            Py_DECREF(name);
            continue;
        }
        value = PyObject_GetAttr(v, name);
        if (value == NULL)
            err = -1;
        else if (PyDict_CheckExact(locals))
            err = PyDict_SetItem(locals, name, value);
        else
            err = PyObject_SetItem(locals, name, value);
        Py_DECREF(name);
        Py_XDECREF(value);
        if (err != 0)
            break;
    }
    Py_DECREF(all);
    return err;
}
static int __pyx_import_star(PyObject* m) {
    int i;
    int ret = -1;
    char* s;
    PyObject *locals = 0;
    PyObject *list = 0;
#if PY_MAJOR_VERSION >= 3
    PyObject *utf8_name = 0;
#endif
    PyObject *name;
    PyObject *item;
    locals = PyDict_New();              if (!locals) goto bad;
    if (__Pyx_import_all_from(locals, m) < 0) goto bad;
    list = PyDict_Items(locals);        if (!list) goto bad;
    for(i=0; i<PyList_GET_SIZE(list); i++) {
        name = PyTuple_GET_ITEM(PyList_GET_ITEM(list, i), 0);
        item = PyTuple_GET_ITEM(PyList_GET_ITEM(list, i), 1);
#if PY_MAJOR_VERSION >= 3
        utf8_name = PyUnicode_AsUTF8String(name);
        if (!utf8_name) goto bad;
        s = PyBytes_AS_STRING(utf8_name);
        if (__pyx_import_star_set(item, name, s) < 0) goto bad;
        Py_DECREF(utf8_name); utf8_name = 0;
#else
        s = PyString_AsString(name);
        if (!s) goto bad;
        if (__pyx_import_star_set(item, name, s) < 0) goto bad;
#endif
    }
    ret = 0;
bad:
    Py_XDECREF(locals);
    Py_XDECREF(list);
#if PY_MAJOR_VERSION >= 3
    Py_XDECREF(utf8_name);
#endif
    return ret;
}



#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_DXFeedPy(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_DXFeedPy},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "DXFeedPy",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_CANDLE_COLUMNS, __pyx_k_CANDLE_COLUMNS, sizeof(__pyx_k_CANDLE_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_s_CONFIGURATION_COLUMNS, __pyx_k_CONFIGURATION_COLUMNS, sizeof(__pyx_k_CONFIGURATION_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_u_Candle, __pyx_k_Candle, sizeof(__pyx_k_Candle), 0, 1, 0, 1},
  {&__pyx_n_u_Configuration, __pyx_k_Configuration, sizeof(__pyx_k_Configuration), 0, 1, 0, 1},
  {&__pyx_n_u_Connected, __pyx_k_Connected, sizeof(__pyx_k_Connected), 0, 1, 0, 1},
  {&__pyx_kp_u_Connected_and_authorized, __pyx_k_Connected_and_authorized, sizeof(__pyx_k_Connected_and_authorized), 0, 1, 0, 0},
  {&__pyx_n_s_ConnectionClass, __pyx_k_ConnectionClass, sizeof(__pyx_k_ConnectionClass), 0, 0, 1, 1},
  {&__pyx_kp_u_Connection_is_not_valid, __pyx_k_Connection_is_not_valid, sizeof(__pyx_k_Connection_is_not_valid), 0, 1, 0, 0},
  {&__pyx_n_s_DataFrame, __pyx_k_DataFrame, sizeof(__pyx_k_DataFrame), 0, 0, 1, 1},
  {&__pyx_n_s_DequeWithLock, __pyx_k_DequeWithLock, sizeof(__pyx_k_DequeWithLock), 0, 0, 1, 1},
  {&__pyx_kp_u_Error_occurred_and_successfully, __pyx_k_Error_occurred_and_successfully, sizeof(__pyx_k_Error_occurred_and_successfully), 0, 1, 0, 0},
  {&__pyx_n_s_GREEKS_COLUMNS, __pyx_k_GREEKS_COLUMNS, sizeof(__pyx_k_GREEKS_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_u_Greeks, __pyx_k_Greeks, sizeof(__pyx_k_Greeks), 0, 1, 0, 1},
  {&__pyx_kp_u_In_underlying_C_API_library_erro, __pyx_k_In_underlying_C_API_library_erro, sizeof(__pyx_k_In_underlying_C_API_library_erro), 0, 1, 0, 0},
  {&__pyx_kp_u_Inapropriate_date_format_should, __pyx_k_Inapropriate_date_format_should, sizeof(__pyx_k_Inapropriate_date_format_should), 0, 1, 0, 0},
  {&__pyx_kp_u_Incorrect_event_type, __pyx_k_Incorrect_event_type, sizeof(__pyx_k_Incorrect_event_type), 0, 1, 0, 0},
  {&__pyx_kp_u_Invalid_subscription, __pyx_k_Invalid_subscription, sizeof(__pyx_k_Invalid_subscription), 0, 1, 0, 0},
  {&__pyx_n_s_Iterable, __pyx_k_Iterable, sizeof(__pyx_k_Iterable), 0, 0, 1, 1},
  {&__pyx_kp_u_Login_required, __pyx_k_Login_required, sizeof(__pyx_k_Login_required), 0, 1, 0, 0},
  {&__pyx_kp_u_M8_ms, __pyx_k_M8_ms, sizeof(__pyx_k_M8_ms), 0, 1, 0, 0},
  {&__pyx_kp_u_No_default_listener_for, __pyx_k_No_default_listener_for, sizeof(__pyx_k_No_default_listener_for), 0, 1, 0, 0},
  {&__pyx_kp_u_Not_connected, __pyx_k_Not_connected, sizeof(__pyx_k_Not_connected), 0, 1, 0, 0},
  {&__pyx_n_s_ORDER_COLUMNS, __pyx_k_ORDER_COLUMNS, sizeof(__pyx_k_ORDER_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_s_Optional, __pyx_k_Optional, sizeof(__pyx_k_Optional), 0, 0, 1, 1},
  {&__pyx_n_u_Order, __pyx_k_Order, sizeof(__pyx_k_Order), 0, 1, 0, 1},
  {&__pyx_n_s_PROFILE_COLUMNS, __pyx_k_PROFILE_COLUMNS, sizeof(__pyx_k_PROFILE_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_u_Profile, __pyx_k_Profile, sizeof(__pyx_k_Profile), 0, 1, 0, 1},
  {&__pyx_n_s_QUOTE_COLUMNS, __pyx_k_QUOTE_COLUMNS, sizeof(__pyx_k_QUOTE_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_u_Quote, __pyx_k_Quote, sizeof(__pyx_k_Quote), 0, 1, 0, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_SERIES_COLUMNS, __pyx_k_SERIES_COLUMNS, sizeof(__pyx_k_SERIES_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_s_SUMMARY_COLUMNS, __pyx_k_SUMMARY_COLUMNS, sizeof(__pyx_k_SUMMARY_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_u_Series, __pyx_k_Series, sizeof(__pyx_k_Series), 0, 1, 0, 1},
  {&__pyx_n_u_SpreadOrder, __pyx_k_SpreadOrder, sizeof(__pyx_k_SpreadOrder), 0, 1, 0, 1},
  {&__pyx_n_s_SubscriptionClass, __pyx_k_SubscriptionClass, sizeof(__pyx_k_SubscriptionClass), 0, 0, 1, 1},
  {&__pyx_kp_u_Subscription_is_not_valid, __pyx_k_Subscription_is_not_valid, sizeof(__pyx_k_Subscription_is_not_valid), 0, 1, 0, 0},
  {&__pyx_n_u_Summary, __pyx_k_Summary, sizeof(__pyx_k_Summary), 0, 1, 0, 1},
  {&__pyx_n_s_THEO_PRICE_COLUMNS, __pyx_k_THEO_PRICE_COLUMNS, sizeof(__pyx_k_THEO_PRICE_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_s_TIME_AND_SALE_COLUMNS, __pyx_k_TIME_AND_SALE_COLUMNS, sizeof(__pyx_k_TIME_AND_SALE_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_s_TRADE_COLUMNS, __pyx_k_TRADE_COLUMNS, sizeof(__pyx_k_TRADE_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_u_TheoPrice, __pyx_k_TheoPrice, sizeof(__pyx_k_TheoPrice), 0, 1, 0, 1},
  {&__pyx_n_u_Time, __pyx_k_Time, sizeof(__pyx_k_Time), 0, 1, 0, 1},
  {&__pyx_n_u_TimeAndSale, __pyx_k_TimeAndSale, sizeof(__pyx_k_TimeAndSale), 0, 1, 0, 1},
  {&__pyx_n_u_Trade, __pyx_k_Trade, sizeof(__pyx_k_Trade), 0, 1, 0, 1},
  {&__pyx_n_u_TradeETH, __pyx_k_TradeETH, sizeof(__pyx_k_TradeETH), 0, 1, 0, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_UNDERLYING_COLUMNS, __pyx_k_UNDERLYING_COLUMNS, sizeof(__pyx_k_UNDERLYING_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_u_Underlying, __pyx_k_Underlying, sizeof(__pyx_k_Underlying), 0, 1, 0, 1},
  {&__pyx_n_s_Union, __pyx_k_Union, sizeof(__pyx_k_Union), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_u_Y_m_d_H_M_S, __pyx_k_Y_m_d_H_M_S, sizeof(__pyx_k_Y_m_d_H_M_S), 0, 1, 0, 0},
  {&__pyx_kp_u__13, __pyx_k__13, sizeof(__pyx_k__13), 0, 1, 0, 0},
  {&__pyx_n_s__14, __pyx_k__14, sizeof(__pyx_k__14), 0, 0, 1, 1},
  {&__pyx_n_s_address, __pyx_k_address, sizeof(__pyx_k_address), 0, 0, 1, 1},
  {&__pyx_n_s_astype, __pyx_k_astype, sizeof(__pyx_k_astype), 0, 0, 1, 1},
  {&__pyx_n_s_candle_time, __pyx_k_candle_time, sizeof(__pyx_k_candle_time), 0, 0, 1, 1},
  {&__pyx_n_s_cc, __pyx_k_cc, sizeof(__pyx_k_cc), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_columns, __pyx_k_columns, sizeof(__pyx_k_columns), 0, 0, 1, 1},
  {&__pyx_n_s_contains, __pyx_k_contains, sizeof(__pyx_k_contains), 0, 0, 1, 1},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_data_len, __pyx_k_data_len, sizeof(__pyx_k_data_len), 0, 0, 1, 1},
  {&__pyx_n_s_datetime, __pyx_k_datetime, sizeof(__pyx_k_datetime), 0, 0, 1, 1},
  {&__pyx_kp_u_demo_dxfeed_com_7300, __pyx_k_demo_dxfeed_com_7300, sizeof(__pyx_k_demo_dxfeed_com_7300), 0, 1, 0, 0},
  {&__pyx_n_s_deque_wl, __pyx_k_deque_wl, sizeof(__pyx_k_deque_wl), 0, 0, 1, 1},
  {&__pyx_kp_u_description, __pyx_k_description, sizeof(__pyx_k_description), 0, 1, 0, 0},
  {&__pyx_n_s_dxf_add_symbols, __pyx_k_dxf_add_symbols, sizeof(__pyx_k_dxf_add_symbols), 0, 0, 1, 1},
  {&__pyx_n_s_dxf_attach_custom_listener, __pyx_k_dxf_attach_custom_listener, sizeof(__pyx_k_dxf_attach_custom_listener), 0, 0, 1, 1},
  {&__pyx_n_s_dxf_attach_listener, __pyx_k_dxf_attach_listener, sizeof(__pyx_k_dxf_attach_listener), 0, 0, 1, 1},
  {&__pyx_n_s_dxf_clear_symbols, __pyx_k_dxf_clear_symbols, sizeof(__pyx_k_dxf_clear_symbols), 0, 0, 1, 1},
  {&__pyx_n_s_dxf_close_connection, __pyx_k_dxf_close_connection, sizeof(__pyx_k_dxf_close_connection), 0, 0, 1, 1},
  {&__pyx_n_s_dxf_close_subscription, __pyx_k_dxf_close_subscription, sizeof(__pyx_k_dxf_close_subscription), 0, 0, 1, 1},
  {&__pyx_n_s_dxf_create_connection, __pyx_k_dxf_create_connection, sizeof(__pyx_k_dxf_create_connection), 0, 0, 1, 1},
  {&__pyx_n_s_dxf_create_connection_auth_beare, __pyx_k_dxf_create_connection_auth_beare, sizeof(__pyx_k_dxf_create_connection_auth_beare), 0, 0, 1, 1},
  {&__pyx_n_s_dxf_create_subscription, __pyx_k_dxf_create_subscription, sizeof(__pyx_k_dxf_create_subscription), 0, 0, 1, 1},
  {&__pyx_n_s_dxf_detach_listener, __pyx_k_dxf_detach_listener, sizeof(__pyx_k_dxf_detach_listener), 0, 0, 1, 1},
  {&__pyx_n_s_dxf_get_current_connected_addres, __pyx_k_dxf_get_current_connected_addres, sizeof(__pyx_k_dxf_get_current_connected_addres), 0, 0, 1, 1},
  {&__pyx_n_s_dxf_get_current_connection_statu, __pyx_k_dxf_get_current_connection_statu, sizeof(__pyx_k_dxf_get_current_connection_statu), 0, 0, 1, 1},
  {&__pyx_n_s_dxf_get_subscription_event_types, __pyx_k_dxf_get_subscription_event_types, sizeof(__pyx_k_dxf_get_subscription_event_types), 0, 0, 1, 1},
  {&__pyx_n_s_dxf_get_symbols, __pyx_k_dxf_get_symbols, sizeof(__pyx_k_dxf_get_symbols), 0, 0, 1, 1},
  {&__pyx_n_s_dxf_initialize_logger, __pyx_k_dxf_initialize_logger, sizeof(__pyx_k_dxf_initialize_logger), 0, 0, 1, 1},
  {&__pyx_n_s_dxf_remove_symbols, __pyx_k_dxf_remove_symbols, sizeof(__pyx_k_dxf_remove_symbols), 0, 0, 1, 1},
  {&__pyx_n_s_dxfeed_core_DXFeedPy, __pyx_k_dxfeed_core_DXFeedPy, sizeof(__pyx_k_dxfeed_core_DXFeedPy), 0, 0, 1, 1},
  {&__pyx_kp_s_dxfeed_core_DXFeedPy_pyx, __pyx_k_dxfeed_core_DXFeedPy_pyx, sizeof(__pyx_k_dxfeed_core_DXFeedPy_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_dxfeed_core_listeners_listener, __pyx_k_dxfeed_core_listeners_listener, sizeof(__pyx_k_dxfeed_core_listeners_listener), 0, 0, 1, 1},
  {&__pyx_n_s_dxfeed_core_utils_data_class, __pyx_k_dxfeed_core_utils_data_class, sizeof(__pyx_k_dxfeed_core_utils_data_class), 0, 0, 1, 1},
  {&__pyx_n_s_dxfeed_core_utils_helpers, __pyx_k_dxfeed_core_utils_helpers, sizeof(__pyx_k_dxfeed_core_utils_helpers), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_kp_u_error_code, __pyx_k_error_code, sizeof(__pyx_k_error_code), 0, 1, 0, 0},
  {&__pyx_n_s_error_code_2, __pyx_k_error_code_2, sizeof(__pyx_k_error_code_2), 0, 0, 1, 1},
  {&__pyx_n_s_et_mapping, __pyx_k_et_mapping, sizeof(__pyx_k_et_mapping), 0, 0, 1, 1},
  {&__pyx_n_s_et_type_int, __pyx_k_et_type_int, sizeof(__pyx_k_et_type_int), 0, 0, 1, 1},
  {&__pyx_n_s_event_type, __pyx_k_event_type, sizeof(__pyx_k_event_type), 0, 0, 1, 1},
  {&__pyx_kp_u_event_type_2, __pyx_k_event_type_2, sizeof(__pyx_k_event_type_2), 0, 1, 0, 0},
  {&__pyx_n_s_event_type_convert, __pyx_k_event_type_convert, sizeof(__pyx_k_event_type_convert), 0, 0, 1, 1},
  {&__pyx_n_s_file_name, __pyx_k_file_name, sizeof(__pyx_k_file_name), 0, 0, 1, 1},
  {&__pyx_n_s_fw, __pyx_k_fw, sizeof(__pyx_k_fw), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_kp_u_has_type_different_from_string, __pyx_k_has_type_different_from_string, sizeof(__pyx_k_has_type_different_from_string), 0, 1, 0, 0},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_idx, __pyx_k_idx, sizeof(__pyx_k_idx), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_keep, __pyx_k_keep, sizeof(__pyx_k_keep), 0, 0, 1, 1},
  {&__pyx_n_s_lis, __pyx_k_lis, sizeof(__pyx_k_lis), 0, 0, 1, 1},
  {&__pyx_n_s_loc, __pyx_k_loc, sizeof(__pyx_k_loc), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_make_new_subscription, __pyx_k_make_new_subscription, sizeof(__pyx_k_make_new_subscription), 0, 0, 1, 1},
  {&__pyx_n_s_maxlen, __pyx_k_maxlen, sizeof(__pyx_k_maxlen), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_kp_u_no_error_information_is_stored, __pyx_k_no_error_information_is_stored, sizeof(__pyx_k_no_error_information_is_stored), 0, 1, 0, 0},
  {&__pyx_kp_u_occurred, __pyx_k_occurred, sizeof(__pyx_k_occurred), 0, 1, 0, 0},
  {&__pyx_n_s_pandas, __pyx_k_pandas, sizeof(__pyx_k_pandas), 0, 0, 1, 1},
  {&__pyx_n_s_pd, __pyx_k_pd, sizeof(__pyx_k_pd), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
  {&__pyx_n_s_return_str, __pyx_k_return_str, sizeof(__pyx_k_return_str), 0, 0, 1, 1},
  {&__pyx_n_s_rewrite_file, __pyx_k_rewrite_file, sizeof(__pyx_k_rewrite_file), 0, 0, 1, 1},
  {&__pyx_n_s_safe_get, __pyx_k_safe_get, sizeof(__pyx_k_safe_get), 0, 0, 1, 1},
  {&__pyx_n_s_sc, __pyx_k_sc, sizeof(__pyx_k_sc), 0, 0, 1, 1},
  {&__pyx_kp_s_self_con_sub_list_ptr_self_conne, __pyx_k_self_con_sub_list_ptr_self_conne, sizeof(__pyx_k_self_con_sub_list_ptr_self_conne), 0, 0, 1, 0},
  {&__pyx_kp_s_self_connection_self_sub_ptr_lis, __pyx_k_self_connection_self_sub_ptr_lis, sizeof(__pyx_k_self_connection_self_sub_ptr_lis), 0, 0, 1, 0},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_show_timezone_info, __pyx_k_show_timezone_info, sizeof(__pyx_k_show_timezone_info), 0, 0, 1, 1},
  {&__pyx_n_s_status, __pyx_k_status, sizeof(__pyx_k_status), 0, 0, 1, 1},
  {&__pyx_n_s_status_mapping, __pyx_k_status_mapping, sizeof(__pyx_k_status_mapping), 0, 0, 1, 1},
  {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
  {&__pyx_n_s_strptime, __pyx_k_strptime, sizeof(__pyx_k_strptime), 0, 0, 1, 1},
  {&__pyx_n_s_sym, __pyx_k_sym, sizeof(__pyx_k_sym), 0, 0, 1, 1},
  {&__pyx_n_s_symbol, __pyx_k_symbol, sizeof(__pyx_k_symbol), 0, 0, 1, 1},
  {&__pyx_n_s_symbol_count, __pyx_k_symbol_count, sizeof(__pyx_k_symbol_count), 0, 0, 1, 1},
  {&__pyx_n_s_symbols, __pyx_k_symbols, sizeof(__pyx_k_symbols), 0, 0, 1, 1},
  {&__pyx_n_s_symbols_list, __pyx_k_symbols_list, sizeof(__pyx_k_symbols_list), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_timestamp, __pyx_k_timestamp, sizeof(__pyx_k_timestamp), 0, 0, 1, 1},
  {&__pyx_n_s_token, __pyx_k_token, sizeof(__pyx_k_token), 0, 0, 1, 1},
  {&__pyx_n_s_total_seconds, __pyx_k_total_seconds, sizeof(__pyx_k_total_seconds), 0, 0, 1, 1},
  {&__pyx_n_s_typing, __pyx_k_typing, sizeof(__pyx_k_typing), 0, 0, 1, 1},
  {&__pyx_n_s_utcnow, __pyx_k_utcnow, sizeof(__pyx_k_utcnow), 0, 0, 1, 1},
  {&__pyx_kp_u_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 1, 0, 0},
  {&__pyx_n_s_verbose, __pyx_k_verbose, sizeof(__pyx_k_verbose), 0, 0, 1, 1},
  {&__pyx_n_s_warn, __pyx_k_warn, sizeof(__pyx_k_warn), 0, 0, 1, 1},
  {&__pyx_n_s_warnings, __pyx_k_warnings, sizeof(__pyx_k_warnings), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_print = __Pyx_GetBuiltinName(__pyx_n_s_print); if (!__pyx_builtin_print) __PYX_ERR(0, 42, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 166, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 219, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 257, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 373, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "dxfeed/core/DXFeedPy.pyx":42
 *     if res == clib.DXF_SUCCESS:
 *         if error_code == dxec.dx_ec_success and verbose:
 *             print("no error information is stored")             # <<<<<<<<<<<<<<
 * 
 *         if verbose:
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_u_no_error_information_is_stored); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.connection,self.sub_ptr_list cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.connection,self.sub_ptr_list cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_self_connection_self_sub_ptr_lis); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "(tree fragment)":4
 *     raise TypeError("self.connection,self.sub_ptr_list cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.connection,self.sub_ptr_list cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_self_connection_self_sub_ptr_lis); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "dxfeed/core/DXFeedPy.pyx":144
 *         time_columns = df.columns[df.columns.str.contains('Time')]
 *         for column in time_columns:
 *             df.loc[:, column] = df.loc[:, column].astype('<M8[ms]')             # <<<<<<<<<<<<<<
 *         return df
 * 
 */
  __pyx_slice__4 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__4)) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__4);
  __Pyx_GIVEREF(__pyx_slice__4);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.con_sub_list_ptr,self.connection,self.listener,self.subscription,self.u_data cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.con_sub_list_ptr,self.connection,self.listener,self.subscription,self.u_data cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_self_con_sub_list_ptr_self_conne); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "(tree fragment)":4
 *     raise TypeError("self.con_sub_list_ptr,self.connection,self.listener,self.subscription,self.u_data cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.con_sub_list_ptr,self.connection,self.listener,self.subscription,self.u_data cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_self_con_sub_list_ptr_self_conne); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "dxfeed/core/DXFeedPy.pyx":219
 *     """
 *     if not cc.connection:
 *         raise ValueError('Connection is not valid')             # <<<<<<<<<<<<<<
 *     if event_type not in ['Trade', 'Quote', 'Summary', 'Profile', 'Order', 'TimeAndSale', 'Candle', 'TradeETH',
 *                           'SpreadOrder', 'Greeks', 'TheoPrice', 'Underlying', 'Series', 'Configuration', ]:
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_u_Connection_is_not_valid); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "dxfeed/core/DXFeedPy.pyx":222
 *     if event_type not in ['Trade', 'Quote', 'Summary', 'Profile', 'Order', 'TimeAndSale', 'Candle', 'TradeETH',
 *                           'SpreadOrder', 'Greeks', 'TheoPrice', 'Underlying', 'Series', 'Configuration', ]:
 *         raise ValueError('Incorrect event type!')             # <<<<<<<<<<<<<<
 * 
 *     sc = cc.make_new_subscription(data_len=data_len)
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_u_Incorrect_event_type); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "dxfeed/core/DXFeedPy.pyx":230
 *     try:
 *         candle_time = datetime.strptime(candle_time, '%Y-%m-%d %H:%M:%S') if candle_time else datetime.utcnow()
 *         timestamp = int((candle_time - datetime(1970, 1, 1)).total_seconds()) * 1000 - 5000             # <<<<<<<<<<<<<<
 *     except ValueError:
 *         raise Exception("Inapropriate date format, should be %Y-%m-%d %H:%M:%S")
 */
  __pyx_tuple__9 = PyTuple_Pack(3, __pyx_int_1970, __pyx_int_1, __pyx_int_1); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "dxfeed/core/DXFeedPy.pyx":232
 *         timestamp = int((candle_time - datetime(1970, 1, 1)).total_seconds()) * 1000 - 5000
 *     except ValueError:
 *         raise Exception("Inapropriate date format, should be %Y-%m-%d %H:%M:%S")             # <<<<<<<<<<<<<<
 * 
 *     if event_type == 'Candle':
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_u_Inapropriate_date_format_should); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "dxfeed/core/DXFeedPy.pyx":256
 *     """
 *     if not sc.subscription:
 *         raise ValueError('Subscription is not valid')             # <<<<<<<<<<<<<<
 *     for idx, sym in enumerate(symbols):
 *         if not isinstance(sym, str):
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_u_Subscription_is_not_valid); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "dxfeed/core/DXFeedPy.pyx":484
 *     """
 *     if not sc.subscription:
 *         raise ValueError('Invalid subscription')             # <<<<<<<<<<<<<<
 * 
 *     cdef int event_type
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_u_Invalid_subscription); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "dxfeed/core/DXFeedPy.pyx":147
 *         return df
 * 
 * def dxf_create_connection(address: Union[str, unicode, bytes] = 'demo.dxfeed.com:7300'):             # <<<<<<<<<<<<<<
 *     """
 *     Function creates connection to dxfeed given url address
 */
  __pyx_tuple__15 = PyTuple_Pack(3, __pyx_n_s_address, __pyx_n_s_cc, __pyx_n_s_error_code_2); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);
  __pyx_codeobj__16 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__15, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_dxfeed_core_DXFeedPy_pyx, __pyx_n_s_dxf_create_connection, 147, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__16)) __PYX_ERR(0, 147, __pyx_L1_error)

  /* "dxfeed/core/DXFeedPy.pyx":169
 *     return cc
 * 
 * def dxf_create_connection_auth_bearer(address: Union[str, unicode, bytes],             # <<<<<<<<<<<<<<
 *                                      token: Union[str, unicode, bytes]):
 *     """
 */
  __pyx_tuple__17 = PyTuple_Pack(4, __pyx_n_s_address, __pyx_n_s_token, __pyx_n_s_cc, __pyx_n_s_error_code_2); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);
  __pyx_codeobj__18 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__17, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_dxfeed_core_DXFeedPy_pyx, __pyx_n_s_dxf_create_connection_auth_beare, 169, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__18)) __PYX_ERR(0, 169, __pyx_L1_error)

  /* "dxfeed/core/DXFeedPy.pyx":196
 *     return cc
 * 
 * def dxf_create_subscription(ConnectionClass cc, event_type: str, candle_time: Optional[str] = None,             # <<<<<<<<<<<<<<
 *                             data_len: int = 100000):
 *     """
 */
  __pyx_tuple__19 = PyTuple_Pack(8, __pyx_n_s_cc, __pyx_n_s_event_type, __pyx_n_s_candle_time, __pyx_n_s_data_len, __pyx_n_s_sc, __pyx_n_s_et_type_int, __pyx_n_s_timestamp, __pyx_n_s_error_code_2); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);
  __pyx_codeobj__20 = (PyObject*)__Pyx_PyCode_New(4, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__19, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_dxfeed_core_DXFeedPy_pyx, __pyx_n_s_dxf_create_subscription, 196, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__20)) __PYX_ERR(0, 196, __pyx_L1_error)

  /* "dxfeed/core/DXFeedPy.pyx":244
 *     return sc
 * 
 * def dxf_add_symbols(SubscriptionClass sc, symbols: Iterable[str]):             # <<<<<<<<<<<<<<
 *     """
 *     Adds symbols to subscription
 */
  __pyx_tuple__21 = PyTuple_Pack(4, __pyx_n_s_sc, __pyx_n_s_symbols, __pyx_n_s_idx, __pyx_n_s_sym); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_codeobj__22 = (PyObject*)__Pyx_PyCode_New(2, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__21, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_dxfeed_core_DXFeedPy_pyx, __pyx_n_s_dxf_add_symbols, 244, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__22)) __PYX_ERR(0, 244, __pyx_L1_error)

  /* "dxfeed/core/DXFeedPy.pyx":264
 *             process_last_error()
 * 
 * def dxf_attach_listener(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Function attaches default listener according to subscription type
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_n_s_sc); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__23, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_dxfeed_core_DXFeedPy_pyx, __pyx_n_s_dxf_attach_listener, 264, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(0, 264, __pyx_L1_error)

  /* "dxfeed/core/DXFeedPy.pyx":323
 *         process_last_error()
 * 
 * def dxf_attach_custom_listener(SubscriptionClass sc, lis.FuncWrapper fw, columns: Iterable[str], data: Iterable = None):             # <<<<<<<<<<<<<<
 *     """
 *     Attaches custom listener
 */
  __pyx_tuple__25 = PyTuple_Pack(4, __pyx_n_s_sc, __pyx_n_s_fw, __pyx_n_s_columns, __pyx_n_s_data); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_codeobj__26 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_dxfeed_core_DXFeedPy_pyx, __pyx_n_s_dxf_attach_custom_listener, 323, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__26)) __PYX_ERR(0, 323, __pyx_L1_error)

  /* "dxfeed/core/DXFeedPy.pyx":347
 *         process_last_error()
 * 
 * def dxf_detach_listener(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Detaches any listener
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_n_s_sc); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_dxfeed_core_DXFeedPy_pyx, __pyx_n_s_dxf_detach_listener, 347, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(0, 347, __pyx_L1_error)

  /* "dxfeed/core/DXFeedPy.pyx":361
 *         process_last_error()
 * 
 * def dxf_close_connection(ConnectionClass cc):             # <<<<<<<<<<<<<<
 *     """
 *     Closes connection
 */
  __pyx_tuple__29 = PyTuple_Pack(2, __pyx_n_s_cc, __pyx_n_s_i); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__29, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_dxfeed_core_DXFeedPy_pyx, __pyx_n_s_dxf_close_connection, 361, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(0, 361, __pyx_L1_error)

  /* "dxfeed/core/DXFeedPy.pyx":384
 *         cc.connection = NULL
 * 
 * def dxf_close_subscription(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Closes subscription
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_n_s_sc); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_dxfeed_core_DXFeedPy_pyx, __pyx_n_s_dxf_close_subscription, 384, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 384, __pyx_L1_error)

  /* "dxfeed/core/DXFeedPy.pyx":397
 *         sc.subscription = NULL
 * 
 * def dxf_get_current_connection_status(ConnectionClass cc, return_str: bool=True):             # <<<<<<<<<<<<<<
 *     """
 *     Returns one of four possible statuses
 */
  __pyx_tuple__33 = PyTuple_Pack(5, __pyx_n_s_cc, __pyx_n_s_return_str, __pyx_n_s_status_mapping, __pyx_n_s_status, __pyx_n_s_result); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_dxfeed_core_DXFeedPy_pyx, __pyx_n_s_dxf_get_current_connection_statu, 397, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 397, __pyx_L1_error)

  /* "dxfeed/core/DXFeedPy.pyx":425
 *     return result
 * 
 * def dxf_get_current_connected_address(ConnectionClass cc):             # <<<<<<<<<<<<<<
 *     """
 *     Returns current connected address
 */
  __pyx_tuple__35 = PyTuple_Pack(2, __pyx_n_s_cc, __pyx_n_s_address); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__35, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_dxfeed_core_DXFeedPy_pyx, __pyx_n_s_dxf_get_current_connected_addres, 425, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 425, __pyx_L1_error)

  /* "dxfeed/core/DXFeedPy.pyx":446
 *     return (<bytes>address).decode('UTF-8')
 * 
 * def dxf_initialize_logger(file_name: str, rewrite_file: bool, show_timezone_info: bool, verbose: bool):             # <<<<<<<<<<<<<<
 *     """
 *     Initializes the internal logger.
 */
  __pyx_tuple__37 = PyTuple_Pack(4, __pyx_n_s_file_name, __pyx_n_s_rewrite_file, __pyx_n_s_show_timezone_info, __pyx_n_s_verbose); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__37, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_dxfeed_core_DXFeedPy_pyx, __pyx_n_s_dxf_initialize_logger, 446, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 446, __pyx_L1_error)

  /* "dxfeed/core/DXFeedPy.pyx":467
 *     clib.dxf_initialize_logger(file_name.encode('UTF-8'), int(rewrite_file), int(show_timezone_info), int(verbose))
 * 
 * def dxf_get_subscription_event_types(SubscriptionClass sc, return_str: bool=True):             # <<<<<<<<<<<<<<
 *     """
 *     Gets subscription event type
 */
  __pyx_tuple__39 = PyTuple_Pack(5, __pyx_n_s_sc, __pyx_n_s_return_str, __pyx_n_s_event_type, __pyx_n_s_et_mapping, __pyx_n_s_result); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_dxfeed_core_DXFeedPy_pyx, __pyx_n_s_dxf_get_subscription_event_types, 467, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(0, 467, __pyx_L1_error)

  /* "dxfeed/core/DXFeedPy.pyx":513
 *     return result
 * 
 * def dxf_get_symbols(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Retrieves the list of symbols currently added to the subscription.
 */
  __pyx_tuple__41 = PyTuple_Pack(5, __pyx_n_s_sc, __pyx_n_s_symbols, __pyx_n_s_symbols_list, __pyx_n_s_symbol_count, __pyx_n_s_i); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_dxfeed_core_DXFeedPy_pyx, __pyx_n_s_dxf_get_symbols, 513, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 513, __pyx_L1_error)

  /* "dxfeed/core/DXFeedPy.pyx":541
 *     return symbols_list
 * 
 * def dxf_remove_symbols(SubscriptionClass sc, symbols: Iterable[str]):             # <<<<<<<<<<<<<<
 *     """
 *     Removes several symbols from the subscription
 */
  __pyx_tuple__43 = PyTuple_Pack(3, __pyx_n_s_sc, __pyx_n_s_symbols, __pyx_n_s_symbol); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_dxfeed_core_DXFeedPy_pyx, __pyx_n_s_dxf_remove_symbols, 541, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 541, __pyx_L1_error)

  /* "dxfeed/core/DXFeedPy.pyx":558
 *         clib.dxf_remove_symbol(sc.subscription, dxf_const_string_t_from_unicode(symbol))
 * 
 * def dxf_clear_symbols(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Removes all symbols from the subscription
 */
  __pyx_tuple__45 = PyTuple_Pack(1, __pyx_n_s_sc); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__45, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_dxfeed_core_DXFeedPy_pyx, __pyx_n_s_dxf_clear_symbols, 558, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_16 = PyInt_FromLong(16); if (unlikely(!__pyx_int_16)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_32 = PyInt_FromLong(32); if (unlikely(!__pyx_int_32)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_64 = PyInt_FromLong(64); if (unlikely(!__pyx_int_64)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_128 = PyInt_FromLong(128); if (unlikely(!__pyx_int_128)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_256 = PyInt_FromLong(256); if (unlikely(!__pyx_int_256)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_512 = PyInt_FromLong(512); if (unlikely(!__pyx_int_512)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1000 = PyInt_FromLong(1000); if (unlikely(!__pyx_int_1000)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1024 = PyInt_FromLong(1024); if (unlikely(!__pyx_int_1024)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1970 = PyInt_FromLong(1970); if (unlikely(!__pyx_int_1970)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2048 = PyInt_FromLong(2048); if (unlikely(!__pyx_int_2048)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4096 = PyInt_FromLong(4096); if (unlikely(!__pyx_int_4096)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5000 = PyInt_FromLong(5000); if (unlikely(!__pyx_int_5000)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8192 = PyInt_FromLong(8192); if (unlikely(!__pyx_int_8192)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_100000 = PyInt_FromLong(100000L); if (unlikely(!__pyx_int_100000)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_16384 = PyInt_FromLong(-16384L); if (unlikely(!__pyx_int_neg_16384)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_6dxfeed_4core_8DXFeedPy_ConnectionClass = &__pyx_vtable_6dxfeed_4core_8DXFeedPy_ConnectionClass;
  __pyx_vtable_6dxfeed_4core_8DXFeedPy_ConnectionClass.make_new_subscription = (struct __pyx_obj_6dxfeed_4core_8DXFeedPy_SubscriptionClass *(*)(struct __pyx_obj_6dxfeed_4core_8DXFeedPy_ConnectionClass *, PyObject *, int __pyx_skip_dispatch))__pyx_f_6dxfeed_4core_8DXFeedPy_15ConnectionClass_make_new_subscription;
  if (PyType_Ready(&__pyx_type_6dxfeed_4core_8DXFeedPy_ConnectionClass) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6dxfeed_4core_8DXFeedPy_ConnectionClass.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6dxfeed_4core_8DXFeedPy_ConnectionClass.tp_dictoffset && __pyx_type_6dxfeed_4core_8DXFeedPy_ConnectionClass.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6dxfeed_4core_8DXFeedPy_ConnectionClass.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6dxfeed_4core_8DXFeedPy_ConnectionClass.tp_dict, __pyx_vtabptr_6dxfeed_4core_8DXFeedPy_ConnectionClass) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ConnectionClass, (PyObject *)&__pyx_type_6dxfeed_4core_8DXFeedPy_ConnectionClass) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6dxfeed_4core_8DXFeedPy_ConnectionClass) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  __pyx_ptype_6dxfeed_4core_8DXFeedPy_ConnectionClass = &__pyx_type_6dxfeed_4core_8DXFeedPy_ConnectionClass;
  if (PyType_Ready(&__pyx_type_6dxfeed_4core_8DXFeedPy_SubscriptionClass) < 0) __PYX_ERR(0, 77, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6dxfeed_4core_8DXFeedPy_SubscriptionClass.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6dxfeed_4core_8DXFeedPy_SubscriptionClass.tp_dictoffset && __pyx_type_6dxfeed_4core_8DXFeedPy_SubscriptionClass.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6dxfeed_4core_8DXFeedPy_SubscriptionClass.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_6dxfeed_4core_8DXFeedPy_SubscriptionClass, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 77, __pyx_L1_error)
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_6dxfeed_4core_8DXFeedPy_17SubscriptionClass___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_6dxfeed_4core_8DXFeedPy_17SubscriptionClass___init__.doc = __pyx_doc_6dxfeed_4core_8DXFeedPy_17SubscriptionClass___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_6dxfeed_4core_8DXFeedPy_17SubscriptionClass___init__;
    }
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SubscriptionClass, (PyObject *)&__pyx_type_6dxfeed_4core_8DXFeedPy_SubscriptionClass) < 0) __PYX_ERR(0, 77, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6dxfeed_4core_8DXFeedPy_SubscriptionClass) < 0) __PYX_ERR(0, 77, __pyx_L1_error)
  __pyx_ptype_6dxfeed_4core_8DXFeedPy_SubscriptionClass = &__pyx_type_6dxfeed_4core_8DXFeedPy_SubscriptionClass;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("dxfeed.core.listeners.listener"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6dxfeed_4core_9listeners_8listener_FuncWrapper = __Pyx_ImportType(__pyx_t_1, "dxfeed.core.listeners.listener", "FuncWrapper", sizeof(struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6dxfeed_4core_9listeners_8listener_FuncWrapper) __PYX_ERR(3, 5, __pyx_L1_error)
  __pyx_vtabptr_6dxfeed_4core_9listeners_8listener_FuncWrapper = (struct __pyx_vtabstruct_6dxfeed_4core_9listeners_8listener_FuncWrapper*)__Pyx_GetVtable(__pyx_ptype_6dxfeed_4core_9listeners_8listener_FuncWrapper->tp_dict); if (unlikely(!__pyx_vtabptr_6dxfeed_4core_9listeners_8listener_FuncWrapper)) __PYX_ERR(3, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("dxfeed.core.utils.helpers"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "unicode_from_dxf_const_string_t", (void (**)(void))&__pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t, "PyObject *(dxf_const_string_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "dxf_const_string_t_from_unicode", (void (**)(void))&__pyx_f_6dxfeed_4core_5utils_7helpers_dxf_const_string_t_from_unicode, "dxf_const_string_t (PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = PyImport_ImportModule("dxfeed.core.listeners.listener"); if (!__pyx_t_2) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "trade_default_listener", (void (**)(void))&__pyx_f_6dxfeed_4core_9listeners_8listener_trade_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "quote_default_listener", (void (**)(void))&__pyx_f_6dxfeed_4core_9listeners_8listener_quote_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "summary_default_listener", (void (**)(void))&__pyx_f_6dxfeed_4core_9listeners_8listener_summary_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "profile_default_listener", (void (**)(void))&__pyx_f_6dxfeed_4core_9listeners_8listener_profile_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "time_and_sale_default_listener", (void (**)(void))&__pyx_f_6dxfeed_4core_9listeners_8listener_time_and_sale_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "candle_default_listener", (void (**)(void))&__pyx_f_6dxfeed_4core_9listeners_8listener_candle_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "order_default_listener", (void (**)(void))&__pyx_f_6dxfeed_4core_9listeners_8listener_order_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "greeks_default_listener", (void (**)(void))&__pyx_f_6dxfeed_4core_9listeners_8listener_greeks_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "theo_price_default_listener", (void (**)(void))&__pyx_f_6dxfeed_4core_9listeners_8listener_theo_price_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "underlying_default_listener", (void (**)(void))&__pyx_f_6dxfeed_4core_9listeners_8listener_underlying_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "series_default_listener", (void (**)(void))&__pyx_f_6dxfeed_4core_9listeners_8listener_series_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "configuration_default_listener", (void (**)(void))&__pyx_f_6dxfeed_4core_9listeners_8listener_configuration_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initDXFeedPy(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initDXFeedPy(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_DXFeedPy(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_DXFeedPy(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_DXFeedPy(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'DXFeedPy' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_DXFeedPy(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("DXFeedPy", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_dxfeed__core__DXFeedPy) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "dxfeed.core.DXFeedPy")) {
      if (unlikely(PyDict_SetItemString(modules, "dxfeed.core.DXFeedPy", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) goto __pyx_L1_error;
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) goto __pyx_L1_error;
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() != 0)) goto __pyx_L1_error;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "dxfeed/core/DXFeedPy.pyx":6
 * 
 * from dxfeed.core.utils.helpers cimport *
 * from dxfeed.core.utils.helpers import *             # <<<<<<<<<<<<<<
 * cimport dxfeed.core.pxd_include.DXFeed as clib
 * cimport dxfeed.core.pxd_include.DXErrorCodes as dxec
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s__14);
  __Pyx_GIVEREF(__pyx_n_s__14);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s__14);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_dxfeed_core_utils_helpers, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_import_star(__pyx_t_2) < 0) __PYX_ERR(0, 6, __pyx_L1_error);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":10
 * cimport dxfeed.core.pxd_include.DXErrorCodes as dxec
 * cimport dxfeed.core.listeners.listener as lis
 * from dxfeed.core.utils.data_class import DequeWithLock as deque_wl             # <<<<<<<<<<<<<<
 * from datetime import datetime
 * import pandas as pd
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_DequeWithLock);
  __Pyx_GIVEREF(__pyx_n_s_DequeWithLock);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_DequeWithLock);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_dxfeed_core_utils_data_class, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_DequeWithLock); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_deque_wl, __pyx_t_2) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":11
 * cimport dxfeed.core.listeners.listener as lis
 * from dxfeed.core.utils.data_class import DequeWithLock as deque_wl
 * from datetime import datetime             # <<<<<<<<<<<<<<
 * import pandas as pd
 * from typing import Optional, Union, Iterable
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_datetime);
  __Pyx_GIVEREF(__pyx_n_s_datetime);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_datetime);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_datetime, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_datetime); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_datetime, __pyx_t_1) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":12
 * from dxfeed.core.utils.data_class import DequeWithLock as deque_wl
 * from datetime import datetime
 * import pandas as pd             # <<<<<<<<<<<<<<
 * from typing import Optional, Union, Iterable
 * from warnings import warn
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pandas, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pd, __pyx_t_2) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":13
 * from datetime import datetime
 * import pandas as pd
 * from typing import Optional, Union, Iterable             # <<<<<<<<<<<<<<
 * from warnings import warn
 * 
 */
  __pyx_t_2 = PyList_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_Optional);
  __Pyx_GIVEREF(__pyx_n_s_Optional);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Optional);
  __Pyx_INCREF(__pyx_n_s_Union);
  __Pyx_GIVEREF(__pyx_n_s_Union);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_Union);
  __Pyx_INCREF(__pyx_n_s_Iterable);
  __Pyx_GIVEREF(__pyx_n_s_Iterable);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_Iterable);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_typing, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Optional); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Optional, __pyx_t_2) < 0) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Union); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Union, __pyx_t_2) < 0) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Iterable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Iterable, __pyx_t_2) < 0) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":14
 * import pandas as pd
 * from typing import Optional, Union, Iterable
 * from warnings import warn             # <<<<<<<<<<<<<<
 * 
 * # for importing variables
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_warn);
  __Pyx_GIVEREF(__pyx_n_s_warn);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_warn);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_warnings, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_warn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_warn, __pyx_t_1) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":17
 * 
 * # for importing variables
 * import dxfeed.core.listeners.listener as lis             # <<<<<<<<<<<<<<
 * from dxfeed.core.pxd_include.EventData cimport *
 * 
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s__14);
  __Pyx_GIVEREF(__pyx_n_s__14);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s__14);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_dxfeed_core_listeners_listener, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_lis, __pyx_t_1) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":147
 *         return df
 * 
 * def dxf_create_connection(address: Union[str, unicode, bytes] = 'demo.dxfeed.com:7300'):             # <<<<<<<<<<<<<<
 *     """
 *     Function creates connection to dxfeed given url address
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6dxfeed_4core_8DXFeedPy_3dxf_create_connection, NULL, __pyx_n_s_dxfeed_core_DXFeedPy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dxf_create_connection, __pyx_t_1) < 0) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":169
 *     return cc
 * 
 * def dxf_create_connection_auth_bearer(address: Union[str, unicode, bytes],             # <<<<<<<<<<<<<<
 *                                      token: Union[str, unicode, bytes]):
 *     """
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6dxfeed_4core_8DXFeedPy_5dxf_create_connection_auth_bearer, NULL, __pyx_n_s_dxfeed_core_DXFeedPy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dxf_create_connection_auth_beare, __pyx_t_1) < 0) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":196
 *     return cc
 * 
 * def dxf_create_subscription(ConnectionClass cc, event_type: str, candle_time: Optional[str] = None,             # <<<<<<<<<<<<<<
 *                             data_len: int = 100000):
 *     """
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6dxfeed_4core_8DXFeedPy_7dxf_create_subscription, NULL, __pyx_n_s_dxfeed_core_DXFeedPy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dxf_create_subscription, __pyx_t_1) < 0) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":244
 *     return sc
 * 
 * def dxf_add_symbols(SubscriptionClass sc, symbols: Iterable[str]):             # <<<<<<<<<<<<<<
 *     """
 *     Adds symbols to subscription
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6dxfeed_4core_8DXFeedPy_9dxf_add_symbols, NULL, __pyx_n_s_dxfeed_core_DXFeedPy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dxf_add_symbols, __pyx_t_1) < 0) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":264
 *             process_last_error()
 * 
 * def dxf_attach_listener(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Function attaches default listener according to subscription type
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6dxfeed_4core_8DXFeedPy_11dxf_attach_listener, NULL, __pyx_n_s_dxfeed_core_DXFeedPy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dxf_attach_listener, __pyx_t_1) < 0) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":323
 *         process_last_error()
 * 
 * def dxf_attach_custom_listener(SubscriptionClass sc, lis.FuncWrapper fw, columns: Iterable[str], data: Iterable = None):             # <<<<<<<<<<<<<<
 *     """
 *     Attaches custom listener
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6dxfeed_4core_8DXFeedPy_13dxf_attach_custom_listener, NULL, __pyx_n_s_dxfeed_core_DXFeedPy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dxf_attach_custom_listener, __pyx_t_1) < 0) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":347
 *         process_last_error()
 * 
 * def dxf_detach_listener(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Detaches any listener
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6dxfeed_4core_8DXFeedPy_15dxf_detach_listener, NULL, __pyx_n_s_dxfeed_core_DXFeedPy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dxf_detach_listener, __pyx_t_1) < 0) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":361
 *         process_last_error()
 * 
 * def dxf_close_connection(ConnectionClass cc):             # <<<<<<<<<<<<<<
 *     """
 *     Closes connection
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6dxfeed_4core_8DXFeedPy_17dxf_close_connection, NULL, __pyx_n_s_dxfeed_core_DXFeedPy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dxf_close_connection, __pyx_t_1) < 0) __PYX_ERR(0, 361, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":384
 *         cc.connection = NULL
 * 
 * def dxf_close_subscription(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Closes subscription
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6dxfeed_4core_8DXFeedPy_19dxf_close_subscription, NULL, __pyx_n_s_dxfeed_core_DXFeedPy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dxf_close_subscription, __pyx_t_1) < 0) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":397
 *         sc.subscription = NULL
 * 
 * def dxf_get_current_connection_status(ConnectionClass cc, return_str: bool=True):             # <<<<<<<<<<<<<<
 *     """
 *     Returns one of four possible statuses
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6dxfeed_4core_8DXFeedPy_21dxf_get_current_connection_status, NULL, __pyx_n_s_dxfeed_core_DXFeedPy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dxf_get_current_connection_statu, __pyx_t_1) < 0) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":425
 *     return result
 * 
 * def dxf_get_current_connected_address(ConnectionClass cc):             # <<<<<<<<<<<<<<
 *     """
 *     Returns current connected address
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6dxfeed_4core_8DXFeedPy_23dxf_get_current_connected_address, NULL, __pyx_n_s_dxfeed_core_DXFeedPy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dxf_get_current_connected_addres, __pyx_t_1) < 0) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":446
 *     return (<bytes>address).decode('UTF-8')
 * 
 * def dxf_initialize_logger(file_name: str, rewrite_file: bool, show_timezone_info: bool, verbose: bool):             # <<<<<<<<<<<<<<
 *     """
 *     Initializes the internal logger.
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6dxfeed_4core_8DXFeedPy_25dxf_initialize_logger, NULL, __pyx_n_s_dxfeed_core_DXFeedPy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dxf_initialize_logger, __pyx_t_1) < 0) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":467
 *     clib.dxf_initialize_logger(file_name.encode('UTF-8'), int(rewrite_file), int(show_timezone_info), int(verbose))
 * 
 * def dxf_get_subscription_event_types(SubscriptionClass sc, return_str: bool=True):             # <<<<<<<<<<<<<<
 *     """
 *     Gets subscription event type
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6dxfeed_4core_8DXFeedPy_27dxf_get_subscription_event_types, NULL, __pyx_n_s_dxfeed_core_DXFeedPy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dxf_get_subscription_event_types, __pyx_t_1) < 0) __PYX_ERR(0, 467, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":513
 *     return result
 * 
 * def dxf_get_symbols(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Retrieves the list of symbols currently added to the subscription.
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6dxfeed_4core_8DXFeedPy_29dxf_get_symbols, NULL, __pyx_n_s_dxfeed_core_DXFeedPy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dxf_get_symbols, __pyx_t_1) < 0) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":541
 *     return symbols_list
 * 
 * def dxf_remove_symbols(SubscriptionClass sc, symbols: Iterable[str]):             # <<<<<<<<<<<<<<
 *     """
 *     Removes several symbols from the subscription
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6dxfeed_4core_8DXFeedPy_31dxf_remove_symbols, NULL, __pyx_n_s_dxfeed_core_DXFeedPy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dxf_remove_symbols, __pyx_t_1) < 0) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":558
 *         clib.dxf_remove_symbol(sc.subscription, dxf_const_string_t_from_unicode(symbol))
 * 
 * def dxf_clear_symbols(SubscriptionClass sc):             # <<<<<<<<<<<<<<
 *     """
 *     Removes all symbols from the subscription
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6dxfeed_4core_8DXFeedPy_33dxf_clear_symbols, NULL, __pyx_n_s_dxfeed_core_DXFeedPy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dxf_clear_symbols, __pyx_t_1) < 0) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/DXFeedPy.pyx":1
 * # distutils: language = c++             # <<<<<<<<<<<<<<
 * # cython: always_allow_keywords=True
 * from libcpp.deque cimport deque as cppdq
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init dxfeed.core.DXFeedPy", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init dxfeed.core.DXFeedPy");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* CIntToDigits */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
#if CYTHON_PEP393_ENABLED
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
#else
    Py_UNICODE *udata;
    uval = PyUnicode_FromUnicode(NULL, ulength);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_AS_UNICODE(uval);
#endif
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp;
                    PyObject *repeat = PyInt_FromSize_t(uoffset - prepend_sign);
                    if (unlikely(!repeat)) goto done_or_error;
                    tmp = PyNumber_Multiply(padding, repeat);
                    Py_DECREF(repeat);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyNumber_Add(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyNumber_Add(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* CIntToPyUnicode */
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned short    uint16_t;
        #else
           typedef unsigned __int16  uint16_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define GCC_DIAGNOSTIC
#endif
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(int)*3+2];
    char *dpos, *end = digits + sizeof(int)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    int remaining;
#ifdef GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (int) (remaining / (8*8));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_8)[digit_pos];
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (int) (remaining / (10*10));
            dpos -= 2;
            *(uint16_t*)dpos = ((const uint16_t*)DIGIT_PAIRS_10)[digit_pos];
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (int) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    if (last_one_off) {
        assert(*dpos == '0');
        dpos++;
    }
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      CYTHON_UNUSED Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind;
    Py_ssize_t i, char_pos;
    void *result_udata;
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely(char_pos + ulength < 0))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + char_pos * result_ukind, udata, (size_t) (ulength * result_ukind));
        } else {
            #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    result_ulength++;
    value_count++;
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* decode_c_bytes */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    if (unlikely((start < 0) | (stop < 0))) {
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (stop > length)
        stop = length;
    length = stop - start;
    if (unlikely(length <= 0))
        return PyUnicode_FromUnicode(NULL, 0);
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto __PYX_BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto __PYX_BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_dxf_connection_status_t(dxf_connection_status_t value) {
    const dxf_connection_status_t neg_one = (dxf_connection_status_t) ((dxf_connection_status_t) 0 - (dxf_connection_status_t) 1), const_zero = (dxf_connection_status_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(dxf_connection_status_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(dxf_connection_status_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(dxf_connection_status_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(dxf_connection_status_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(dxf_connection_status_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(dxf_connection_status_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE dxf_long_t __Pyx_PyInt_As_dxf_long_t(PyObject *x) {
    const dxf_long_t neg_one = (dxf_long_t) ((dxf_long_t) 0 - (dxf_long_t) 1), const_zero = (dxf_long_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(dxf_long_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(dxf_long_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (dxf_long_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (dxf_long_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(dxf_long_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(dxf_long_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(dxf_long_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(dxf_long_t) >= 2 * PyLong_SHIFT) {
                            return (dxf_long_t) (((((dxf_long_t)digits[1]) << PyLong_SHIFT) | (dxf_long_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(dxf_long_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(dxf_long_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(dxf_long_t) >= 3 * PyLong_SHIFT) {
                            return (dxf_long_t) (((((((dxf_long_t)digits[2]) << PyLong_SHIFT) | (dxf_long_t)digits[1]) << PyLong_SHIFT) | (dxf_long_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(dxf_long_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(dxf_long_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(dxf_long_t) >= 4 * PyLong_SHIFT) {
                            return (dxf_long_t) (((((((((dxf_long_t)digits[3]) << PyLong_SHIFT) | (dxf_long_t)digits[2]) << PyLong_SHIFT) | (dxf_long_t)digits[1]) << PyLong_SHIFT) | (dxf_long_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (dxf_long_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(dxf_long_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(dxf_long_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(dxf_long_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(dxf_long_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (dxf_long_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(dxf_long_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(dxf_long_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(dxf_long_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(dxf_long_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(dxf_long_t) - 1 > 2 * PyLong_SHIFT) {
                            return (dxf_long_t) (((dxf_long_t)-1)*(((((dxf_long_t)digits[1]) << PyLong_SHIFT) | (dxf_long_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(dxf_long_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(dxf_long_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(dxf_long_t) - 1 > 2 * PyLong_SHIFT) {
                            return (dxf_long_t) ((((((dxf_long_t)digits[1]) << PyLong_SHIFT) | (dxf_long_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(dxf_long_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(dxf_long_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(dxf_long_t) - 1 > 3 * PyLong_SHIFT) {
                            return (dxf_long_t) (((dxf_long_t)-1)*(((((((dxf_long_t)digits[2]) << PyLong_SHIFT) | (dxf_long_t)digits[1]) << PyLong_SHIFT) | (dxf_long_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(dxf_long_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(dxf_long_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(dxf_long_t) - 1 > 3 * PyLong_SHIFT) {
                            return (dxf_long_t) ((((((((dxf_long_t)digits[2]) << PyLong_SHIFT) | (dxf_long_t)digits[1]) << PyLong_SHIFT) | (dxf_long_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(dxf_long_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(dxf_long_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(dxf_long_t) - 1 > 4 * PyLong_SHIFT) {
                            return (dxf_long_t) (((dxf_long_t)-1)*(((((((((dxf_long_t)digits[3]) << PyLong_SHIFT) | (dxf_long_t)digits[2]) << PyLong_SHIFT) | (dxf_long_t)digits[1]) << PyLong_SHIFT) | (dxf_long_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(dxf_long_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(dxf_long_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(dxf_long_t) - 1 > 4 * PyLong_SHIFT) {
                            return (dxf_long_t) ((((((((((dxf_long_t)digits[3]) << PyLong_SHIFT) | (dxf_long_t)digits[2]) << PyLong_SHIFT) | (dxf_long_t)digits[1]) << PyLong_SHIFT) | (dxf_long_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(dxf_long_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(dxf_long_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(dxf_long_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(dxf_long_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            dxf_long_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (dxf_long_t) -1;
        }
    } else {
        dxf_long_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (dxf_long_t) -1;
        val = __Pyx_PyInt_As_dxf_long_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to dxf_long_t");
    return (dxf_long_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to dxf_long_t");
    return (dxf_long_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) ((size_t) 0 - (size_t) 1), const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CStringEquals */
static CYTHON_INLINE int __Pyx_StrEq(const char *s1, const char *s2) {
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    return *s1 == *s2;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
