/* Generated by Cython 0.29.16 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "name": "dxfeed.core.listeners.listener",
        "sources": [
            "dxfeed/core/listeners/listener.pyx"
        ]
    },
    "module_name": "dxfeed.core.listeners.listener"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_16"
#define CYTHON_HEX_VERSION 0x001D10F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__dxfeed__core__listeners__listener
#define __PYX_HAVE_API__dxfeed__core__listeners__listener
/* Early includes */
#include "DXTypes.h"
#include <stddef.h>
#include "RecordData.h"
#include "EventData.h"
#include <string.h>
#include <stdio.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "stringsource",
  "dxfeed\\core\\listeners\\listener.pyx",
  "type.pxd",
};
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()


/*--- Type declarations ---*/
struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper;

/* "dxfeed/core/listeners/listener.pxd":5
 * 
 * 
 * cdef class FuncWrapper:             # <<<<<<<<<<<<<<
 * 	cdef dxf_event_listener_t func
 * 	@staticmethod
 */
struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper {
  PyObject_HEAD
  struct __pyx_vtabstruct_6dxfeed_4core_9listeners_8listener_FuncWrapper *__pyx_vtab;
  dxf_event_listener_t func;
};



/* "dxfeed/core/listeners/listener.pyx":3
 * from dxfeed.core.utils.helpers cimport *
 * 
 * cdef class FuncWrapper:             # <<<<<<<<<<<<<<
 *     def __cinit__(self):
 *         self.func = NULL
 */

struct __pyx_vtabstruct_6dxfeed_4core_9listeners_8listener_FuncWrapper {
  struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *(*make_from_ptr)(dxf_event_listener_t);
};
static struct __pyx_vtabstruct_6dxfeed_4core_9listeners_8listener_FuncWrapper *__pyx_vtabptr_6dxfeed_4core_9listeners_8listener_FuncWrapper;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_wchar_t(wchar_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_dxf_int_t(dxf_int_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_dxf_long_t(dxf_long_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_dxf_bool_t(dxf_bool_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_dxf_dayid_t(dxf_dayid_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_dxf_uint_t(dxf_uint_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_dxf_order_side_t(dxf_order_side_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_dxf_tns_type_t(dxf_tns_type_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_dxf_order_scope_t(dxf_order_scope_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *__pyx_f_6dxfeed_4core_9listeners_8listener_11FuncWrapper_make_from_ptr(dxf_event_listener_t __pyx_v_f); /* proto*/

/* Module declarations from 'libc.stddef' */

/* Module declarations from 'dxfeed.core.pxd_include.DXTypes' */

/* Module declarations from 'dxfeed.core.pxd_include.RecordData' */

/* Module declarations from 'dxfeed.core.pxd_include.EventData' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'dxfeed.core.utils.helpers' */
static PyObject *(*__pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t)(dxf_const_string_t); /*proto*/

/* Module declarations from 'dxfeed.core.listeners.listener' */
static PyTypeObject *__pyx_ptype_6dxfeed_4core_9listeners_8listener_FuncWrapper = 0;
#define __Pyx_MODULE_NAME "dxfeed.core.listeners.listener"
extern int __pyx_module_is_main_dxfeed__core__listeners__listener;
int __pyx_module_is_main_dxfeed__core__listeners__listener = 0;

/* Implementation of 'dxfeed.core.listeners.listener' */
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_range;
static const char __pyx_k_EPS[] = "EPS";
static const char __pyx_k_Low[] = "Low";
static const char __pyx_k_Rho[] = "Rho";
static const char __pyx_k_Beta[] = "Beta";
static const char __pyx_k_High[] = "High";
static const char __pyx_k_Open[] = "Open";
static const char __pyx_k_Side[] = "Side";
static const char __pyx_k_Size[] = "Size";
static const char __pyx_k_Tick[] = "Tick";
static const char __pyx_k_Time[] = "Time";
static const char __pyx_k_Type[] = "Type";
static const char __pyx_k_VWap[] = "VWap";
static const char __pyx_k_Vega[] = "Vega";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_Buyer[] = "Buyer";
static const char __pyx_k_Close[] = "Close";
static const char __pyx_k_Count[] = "Count";
static const char __pyx_k_DayId[] = "DayId";
static const char __pyx_k_Delta[] = "Delta";
static const char __pyx_k_Gamma[] = "Gamma";
static const char __pyx_k_Index[] = "Index";
static const char __pyx_k_IsETH[] = "IsETH";
static const char __pyx_k_Price[] = "Price";
static const char __pyx_k_Scope[] = "Scope";
static const char __pyx_k_Theta[] = "Theta";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_Change[] = "Change";
static const char __pyx_k_Object[] = "Object";
static const char __pyx_k_Seller[] = "Seller";
static const char __pyx_k_Shares[] = "Shares";
static const char __pyx_k_Symbol[] = "Symbol";
static const char __pyx_k_Volume[] = "Volume";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_AskSize[] = "AskSize";
static const char __pyx_k_AskTime[] = "AskTime";
static const char __pyx_k_BidSize[] = "BidSize";
static const char __pyx_k_BidTime[] = "BidTime";
static const char __pyx_k_DivFreq[] = "DivFreq";
static const char __pyx_k_Version[] = "Version";
static const char __pyx_k_AskPrice[] = "AskPrice";
static const char __pyx_k_BidPrice[] = "BidPrice";
static const char __pyx_k_Dividend[] = "Dividend";
static const char __pyx_k_Interest[] = "Interest";
static const char __pyx_k_RawFlags[] = "RawFlags";
static const char __pyx_k_Sequence[] = "Sequence";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_AskVolume[] = "AskVolume";
static const char __pyx_k_BidVolume[] = "BidVolume";
static const char __pyx_k_DayVolume[] = "DayVolume";
static const char __pyx_k_PrevDayId[] = "PrevDayId";
static const char __pyx_k_TimeNanos[] = "TimeNanos";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_52LowPrice[] = "52LowPrice";
static const char __pyx_k_EventFlags[] = "EventFlags";
static const char __pyx_k_ExdDivDate[] = "ExdDivDate";
static const char __pyx_k_Expiration[] = "Expiration";
static const char __pyx_k_IsEthTrade[] = "IsEthTrade";
static const char __pyx_k_Volatility[] = "Volatility";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_52HighPrice[] = "52HighPrice";
static const char __pyx_k_DayLowPrice[] = "DayLowPrice";
static const char __pyx_k_Description[] = "Description";
static const char __pyx_k_FuncWrapper[] = "FuncWrapper";
static const char __pyx_k_IsSpreadLeg[] = "IsSpreadLeg";
static const char __pyx_k_IsValidTick[] = "IsValidTick";
static const char __pyx_k_MarketMaker[] = "MarketMaker";
static const char __pyx_k_safe_append[] = "safe_append";
static const char __pyx_k_DayHighPrice[] = "DayHighPrice";
static const char __pyx_k_ExchangeCode[] = "ExchangeCode";
static const char __pyx_k_ExdDivAmount[] = "ExdDivAmount";
static const char __pyx_k_ForwardPrice[] = "ForwardPrice";
static const char __pyx_k_OpenInterest[] = "OpenInterest";
static const char __pyx_k_PutCallRatio[] = "PutCallRatio";
static const char __pyx_k_SpreadSymbol[] = "SpreadSymbol";
static const char __pyx_k_StatusReason[] = "StatusReason";
static const char __pyx_k_DayClosePrice[] = "DayClosePrice";
static const char __pyx_k_ImpVolatility[] = "ImpVolatility";
static const char __pyx_k_ORDER_COLUMNS[] = "ORDER_COLUMNS";
static const char __pyx_k_PrevDayVolume[] = "PrevDayVolume";
static const char __pyx_k_QUOTE_COLUMNS[] = "QUOTE_COLUMNS";
static const char __pyx_k_TRADE_COLUMNS[] = "TRADE_COLUMNS";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_BackVolatility[] = "BackVolatility";
static const char __pyx_k_CANDLE_COLUMNS[] = "CANDLE_COLUMNS";
static const char __pyx_k_GREEKS_COLUMNS[] = "GREEKS_COLUMNS";
static const char __pyx_k_SERIES_COLUMNS[] = "SERIES_COLUMNS";
static const char __pyx_k_AskExchangeCode[] = "AskExchangeCode";
static const char __pyx_k_BidExchangeCode[] = "BidExchangeCode";
static const char __pyx_k_FrontVolatility[] = "FrontVolatility";
static const char __pyx_k_PROFILE_COLUMNS[] = "PROFILE_COLUMNS";
static const char __pyx_k_SUMMARY_COLUMNS[] = "SUMMARY_COLUMNS";
static const char __pyx_k_UnderlyingPrice[] = "UnderlyingPrice";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_PrevDayClosePrice[] = "PrevDayClosePrice";
static const char __pyx_k_THEO_PRICE_COLUMNS[] = "THEO_PRICE_COLUMNS";
static const char __pyx_k_TradeThroughExempt[] = "TradeThroughExempt";
static const char __pyx_k_UNDERLYING_COLUMNS[] = "UNDERLYING_COLUMNS";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_CONFIGURATION_COLUMNS[] = "CONFIGURATION_COLUMNS";
static const char __pyx_k_TIME_AND_SALE_COLUMNS[] = "TIME_AND_SALE_COLUMNS";
static const char __pyx_k_ExchangeSaleConditions[] = "ExchangeSaleConditions";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static PyObject *__pyx_kp_u_52HighPrice;
static PyObject *__pyx_kp_u_52LowPrice;
static PyObject *__pyx_n_u_AskExchangeCode;
static PyObject *__pyx_n_u_AskPrice;
static PyObject *__pyx_n_u_AskSize;
static PyObject *__pyx_n_u_AskTime;
static PyObject *__pyx_n_u_AskVolume;
static PyObject *__pyx_n_u_BackVolatility;
static PyObject *__pyx_n_u_Beta;
static PyObject *__pyx_n_u_BidExchangeCode;
static PyObject *__pyx_n_u_BidPrice;
static PyObject *__pyx_n_u_BidSize;
static PyObject *__pyx_n_u_BidTime;
static PyObject *__pyx_n_u_BidVolume;
static PyObject *__pyx_n_u_Buyer;
static PyObject *__pyx_n_s_CANDLE_COLUMNS;
static PyObject *__pyx_n_s_CONFIGURATION_COLUMNS;
static PyObject *__pyx_n_u_Change;
static PyObject *__pyx_n_u_Close;
static PyObject *__pyx_n_u_Count;
static PyObject *__pyx_n_u_DayClosePrice;
static PyObject *__pyx_n_u_DayHighPrice;
static PyObject *__pyx_n_u_DayId;
static PyObject *__pyx_n_u_DayLowPrice;
static PyObject *__pyx_n_u_DayVolume;
static PyObject *__pyx_n_u_Delta;
static PyObject *__pyx_n_u_Description;
static PyObject *__pyx_n_u_DivFreq;
static PyObject *__pyx_n_u_Dividend;
static PyObject *__pyx_n_u_EPS;
static PyObject *__pyx_n_u_EventFlags;
static PyObject *__pyx_n_u_ExchangeCode;
static PyObject *__pyx_n_u_ExchangeSaleConditions;
static PyObject *__pyx_n_u_ExdDivAmount;
static PyObject *__pyx_n_u_ExdDivDate;
static PyObject *__pyx_n_u_Expiration;
static PyObject *__pyx_n_u_ForwardPrice;
static PyObject *__pyx_n_u_FrontVolatility;
static PyObject *__pyx_n_s_FuncWrapper;
static PyObject *__pyx_n_s_GREEKS_COLUMNS;
static PyObject *__pyx_n_u_Gamma;
static PyObject *__pyx_n_u_High;
static PyObject *__pyx_n_u_ImpVolatility;
static PyObject *__pyx_n_u_Index;
static PyObject *__pyx_n_u_Interest;
static PyObject *__pyx_n_u_IsETH;
static PyObject *__pyx_n_u_IsEthTrade;
static PyObject *__pyx_n_u_IsSpreadLeg;
static PyObject *__pyx_n_u_IsValidTick;
static PyObject *__pyx_n_u_Low;
static PyObject *__pyx_n_u_MarketMaker;
static PyObject *__pyx_n_s_ORDER_COLUMNS;
static PyObject *__pyx_n_u_Object;
static PyObject *__pyx_n_u_Open;
static PyObject *__pyx_n_u_OpenInterest;
static PyObject *__pyx_n_s_PROFILE_COLUMNS;
static PyObject *__pyx_n_u_PrevDayClosePrice;
static PyObject *__pyx_n_u_PrevDayId;
static PyObject *__pyx_n_u_PrevDayVolume;
static PyObject *__pyx_n_u_Price;
static PyObject *__pyx_n_u_PutCallRatio;
static PyObject *__pyx_n_s_QUOTE_COLUMNS;
static PyObject *__pyx_n_u_RawFlags;
static PyObject *__pyx_n_u_Rho;
static PyObject *__pyx_n_s_SERIES_COLUMNS;
static PyObject *__pyx_n_s_SUMMARY_COLUMNS;
static PyObject *__pyx_n_u_Scope;
static PyObject *__pyx_n_u_Seller;
static PyObject *__pyx_n_u_Sequence;
static PyObject *__pyx_n_u_Shares;
static PyObject *__pyx_n_u_Side;
static PyObject *__pyx_n_u_Size;
static PyObject *__pyx_n_u_SpreadSymbol;
static PyObject *__pyx_n_u_StatusReason;
static PyObject *__pyx_n_u_Symbol;
static PyObject *__pyx_n_s_THEO_PRICE_COLUMNS;
static PyObject *__pyx_n_s_TIME_AND_SALE_COLUMNS;
static PyObject *__pyx_n_s_TRADE_COLUMNS;
static PyObject *__pyx_n_u_Theta;
static PyObject *__pyx_n_u_Tick;
static PyObject *__pyx_n_u_Time;
static PyObject *__pyx_n_u_TimeNanos;
static PyObject *__pyx_n_u_TradeThroughExempt;
static PyObject *__pyx_n_u_Type;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_UNDERLYING_COLUMNS;
static PyObject *__pyx_n_u_UnderlyingPrice;
static PyObject *__pyx_n_u_VWap;
static PyObject *__pyx_n_u_Vega;
static PyObject *__pyx_n_u_Version;
static PyObject *__pyx_n_u_Volatility;
static PyObject *__pyx_n_u_Volume;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_safe_append;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_test;
static int __pyx_pf_6dxfeed_4core_9listeners_8listener_11FuncWrapper___cinit__(struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_9listeners_8listener_11FuncWrapper_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6dxfeed_4core_9listeners_8listener_11FuncWrapper_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_6dxfeed_4core_9listeners_8listener_FuncWrapper(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
/* Late includes */

/* "dxfeed/core/listeners/listener.pyx":4
 * 
 * cdef class FuncWrapper:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.func = NULL
 * 
 */

/* Python wrapper */
static int __pyx_pw_6dxfeed_4core_9listeners_8listener_11FuncWrapper_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6dxfeed_4core_9listeners_8listener_11FuncWrapper_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_6dxfeed_4core_9listeners_8listener_11FuncWrapper___cinit__(((struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6dxfeed_4core_9listeners_8listener_11FuncWrapper___cinit__(struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "dxfeed/core/listeners/listener.pyx":5
 * cdef class FuncWrapper:
 *     def __cinit__(self):
 *         self.func = NULL             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_v_self->func = NULL;

  /* "dxfeed/core/listeners/listener.pyx":4
 * 
 * cdef class FuncWrapper:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.func = NULL
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/listeners/listener.pyx":8
 * 
 *     @staticmethod
 *     cdef FuncWrapper make_from_ptr(dxf_event_listener_t f):             # <<<<<<<<<<<<<<
 *         cdef FuncWrapper out = FuncWrapper()
 *         out.func = f
 */

static struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *__pyx_f_6dxfeed_4core_9listeners_8listener_11FuncWrapper_make_from_ptr(dxf_event_listener_t __pyx_v_f) {
  struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *__pyx_v_out = 0;
  struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("make_from_ptr", 0);

  /* "dxfeed/core/listeners/listener.pyx":9
 *     @staticmethod
 *     cdef FuncWrapper make_from_ptr(dxf_event_listener_t f):
 *         cdef FuncWrapper out = FuncWrapper()             # <<<<<<<<<<<<<<
 *         out.func = f
 *         return out
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6dxfeed_4core_9listeners_8listener_FuncWrapper)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_out = ((struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "dxfeed/core/listeners/listener.pyx":10
 *     cdef FuncWrapper make_from_ptr(dxf_event_listener_t f):
 *         cdef FuncWrapper out = FuncWrapper()
 *         out.func = f             # <<<<<<<<<<<<<<
 *         return out
 * 
 */
  __pyx_v_out->func = __pyx_v_f;

  /* "dxfeed/core/listeners/listener.pyx":11
 *         cdef FuncWrapper out = FuncWrapper()
 *         out.func = f
 *         return out             # <<<<<<<<<<<<<<
 * 
 * TRADE_COLUMNS = ['Symbol', 'Price', 'ExchangeCode', 'Size', 'Tick', 'Change', 'DayVolume', 'Time', 'IsETH']
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_out));
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "dxfeed/core/listeners/listener.pyx":8
 * 
 *     @staticmethod
 *     cdef FuncWrapper make_from_ptr(dxf_event_listener_t f):             # <<<<<<<<<<<<<<
 *         cdef FuncWrapper out = FuncWrapper()
 *         out.func = f
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("dxfeed.core.listeners.listener.FuncWrapper.make_from_ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_out);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_9listeners_8listener_11FuncWrapper_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6dxfeed_4core_9listeners_8listener_11FuncWrapper_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6dxfeed_4core_9listeners_8listener_11FuncWrapper_2__reduce_cython__(((struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_9listeners_8listener_11FuncWrapper_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("dxfeed.core.listeners.listener.FuncWrapper.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6dxfeed_4core_9listeners_8listener_11FuncWrapper_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6dxfeed_4core_9listeners_8listener_11FuncWrapper_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6dxfeed_4core_9listeners_8listener_11FuncWrapper_4__setstate_cython__(((struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6dxfeed_4core_9listeners_8listener_11FuncWrapper_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("dxfeed.core.listeners.listener.FuncWrapper.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "dxfeed/core/listeners/listener.pyx":14
 * 
 * TRADE_COLUMNS = ['Symbol', 'Price', 'ExchangeCode', 'Size', 'Tick', 'Change', 'DayVolume', 'Time', 'IsETH']
 * cdef void trade_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                  dxf_const_string_t symbol_name,
 *                                  const dxf_event_data_t*data,
 */

static void __pyx_f_6dxfeed_4core_9listeners_8listener_trade_default_listener(CYTHON_UNUSED int __pyx_v_event_type, dxf_const_string_t __pyx_v_symbol_name, dxf_event_data_t const *__pyx_v_data, int __pyx_v_data_count, void *__pyx_v_user_data) {
  dxf_trade_t *__pyx_v_trades;
  PyObject *__pyx_v_py_data = NULL;
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("trade_default_listener", 1);

  /* "dxfeed/core/listeners/listener.pyx":18
 *                                  const dxf_event_data_t*data,
 *                                  int data_count, void*user_data) nogil:
 *     cdef dxf_trade_t*trades = <dxf_trade_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*try:*/ {
    __pyx_v_trades = ((dxf_trade_t *)__pyx_v_data);

    /* "dxfeed/core/listeners/listener.pyx":19
 *                                  int data_count, void*user_data) nogil:
 *     cdef dxf_trade_t*trades = <dxf_trade_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 * 
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "dxfeed/core/listeners/listener.pyx":20
 *     cdef dxf_trade_t*trades = <dxf_trade_t*> data
 *     with gil:
 *         py_data = <object> user_data             # <<<<<<<<<<<<<<
 * 
 *         for i in range(data_count):
 */
          __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
          __Pyx_INCREF(__pyx_t_1);
          __pyx_v_py_data = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "dxfeed/core/listeners/listener.pyx":22
 *         py_data = <object> user_data
 * 
 *         for i in range(data_count):             # <<<<<<<<<<<<<<
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  trades[i].price,
 */
          __pyx_t_2 = __pyx_v_data_count;
          __pyx_t_3 = __pyx_t_2;
          for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
            __pyx_v_i = __pyx_t_4;

            /* "dxfeed/core/listeners/listener.pyx":23
 * 
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  trades[i].price,
 *                                  trades[i].exchange_code,
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_data, __pyx_n_s_safe_append); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 23, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_6 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t(__pyx_v_symbol_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 23, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);

            /* "dxfeed/core/listeners/listener.pyx":24
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  trades[i].price,             # <<<<<<<<<<<<<<
 *                                  trades[i].exchange_code,
 *                                  trades[i].size,
 */
            __pyx_t_7 = PyFloat_FromDouble((__pyx_v_trades[__pyx_v_i]).price); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 24, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);

            /* "dxfeed/core/listeners/listener.pyx":25
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  trades[i].price,
 *                                  trades[i].exchange_code,             # <<<<<<<<<<<<<<
 *                                  trades[i].size,
 *                                  trades[i].tick,
 */
            __pyx_t_8 = __Pyx_PyInt_From_wchar_t((__pyx_v_trades[__pyx_v_i]).exchange_code); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 25, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_8);

            /* "dxfeed/core/listeners/listener.pyx":26
 *                                  trades[i].price,
 *                                  trades[i].exchange_code,
 *                                  trades[i].size,             # <<<<<<<<<<<<<<
 *                                  trades[i].tick,
 *                                  trades[i].change,
 */
            __pyx_t_9 = __Pyx_PyInt_From_dxf_int_t((__pyx_v_trades[__pyx_v_i]).size); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 26, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_9);

            /* "dxfeed/core/listeners/listener.pyx":27
 *                                  trades[i].exchange_code,
 *                                  trades[i].size,
 *                                  trades[i].tick,             # <<<<<<<<<<<<<<
 *                                  trades[i].change,
 *                                  trades[i].day_volume,
 */
            __pyx_t_10 = __Pyx_PyInt_From_dxf_int_t((__pyx_v_trades[__pyx_v_i]).tick); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 27, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_10);

            /* "dxfeed/core/listeners/listener.pyx":28
 *                                  trades[i].size,
 *                                  trades[i].tick,
 *                                  trades[i].change,             # <<<<<<<<<<<<<<
 *                                  trades[i].day_volume,
 *                                  trades[i].time,
 */
            __pyx_t_11 = PyFloat_FromDouble((__pyx_v_trades[__pyx_v_i]).change); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 28, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_11);

            /* "dxfeed/core/listeners/listener.pyx":29
 *                                  trades[i].tick,
 *                                  trades[i].change,
 *                                  trades[i].day_volume,             # <<<<<<<<<<<<<<
 *                                  trades[i].time,
 *                                  trades[i].is_eth])
 */
            __pyx_t_12 = PyFloat_FromDouble((__pyx_v_trades[__pyx_v_i]).day_volume); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 29, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);

            /* "dxfeed/core/listeners/listener.pyx":30
 *                                  trades[i].change,
 *                                  trades[i].day_volume,
 *                                  trades[i].time,             # <<<<<<<<<<<<<<
 *                                  trades[i].is_eth])
 * 
 */
            __pyx_t_13 = __Pyx_PyInt_From_dxf_long_t((__pyx_v_trades[__pyx_v_i]).time); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 30, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);

            /* "dxfeed/core/listeners/listener.pyx":31
 *                                  trades[i].day_volume,
 *                                  trades[i].time,
 *                                  trades[i].is_eth])             # <<<<<<<<<<<<<<
 * 
 * QUOTE_COLUMNS = ['Symbol', 'BidTime', 'BidExchangeCode', 'BidPrice', 'BidSize', 'AskTime', 'AskExchangeCode',
 */
            __pyx_t_14 = __Pyx_PyInt_From_dxf_bool_t((__pyx_v_trades[__pyx_v_i]).is_eth); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 31, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_14);

            /* "dxfeed/core/listeners/listener.pyx":23
 * 
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  trades[i].price,
 *                                  trades[i].exchange_code,
 */
            __pyx_t_15 = PyList_New(9); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_GIVEREF(__pyx_t_6);
            PyList_SET_ITEM(__pyx_t_15, 0, __pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_7);
            PyList_SET_ITEM(__pyx_t_15, 1, __pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_8);
            PyList_SET_ITEM(__pyx_t_15, 2, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_9);
            PyList_SET_ITEM(__pyx_t_15, 3, __pyx_t_9);
            __Pyx_GIVEREF(__pyx_t_10);
            PyList_SET_ITEM(__pyx_t_15, 4, __pyx_t_10);
            __Pyx_GIVEREF(__pyx_t_11);
            PyList_SET_ITEM(__pyx_t_15, 5, __pyx_t_11);
            __Pyx_GIVEREF(__pyx_t_12);
            PyList_SET_ITEM(__pyx_t_15, 6, __pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_13);
            PyList_SET_ITEM(__pyx_t_15, 7, __pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_14);
            PyList_SET_ITEM(__pyx_t_15, 8, __pyx_t_14);
            __pyx_t_6 = 0;
            __pyx_t_7 = 0;
            __pyx_t_8 = 0;
            __pyx_t_9 = 0;
            __pyx_t_10 = 0;
            __pyx_t_11 = 0;
            __pyx_t_12 = 0;
            __pyx_t_13 = 0;
            __pyx_t_14 = 0;
            __pyx_t_14 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_14)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_14);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_1 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_14, __pyx_t_15) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_15);
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 23, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
        }

        /* "dxfeed/core/listeners/listener.pyx":19
 *                                  int data_count, void*user_data) nogil:
 *     cdef dxf_trade_t*trades = <dxf_trade_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 * 
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }
  }

  /* "dxfeed/core/listeners/listener.pyx":18
 *                                  const dxf_event_data_t*data,
 *                                  int data_count, void*user_data) nogil:
 *     cdef dxf_trade_t*trades = <dxf_trade_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*finally:*/ {
    /*normal exit:*/{
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L5;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "dxfeed/core/listeners/listener.pyx":14
 * 
 * TRADE_COLUMNS = ['Symbol', 'Price', 'ExchangeCode', 'Size', 'Tick', 'Change', 'DayVolume', 'Time', 'IsETH']
 * cdef void trade_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                  dxf_const_string_t symbol_name,
 *                                  const dxf_event_data_t*data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_WriteUnraisable("dxfeed.core.listeners.listener.trade_default_listener", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_py_data);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "dxfeed/core/listeners/listener.pyx":35
 * QUOTE_COLUMNS = ['Symbol', 'BidTime', 'BidExchangeCode', 'BidPrice', 'BidSize', 'AskTime', 'AskExchangeCode',
 *                  'AskPrice', 'AskSize', 'Scope']
 * cdef void quote_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                  dxf_const_string_t symbol_name,
 *                                  const dxf_event_data_t*data,
 */

static void __pyx_f_6dxfeed_4core_9listeners_8listener_quote_default_listener(CYTHON_UNUSED int __pyx_v_event_type, dxf_const_string_t __pyx_v_symbol_name, dxf_event_data_t const *__pyx_v_data, int __pyx_v_data_count, void *__pyx_v_user_data) {
  dxf_quote_t *__pyx_v_quotes;
  PyObject *__pyx_v_py_data = NULL;
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("quote_default_listener", 1);

  /* "dxfeed/core/listeners/listener.pyx":40
 *                                  int data_count,
 *                                  void*user_data) nogil:
 *     cdef dxf_quote_t*quotes = <dxf_quote_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*try:*/ {
    __pyx_v_quotes = ((dxf_quote_t *)__pyx_v_data);

    /* "dxfeed/core/listeners/listener.pyx":41
 *                                  void*user_data) nogil:
 *     cdef dxf_quote_t*quotes = <dxf_quote_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 * 
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "dxfeed/core/listeners/listener.pyx":42
 *     cdef dxf_quote_t*quotes = <dxf_quote_t*> data
 *     with gil:
 *         py_data = <object> user_data             # <<<<<<<<<<<<<<
 * 
 *         for i in range(data_count):
 */
          __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
          __Pyx_INCREF(__pyx_t_1);
          __pyx_v_py_data = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "dxfeed/core/listeners/listener.pyx":44
 *         py_data = <object> user_data
 * 
 *         for i in range(data_count):             # <<<<<<<<<<<<<<
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  quotes[i].bid_time,
 */
          __pyx_t_2 = __pyx_v_data_count;
          __pyx_t_3 = __pyx_t_2;
          for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
            __pyx_v_i = __pyx_t_4;

            /* "dxfeed/core/listeners/listener.pyx":45
 * 
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  quotes[i].bid_time,
 *                                  quotes[i].bid_exchange_code,
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_data, __pyx_n_s_safe_append); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 45, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_6 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t(__pyx_v_symbol_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 45, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);

            /* "dxfeed/core/listeners/listener.pyx":46
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  quotes[i].bid_time,             # <<<<<<<<<<<<<<
 *                                  quotes[i].bid_exchange_code,
 *                                  quotes[i].bid_price,
 */
            __pyx_t_7 = __Pyx_PyInt_From_dxf_long_t((__pyx_v_quotes[__pyx_v_i]).bid_time); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 46, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);

            /* "dxfeed/core/listeners/listener.pyx":47
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  quotes[i].bid_time,
 *                                  quotes[i].bid_exchange_code,             # <<<<<<<<<<<<<<
 *                                  quotes[i].bid_price,
 *                                  quotes[i].bid_size,
 */
            __pyx_t_8 = __Pyx_PyInt_From_wchar_t((__pyx_v_quotes[__pyx_v_i]).bid_exchange_code); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 47, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_8);

            /* "dxfeed/core/listeners/listener.pyx":48
 *                                  quotes[i].bid_time,
 *                                  quotes[i].bid_exchange_code,
 *                                  quotes[i].bid_price,             # <<<<<<<<<<<<<<
 *                                  quotes[i].bid_size,
 *                                  quotes[i].ask_time,
 */
            __pyx_t_9 = PyFloat_FromDouble((__pyx_v_quotes[__pyx_v_i]).bid_price); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 48, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_9);

            /* "dxfeed/core/listeners/listener.pyx":49
 *                                  quotes[i].bid_exchange_code,
 *                                  quotes[i].bid_price,
 *                                  quotes[i].bid_size,             # <<<<<<<<<<<<<<
 *                                  quotes[i].ask_time,
 *                                  quotes[i].ask_exchange_code,
 */
            __pyx_t_10 = __Pyx_PyInt_From_dxf_int_t((__pyx_v_quotes[__pyx_v_i]).bid_size); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 49, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_10);

            /* "dxfeed/core/listeners/listener.pyx":50
 *                                  quotes[i].bid_price,
 *                                  quotes[i].bid_size,
 *                                  quotes[i].ask_time,             # <<<<<<<<<<<<<<
 *                                  quotes[i].ask_exchange_code,
 *                                  quotes[i].ask_price,
 */
            __pyx_t_11 = __Pyx_PyInt_From_dxf_long_t((__pyx_v_quotes[__pyx_v_i]).ask_time); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 50, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_11);

            /* "dxfeed/core/listeners/listener.pyx":51
 *                                  quotes[i].bid_size,
 *                                  quotes[i].ask_time,
 *                                  quotes[i].ask_exchange_code,             # <<<<<<<<<<<<<<
 *                                  quotes[i].ask_price,
 *                                  quotes[i].ask_size,
 */
            __pyx_t_12 = __Pyx_PyInt_From_wchar_t((__pyx_v_quotes[__pyx_v_i]).ask_exchange_code); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 51, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);

            /* "dxfeed/core/listeners/listener.pyx":52
 *                                  quotes[i].ask_time,
 *                                  quotes[i].ask_exchange_code,
 *                                  quotes[i].ask_price,             # <<<<<<<<<<<<<<
 *                                  quotes[i].ask_size,
 *                                  <int> quotes[i].scope])
 */
            __pyx_t_13 = PyFloat_FromDouble((__pyx_v_quotes[__pyx_v_i]).ask_price); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 52, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);

            /* "dxfeed/core/listeners/listener.pyx":53
 *                                  quotes[i].ask_exchange_code,
 *                                  quotes[i].ask_price,
 *                                  quotes[i].ask_size,             # <<<<<<<<<<<<<<
 *                                  <int> quotes[i].scope])
 * 
 */
            __pyx_t_14 = __Pyx_PyInt_From_dxf_int_t((__pyx_v_quotes[__pyx_v_i]).ask_size); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 53, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_14);

            /* "dxfeed/core/listeners/listener.pyx":54
 *                                  quotes[i].ask_price,
 *                                  quotes[i].ask_size,
 *                                  <int> quotes[i].scope])             # <<<<<<<<<<<<<<
 * 
 * SUMMARY_COLUMNS = ['Symbol', 'DayId', 'DayHighPrice', 'DayLowPrice', 'DayClosePrice', 'PrevDayId', 'PrevDayClosePrice',
 */
            __pyx_t_15 = __Pyx_PyInt_From_int(((int)(__pyx_v_quotes[__pyx_v_i]).scope)); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 54, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_15);

            /* "dxfeed/core/listeners/listener.pyx":45
 * 
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  quotes[i].bid_time,
 *                                  quotes[i].bid_exchange_code,
 */
            __pyx_t_16 = PyList_New(10); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 45, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_16);
            __Pyx_GIVEREF(__pyx_t_6);
            PyList_SET_ITEM(__pyx_t_16, 0, __pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_7);
            PyList_SET_ITEM(__pyx_t_16, 1, __pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_8);
            PyList_SET_ITEM(__pyx_t_16, 2, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_9);
            PyList_SET_ITEM(__pyx_t_16, 3, __pyx_t_9);
            __Pyx_GIVEREF(__pyx_t_10);
            PyList_SET_ITEM(__pyx_t_16, 4, __pyx_t_10);
            __Pyx_GIVEREF(__pyx_t_11);
            PyList_SET_ITEM(__pyx_t_16, 5, __pyx_t_11);
            __Pyx_GIVEREF(__pyx_t_12);
            PyList_SET_ITEM(__pyx_t_16, 6, __pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_13);
            PyList_SET_ITEM(__pyx_t_16, 7, __pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_14);
            PyList_SET_ITEM(__pyx_t_16, 8, __pyx_t_14);
            __Pyx_GIVEREF(__pyx_t_15);
            PyList_SET_ITEM(__pyx_t_16, 9, __pyx_t_15);
            __pyx_t_6 = 0;
            __pyx_t_7 = 0;
            __pyx_t_8 = 0;
            __pyx_t_9 = 0;
            __pyx_t_10 = 0;
            __pyx_t_11 = 0;
            __pyx_t_12 = 0;
            __pyx_t_13 = 0;
            __pyx_t_14 = 0;
            __pyx_t_15 = 0;
            __pyx_t_15 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_15)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_15);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_1 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_15, __pyx_t_16) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_16);
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 45, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
        }

        /* "dxfeed/core/listeners/listener.pyx":41
 *                                  void*user_data) nogil:
 *     cdef dxf_quote_t*quotes = <dxf_quote_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 * 
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }
  }

  /* "dxfeed/core/listeners/listener.pyx":40
 *                                  int data_count,
 *                                  void*user_data) nogil:
 *     cdef dxf_quote_t*quotes = <dxf_quote_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*finally:*/ {
    /*normal exit:*/{
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L5;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "dxfeed/core/listeners/listener.pyx":35
 * QUOTE_COLUMNS = ['Symbol', 'BidTime', 'BidExchangeCode', 'BidPrice', 'BidSize', 'AskTime', 'AskExchangeCode',
 *                  'AskPrice', 'AskSize', 'Scope']
 * cdef void quote_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                  dxf_const_string_t symbol_name,
 *                                  const dxf_event_data_t*data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_WriteUnraisable("dxfeed.core.listeners.listener.quote_default_listener", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_py_data);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "dxfeed/core/listeners/listener.pyx":58
 * SUMMARY_COLUMNS = ['Symbol', 'DayId', 'DayHighPrice', 'DayLowPrice', 'DayClosePrice', 'PrevDayId', 'PrevDayClosePrice',
 *                    'PrevDayVolume', 'OpenInterest']
 * cdef void summary_default_listener(int event_type, dxf_const_string_t symbol_name,             # <<<<<<<<<<<<<<
 *                                    const dxf_event_data_t*data, int data_count, void*user_data) nogil:
 *     cdef dxf_summary_t*summary = <dxf_summary_t*> data
 */

static void __pyx_f_6dxfeed_4core_9listeners_8listener_summary_default_listener(CYTHON_UNUSED int __pyx_v_event_type, dxf_const_string_t __pyx_v_symbol_name, dxf_event_data_t const *__pyx_v_data, int __pyx_v_data_count, void *__pyx_v_user_data) {
  dxf_summary_t *__pyx_v_summary;
  PyObject *__pyx_v_py_data = NULL;
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("summary_default_listener", 1);

  /* "dxfeed/core/listeners/listener.pyx":60
 * cdef void summary_default_listener(int event_type, dxf_const_string_t symbol_name,
 *                                    const dxf_event_data_t*data, int data_count, void*user_data) nogil:
 *     cdef dxf_summary_t*summary = <dxf_summary_t*> data             # <<<<<<<<<<<<<<
 * 
 *     with gil:
 */
  /*try:*/ {
    __pyx_v_summary = ((dxf_summary_t *)__pyx_v_data);

    /* "dxfeed/core/listeners/listener.pyx":62
 *     cdef dxf_summary_t*summary = <dxf_summary_t*> data
 * 
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "dxfeed/core/listeners/listener.pyx":63
 * 
 *     with gil:
 *         py_data = <object> user_data             # <<<<<<<<<<<<<<
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 */
          __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
          __Pyx_INCREF(__pyx_t_1);
          __pyx_v_py_data = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "dxfeed/core/listeners/listener.pyx":64
 *     with gil:
 *         py_data = <object> user_data
 *         for i in range(data_count):             # <<<<<<<<<<<<<<
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  summary[i].day_id,
 */
          __pyx_t_2 = __pyx_v_data_count;
          __pyx_t_3 = __pyx_t_2;
          for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
            __pyx_v_i = __pyx_t_4;

            /* "dxfeed/core/listeners/listener.pyx":65
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  summary[i].day_id,
 *                                  summary[i].day_high_price,
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_data, __pyx_n_s_safe_append); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 65, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_6 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t(__pyx_v_symbol_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 65, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);

            /* "dxfeed/core/listeners/listener.pyx":66
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  summary[i].day_id,             # <<<<<<<<<<<<<<
 *                                  summary[i].day_high_price,
 *                                  summary[i].day_low_price,
 */
            __pyx_t_7 = __Pyx_PyInt_From_dxf_dayid_t((__pyx_v_summary[__pyx_v_i]).day_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 66, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);

            /* "dxfeed/core/listeners/listener.pyx":67
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  summary[i].day_id,
 *                                  summary[i].day_high_price,             # <<<<<<<<<<<<<<
 *                                  summary[i].day_low_price,
 *                                  summary[i].day_close_price,
 */
            __pyx_t_8 = PyFloat_FromDouble((__pyx_v_summary[__pyx_v_i]).day_high_price); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 67, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_8);

            /* "dxfeed/core/listeners/listener.pyx":68
 *                                  summary[i].day_id,
 *                                  summary[i].day_high_price,
 *                                  summary[i].day_low_price,             # <<<<<<<<<<<<<<
 *                                  summary[i].day_close_price,
 *                                  summary[i].prev_day_id,
 */
            __pyx_t_9 = PyFloat_FromDouble((__pyx_v_summary[__pyx_v_i]).day_low_price); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 68, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_9);

            /* "dxfeed/core/listeners/listener.pyx":69
 *                                  summary[i].day_high_price,
 *                                  summary[i].day_low_price,
 *                                  summary[i].day_close_price,             # <<<<<<<<<<<<<<
 *                                  summary[i].prev_day_id,
 *                                  summary[i].prev_day_close_price,
 */
            __pyx_t_10 = PyFloat_FromDouble((__pyx_v_summary[__pyx_v_i]).day_close_price); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 69, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_10);

            /* "dxfeed/core/listeners/listener.pyx":70
 *                                  summary[i].day_low_price,
 *                                  summary[i].day_close_price,
 *                                  summary[i].prev_day_id,             # <<<<<<<<<<<<<<
 *                                  summary[i].prev_day_close_price,
 *                                  summary[i].prev_day_volume,
 */
            __pyx_t_11 = __Pyx_PyInt_From_dxf_dayid_t((__pyx_v_summary[__pyx_v_i]).prev_day_id); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 70, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_11);

            /* "dxfeed/core/listeners/listener.pyx":71
 *                                  summary[i].day_close_price,
 *                                  summary[i].prev_day_id,
 *                                  summary[i].prev_day_close_price,             # <<<<<<<<<<<<<<
 *                                  summary[i].prev_day_volume,
 *                                  summary[i].open_interest])
 */
            __pyx_t_12 = PyFloat_FromDouble((__pyx_v_summary[__pyx_v_i]).prev_day_close_price); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 71, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);

            /* "dxfeed/core/listeners/listener.pyx":72
 *                                  summary[i].prev_day_id,
 *                                  summary[i].prev_day_close_price,
 *                                  summary[i].prev_day_volume,             # <<<<<<<<<<<<<<
 *                                  summary[i].open_interest])
 * 
 */
            __pyx_t_13 = PyFloat_FromDouble((__pyx_v_summary[__pyx_v_i]).prev_day_volume); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 72, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);

            /* "dxfeed/core/listeners/listener.pyx":73
 *                                  summary[i].prev_day_close_price,
 *                                  summary[i].prev_day_volume,
 *                                  summary[i].open_interest])             # <<<<<<<<<<<<<<
 * 
 * PROFILE_COLUMNS = ['Symbol', 'Beta', 'EPS', 'DivFreq', 'ExdDivAmount', 'ExdDivDate', '52HighPrice', '52LowPrice',
 */
            __pyx_t_14 = __Pyx_PyInt_From_dxf_int_t((__pyx_v_summary[__pyx_v_i]).open_interest); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 73, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_14);

            /* "dxfeed/core/listeners/listener.pyx":65
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  summary[i].day_id,
 *                                  summary[i].day_high_price,
 */
            __pyx_t_15 = PyList_New(9); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 65, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_GIVEREF(__pyx_t_6);
            PyList_SET_ITEM(__pyx_t_15, 0, __pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_7);
            PyList_SET_ITEM(__pyx_t_15, 1, __pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_8);
            PyList_SET_ITEM(__pyx_t_15, 2, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_9);
            PyList_SET_ITEM(__pyx_t_15, 3, __pyx_t_9);
            __Pyx_GIVEREF(__pyx_t_10);
            PyList_SET_ITEM(__pyx_t_15, 4, __pyx_t_10);
            __Pyx_GIVEREF(__pyx_t_11);
            PyList_SET_ITEM(__pyx_t_15, 5, __pyx_t_11);
            __Pyx_GIVEREF(__pyx_t_12);
            PyList_SET_ITEM(__pyx_t_15, 6, __pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_13);
            PyList_SET_ITEM(__pyx_t_15, 7, __pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_14);
            PyList_SET_ITEM(__pyx_t_15, 8, __pyx_t_14);
            __pyx_t_6 = 0;
            __pyx_t_7 = 0;
            __pyx_t_8 = 0;
            __pyx_t_9 = 0;
            __pyx_t_10 = 0;
            __pyx_t_11 = 0;
            __pyx_t_12 = 0;
            __pyx_t_13 = 0;
            __pyx_t_14 = 0;
            __pyx_t_14 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_14)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_14);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_1 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_14, __pyx_t_15) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_15);
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 65, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
        }

        /* "dxfeed/core/listeners/listener.pyx":62
 *     cdef dxf_summary_t*summary = <dxf_summary_t*> data
 * 
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }
  }

  /* "dxfeed/core/listeners/listener.pyx":60
 * cdef void summary_default_listener(int event_type, dxf_const_string_t symbol_name,
 *                                    const dxf_event_data_t*data, int data_count, void*user_data) nogil:
 *     cdef dxf_summary_t*summary = <dxf_summary_t*> data             # <<<<<<<<<<<<<<
 * 
 *     with gil:
 */
  /*finally:*/ {
    /*normal exit:*/{
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L5;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "dxfeed/core/listeners/listener.pyx":58
 * SUMMARY_COLUMNS = ['Symbol', 'DayId', 'DayHighPrice', 'DayLowPrice', 'DayClosePrice', 'PrevDayId', 'PrevDayClosePrice',
 *                    'PrevDayVolume', 'OpenInterest']
 * cdef void summary_default_listener(int event_type, dxf_const_string_t symbol_name,             # <<<<<<<<<<<<<<
 *                                    const dxf_event_data_t*data, int data_count, void*user_data) nogil:
 *     cdef dxf_summary_t*summary = <dxf_summary_t*> data
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_WriteUnraisable("dxfeed.core.listeners.listener.summary_default_listener", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_py_data);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "dxfeed/core/listeners/listener.pyx":77
 * PROFILE_COLUMNS = ['Symbol', 'Beta', 'EPS', 'DivFreq', 'ExdDivAmount', 'ExdDivDate', '52HighPrice', '52LowPrice',
 *                    'Shares', 'Description', 'RawFlags', 'StatusReason']
 * cdef void profile_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                    dxf_const_string_t symbol_name,
 *                                    const dxf_event_data_t*data,
 */

static void __pyx_f_6dxfeed_4core_9listeners_8listener_profile_default_listener(CYTHON_UNUSED int __pyx_v_event_type, dxf_const_string_t __pyx_v_symbol_name, dxf_event_data_t const *__pyx_v_data, int __pyx_v_data_count, void *__pyx_v_user_data) {
  dxf_profile_t *__pyx_v_p;
  PyObject *__pyx_v_py_data = NULL;
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("profile_default_listener", 1);

  /* "dxfeed/core/listeners/listener.pyx":82
 *                                    int data_count,
 *                                    void*user_data)  nogil:
 *     cdef dxf_profile_t*p = <dxf_profile_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*try:*/ {
    __pyx_v_p = ((dxf_profile_t *)__pyx_v_data);

    /* "dxfeed/core/listeners/listener.pyx":83
 *                                    void*user_data)  nogil:
 *     cdef dxf_profile_t*p = <dxf_profile_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "dxfeed/core/listeners/listener.pyx":84
 *     cdef dxf_profile_t*p = <dxf_profile_t*> data
 *     with gil:
 *         py_data = <object> user_data             # <<<<<<<<<<<<<<
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 */
          __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
          __Pyx_INCREF(__pyx_t_1);
          __pyx_v_py_data = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "dxfeed/core/listeners/listener.pyx":85
 *     with gil:
 *         py_data = <object> user_data
 *         for i in range(data_count):             # <<<<<<<<<<<<<<
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  p[i].beta,
 */
          __pyx_t_2 = __pyx_v_data_count;
          __pyx_t_3 = __pyx_t_2;
          for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
            __pyx_v_i = __pyx_t_4;

            /* "dxfeed/core/listeners/listener.pyx":86
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  p[i].beta,
 *                                  p[i].eps,
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_data, __pyx_n_s_safe_append); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 86, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_6 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t(__pyx_v_symbol_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 86, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);

            /* "dxfeed/core/listeners/listener.pyx":87
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  p[i].beta,             # <<<<<<<<<<<<<<
 *                                  p[i].eps,
 *                                  p[i].div_freq,
 */
            __pyx_t_7 = PyFloat_FromDouble((__pyx_v_p[__pyx_v_i]).beta); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 87, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);

            /* "dxfeed/core/listeners/listener.pyx":88
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  p[i].beta,
 *                                  p[i].eps,             # <<<<<<<<<<<<<<
 *                                  p[i].div_freq,
 *                                  p[i].exd_div_amount,
 */
            __pyx_t_8 = PyFloat_FromDouble((__pyx_v_p[__pyx_v_i]).eps); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 88, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_8);

            /* "dxfeed/core/listeners/listener.pyx":89
 *                                  p[i].beta,
 *                                  p[i].eps,
 *                                  p[i].div_freq,             # <<<<<<<<<<<<<<
 *                                  p[i].exd_div_amount,
 *                                  p[i].exd_div_date,
 */
            __pyx_t_9 = __Pyx_PyInt_From_dxf_int_t((__pyx_v_p[__pyx_v_i]).div_freq); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 89, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_9);

            /* "dxfeed/core/listeners/listener.pyx":90
 *                                  p[i].eps,
 *                                  p[i].div_freq,
 *                                  p[i].exd_div_amount,             # <<<<<<<<<<<<<<
 *                                  p[i].exd_div_date,
 *                                  p[i]._52_high_price,
 */
            __pyx_t_10 = PyFloat_FromDouble((__pyx_v_p[__pyx_v_i]).exd_div_amount); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 90, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_10);

            /* "dxfeed/core/listeners/listener.pyx":91
 *                                  p[i].div_freq,
 *                                  p[i].exd_div_amount,
 *                                  p[i].exd_div_date,             # <<<<<<<<<<<<<<
 *                                  p[i]._52_high_price,
 *                                  p[i]._52_low_price,
 */
            __pyx_t_11 = __Pyx_PyInt_From_dxf_dayid_t((__pyx_v_p[__pyx_v_i]).exd_div_date); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 91, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_11);

            /* "dxfeed/core/listeners/listener.pyx":92
 *                                  p[i].exd_div_amount,
 *                                  p[i].exd_div_date,
 *                                  p[i]._52_high_price,             # <<<<<<<<<<<<<<
 *                                  p[i]._52_low_price,
 *                                  p[i].shares,
 */
            __pyx_t_12 = PyFloat_FromDouble((__pyx_v_p[__pyx_v_i])._52_high_price); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 92, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);

            /* "dxfeed/core/listeners/listener.pyx":93
 *                                  p[i].exd_div_date,
 *                                  p[i]._52_high_price,
 *                                  p[i]._52_low_price,             # <<<<<<<<<<<<<<
 *                                  p[i].shares,
 *                                  unicode_from_dxf_const_string_t(p[i].description),
 */
            __pyx_t_13 = PyFloat_FromDouble((__pyx_v_p[__pyx_v_i])._52_low_price); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 93, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);

            /* "dxfeed/core/listeners/listener.pyx":94
 *                                  p[i]._52_high_price,
 *                                  p[i]._52_low_price,
 *                                  p[i].shares,             # <<<<<<<<<<<<<<
 *                                  unicode_from_dxf_const_string_t(p[i].description),
 *                                  p[i].raw_flags,
 */
            __pyx_t_14 = PyFloat_FromDouble((__pyx_v_p[__pyx_v_i]).shares); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 94, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_14);

            /* "dxfeed/core/listeners/listener.pyx":95
 *                                  p[i]._52_low_price,
 *                                  p[i].shares,
 *                                  unicode_from_dxf_const_string_t(p[i].description),             # <<<<<<<<<<<<<<
 *                                  p[i].raw_flags,
 *                                  unicode_from_dxf_const_string_t(p[i].status_reason)])
 */
            __pyx_t_15 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t((__pyx_v_p[__pyx_v_i]).description); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 95, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_15);

            /* "dxfeed/core/listeners/listener.pyx":96
 *                                  p[i].shares,
 *                                  unicode_from_dxf_const_string_t(p[i].description),
 *                                  p[i].raw_flags,             # <<<<<<<<<<<<<<
 *                                  unicode_from_dxf_const_string_t(p[i].status_reason)])
 * 
 */
            __pyx_t_16 = __Pyx_PyInt_From_dxf_int_t((__pyx_v_p[__pyx_v_i]).raw_flags); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 96, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_16);

            /* "dxfeed/core/listeners/listener.pyx":97
 *                                  unicode_from_dxf_const_string_t(p[i].description),
 *                                  p[i].raw_flags,
 *                                  unicode_from_dxf_const_string_t(p[i].status_reason)])             # <<<<<<<<<<<<<<
 * 
 * TIME_AND_SALE_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'ExchangeCode', 'Price', 'Size', 'BidPrice',
 */
            __pyx_t_17 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t((__pyx_v_p[__pyx_v_i]).status_reason); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 97, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_17);

            /* "dxfeed/core/listeners/listener.pyx":86
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  p[i].beta,
 *                                  p[i].eps,
 */
            __pyx_t_18 = PyList_New(12); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 86, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_18);
            __Pyx_GIVEREF(__pyx_t_6);
            PyList_SET_ITEM(__pyx_t_18, 0, __pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_7);
            PyList_SET_ITEM(__pyx_t_18, 1, __pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_8);
            PyList_SET_ITEM(__pyx_t_18, 2, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_9);
            PyList_SET_ITEM(__pyx_t_18, 3, __pyx_t_9);
            __Pyx_GIVEREF(__pyx_t_10);
            PyList_SET_ITEM(__pyx_t_18, 4, __pyx_t_10);
            __Pyx_GIVEREF(__pyx_t_11);
            PyList_SET_ITEM(__pyx_t_18, 5, __pyx_t_11);
            __Pyx_GIVEREF(__pyx_t_12);
            PyList_SET_ITEM(__pyx_t_18, 6, __pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_13);
            PyList_SET_ITEM(__pyx_t_18, 7, __pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_14);
            PyList_SET_ITEM(__pyx_t_18, 8, __pyx_t_14);
            __Pyx_GIVEREF(__pyx_t_15);
            PyList_SET_ITEM(__pyx_t_18, 9, __pyx_t_15);
            __Pyx_GIVEREF(__pyx_t_16);
            PyList_SET_ITEM(__pyx_t_18, 10, __pyx_t_16);
            __Pyx_GIVEREF(__pyx_t_17);
            PyList_SET_ITEM(__pyx_t_18, 11, __pyx_t_17);
            __pyx_t_6 = 0;
            __pyx_t_7 = 0;
            __pyx_t_8 = 0;
            __pyx_t_9 = 0;
            __pyx_t_10 = 0;
            __pyx_t_11 = 0;
            __pyx_t_12 = 0;
            __pyx_t_13 = 0;
            __pyx_t_14 = 0;
            __pyx_t_15 = 0;
            __pyx_t_16 = 0;
            __pyx_t_17 = 0;
            __pyx_t_17 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_17)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_17);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_1 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_17, __pyx_t_18) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_18);
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 86, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
        }

        /* "dxfeed/core/listeners/listener.pyx":83
 *                                    void*user_data)  nogil:
 *     cdef dxf_profile_t*p = <dxf_profile_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }
  }

  /* "dxfeed/core/listeners/listener.pyx":82
 *                                    int data_count,
 *                                    void*user_data)  nogil:
 *     cdef dxf_profile_t*p = <dxf_profile_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*finally:*/ {
    /*normal exit:*/{
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L5;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "dxfeed/core/listeners/listener.pyx":77
 * PROFILE_COLUMNS = ['Symbol', 'Beta', 'EPS', 'DivFreq', 'ExdDivAmount', 'ExdDivDate', '52HighPrice', '52LowPrice',
 *                    'Shares', 'Description', 'RawFlags', 'StatusReason']
 * cdef void profile_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                    dxf_const_string_t symbol_name,
 *                                    const dxf_event_data_t*data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_WriteUnraisable("dxfeed.core.listeners.listener.profile_default_listener", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_py_data);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "dxfeed/core/listeners/listener.pyx":102
 *                          'AskPrice', 'ExchangeSaleConditions', 'RawFlags', 'Buyer', 'Seller', 'Side', 'Type',
 *                          'IsValidTick', 'IsEthTrade', 'TradeThroughExempt', 'IsSpreadLeg']
 * cdef void time_and_sale_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                          dxf_const_string_t symbol_name,
 *                                          const dxf_event_data_t*data,
 */

static void __pyx_f_6dxfeed_4core_9listeners_8listener_time_and_sale_default_listener(CYTHON_UNUSED int __pyx_v_event_type, dxf_const_string_t __pyx_v_symbol_name, dxf_event_data_t const *__pyx_v_data, int __pyx_v_data_count, void *__pyx_v_user_data) {
  dxf_time_and_sale_t *__pyx_v_tns;
  PyObject *__pyx_v_py_data = NULL;
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("time_and_sale_default_listener", 1);

  /* "dxfeed/core/listeners/listener.pyx":107
 *                                          int data_count,
 *                                          void*user_data) nogil:
 *     cdef dxf_time_and_sale_t*tns = <dxf_time_and_sale_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*try:*/ {
    __pyx_v_tns = ((dxf_time_and_sale_t *)__pyx_v_data);

    /* "dxfeed/core/listeners/listener.pyx":108
 *                                          void*user_data) nogil:
 *     cdef dxf_time_and_sale_t*tns = <dxf_time_and_sale_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "dxfeed/core/listeners/listener.pyx":109
 *     cdef dxf_time_and_sale_t*tns = <dxf_time_and_sale_t*> data
 *     with gil:
 *         py_data = <object> user_data             # <<<<<<<<<<<<<<
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 */
          __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
          __Pyx_INCREF(__pyx_t_1);
          __pyx_v_py_data = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "dxfeed/core/listeners/listener.pyx":110
 *     with gil:
 *         py_data = <object> user_data
 *         for i in range(data_count):             # <<<<<<<<<<<<<<
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  tns[i].event_flags,
 */
          __pyx_t_2 = __pyx_v_data_count;
          __pyx_t_3 = __pyx_t_2;
          for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
            __pyx_v_i = __pyx_t_4;

            /* "dxfeed/core/listeners/listener.pyx":111
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  tns[i].event_flags,
 *                                  tns[i].index,
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_data, __pyx_n_s_safe_append); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 111, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_6 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t(__pyx_v_symbol_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 111, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);

            /* "dxfeed/core/listeners/listener.pyx":112
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  tns[i].event_flags,             # <<<<<<<<<<<<<<
 *                                  tns[i].index,
 *                                  tns[i].time,
 */
            __pyx_t_7 = __Pyx_PyInt_From_dxf_uint_t((__pyx_v_tns[__pyx_v_i]).event_flags); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 112, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);

            /* "dxfeed/core/listeners/listener.pyx":113
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  tns[i].event_flags,
 *                                  tns[i].index,             # <<<<<<<<<<<<<<
 *                                  tns[i].time,
 *                                  tns[i].exchange_code,
 */
            __pyx_t_8 = __Pyx_PyInt_From_dxf_long_t((__pyx_v_tns[__pyx_v_i]).index); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 113, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_8);

            /* "dxfeed/core/listeners/listener.pyx":114
 *                                  tns[i].event_flags,
 *                                  tns[i].index,
 *                                  tns[i].time,             # <<<<<<<<<<<<<<
 *                                  tns[i].exchange_code,
 *                                  tns[i].price,
 */
            __pyx_t_9 = __Pyx_PyInt_From_dxf_long_t((__pyx_v_tns[__pyx_v_i]).time); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 114, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_9);

            /* "dxfeed/core/listeners/listener.pyx":115
 *                                  tns[i].index,
 *                                  tns[i].time,
 *                                  tns[i].exchange_code,             # <<<<<<<<<<<<<<
 *                                  tns[i].price,
 *                                  tns[i].size,
 */
            __pyx_t_10 = __Pyx_PyInt_From_wchar_t((__pyx_v_tns[__pyx_v_i]).exchange_code); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 115, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_10);

            /* "dxfeed/core/listeners/listener.pyx":116
 *                                  tns[i].time,
 *                                  tns[i].exchange_code,
 *                                  tns[i].price,             # <<<<<<<<<<<<<<
 *                                  tns[i].size,
 *                                  tns[i].bid_price,
 */
            __pyx_t_11 = PyFloat_FromDouble((__pyx_v_tns[__pyx_v_i]).price); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 116, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_11);

            /* "dxfeed/core/listeners/listener.pyx":117
 *                                  tns[i].exchange_code,
 *                                  tns[i].price,
 *                                  tns[i].size,             # <<<<<<<<<<<<<<
 *                                  tns[i].bid_price,
 *                                  tns[i].ask_price,
 */
            __pyx_t_12 = __Pyx_PyInt_From_dxf_int_t((__pyx_v_tns[__pyx_v_i]).size); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 117, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);

            /* "dxfeed/core/listeners/listener.pyx":118
 *                                  tns[i].price,
 *                                  tns[i].size,
 *                                  tns[i].bid_price,             # <<<<<<<<<<<<<<
 *                                  tns[i].ask_price,
 *                                  unicode_from_dxf_const_string_t(tns[i].exchange_sale_conditions),
 */
            __pyx_t_13 = PyFloat_FromDouble((__pyx_v_tns[__pyx_v_i]).bid_price); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 118, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);

            /* "dxfeed/core/listeners/listener.pyx":119
 *                                  tns[i].size,
 *                                  tns[i].bid_price,
 *                                  tns[i].ask_price,             # <<<<<<<<<<<<<<
 *                                  unicode_from_dxf_const_string_t(tns[i].exchange_sale_conditions),
 *                                  tns[i].raw_flags,
 */
            __pyx_t_14 = PyFloat_FromDouble((__pyx_v_tns[__pyx_v_i]).ask_price); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 119, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_14);

            /* "dxfeed/core/listeners/listener.pyx":120
 *                                  tns[i].bid_price,
 *                                  tns[i].ask_price,
 *                                  unicode_from_dxf_const_string_t(tns[i].exchange_sale_conditions),             # <<<<<<<<<<<<<<
 *                                  tns[i].raw_flags,
 *                                  unicode_from_dxf_const_string_t(tns[i].buyer),
 */
            __pyx_t_15 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t((__pyx_v_tns[__pyx_v_i]).exchange_sale_conditions); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 120, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_15);

            /* "dxfeed/core/listeners/listener.pyx":121
 *                                  tns[i].ask_price,
 *                                  unicode_from_dxf_const_string_t(tns[i].exchange_sale_conditions),
 *                                  tns[i].raw_flags,             # <<<<<<<<<<<<<<
 *                                  unicode_from_dxf_const_string_t(tns[i].buyer),
 *                                  unicode_from_dxf_const_string_t(tns[i].seller),
 */
            __pyx_t_16 = __Pyx_PyInt_From_dxf_int_t((__pyx_v_tns[__pyx_v_i]).raw_flags); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 121, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_16);

            /* "dxfeed/core/listeners/listener.pyx":122
 *                                  unicode_from_dxf_const_string_t(tns[i].exchange_sale_conditions),
 *                                  tns[i].raw_flags,
 *                                  unicode_from_dxf_const_string_t(tns[i].buyer),             # <<<<<<<<<<<<<<
 *                                  unicode_from_dxf_const_string_t(tns[i].seller),
 *                                  tns[i].side,
 */
            __pyx_t_17 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t((__pyx_v_tns[__pyx_v_i]).buyer); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 122, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_17);

            /* "dxfeed/core/listeners/listener.pyx":123
 *                                  tns[i].raw_flags,
 *                                  unicode_from_dxf_const_string_t(tns[i].buyer),
 *                                  unicode_from_dxf_const_string_t(tns[i].seller),             # <<<<<<<<<<<<<<
 *                                  tns[i].side,
 *                                  tns[i].type,
 */
            __pyx_t_18 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t((__pyx_v_tns[__pyx_v_i]).seller); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 123, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_18);

            /* "dxfeed/core/listeners/listener.pyx":124
 *                                  unicode_from_dxf_const_string_t(tns[i].buyer),
 *                                  unicode_from_dxf_const_string_t(tns[i].seller),
 *                                  tns[i].side,             # <<<<<<<<<<<<<<
 *                                  tns[i].type,
 *                                  tns[i].is_valid_tick,
 */
            __pyx_t_19 = __Pyx_PyInt_From_dxf_order_side_t((__pyx_v_tns[__pyx_v_i]).side); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 124, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_19);

            /* "dxfeed/core/listeners/listener.pyx":125
 *                                  unicode_from_dxf_const_string_t(tns[i].seller),
 *                                  tns[i].side,
 *                                  tns[i].type,             # <<<<<<<<<<<<<<
 *                                  tns[i].is_valid_tick,
 *                                  tns[i].is_eth_trade,
 */
            __pyx_t_20 = __Pyx_PyInt_From_dxf_tns_type_t((__pyx_v_tns[__pyx_v_i]).type); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 125, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_20);

            /* "dxfeed/core/listeners/listener.pyx":126
 *                                  tns[i].side,
 *                                  tns[i].type,
 *                                  tns[i].is_valid_tick,             # <<<<<<<<<<<<<<
 *                                  tns[i].is_eth_trade,
 *                                  tns[i].trade_through_exempt,
 */
            __pyx_t_21 = __Pyx_PyInt_From_dxf_bool_t((__pyx_v_tns[__pyx_v_i]).is_valid_tick); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 126, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_21);

            /* "dxfeed/core/listeners/listener.pyx":127
 *                                  tns[i].type,
 *                                  tns[i].is_valid_tick,
 *                                  tns[i].is_eth_trade,             # <<<<<<<<<<<<<<
 *                                  tns[i].trade_through_exempt,
 *                                  tns[i].is_spread_leg])
 */
            __pyx_t_22 = __Pyx_PyInt_From_dxf_bool_t((__pyx_v_tns[__pyx_v_i]).is_eth_trade); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 127, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_22);

            /* "dxfeed/core/listeners/listener.pyx":128
 *                                  tns[i].is_valid_tick,
 *                                  tns[i].is_eth_trade,
 *                                  tns[i].trade_through_exempt,             # <<<<<<<<<<<<<<
 *                                  tns[i].is_spread_leg])
 * 
 */
            __pyx_t_23 = __Pyx_PyInt_From_wchar_t((__pyx_v_tns[__pyx_v_i]).trade_through_exempt); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 128, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_23);

            /* "dxfeed/core/listeners/listener.pyx":129
 *                                  tns[i].is_eth_trade,
 *                                  tns[i].trade_through_exempt,
 *                                  tns[i].is_spread_leg])             # <<<<<<<<<<<<<<
 * 
 * CANDLE_COLUMNS = ['Symbol', 'Index', 'Time', 'Sequence', 'Count', 'Open', 'High', 'Low', 'Close', 'Volume', 'VWap',
 */
            __pyx_t_24 = __Pyx_PyInt_From_dxf_bool_t((__pyx_v_tns[__pyx_v_i]).is_spread_leg); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 129, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_24);

            /* "dxfeed/core/listeners/listener.pyx":111
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  tns[i].event_flags,
 *                                  tns[i].index,
 */
            __pyx_t_25 = PyList_New(19); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 111, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_25);
            __Pyx_GIVEREF(__pyx_t_6);
            PyList_SET_ITEM(__pyx_t_25, 0, __pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_7);
            PyList_SET_ITEM(__pyx_t_25, 1, __pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_8);
            PyList_SET_ITEM(__pyx_t_25, 2, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_9);
            PyList_SET_ITEM(__pyx_t_25, 3, __pyx_t_9);
            __Pyx_GIVEREF(__pyx_t_10);
            PyList_SET_ITEM(__pyx_t_25, 4, __pyx_t_10);
            __Pyx_GIVEREF(__pyx_t_11);
            PyList_SET_ITEM(__pyx_t_25, 5, __pyx_t_11);
            __Pyx_GIVEREF(__pyx_t_12);
            PyList_SET_ITEM(__pyx_t_25, 6, __pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_13);
            PyList_SET_ITEM(__pyx_t_25, 7, __pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_14);
            PyList_SET_ITEM(__pyx_t_25, 8, __pyx_t_14);
            __Pyx_GIVEREF(__pyx_t_15);
            PyList_SET_ITEM(__pyx_t_25, 9, __pyx_t_15);
            __Pyx_GIVEREF(__pyx_t_16);
            PyList_SET_ITEM(__pyx_t_25, 10, __pyx_t_16);
            __Pyx_GIVEREF(__pyx_t_17);
            PyList_SET_ITEM(__pyx_t_25, 11, __pyx_t_17);
            __Pyx_GIVEREF(__pyx_t_18);
            PyList_SET_ITEM(__pyx_t_25, 12, __pyx_t_18);
            __Pyx_GIVEREF(__pyx_t_19);
            PyList_SET_ITEM(__pyx_t_25, 13, __pyx_t_19);
            __Pyx_GIVEREF(__pyx_t_20);
            PyList_SET_ITEM(__pyx_t_25, 14, __pyx_t_20);
            __Pyx_GIVEREF(__pyx_t_21);
            PyList_SET_ITEM(__pyx_t_25, 15, __pyx_t_21);
            __Pyx_GIVEREF(__pyx_t_22);
            PyList_SET_ITEM(__pyx_t_25, 16, __pyx_t_22);
            __Pyx_GIVEREF(__pyx_t_23);
            PyList_SET_ITEM(__pyx_t_25, 17, __pyx_t_23);
            __Pyx_GIVEREF(__pyx_t_24);
            PyList_SET_ITEM(__pyx_t_25, 18, __pyx_t_24);
            __pyx_t_6 = 0;
            __pyx_t_7 = 0;
            __pyx_t_8 = 0;
            __pyx_t_9 = 0;
            __pyx_t_10 = 0;
            __pyx_t_11 = 0;
            __pyx_t_12 = 0;
            __pyx_t_13 = 0;
            __pyx_t_14 = 0;
            __pyx_t_15 = 0;
            __pyx_t_16 = 0;
            __pyx_t_17 = 0;
            __pyx_t_18 = 0;
            __pyx_t_19 = 0;
            __pyx_t_20 = 0;
            __pyx_t_21 = 0;
            __pyx_t_22 = 0;
            __pyx_t_23 = 0;
            __pyx_t_24 = 0;
            __pyx_t_24 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_24 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_24)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_24);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_1 = (__pyx_t_24) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_24, __pyx_t_25) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_25);
            __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
            __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 111, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
        }

        /* "dxfeed/core/listeners/listener.pyx":108
 *                                          void*user_data) nogil:
 *     cdef dxf_time_and_sale_t*tns = <dxf_time_and_sale_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }
  }

  /* "dxfeed/core/listeners/listener.pyx":107
 *                                          int data_count,
 *                                          void*user_data) nogil:
 *     cdef dxf_time_and_sale_t*tns = <dxf_time_and_sale_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*finally:*/ {
    /*normal exit:*/{
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L5;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "dxfeed/core/listeners/listener.pyx":102
 *                          'AskPrice', 'ExchangeSaleConditions', 'RawFlags', 'Buyer', 'Seller', 'Side', 'Type',
 *                          'IsValidTick', 'IsEthTrade', 'TradeThroughExempt', 'IsSpreadLeg']
 * cdef void time_and_sale_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                          dxf_const_string_t symbol_name,
 *                                          const dxf_event_data_t*data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_WriteUnraisable("dxfeed.core.listeners.listener.time_and_sale_default_listener", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_py_data);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "dxfeed/core/listeners/listener.pyx":133
 * CANDLE_COLUMNS = ['Symbol', 'Index', 'Time', 'Sequence', 'Count', 'Open', 'High', 'Low', 'Close', 'Volume', 'VWap',
 *                   'BidVolume', 'AskVolume', 'OpenInterest', 'ImpVolatility']
 * cdef void candle_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                   dxf_const_string_t symbol_name,
 *                                   const dxf_event_data_t*data,
 */

static void __pyx_f_6dxfeed_4core_9listeners_8listener_candle_default_listener(CYTHON_UNUSED int __pyx_v_event_type, dxf_const_string_t __pyx_v_symbol_name, dxf_event_data_t const *__pyx_v_data, int __pyx_v_data_count, void *__pyx_v_user_data) {
  dxf_candle_t *__pyx_v_candle;
  PyObject *__pyx_v_py_data = NULL;
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("candle_default_listener", 1);

  /* "dxfeed/core/listeners/listener.pyx":138
 *                                   int data_count,
 *                                   void*user_data) nogil:
 *     cdef dxf_candle_t*candle = <dxf_candle_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*try:*/ {
    __pyx_v_candle = ((dxf_candle_t *)__pyx_v_data);

    /* "dxfeed/core/listeners/listener.pyx":139
 *                                   void*user_data) nogil:
 *     cdef dxf_candle_t*candle = <dxf_candle_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "dxfeed/core/listeners/listener.pyx":140
 *     cdef dxf_candle_t*candle = <dxf_candle_t*> data
 *     with gil:
 *         py_data = <object> user_data             # <<<<<<<<<<<<<<
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 */
          __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
          __Pyx_INCREF(__pyx_t_1);
          __pyx_v_py_data = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "dxfeed/core/listeners/listener.pyx":141
 *     with gil:
 *         py_data = <object> user_data
 *         for i in range(data_count):             # <<<<<<<<<<<<<<
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  candle[i].index,
 */
          __pyx_t_2 = __pyx_v_data_count;
          __pyx_t_3 = __pyx_t_2;
          for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
            __pyx_v_i = __pyx_t_4;

            /* "dxfeed/core/listeners/listener.pyx":142
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  candle[i].index,
 *                                  candle[i].time,
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_data, __pyx_n_s_safe_append); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 142, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_6 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t(__pyx_v_symbol_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 142, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);

            /* "dxfeed/core/listeners/listener.pyx":143
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  candle[i].index,             # <<<<<<<<<<<<<<
 *                                  candle[i].time,
 *                                  candle[i].sequence,
 */
            __pyx_t_7 = __Pyx_PyInt_From_dxf_long_t((__pyx_v_candle[__pyx_v_i]).index); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 143, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);

            /* "dxfeed/core/listeners/listener.pyx":144
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  candle[i].index,
 *                                  candle[i].time,             # <<<<<<<<<<<<<<
 *                                  candle[i].sequence,
 *                                  candle[i].count,
 */
            __pyx_t_8 = __Pyx_PyInt_From_dxf_long_t((__pyx_v_candle[__pyx_v_i]).time); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 144, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_8);

            /* "dxfeed/core/listeners/listener.pyx":145
 *                                  candle[i].index,
 *                                  candle[i].time,
 *                                  candle[i].sequence,             # <<<<<<<<<<<<<<
 *                                  candle[i].count,
 *                                  candle[i].open,
 */
            __pyx_t_9 = __Pyx_PyInt_From_dxf_int_t((__pyx_v_candle[__pyx_v_i]).sequence); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 145, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_9);

            /* "dxfeed/core/listeners/listener.pyx":146
 *                                  candle[i].time,
 *                                  candle[i].sequence,
 *                                  candle[i].count,             # <<<<<<<<<<<<<<
 *                                  candle[i].open,
 *                                  candle[i].high,
 */
            __pyx_t_10 = PyFloat_FromDouble((__pyx_v_candle[__pyx_v_i]).count); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 146, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_10);

            /* "dxfeed/core/listeners/listener.pyx":147
 *                                  candle[i].sequence,
 *                                  candle[i].count,
 *                                  candle[i].open,             # <<<<<<<<<<<<<<
 *                                  candle[i].high,
 *                                  candle[i].low,
 */
            __pyx_t_11 = PyFloat_FromDouble((__pyx_v_candle[__pyx_v_i]).open); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 147, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_11);

            /* "dxfeed/core/listeners/listener.pyx":148
 *                                  candle[i].count,
 *                                  candle[i].open,
 *                                  candle[i].high,             # <<<<<<<<<<<<<<
 *                                  candle[i].low,
 *                                  candle[i].close,
 */
            __pyx_t_12 = PyFloat_FromDouble((__pyx_v_candle[__pyx_v_i]).high); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 148, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);

            /* "dxfeed/core/listeners/listener.pyx":149
 *                                  candle[i].open,
 *                                  candle[i].high,
 *                                  candle[i].low,             # <<<<<<<<<<<<<<
 *                                  candle[i].close,
 *                                  candle[i].volume,
 */
            __pyx_t_13 = PyFloat_FromDouble((__pyx_v_candle[__pyx_v_i]).low); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 149, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);

            /* "dxfeed/core/listeners/listener.pyx":150
 *                                  candle[i].high,
 *                                  candle[i].low,
 *                                  candle[i].close,             # <<<<<<<<<<<<<<
 *                                  candle[i].volume,
 *                                  candle[i].vwap,
 */
            __pyx_t_14 = PyFloat_FromDouble((__pyx_v_candle[__pyx_v_i]).close); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 150, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_14);

            /* "dxfeed/core/listeners/listener.pyx":151
 *                                  candle[i].low,
 *                                  candle[i].close,
 *                                  candle[i].volume,             # <<<<<<<<<<<<<<
 *                                  candle[i].vwap,
 *                                  candle[i].bid_volume,
 */
            __pyx_t_15 = PyFloat_FromDouble((__pyx_v_candle[__pyx_v_i]).volume); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 151, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_15);

            /* "dxfeed/core/listeners/listener.pyx":152
 *                                  candle[i].close,
 *                                  candle[i].volume,
 *                                  candle[i].vwap,             # <<<<<<<<<<<<<<
 *                                  candle[i].bid_volume,
 *                                  candle[i].ask_volume,
 */
            __pyx_t_16 = PyFloat_FromDouble((__pyx_v_candle[__pyx_v_i]).vwap); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 152, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_16);

            /* "dxfeed/core/listeners/listener.pyx":153
 *                                  candle[i].volume,
 *                                  candle[i].vwap,
 *                                  candle[i].bid_volume,             # <<<<<<<<<<<<<<
 *                                  candle[i].ask_volume,
 *                                  candle[i].open_interest,
 */
            __pyx_t_17 = PyFloat_FromDouble((__pyx_v_candle[__pyx_v_i]).bid_volume); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 153, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_17);

            /* "dxfeed/core/listeners/listener.pyx":154
 *                                  candle[i].vwap,
 *                                  candle[i].bid_volume,
 *                                  candle[i].ask_volume,             # <<<<<<<<<<<<<<
 *                                  candle[i].open_interest,
 *                                  candle[i].imp_volatility])
 */
            __pyx_t_18 = PyFloat_FromDouble((__pyx_v_candle[__pyx_v_i]).ask_volume); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 154, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_18);

            /* "dxfeed/core/listeners/listener.pyx":155
 *                                  candle[i].bid_volume,
 *                                  candle[i].ask_volume,
 *                                  candle[i].open_interest,             # <<<<<<<<<<<<<<
 *                                  candle[i].imp_volatility])
 * 
 */
            __pyx_t_19 = __Pyx_PyInt_From_dxf_int_t((__pyx_v_candle[__pyx_v_i]).open_interest); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 155, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_19);

            /* "dxfeed/core/listeners/listener.pyx":156
 *                                  candle[i].ask_volume,
 *                                  candle[i].open_interest,
 *                                  candle[i].imp_volatility])             # <<<<<<<<<<<<<<
 * 
 * ORDER_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'TimeNanos', 'Sequence', 'Price', 'Size', 'Count', 'Scope',
 */
            __pyx_t_20 = PyFloat_FromDouble((__pyx_v_candle[__pyx_v_i]).imp_volatility); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 156, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_20);

            /* "dxfeed/core/listeners/listener.pyx":142
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  candle[i].index,
 *                                  candle[i].time,
 */
            __pyx_t_21 = PyList_New(15); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 142, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_21);
            __Pyx_GIVEREF(__pyx_t_6);
            PyList_SET_ITEM(__pyx_t_21, 0, __pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_7);
            PyList_SET_ITEM(__pyx_t_21, 1, __pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_8);
            PyList_SET_ITEM(__pyx_t_21, 2, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_9);
            PyList_SET_ITEM(__pyx_t_21, 3, __pyx_t_9);
            __Pyx_GIVEREF(__pyx_t_10);
            PyList_SET_ITEM(__pyx_t_21, 4, __pyx_t_10);
            __Pyx_GIVEREF(__pyx_t_11);
            PyList_SET_ITEM(__pyx_t_21, 5, __pyx_t_11);
            __Pyx_GIVEREF(__pyx_t_12);
            PyList_SET_ITEM(__pyx_t_21, 6, __pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_13);
            PyList_SET_ITEM(__pyx_t_21, 7, __pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_14);
            PyList_SET_ITEM(__pyx_t_21, 8, __pyx_t_14);
            __Pyx_GIVEREF(__pyx_t_15);
            PyList_SET_ITEM(__pyx_t_21, 9, __pyx_t_15);
            __Pyx_GIVEREF(__pyx_t_16);
            PyList_SET_ITEM(__pyx_t_21, 10, __pyx_t_16);
            __Pyx_GIVEREF(__pyx_t_17);
            PyList_SET_ITEM(__pyx_t_21, 11, __pyx_t_17);
            __Pyx_GIVEREF(__pyx_t_18);
            PyList_SET_ITEM(__pyx_t_21, 12, __pyx_t_18);
            __Pyx_GIVEREF(__pyx_t_19);
            PyList_SET_ITEM(__pyx_t_21, 13, __pyx_t_19);
            __Pyx_GIVEREF(__pyx_t_20);
            PyList_SET_ITEM(__pyx_t_21, 14, __pyx_t_20);
            __pyx_t_6 = 0;
            __pyx_t_7 = 0;
            __pyx_t_8 = 0;
            __pyx_t_9 = 0;
            __pyx_t_10 = 0;
            __pyx_t_11 = 0;
            __pyx_t_12 = 0;
            __pyx_t_13 = 0;
            __pyx_t_14 = 0;
            __pyx_t_15 = 0;
            __pyx_t_16 = 0;
            __pyx_t_17 = 0;
            __pyx_t_18 = 0;
            __pyx_t_19 = 0;
            __pyx_t_20 = 0;
            __pyx_t_20 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_20)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_20);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_1 = (__pyx_t_20) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_20, __pyx_t_21) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_21);
            __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
            __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 142, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
        }

        /* "dxfeed/core/listeners/listener.pyx":139
 *                                   void*user_data) nogil:
 *     cdef dxf_candle_t*candle = <dxf_candle_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }
  }

  /* "dxfeed/core/listeners/listener.pyx":138
 *                                   int data_count,
 *                                   void*user_data) nogil:
 *     cdef dxf_candle_t*candle = <dxf_candle_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*finally:*/ {
    /*normal exit:*/{
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L5;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "dxfeed/core/listeners/listener.pyx":133
 * CANDLE_COLUMNS = ['Symbol', 'Index', 'Time', 'Sequence', 'Count', 'Open', 'High', 'Low', 'Close', 'Volume', 'VWap',
 *                   'BidVolume', 'AskVolume', 'OpenInterest', 'ImpVolatility']
 * cdef void candle_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                   dxf_const_string_t symbol_name,
 *                                   const dxf_event_data_t*data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_WriteUnraisable("dxfeed.core.listeners.listener.candle_default_listener", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_py_data);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "dxfeed/core/listeners/listener.pyx":160
 * ORDER_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'TimeNanos', 'Sequence', 'Price', 'Size', 'Count', 'Scope',
 *                  'Side', 'ExchangeCode', 'MarketMaker', 'SpreadSymbol']
 * cdef void order_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                  dxf_const_string_t symbol_name,
 *                                  const dxf_event_data_t*data,
 */

static void __pyx_f_6dxfeed_4core_9listeners_8listener_order_default_listener(CYTHON_UNUSED int __pyx_v_event_type, dxf_const_string_t __pyx_v_symbol_name, dxf_event_data_t const *__pyx_v_data, int __pyx_v_data_count, void *__pyx_v_user_data) {
  dxf_order_t *__pyx_v_order;
  PyObject *__pyx_v_py_data = NULL;
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("order_default_listener", 1);

  /* "dxfeed/core/listeners/listener.pyx":165
 *                                  int data_count,
 *                                  void*user_data) nogil:
 *     cdef dxf_order_t*order = <dxf_order_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*try:*/ {
    __pyx_v_order = ((dxf_order_t *)__pyx_v_data);

    /* "dxfeed/core/listeners/listener.pyx":166
 *                                  void*user_data) nogil:
 *     cdef dxf_order_t*order = <dxf_order_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "dxfeed/core/listeners/listener.pyx":167
 *     cdef dxf_order_t*order = <dxf_order_t*> data
 *     with gil:
 *         py_data = <object> user_data             # <<<<<<<<<<<<<<
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 */
          __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
          __Pyx_INCREF(__pyx_t_1);
          __pyx_v_py_data = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "dxfeed/core/listeners/listener.pyx":168
 *     with gil:
 *         py_data = <object> user_data
 *         for i in range(data_count):             # <<<<<<<<<<<<<<
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  order[i].event_flags,
 */
          __pyx_t_2 = __pyx_v_data_count;
          __pyx_t_3 = __pyx_t_2;
          for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
            __pyx_v_i = __pyx_t_4;

            /* "dxfeed/core/listeners/listener.pyx":169
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  order[i].event_flags,
 *                                  order[i].index,
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_data, __pyx_n_s_safe_append); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 169, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_6 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t(__pyx_v_symbol_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 169, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);

            /* "dxfeed/core/listeners/listener.pyx":170
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  order[i].event_flags,             # <<<<<<<<<<<<<<
 *                                  order[i].index,
 *                                  order[i].time,
 */
            __pyx_t_7 = __Pyx_PyInt_From_dxf_uint_t((__pyx_v_order[__pyx_v_i]).event_flags); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 170, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);

            /* "dxfeed/core/listeners/listener.pyx":171
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  order[i].event_flags,
 *                                  order[i].index,             # <<<<<<<<<<<<<<
 *                                  order[i].time,
 *                                  order[i].time_nanos,
 */
            __pyx_t_8 = __Pyx_PyInt_From_dxf_long_t((__pyx_v_order[__pyx_v_i]).index); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 171, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_8);

            /* "dxfeed/core/listeners/listener.pyx":172
 *                                  order[i].event_flags,
 *                                  order[i].index,
 *                                  order[i].time,             # <<<<<<<<<<<<<<
 *                                  order[i].time_nanos,
 *                                  order[i].sequence,
 */
            __pyx_t_9 = __Pyx_PyInt_From_dxf_long_t((__pyx_v_order[__pyx_v_i]).time); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 172, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_9);

            /* "dxfeed/core/listeners/listener.pyx":173
 *                                  order[i].index,
 *                                  order[i].time,
 *                                  order[i].time_nanos,             # <<<<<<<<<<<<<<
 *                                  order[i].sequence,
 *                                  order[i].price,
 */
            __pyx_t_10 = __Pyx_PyInt_From_dxf_int_t((__pyx_v_order[__pyx_v_i]).time_nanos); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 173, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_10);

            /* "dxfeed/core/listeners/listener.pyx":174
 *                                  order[i].time,
 *                                  order[i].time_nanos,
 *                                  order[i].sequence,             # <<<<<<<<<<<<<<
 *                                  order[i].price,
 *                                  order[i].size,
 */
            __pyx_t_11 = __Pyx_PyInt_From_dxf_int_t((__pyx_v_order[__pyx_v_i]).sequence); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 174, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_11);

            /* "dxfeed/core/listeners/listener.pyx":175
 *                                  order[i].time_nanos,
 *                                  order[i].sequence,
 *                                  order[i].price,             # <<<<<<<<<<<<<<
 *                                  order[i].size,
 *                                  order[i].count,
 */
            __pyx_t_12 = PyFloat_FromDouble((__pyx_v_order[__pyx_v_i]).price); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 175, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);

            /* "dxfeed/core/listeners/listener.pyx":176
 *                                  order[i].sequence,
 *                                  order[i].price,
 *                                  order[i].size,             # <<<<<<<<<<<<<<
 *                                  order[i].count,
 *                                  order[i].scope,
 */
            __pyx_t_13 = __Pyx_PyInt_From_dxf_int_t((__pyx_v_order[__pyx_v_i]).size); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 176, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);

            /* "dxfeed/core/listeners/listener.pyx":177
 *                                  order[i].price,
 *                                  order[i].size,
 *                                  order[i].count,             # <<<<<<<<<<<<<<
 *                                  order[i].scope,
 *                                  order[i].side,
 */
            __pyx_t_14 = __Pyx_PyInt_From_dxf_int_t((__pyx_v_order[__pyx_v_i]).count); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 177, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_14);

            /* "dxfeed/core/listeners/listener.pyx":178
 *                                  order[i].size,
 *                                  order[i].count,
 *                                  order[i].scope,             # <<<<<<<<<<<<<<
 *                                  order[i].side,
 *                                  order[i].exchange_code,
 */
            __pyx_t_15 = __Pyx_PyInt_From_dxf_order_scope_t((__pyx_v_order[__pyx_v_i]).scope); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 178, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_15);

            /* "dxfeed/core/listeners/listener.pyx":179
 *                                  order[i].count,
 *                                  order[i].scope,
 *                                  order[i].side,             # <<<<<<<<<<<<<<
 *                                  order[i].exchange_code,
 *                                  unicode_from_dxf_const_string_t(order[i].market_maker),
 */
            __pyx_t_16 = __Pyx_PyInt_From_dxf_order_side_t((__pyx_v_order[__pyx_v_i]).side); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 179, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_16);

            /* "dxfeed/core/listeners/listener.pyx":180
 *                                  order[i].scope,
 *                                  order[i].side,
 *                                  order[i].exchange_code,             # <<<<<<<<<<<<<<
 *                                  unicode_from_dxf_const_string_t(order[i].market_maker),
 *                                  unicode_from_dxf_const_string_t(order[i].spread_symbol)])
 */
            __pyx_t_17 = __Pyx_PyInt_From_wchar_t((__pyx_v_order[__pyx_v_i]).exchange_code); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 180, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_17);

            /* "dxfeed/core/listeners/listener.pyx":181
 *                                  order[i].side,
 *                                  order[i].exchange_code,
 *                                  unicode_from_dxf_const_string_t(order[i].market_maker),             # <<<<<<<<<<<<<<
 *                                  unicode_from_dxf_const_string_t(order[i].spread_symbol)])
 * 
 */
            __pyx_t_18 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t((__pyx_v_order[__pyx_v_i]).market_maker); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 181, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_18);

            /* "dxfeed/core/listeners/listener.pyx":182
 *                                  order[i].exchange_code,
 *                                  unicode_from_dxf_const_string_t(order[i].market_maker),
 *                                  unicode_from_dxf_const_string_t(order[i].spread_symbol)])             # <<<<<<<<<<<<<<
 * 
 * GREEKS_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'Price', 'Volatility', 'Delta', 'Gamma', 'Theta', 'Rho',
 */
            __pyx_t_19 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t((__pyx_v_order[__pyx_v_i]).spread_symbol); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 182, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_19);

            /* "dxfeed/core/listeners/listener.pyx":169
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  order[i].event_flags,
 *                                  order[i].index,
 */
            __pyx_t_20 = PyList_New(14); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 169, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_20);
            __Pyx_GIVEREF(__pyx_t_6);
            PyList_SET_ITEM(__pyx_t_20, 0, __pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_7);
            PyList_SET_ITEM(__pyx_t_20, 1, __pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_8);
            PyList_SET_ITEM(__pyx_t_20, 2, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_9);
            PyList_SET_ITEM(__pyx_t_20, 3, __pyx_t_9);
            __Pyx_GIVEREF(__pyx_t_10);
            PyList_SET_ITEM(__pyx_t_20, 4, __pyx_t_10);
            __Pyx_GIVEREF(__pyx_t_11);
            PyList_SET_ITEM(__pyx_t_20, 5, __pyx_t_11);
            __Pyx_GIVEREF(__pyx_t_12);
            PyList_SET_ITEM(__pyx_t_20, 6, __pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_13);
            PyList_SET_ITEM(__pyx_t_20, 7, __pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_14);
            PyList_SET_ITEM(__pyx_t_20, 8, __pyx_t_14);
            __Pyx_GIVEREF(__pyx_t_15);
            PyList_SET_ITEM(__pyx_t_20, 9, __pyx_t_15);
            __Pyx_GIVEREF(__pyx_t_16);
            PyList_SET_ITEM(__pyx_t_20, 10, __pyx_t_16);
            __Pyx_GIVEREF(__pyx_t_17);
            PyList_SET_ITEM(__pyx_t_20, 11, __pyx_t_17);
            __Pyx_GIVEREF(__pyx_t_18);
            PyList_SET_ITEM(__pyx_t_20, 12, __pyx_t_18);
            __Pyx_GIVEREF(__pyx_t_19);
            PyList_SET_ITEM(__pyx_t_20, 13, __pyx_t_19);
            __pyx_t_6 = 0;
            __pyx_t_7 = 0;
            __pyx_t_8 = 0;
            __pyx_t_9 = 0;
            __pyx_t_10 = 0;
            __pyx_t_11 = 0;
            __pyx_t_12 = 0;
            __pyx_t_13 = 0;
            __pyx_t_14 = 0;
            __pyx_t_15 = 0;
            __pyx_t_16 = 0;
            __pyx_t_17 = 0;
            __pyx_t_18 = 0;
            __pyx_t_19 = 0;
            __pyx_t_19 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_19)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_19);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_1 = (__pyx_t_19) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_19, __pyx_t_20) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_20);
            __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
            __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 169, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
        }

        /* "dxfeed/core/listeners/listener.pyx":166
 *                                  void*user_data) nogil:
 *     cdef dxf_order_t*order = <dxf_order_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }
  }

  /* "dxfeed/core/listeners/listener.pyx":165
 *                                  int data_count,
 *                                  void*user_data) nogil:
 *     cdef dxf_order_t*order = <dxf_order_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*finally:*/ {
    /*normal exit:*/{
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L5;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "dxfeed/core/listeners/listener.pyx":160
 * ORDER_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'TimeNanos', 'Sequence', 'Price', 'Size', 'Count', 'Scope',
 *                  'Side', 'ExchangeCode', 'MarketMaker', 'SpreadSymbol']
 * cdef void order_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                  dxf_const_string_t symbol_name,
 *                                  const dxf_event_data_t*data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_WriteUnraisable("dxfeed.core.listeners.listener.order_default_listener", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_py_data);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "dxfeed/core/listeners/listener.pyx":186
 * GREEKS_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'Price', 'Volatility', 'Delta', 'Gamma', 'Theta', 'Rho',
 *                   'Vega']
 * cdef void greeks_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                   dxf_const_string_t symbol_name,
 *                                   const dxf_event_data_t*data,
 */

static void __pyx_f_6dxfeed_4core_9listeners_8listener_greeks_default_listener(CYTHON_UNUSED int __pyx_v_event_type, dxf_const_string_t __pyx_v_symbol_name, dxf_event_data_t const *__pyx_v_data, int __pyx_v_data_count, void *__pyx_v_user_data) {
  dxf_greeks_t *__pyx_v_greeks;
  PyObject *__pyx_v_py_data = NULL;
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("greeks_default_listener", 1);

  /* "dxfeed/core/listeners/listener.pyx":191
 *                                   int data_count,
 *                                   void*user_data) nogil:
 *     cdef dxf_greeks_t*greeks = <dxf_greeks_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*try:*/ {
    __pyx_v_greeks = ((dxf_greeks_t *)__pyx_v_data);

    /* "dxfeed/core/listeners/listener.pyx":192
 *                                   void*user_data) nogil:
 *     cdef dxf_greeks_t*greeks = <dxf_greeks_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "dxfeed/core/listeners/listener.pyx":193
 *     cdef dxf_greeks_t*greeks = <dxf_greeks_t*> data
 *     with gil:
 *         py_data = <object> user_data             # <<<<<<<<<<<<<<
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 */
          __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
          __Pyx_INCREF(__pyx_t_1);
          __pyx_v_py_data = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "dxfeed/core/listeners/listener.pyx":194
 *     with gil:
 *         py_data = <object> user_data
 *         for i in range(data_count):             # <<<<<<<<<<<<<<
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  greeks[i].event_flags,
 */
          __pyx_t_2 = __pyx_v_data_count;
          __pyx_t_3 = __pyx_t_2;
          for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
            __pyx_v_i = __pyx_t_4;

            /* "dxfeed/core/listeners/listener.pyx":195
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  greeks[i].event_flags,
 *                                  greeks[i].index,
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_data, __pyx_n_s_safe_append); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 195, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_6 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t(__pyx_v_symbol_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 195, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);

            /* "dxfeed/core/listeners/listener.pyx":196
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  greeks[i].event_flags,             # <<<<<<<<<<<<<<
 *                                  greeks[i].index,
 *                                  greeks[i].time,
 */
            __pyx_t_7 = __Pyx_PyInt_From_dxf_uint_t((__pyx_v_greeks[__pyx_v_i]).event_flags); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 196, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);

            /* "dxfeed/core/listeners/listener.pyx":197
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  greeks[i].event_flags,
 *                                  greeks[i].index,             # <<<<<<<<<<<<<<
 *                                  greeks[i].time,
 *                                  greeks[i].price,
 */
            __pyx_t_8 = __Pyx_PyInt_From_dxf_long_t((__pyx_v_greeks[__pyx_v_i]).index); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 197, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_8);

            /* "dxfeed/core/listeners/listener.pyx":198
 *                                  greeks[i].event_flags,
 *                                  greeks[i].index,
 *                                  greeks[i].time,             # <<<<<<<<<<<<<<
 *                                  greeks[i].price,
 *                                  greeks[i].volatility,
 */
            __pyx_t_9 = __Pyx_PyInt_From_dxf_long_t((__pyx_v_greeks[__pyx_v_i]).time); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 198, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_9);

            /* "dxfeed/core/listeners/listener.pyx":199
 *                                  greeks[i].index,
 *                                  greeks[i].time,
 *                                  greeks[i].price,             # <<<<<<<<<<<<<<
 *                                  greeks[i].volatility,
 *                                  greeks[i].delta,
 */
            __pyx_t_10 = PyFloat_FromDouble((__pyx_v_greeks[__pyx_v_i]).price); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 199, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_10);

            /* "dxfeed/core/listeners/listener.pyx":200
 *                                  greeks[i].time,
 *                                  greeks[i].price,
 *                                  greeks[i].volatility,             # <<<<<<<<<<<<<<
 *                                  greeks[i].delta,
 *                                  greeks[i].gamma,
 */
            __pyx_t_11 = PyFloat_FromDouble((__pyx_v_greeks[__pyx_v_i]).volatility); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 200, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_11);

            /* "dxfeed/core/listeners/listener.pyx":201
 *                                  greeks[i].price,
 *                                  greeks[i].volatility,
 *                                  greeks[i].delta,             # <<<<<<<<<<<<<<
 *                                  greeks[i].gamma,
 *                                  greeks[i].theta,
 */
            __pyx_t_12 = PyFloat_FromDouble((__pyx_v_greeks[__pyx_v_i]).delta); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 201, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);

            /* "dxfeed/core/listeners/listener.pyx":202
 *                                  greeks[i].volatility,
 *                                  greeks[i].delta,
 *                                  greeks[i].gamma,             # <<<<<<<<<<<<<<
 *                                  greeks[i].theta,
 *                                  greeks[i].rho,
 */
            __pyx_t_13 = PyFloat_FromDouble((__pyx_v_greeks[__pyx_v_i]).gamma); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 202, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);

            /* "dxfeed/core/listeners/listener.pyx":203
 *                                  greeks[i].delta,
 *                                  greeks[i].gamma,
 *                                  greeks[i].theta,             # <<<<<<<<<<<<<<
 *                                  greeks[i].rho,
 *                                  greeks[i].vega])
 */
            __pyx_t_14 = PyFloat_FromDouble((__pyx_v_greeks[__pyx_v_i]).theta); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 203, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_14);

            /* "dxfeed/core/listeners/listener.pyx":204
 *                                  greeks[i].gamma,
 *                                  greeks[i].theta,
 *                                  greeks[i].rho,             # <<<<<<<<<<<<<<
 *                                  greeks[i].vega])
 * 
 */
            __pyx_t_15 = PyFloat_FromDouble((__pyx_v_greeks[__pyx_v_i]).rho); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 204, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_15);

            /* "dxfeed/core/listeners/listener.pyx":205
 *                                  greeks[i].theta,
 *                                  greeks[i].rho,
 *                                  greeks[i].vega])             # <<<<<<<<<<<<<<
 * 
 * THEO_PRICE_COLUMNS = ['Symbol', 'Time', 'Price', 'UnderlyingPrice', 'Delta', 'Gamma', 'Dividend', 'Interest']
 */
            __pyx_t_16 = PyFloat_FromDouble((__pyx_v_greeks[__pyx_v_i]).vega); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 205, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_16);

            /* "dxfeed/core/listeners/listener.pyx":195
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  greeks[i].event_flags,
 *                                  greeks[i].index,
 */
            __pyx_t_17 = PyList_New(11); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 195, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_GIVEREF(__pyx_t_6);
            PyList_SET_ITEM(__pyx_t_17, 0, __pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_7);
            PyList_SET_ITEM(__pyx_t_17, 1, __pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_8);
            PyList_SET_ITEM(__pyx_t_17, 2, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_9);
            PyList_SET_ITEM(__pyx_t_17, 3, __pyx_t_9);
            __Pyx_GIVEREF(__pyx_t_10);
            PyList_SET_ITEM(__pyx_t_17, 4, __pyx_t_10);
            __Pyx_GIVEREF(__pyx_t_11);
            PyList_SET_ITEM(__pyx_t_17, 5, __pyx_t_11);
            __Pyx_GIVEREF(__pyx_t_12);
            PyList_SET_ITEM(__pyx_t_17, 6, __pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_13);
            PyList_SET_ITEM(__pyx_t_17, 7, __pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_14);
            PyList_SET_ITEM(__pyx_t_17, 8, __pyx_t_14);
            __Pyx_GIVEREF(__pyx_t_15);
            PyList_SET_ITEM(__pyx_t_17, 9, __pyx_t_15);
            __Pyx_GIVEREF(__pyx_t_16);
            PyList_SET_ITEM(__pyx_t_17, 10, __pyx_t_16);
            __pyx_t_6 = 0;
            __pyx_t_7 = 0;
            __pyx_t_8 = 0;
            __pyx_t_9 = 0;
            __pyx_t_10 = 0;
            __pyx_t_11 = 0;
            __pyx_t_12 = 0;
            __pyx_t_13 = 0;
            __pyx_t_14 = 0;
            __pyx_t_15 = 0;
            __pyx_t_16 = 0;
            __pyx_t_16 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_16)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_16);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_1 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_16, __pyx_t_17) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_17);
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 195, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
        }

        /* "dxfeed/core/listeners/listener.pyx":192
 *                                   void*user_data) nogil:
 *     cdef dxf_greeks_t*greeks = <dxf_greeks_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }
  }

  /* "dxfeed/core/listeners/listener.pyx":191
 *                                   int data_count,
 *                                   void*user_data) nogil:
 *     cdef dxf_greeks_t*greeks = <dxf_greeks_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*finally:*/ {
    /*normal exit:*/{
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L5;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "dxfeed/core/listeners/listener.pyx":186
 * GREEKS_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'Price', 'Volatility', 'Delta', 'Gamma', 'Theta', 'Rho',
 *                   'Vega']
 * cdef void greeks_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                   dxf_const_string_t symbol_name,
 *                                   const dxf_event_data_t*data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_WriteUnraisable("dxfeed.core.listeners.listener.greeks_default_listener", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_py_data);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "dxfeed/core/listeners/listener.pyx":208
 * 
 * THEO_PRICE_COLUMNS = ['Symbol', 'Time', 'Price', 'UnderlyingPrice', 'Delta', 'Gamma', 'Dividend', 'Interest']
 * cdef void theo_price_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                       dxf_const_string_t symbol_name,
 *                                       const dxf_event_data_t*data,
 */

static void __pyx_f_6dxfeed_4core_9listeners_8listener_theo_price_default_listener(CYTHON_UNUSED int __pyx_v_event_type, dxf_const_string_t __pyx_v_symbol_name, dxf_event_data_t const *__pyx_v_data, int __pyx_v_data_count, void *__pyx_v_user_data) {
  dxf_theo_price_t *__pyx_v_theo_price;
  PyObject *__pyx_v_py_data = NULL;
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("theo_price_default_listener", 1);

  /* "dxfeed/core/listeners/listener.pyx":213
 *                                       int data_count,
 *                                       void*user_data) nogil:
 *     cdef dxf_theo_price_t*theo_price = <dxf_theo_price_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*try:*/ {
    __pyx_v_theo_price = ((dxf_theo_price_t *)__pyx_v_data);

    /* "dxfeed/core/listeners/listener.pyx":214
 *                                       void*user_data) nogil:
 *     cdef dxf_theo_price_t*theo_price = <dxf_theo_price_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "dxfeed/core/listeners/listener.pyx":215
 *     cdef dxf_theo_price_t*theo_price = <dxf_theo_price_t*> data
 *     with gil:
 *         py_data = <object> user_data             # <<<<<<<<<<<<<<
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 */
          __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
          __Pyx_INCREF(__pyx_t_1);
          __pyx_v_py_data = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "dxfeed/core/listeners/listener.pyx":216
 *     with gil:
 *         py_data = <object> user_data
 *         for i in range(data_count):             # <<<<<<<<<<<<<<
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  theo_price[i].time,
 */
          __pyx_t_2 = __pyx_v_data_count;
          __pyx_t_3 = __pyx_t_2;
          for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
            __pyx_v_i = __pyx_t_4;

            /* "dxfeed/core/listeners/listener.pyx":217
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  theo_price[i].time,
 *                                  theo_price[i].price,
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_data, __pyx_n_s_safe_append); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 217, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_6 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t(__pyx_v_symbol_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 217, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);

            /* "dxfeed/core/listeners/listener.pyx":218
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  theo_price[i].time,             # <<<<<<<<<<<<<<
 *                                  theo_price[i].price,
 *                                  theo_price[i].underlying_price,
 */
            __pyx_t_7 = __Pyx_PyInt_From_dxf_long_t((__pyx_v_theo_price[__pyx_v_i]).time); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 218, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);

            /* "dxfeed/core/listeners/listener.pyx":219
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  theo_price[i].time,
 *                                  theo_price[i].price,             # <<<<<<<<<<<<<<
 *                                  theo_price[i].underlying_price,
 *                                  theo_price[i].delta,
 */
            __pyx_t_8 = PyFloat_FromDouble((__pyx_v_theo_price[__pyx_v_i]).price); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 219, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_8);

            /* "dxfeed/core/listeners/listener.pyx":220
 *                                  theo_price[i].time,
 *                                  theo_price[i].price,
 *                                  theo_price[i].underlying_price,             # <<<<<<<<<<<<<<
 *                                  theo_price[i].delta,
 *                                  theo_price[i].gamma,
 */
            __pyx_t_9 = PyFloat_FromDouble((__pyx_v_theo_price[__pyx_v_i]).underlying_price); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 220, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_9);

            /* "dxfeed/core/listeners/listener.pyx":221
 *                                  theo_price[i].price,
 *                                  theo_price[i].underlying_price,
 *                                  theo_price[i].delta,             # <<<<<<<<<<<<<<
 *                                  theo_price[i].gamma,
 *                                  theo_price[i].dividend,
 */
            __pyx_t_10 = PyFloat_FromDouble((__pyx_v_theo_price[__pyx_v_i]).delta); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 221, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_10);

            /* "dxfeed/core/listeners/listener.pyx":222
 *                                  theo_price[i].underlying_price,
 *                                  theo_price[i].delta,
 *                                  theo_price[i].gamma,             # <<<<<<<<<<<<<<
 *                                  theo_price[i].dividend,
 *                                  theo_price[i].interest])
 */
            __pyx_t_11 = PyFloat_FromDouble((__pyx_v_theo_price[__pyx_v_i]).gamma); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 222, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_11);

            /* "dxfeed/core/listeners/listener.pyx":223
 *                                  theo_price[i].delta,
 *                                  theo_price[i].gamma,
 *                                  theo_price[i].dividend,             # <<<<<<<<<<<<<<
 *                                  theo_price[i].interest])
 * 
 */
            __pyx_t_12 = PyFloat_FromDouble((__pyx_v_theo_price[__pyx_v_i]).dividend); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 223, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);

            /* "dxfeed/core/listeners/listener.pyx":224
 *                                  theo_price[i].gamma,
 *                                  theo_price[i].dividend,
 *                                  theo_price[i].interest])             # <<<<<<<<<<<<<<
 * 
 * UNDERLYING_COLUMNS = ['Symbol', 'Volatility', 'FrontVolatility', 'BackVolatility', 'PutCallRatio']
 */
            __pyx_t_13 = PyFloat_FromDouble((__pyx_v_theo_price[__pyx_v_i]).interest); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 224, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);

            /* "dxfeed/core/listeners/listener.pyx":217
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  theo_price[i].time,
 *                                  theo_price[i].price,
 */
            __pyx_t_14 = PyList_New(8); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 217, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_GIVEREF(__pyx_t_6);
            PyList_SET_ITEM(__pyx_t_14, 0, __pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_7);
            PyList_SET_ITEM(__pyx_t_14, 1, __pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_8);
            PyList_SET_ITEM(__pyx_t_14, 2, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_9);
            PyList_SET_ITEM(__pyx_t_14, 3, __pyx_t_9);
            __Pyx_GIVEREF(__pyx_t_10);
            PyList_SET_ITEM(__pyx_t_14, 4, __pyx_t_10);
            __Pyx_GIVEREF(__pyx_t_11);
            PyList_SET_ITEM(__pyx_t_14, 5, __pyx_t_11);
            __Pyx_GIVEREF(__pyx_t_12);
            PyList_SET_ITEM(__pyx_t_14, 6, __pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_13);
            PyList_SET_ITEM(__pyx_t_14, 7, __pyx_t_13);
            __pyx_t_6 = 0;
            __pyx_t_7 = 0;
            __pyx_t_8 = 0;
            __pyx_t_9 = 0;
            __pyx_t_10 = 0;
            __pyx_t_11 = 0;
            __pyx_t_12 = 0;
            __pyx_t_13 = 0;
            __pyx_t_13 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_13)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_13);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_1 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_13, __pyx_t_14) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_14);
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 217, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
        }

        /* "dxfeed/core/listeners/listener.pyx":214
 *                                       void*user_data) nogil:
 *     cdef dxf_theo_price_t*theo_price = <dxf_theo_price_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }
  }

  /* "dxfeed/core/listeners/listener.pyx":213
 *                                       int data_count,
 *                                       void*user_data) nogil:
 *     cdef dxf_theo_price_t*theo_price = <dxf_theo_price_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*finally:*/ {
    /*normal exit:*/{
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L5;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "dxfeed/core/listeners/listener.pyx":208
 * 
 * THEO_PRICE_COLUMNS = ['Symbol', 'Time', 'Price', 'UnderlyingPrice', 'Delta', 'Gamma', 'Dividend', 'Interest']
 * cdef void theo_price_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                       dxf_const_string_t symbol_name,
 *                                       const dxf_event_data_t*data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_WriteUnraisable("dxfeed.core.listeners.listener.theo_price_default_listener", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_py_data);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "dxfeed/core/listeners/listener.pyx":227
 * 
 * UNDERLYING_COLUMNS = ['Symbol', 'Volatility', 'FrontVolatility', 'BackVolatility', 'PutCallRatio']
 * cdef void underlying_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                       dxf_const_string_t symbol_name,
 *                                       const dxf_event_data_t*data,
 */

static void __pyx_f_6dxfeed_4core_9listeners_8listener_underlying_default_listener(CYTHON_UNUSED int __pyx_v_event_type, dxf_const_string_t __pyx_v_symbol_name, dxf_event_data_t const *__pyx_v_data, int __pyx_v_data_count, void *__pyx_v_user_data) {
  dxf_underlying_t *__pyx_v_underlying;
  PyObject *__pyx_v_py_data = NULL;
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("underlying_default_listener", 1);

  /* "dxfeed/core/listeners/listener.pyx":232
 *                                       int data_count,
 *                                       void*user_data) nogil:
 *     cdef dxf_underlying_t*underlying = <dxf_underlying_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*try:*/ {
    __pyx_v_underlying = ((dxf_underlying_t *)__pyx_v_data);

    /* "dxfeed/core/listeners/listener.pyx":233
 *                                       void*user_data) nogil:
 *     cdef dxf_underlying_t*underlying = <dxf_underlying_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "dxfeed/core/listeners/listener.pyx":234
 *     cdef dxf_underlying_t*underlying = <dxf_underlying_t*> data
 *     with gil:
 *         py_data = <object> user_data             # <<<<<<<<<<<<<<
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 */
          __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
          __Pyx_INCREF(__pyx_t_1);
          __pyx_v_py_data = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "dxfeed/core/listeners/listener.pyx":235
 *     with gil:
 *         py_data = <object> user_data
 *         for i in range(data_count):             # <<<<<<<<<<<<<<
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  underlying[i].volatility,
 */
          __pyx_t_2 = __pyx_v_data_count;
          __pyx_t_3 = __pyx_t_2;
          for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
            __pyx_v_i = __pyx_t_4;

            /* "dxfeed/core/listeners/listener.pyx":236
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  underlying[i].volatility,
 *                                  underlying[i].front_volatility,
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_data, __pyx_n_s_safe_append); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 236, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_6 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t(__pyx_v_symbol_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 236, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);

            /* "dxfeed/core/listeners/listener.pyx":237
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  underlying[i].volatility,             # <<<<<<<<<<<<<<
 *                                  underlying[i].front_volatility,
 *                                  underlying[i].back_volatility,
 */
            __pyx_t_7 = PyFloat_FromDouble((__pyx_v_underlying[__pyx_v_i]).volatility); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 237, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);

            /* "dxfeed/core/listeners/listener.pyx":238
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  underlying[i].volatility,
 *                                  underlying[i].front_volatility,             # <<<<<<<<<<<<<<
 *                                  underlying[i].back_volatility,
 *                                  underlying[i].put_call_ratio])
 */
            __pyx_t_8 = PyFloat_FromDouble((__pyx_v_underlying[__pyx_v_i]).front_volatility); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 238, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_8);

            /* "dxfeed/core/listeners/listener.pyx":239
 *                                  underlying[i].volatility,
 *                                  underlying[i].front_volatility,
 *                                  underlying[i].back_volatility,             # <<<<<<<<<<<<<<
 *                                  underlying[i].put_call_ratio])
 * 
 */
            __pyx_t_9 = PyFloat_FromDouble((__pyx_v_underlying[__pyx_v_i]).back_volatility); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 239, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_9);

            /* "dxfeed/core/listeners/listener.pyx":240
 *                                  underlying[i].front_volatility,
 *                                  underlying[i].back_volatility,
 *                                  underlying[i].put_call_ratio])             # <<<<<<<<<<<<<<
 * 
 * SERIES_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'Sequence', 'Sequence', 'Expiration', 'Volatility',
 */
            __pyx_t_10 = PyFloat_FromDouble((__pyx_v_underlying[__pyx_v_i]).put_call_ratio); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 240, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_10);

            /* "dxfeed/core/listeners/listener.pyx":236
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  underlying[i].volatility,
 *                                  underlying[i].front_volatility,
 */
            __pyx_t_11 = PyList_New(5); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 236, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_GIVEREF(__pyx_t_6);
            PyList_SET_ITEM(__pyx_t_11, 0, __pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_7);
            PyList_SET_ITEM(__pyx_t_11, 1, __pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_8);
            PyList_SET_ITEM(__pyx_t_11, 2, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_9);
            PyList_SET_ITEM(__pyx_t_11, 3, __pyx_t_9);
            __Pyx_GIVEREF(__pyx_t_10);
            PyList_SET_ITEM(__pyx_t_11, 4, __pyx_t_10);
            __pyx_t_6 = 0;
            __pyx_t_7 = 0;
            __pyx_t_8 = 0;
            __pyx_t_9 = 0;
            __pyx_t_10 = 0;
            __pyx_t_10 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_10)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_10);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_10, __pyx_t_11) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_11);
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 236, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
        }

        /* "dxfeed/core/listeners/listener.pyx":233
 *                                       void*user_data) nogil:
 *     cdef dxf_underlying_t*underlying = <dxf_underlying_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }
  }

  /* "dxfeed/core/listeners/listener.pyx":232
 *                                       int data_count,
 *                                       void*user_data) nogil:
 *     cdef dxf_underlying_t*underlying = <dxf_underlying_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*finally:*/ {
    /*normal exit:*/{
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L5;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "dxfeed/core/listeners/listener.pyx":227
 * 
 * UNDERLYING_COLUMNS = ['Symbol', 'Volatility', 'FrontVolatility', 'BackVolatility', 'PutCallRatio']
 * cdef void underlying_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                       dxf_const_string_t symbol_name,
 *                                       const dxf_event_data_t*data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_WriteUnraisable("dxfeed.core.listeners.listener.underlying_default_listener", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_py_data);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "dxfeed/core/listeners/listener.pyx":244
 * SERIES_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'Sequence', 'Sequence', 'Expiration', 'Volatility',
 *                   'PutCallRatio', 'ForwardPrice', 'Dividend', 'Interest']
 * cdef void series_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                   dxf_const_string_t symbol_name,
 *                                   const dxf_event_data_t*data,
 */

static void __pyx_f_6dxfeed_4core_9listeners_8listener_series_default_listener(CYTHON_UNUSED int __pyx_v_event_type, dxf_const_string_t __pyx_v_symbol_name, dxf_event_data_t const *__pyx_v_data, int __pyx_v_data_count, void *__pyx_v_user_data) {
  dxf_series_t *__pyx_v_series;
  PyObject *__pyx_v_py_data = NULL;
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("series_default_listener", 1);

  /* "dxfeed/core/listeners/listener.pyx":249
 *                                   int data_count,
 *                                   void*user_data) nogil:
 *     cdef dxf_series_t*series = <dxf_series_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*try:*/ {
    __pyx_v_series = ((dxf_series_t *)__pyx_v_data);

    /* "dxfeed/core/listeners/listener.pyx":250
 *                                   void*user_data) nogil:
 *     cdef dxf_series_t*series = <dxf_series_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "dxfeed/core/listeners/listener.pyx":251
 *     cdef dxf_series_t*series = <dxf_series_t*> data
 *     with gil:
 *         py_data = <object> user_data             # <<<<<<<<<<<<<<
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 */
          __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
          __Pyx_INCREF(__pyx_t_1);
          __pyx_v_py_data = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "dxfeed/core/listeners/listener.pyx":252
 *     with gil:
 *         py_data = <object> user_data
 *         for i in range(data_count):             # <<<<<<<<<<<<<<
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  series[i].event_flags,
 */
          __pyx_t_2 = __pyx_v_data_count;
          __pyx_t_3 = __pyx_t_2;
          for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
            __pyx_v_i = __pyx_t_4;

            /* "dxfeed/core/listeners/listener.pyx":253
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  series[i].event_flags,
 *                                  series[i].index,
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_data, __pyx_n_s_safe_append); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 253, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_6 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t(__pyx_v_symbol_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 253, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);

            /* "dxfeed/core/listeners/listener.pyx":254
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  series[i].event_flags,             # <<<<<<<<<<<<<<
 *                                  series[i].index,
 *                                  series[i].time,
 */
            __pyx_t_7 = __Pyx_PyInt_From_dxf_uint_t((__pyx_v_series[__pyx_v_i]).event_flags); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 254, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);

            /* "dxfeed/core/listeners/listener.pyx":255
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  series[i].event_flags,
 *                                  series[i].index,             # <<<<<<<<<<<<<<
 *                                  series[i].time,
 *                                  series[i].sequence,
 */
            __pyx_t_8 = __Pyx_PyInt_From_dxf_long_t((__pyx_v_series[__pyx_v_i]).index); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 255, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_8);

            /* "dxfeed/core/listeners/listener.pyx":256
 *                                  series[i].event_flags,
 *                                  series[i].index,
 *                                  series[i].time,             # <<<<<<<<<<<<<<
 *                                  series[i].sequence,
 *                                  series[i].sequence,
 */
            __pyx_t_9 = __Pyx_PyInt_From_dxf_long_t((__pyx_v_series[__pyx_v_i]).time); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 256, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_9);

            /* "dxfeed/core/listeners/listener.pyx":257
 *                                  series[i].index,
 *                                  series[i].time,
 *                                  series[i].sequence,             # <<<<<<<<<<<<<<
 *                                  series[i].sequence,
 *                                  series[i].expiration,
 */
            __pyx_t_10 = __Pyx_PyInt_From_dxf_int_t((__pyx_v_series[__pyx_v_i]).sequence); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 257, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_10);

            /* "dxfeed/core/listeners/listener.pyx":258
 *                                  series[i].time,
 *                                  series[i].sequence,
 *                                  series[i].sequence,             # <<<<<<<<<<<<<<
 *                                  series[i].expiration,
 *                                  series[i].volatility,
 */
            __pyx_t_11 = __Pyx_PyInt_From_dxf_int_t((__pyx_v_series[__pyx_v_i]).sequence); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 258, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_11);

            /* "dxfeed/core/listeners/listener.pyx":259
 *                                  series[i].sequence,
 *                                  series[i].sequence,
 *                                  series[i].expiration,             # <<<<<<<<<<<<<<
 *                                  series[i].volatility,
 *                                  series[i].put_call_ratio,
 */
            __pyx_t_12 = __Pyx_PyInt_From_dxf_dayid_t((__pyx_v_series[__pyx_v_i]).expiration); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 259, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);

            /* "dxfeed/core/listeners/listener.pyx":260
 *                                  series[i].sequence,
 *                                  series[i].expiration,
 *                                  series[i].volatility,             # <<<<<<<<<<<<<<
 *                                  series[i].put_call_ratio,
 *                                  series[i].forward_price,
 */
            __pyx_t_13 = PyFloat_FromDouble((__pyx_v_series[__pyx_v_i]).volatility); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 260, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_13);

            /* "dxfeed/core/listeners/listener.pyx":261
 *                                  series[i].expiration,
 *                                  series[i].volatility,
 *                                  series[i].put_call_ratio,             # <<<<<<<<<<<<<<
 *                                  series[i].forward_price,
 *                                  series[i].dividend,
 */
            __pyx_t_14 = PyFloat_FromDouble((__pyx_v_series[__pyx_v_i]).put_call_ratio); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 261, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_14);

            /* "dxfeed/core/listeners/listener.pyx":262
 *                                  series[i].volatility,
 *                                  series[i].put_call_ratio,
 *                                  series[i].forward_price,             # <<<<<<<<<<<<<<
 *                                  series[i].dividend,
 *                                  series[i].interest])
 */
            __pyx_t_15 = PyFloat_FromDouble((__pyx_v_series[__pyx_v_i]).forward_price); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 262, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_15);

            /* "dxfeed/core/listeners/listener.pyx":263
 *                                  series[i].put_call_ratio,
 *                                  series[i].forward_price,
 *                                  series[i].dividend,             # <<<<<<<<<<<<<<
 *                                  series[i].interest])
 * 
 */
            __pyx_t_16 = PyFloat_FromDouble((__pyx_v_series[__pyx_v_i]).dividend); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 263, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_16);

            /* "dxfeed/core/listeners/listener.pyx":264
 *                                  series[i].forward_price,
 *                                  series[i].dividend,
 *                                  series[i].interest])             # <<<<<<<<<<<<<<
 * 
 * CONFIGURATION_COLUMNS = ['Symbol', 'Version', 'Object']
 */
            __pyx_t_17 = PyFloat_FromDouble((__pyx_v_series[__pyx_v_i]).interest); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 264, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_17);

            /* "dxfeed/core/listeners/listener.pyx":253
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  series[i].event_flags,
 *                                  series[i].index,
 */
            __pyx_t_18 = PyList_New(12); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 253, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_18);
            __Pyx_GIVEREF(__pyx_t_6);
            PyList_SET_ITEM(__pyx_t_18, 0, __pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_7);
            PyList_SET_ITEM(__pyx_t_18, 1, __pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_8);
            PyList_SET_ITEM(__pyx_t_18, 2, __pyx_t_8);
            __Pyx_GIVEREF(__pyx_t_9);
            PyList_SET_ITEM(__pyx_t_18, 3, __pyx_t_9);
            __Pyx_GIVEREF(__pyx_t_10);
            PyList_SET_ITEM(__pyx_t_18, 4, __pyx_t_10);
            __Pyx_GIVEREF(__pyx_t_11);
            PyList_SET_ITEM(__pyx_t_18, 5, __pyx_t_11);
            __Pyx_GIVEREF(__pyx_t_12);
            PyList_SET_ITEM(__pyx_t_18, 6, __pyx_t_12);
            __Pyx_GIVEREF(__pyx_t_13);
            PyList_SET_ITEM(__pyx_t_18, 7, __pyx_t_13);
            __Pyx_GIVEREF(__pyx_t_14);
            PyList_SET_ITEM(__pyx_t_18, 8, __pyx_t_14);
            __Pyx_GIVEREF(__pyx_t_15);
            PyList_SET_ITEM(__pyx_t_18, 9, __pyx_t_15);
            __Pyx_GIVEREF(__pyx_t_16);
            PyList_SET_ITEM(__pyx_t_18, 10, __pyx_t_16);
            __Pyx_GIVEREF(__pyx_t_17);
            PyList_SET_ITEM(__pyx_t_18, 11, __pyx_t_17);
            __pyx_t_6 = 0;
            __pyx_t_7 = 0;
            __pyx_t_8 = 0;
            __pyx_t_9 = 0;
            __pyx_t_10 = 0;
            __pyx_t_11 = 0;
            __pyx_t_12 = 0;
            __pyx_t_13 = 0;
            __pyx_t_14 = 0;
            __pyx_t_15 = 0;
            __pyx_t_16 = 0;
            __pyx_t_17 = 0;
            __pyx_t_17 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_17)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_17);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_1 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_17, __pyx_t_18) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_18);
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 253, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
        }

        /* "dxfeed/core/listeners/listener.pyx":250
 *                                   void*user_data) nogil:
 *     cdef dxf_series_t*series = <dxf_series_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }
  }

  /* "dxfeed/core/listeners/listener.pyx":249
 *                                   int data_count,
 *                                   void*user_data) nogil:
 *     cdef dxf_series_t*series = <dxf_series_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*finally:*/ {
    /*normal exit:*/{
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L5;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "dxfeed/core/listeners/listener.pyx":244
 * SERIES_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'Sequence', 'Sequence', 'Expiration', 'Volatility',
 *                   'PutCallRatio', 'ForwardPrice', 'Dividend', 'Interest']
 * cdef void series_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                   dxf_const_string_t symbol_name,
 *                                   const dxf_event_data_t*data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_WriteUnraisable("dxfeed.core.listeners.listener.series_default_listener", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_py_data);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "dxfeed/core/listeners/listener.pyx":267
 * 
 * CONFIGURATION_COLUMNS = ['Symbol', 'Version', 'Object']
 * cdef void configuration_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                          dxf_const_string_t symbol_name,
 *                                          const dxf_event_data_t*data,
 */

static void __pyx_f_6dxfeed_4core_9listeners_8listener_configuration_default_listener(CYTHON_UNUSED int __pyx_v_event_type, dxf_const_string_t __pyx_v_symbol_name, dxf_event_data_t const *__pyx_v_data, int __pyx_v_data_count, void *__pyx_v_user_data) {
  dxf_configuration_t *__pyx_v_config;
  PyObject *__pyx_v_py_data = NULL;
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("configuration_default_listener", 1);

  /* "dxfeed/core/listeners/listener.pyx":272
 *                                          int data_count,
 *                                          void*user_data) nogil:
 *     cdef dxf_configuration_t*config = <dxf_configuration_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*try:*/ {
    __pyx_v_config = ((dxf_configuration_t *)__pyx_v_data);

    /* "dxfeed/core/listeners/listener.pyx":273
 *                                          void*user_data) nogil:
 *     cdef dxf_configuration_t*config = <dxf_configuration_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "dxfeed/core/listeners/listener.pyx":274
 *     cdef dxf_configuration_t*config = <dxf_configuration_t*> data
 *     with gil:
 *         py_data = <object> user_data             # <<<<<<<<<<<<<<
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 */
          __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
          __Pyx_INCREF(__pyx_t_1);
          __pyx_v_py_data = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "dxfeed/core/listeners/listener.pyx":275
 *     with gil:
 *         py_data = <object> user_data
 *         for i in range(data_count):             # <<<<<<<<<<<<<<
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  config[i].version,
 */
          __pyx_t_2 = __pyx_v_data_count;
          __pyx_t_3 = __pyx_t_2;
          for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
            __pyx_v_i = __pyx_t_4;

            /* "dxfeed/core/listeners/listener.pyx":276
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  config[i].version,
 *                                  unicode_from_dxf_const_string_t(config[i].object)])
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_data, __pyx_n_s_safe_append); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 276, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_6 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t(__pyx_v_symbol_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 276, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_6);

            /* "dxfeed/core/listeners/listener.pyx":277
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  config[i].version,             # <<<<<<<<<<<<<<
 *                                  unicode_from_dxf_const_string_t(config[i].object)])
 */
            __pyx_t_7 = __Pyx_PyInt_From_dxf_int_t((__pyx_v_config[__pyx_v_i]).version); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 277, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_7);

            /* "dxfeed/core/listeners/listener.pyx":278
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),
 *                                  config[i].version,
 *                                  unicode_from_dxf_const_string_t(config[i].object)])             # <<<<<<<<<<<<<<
 */
            __pyx_t_8 = __pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t((__pyx_v_config[__pyx_v_i]).object); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 278, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_8);

            /* "dxfeed/core/listeners/listener.pyx":276
 *         py_data = <object> user_data
 *         for i in range(data_count):
 *             py_data.safe_append([unicode_from_dxf_const_string_t(symbol_name),             # <<<<<<<<<<<<<<
 *                                  config[i].version,
 *                                  unicode_from_dxf_const_string_t(config[i].object)])
 */
            __pyx_t_9 = PyList_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 276, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_GIVEREF(__pyx_t_6);
            PyList_SET_ITEM(__pyx_t_9, 0, __pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_7);
            PyList_SET_ITEM(__pyx_t_9, 1, __pyx_t_7);
            __Pyx_GIVEREF(__pyx_t_8);
            PyList_SET_ITEM(__pyx_t_9, 2, __pyx_t_8);
            __pyx_t_6 = 0;
            __pyx_t_7 = 0;
            __pyx_t_8 = 0;
            __pyx_t_8 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_8)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_8);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_9);
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 276, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
        }

        /* "dxfeed/core/listeners/listener.pyx":273
 *                                          void*user_data) nogil:
 *     cdef dxf_configuration_t*config = <dxf_configuration_t*> data
 *     with gil:             # <<<<<<<<<<<<<<
 *         py_data = <object> user_data
 *         for i in range(data_count):
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }
  }

  /* "dxfeed/core/listeners/listener.pyx":272
 *                                          int data_count,
 *                                          void*user_data) nogil:
 *     cdef dxf_configuration_t*config = <dxf_configuration_t*> data             # <<<<<<<<<<<<<<
 *     with gil:
 *         py_data = <object> user_data
 */
  /*finally:*/ {
    /*normal exit:*/{
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L5;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "dxfeed/core/listeners/listener.pyx":267
 * 
 * CONFIGURATION_COLUMNS = ['Symbol', 'Version', 'Object']
 * cdef void configuration_default_listener(int event_type,             # <<<<<<<<<<<<<<
 *                                          dxf_const_string_t symbol_name,
 *                                          const dxf_event_data_t*data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_WriteUnraisable("dxfeed.core.listeners.listener.configuration_default_listener", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_py_data);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}
static struct __pyx_vtabstruct_6dxfeed_4core_9listeners_8listener_FuncWrapper __pyx_vtable_6dxfeed_4core_9listeners_8listener_FuncWrapper;

static PyObject *__pyx_tp_new_6dxfeed_4core_9listeners_8listener_FuncWrapper(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *)o);
  p->__pyx_vtab = __pyx_vtabptr_6dxfeed_4core_9listeners_8listener_FuncWrapper;
  if (unlikely(__pyx_pw_6dxfeed_4core_9listeners_8listener_11FuncWrapper_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_6dxfeed_4core_9listeners_8listener_FuncWrapper(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyMethodDef __pyx_methods_6dxfeed_4core_9listeners_8listener_FuncWrapper[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6dxfeed_4core_9listeners_8listener_11FuncWrapper_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6dxfeed_4core_9listeners_8listener_11FuncWrapper_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6dxfeed_4core_9listeners_8listener_FuncWrapper = {
  PyVarObject_HEAD_INIT(0, 0)
  "dxfeed.core.listeners.listener.FuncWrapper", /*tp_name*/
  sizeof(struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6dxfeed_4core_9listeners_8listener_FuncWrapper, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6dxfeed_4core_9listeners_8listener_FuncWrapper, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6dxfeed_4core_9listeners_8listener_FuncWrapper, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_listener(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_listener},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "listener",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_52HighPrice, __pyx_k_52HighPrice, sizeof(__pyx_k_52HighPrice), 0, 1, 0, 0},
  {&__pyx_kp_u_52LowPrice, __pyx_k_52LowPrice, sizeof(__pyx_k_52LowPrice), 0, 1, 0, 0},
  {&__pyx_n_u_AskExchangeCode, __pyx_k_AskExchangeCode, sizeof(__pyx_k_AskExchangeCode), 0, 1, 0, 1},
  {&__pyx_n_u_AskPrice, __pyx_k_AskPrice, sizeof(__pyx_k_AskPrice), 0, 1, 0, 1},
  {&__pyx_n_u_AskSize, __pyx_k_AskSize, sizeof(__pyx_k_AskSize), 0, 1, 0, 1},
  {&__pyx_n_u_AskTime, __pyx_k_AskTime, sizeof(__pyx_k_AskTime), 0, 1, 0, 1},
  {&__pyx_n_u_AskVolume, __pyx_k_AskVolume, sizeof(__pyx_k_AskVolume), 0, 1, 0, 1},
  {&__pyx_n_u_BackVolatility, __pyx_k_BackVolatility, sizeof(__pyx_k_BackVolatility), 0, 1, 0, 1},
  {&__pyx_n_u_Beta, __pyx_k_Beta, sizeof(__pyx_k_Beta), 0, 1, 0, 1},
  {&__pyx_n_u_BidExchangeCode, __pyx_k_BidExchangeCode, sizeof(__pyx_k_BidExchangeCode), 0, 1, 0, 1},
  {&__pyx_n_u_BidPrice, __pyx_k_BidPrice, sizeof(__pyx_k_BidPrice), 0, 1, 0, 1},
  {&__pyx_n_u_BidSize, __pyx_k_BidSize, sizeof(__pyx_k_BidSize), 0, 1, 0, 1},
  {&__pyx_n_u_BidTime, __pyx_k_BidTime, sizeof(__pyx_k_BidTime), 0, 1, 0, 1},
  {&__pyx_n_u_BidVolume, __pyx_k_BidVolume, sizeof(__pyx_k_BidVolume), 0, 1, 0, 1},
  {&__pyx_n_u_Buyer, __pyx_k_Buyer, sizeof(__pyx_k_Buyer), 0, 1, 0, 1},
  {&__pyx_n_s_CANDLE_COLUMNS, __pyx_k_CANDLE_COLUMNS, sizeof(__pyx_k_CANDLE_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_s_CONFIGURATION_COLUMNS, __pyx_k_CONFIGURATION_COLUMNS, sizeof(__pyx_k_CONFIGURATION_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_u_Change, __pyx_k_Change, sizeof(__pyx_k_Change), 0, 1, 0, 1},
  {&__pyx_n_u_Close, __pyx_k_Close, sizeof(__pyx_k_Close), 0, 1, 0, 1},
  {&__pyx_n_u_Count, __pyx_k_Count, sizeof(__pyx_k_Count), 0, 1, 0, 1},
  {&__pyx_n_u_DayClosePrice, __pyx_k_DayClosePrice, sizeof(__pyx_k_DayClosePrice), 0, 1, 0, 1},
  {&__pyx_n_u_DayHighPrice, __pyx_k_DayHighPrice, sizeof(__pyx_k_DayHighPrice), 0, 1, 0, 1},
  {&__pyx_n_u_DayId, __pyx_k_DayId, sizeof(__pyx_k_DayId), 0, 1, 0, 1},
  {&__pyx_n_u_DayLowPrice, __pyx_k_DayLowPrice, sizeof(__pyx_k_DayLowPrice), 0, 1, 0, 1},
  {&__pyx_n_u_DayVolume, __pyx_k_DayVolume, sizeof(__pyx_k_DayVolume), 0, 1, 0, 1},
  {&__pyx_n_u_Delta, __pyx_k_Delta, sizeof(__pyx_k_Delta), 0, 1, 0, 1},
  {&__pyx_n_u_Description, __pyx_k_Description, sizeof(__pyx_k_Description), 0, 1, 0, 1},
  {&__pyx_n_u_DivFreq, __pyx_k_DivFreq, sizeof(__pyx_k_DivFreq), 0, 1, 0, 1},
  {&__pyx_n_u_Dividend, __pyx_k_Dividend, sizeof(__pyx_k_Dividend), 0, 1, 0, 1},
  {&__pyx_n_u_EPS, __pyx_k_EPS, sizeof(__pyx_k_EPS), 0, 1, 0, 1},
  {&__pyx_n_u_EventFlags, __pyx_k_EventFlags, sizeof(__pyx_k_EventFlags), 0, 1, 0, 1},
  {&__pyx_n_u_ExchangeCode, __pyx_k_ExchangeCode, sizeof(__pyx_k_ExchangeCode), 0, 1, 0, 1},
  {&__pyx_n_u_ExchangeSaleConditions, __pyx_k_ExchangeSaleConditions, sizeof(__pyx_k_ExchangeSaleConditions), 0, 1, 0, 1},
  {&__pyx_n_u_ExdDivAmount, __pyx_k_ExdDivAmount, sizeof(__pyx_k_ExdDivAmount), 0, 1, 0, 1},
  {&__pyx_n_u_ExdDivDate, __pyx_k_ExdDivDate, sizeof(__pyx_k_ExdDivDate), 0, 1, 0, 1},
  {&__pyx_n_u_Expiration, __pyx_k_Expiration, sizeof(__pyx_k_Expiration), 0, 1, 0, 1},
  {&__pyx_n_u_ForwardPrice, __pyx_k_ForwardPrice, sizeof(__pyx_k_ForwardPrice), 0, 1, 0, 1},
  {&__pyx_n_u_FrontVolatility, __pyx_k_FrontVolatility, sizeof(__pyx_k_FrontVolatility), 0, 1, 0, 1},
  {&__pyx_n_s_FuncWrapper, __pyx_k_FuncWrapper, sizeof(__pyx_k_FuncWrapper), 0, 0, 1, 1},
  {&__pyx_n_s_GREEKS_COLUMNS, __pyx_k_GREEKS_COLUMNS, sizeof(__pyx_k_GREEKS_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_u_Gamma, __pyx_k_Gamma, sizeof(__pyx_k_Gamma), 0, 1, 0, 1},
  {&__pyx_n_u_High, __pyx_k_High, sizeof(__pyx_k_High), 0, 1, 0, 1},
  {&__pyx_n_u_ImpVolatility, __pyx_k_ImpVolatility, sizeof(__pyx_k_ImpVolatility), 0, 1, 0, 1},
  {&__pyx_n_u_Index, __pyx_k_Index, sizeof(__pyx_k_Index), 0, 1, 0, 1},
  {&__pyx_n_u_Interest, __pyx_k_Interest, sizeof(__pyx_k_Interest), 0, 1, 0, 1},
  {&__pyx_n_u_IsETH, __pyx_k_IsETH, sizeof(__pyx_k_IsETH), 0, 1, 0, 1},
  {&__pyx_n_u_IsEthTrade, __pyx_k_IsEthTrade, sizeof(__pyx_k_IsEthTrade), 0, 1, 0, 1},
  {&__pyx_n_u_IsSpreadLeg, __pyx_k_IsSpreadLeg, sizeof(__pyx_k_IsSpreadLeg), 0, 1, 0, 1},
  {&__pyx_n_u_IsValidTick, __pyx_k_IsValidTick, sizeof(__pyx_k_IsValidTick), 0, 1, 0, 1},
  {&__pyx_n_u_Low, __pyx_k_Low, sizeof(__pyx_k_Low), 0, 1, 0, 1},
  {&__pyx_n_u_MarketMaker, __pyx_k_MarketMaker, sizeof(__pyx_k_MarketMaker), 0, 1, 0, 1},
  {&__pyx_n_s_ORDER_COLUMNS, __pyx_k_ORDER_COLUMNS, sizeof(__pyx_k_ORDER_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_u_Object, __pyx_k_Object, sizeof(__pyx_k_Object), 0, 1, 0, 1},
  {&__pyx_n_u_Open, __pyx_k_Open, sizeof(__pyx_k_Open), 0, 1, 0, 1},
  {&__pyx_n_u_OpenInterest, __pyx_k_OpenInterest, sizeof(__pyx_k_OpenInterest), 0, 1, 0, 1},
  {&__pyx_n_s_PROFILE_COLUMNS, __pyx_k_PROFILE_COLUMNS, sizeof(__pyx_k_PROFILE_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_u_PrevDayClosePrice, __pyx_k_PrevDayClosePrice, sizeof(__pyx_k_PrevDayClosePrice), 0, 1, 0, 1},
  {&__pyx_n_u_PrevDayId, __pyx_k_PrevDayId, sizeof(__pyx_k_PrevDayId), 0, 1, 0, 1},
  {&__pyx_n_u_PrevDayVolume, __pyx_k_PrevDayVolume, sizeof(__pyx_k_PrevDayVolume), 0, 1, 0, 1},
  {&__pyx_n_u_Price, __pyx_k_Price, sizeof(__pyx_k_Price), 0, 1, 0, 1},
  {&__pyx_n_u_PutCallRatio, __pyx_k_PutCallRatio, sizeof(__pyx_k_PutCallRatio), 0, 1, 0, 1},
  {&__pyx_n_s_QUOTE_COLUMNS, __pyx_k_QUOTE_COLUMNS, sizeof(__pyx_k_QUOTE_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_u_RawFlags, __pyx_k_RawFlags, sizeof(__pyx_k_RawFlags), 0, 1, 0, 1},
  {&__pyx_n_u_Rho, __pyx_k_Rho, sizeof(__pyx_k_Rho), 0, 1, 0, 1},
  {&__pyx_n_s_SERIES_COLUMNS, __pyx_k_SERIES_COLUMNS, sizeof(__pyx_k_SERIES_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_s_SUMMARY_COLUMNS, __pyx_k_SUMMARY_COLUMNS, sizeof(__pyx_k_SUMMARY_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_u_Scope, __pyx_k_Scope, sizeof(__pyx_k_Scope), 0, 1, 0, 1},
  {&__pyx_n_u_Seller, __pyx_k_Seller, sizeof(__pyx_k_Seller), 0, 1, 0, 1},
  {&__pyx_n_u_Sequence, __pyx_k_Sequence, sizeof(__pyx_k_Sequence), 0, 1, 0, 1},
  {&__pyx_n_u_Shares, __pyx_k_Shares, sizeof(__pyx_k_Shares), 0, 1, 0, 1},
  {&__pyx_n_u_Side, __pyx_k_Side, sizeof(__pyx_k_Side), 0, 1, 0, 1},
  {&__pyx_n_u_Size, __pyx_k_Size, sizeof(__pyx_k_Size), 0, 1, 0, 1},
  {&__pyx_n_u_SpreadSymbol, __pyx_k_SpreadSymbol, sizeof(__pyx_k_SpreadSymbol), 0, 1, 0, 1},
  {&__pyx_n_u_StatusReason, __pyx_k_StatusReason, sizeof(__pyx_k_StatusReason), 0, 1, 0, 1},
  {&__pyx_n_u_Symbol, __pyx_k_Symbol, sizeof(__pyx_k_Symbol), 0, 1, 0, 1},
  {&__pyx_n_s_THEO_PRICE_COLUMNS, __pyx_k_THEO_PRICE_COLUMNS, sizeof(__pyx_k_THEO_PRICE_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_s_TIME_AND_SALE_COLUMNS, __pyx_k_TIME_AND_SALE_COLUMNS, sizeof(__pyx_k_TIME_AND_SALE_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_s_TRADE_COLUMNS, __pyx_k_TRADE_COLUMNS, sizeof(__pyx_k_TRADE_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_u_Theta, __pyx_k_Theta, sizeof(__pyx_k_Theta), 0, 1, 0, 1},
  {&__pyx_n_u_Tick, __pyx_k_Tick, sizeof(__pyx_k_Tick), 0, 1, 0, 1},
  {&__pyx_n_u_Time, __pyx_k_Time, sizeof(__pyx_k_Time), 0, 1, 0, 1},
  {&__pyx_n_u_TimeNanos, __pyx_k_TimeNanos, sizeof(__pyx_k_TimeNanos), 0, 1, 0, 1},
  {&__pyx_n_u_TradeThroughExempt, __pyx_k_TradeThroughExempt, sizeof(__pyx_k_TradeThroughExempt), 0, 1, 0, 1},
  {&__pyx_n_u_Type, __pyx_k_Type, sizeof(__pyx_k_Type), 0, 1, 0, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_UNDERLYING_COLUMNS, __pyx_k_UNDERLYING_COLUMNS, sizeof(__pyx_k_UNDERLYING_COLUMNS), 0, 0, 1, 1},
  {&__pyx_n_u_UnderlyingPrice, __pyx_k_UnderlyingPrice, sizeof(__pyx_k_UnderlyingPrice), 0, 1, 0, 1},
  {&__pyx_n_u_VWap, __pyx_k_VWap, sizeof(__pyx_k_VWap), 0, 1, 0, 1},
  {&__pyx_n_u_Vega, __pyx_k_Vega, sizeof(__pyx_k_Vega), 0, 1, 0, 1},
  {&__pyx_n_u_Version, __pyx_k_Version, sizeof(__pyx_k_Version), 0, 1, 0, 1},
  {&__pyx_n_u_Volatility, __pyx_k_Volatility, sizeof(__pyx_k_Volatility), 0, 1, 0, 1},
  {&__pyx_n_u_Volume, __pyx_k_Volume, sizeof(__pyx_k_Volume), 0, 1, 0, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_safe_append, __pyx_k_safe_append, sizeof(__pyx_k_safe_append), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(1, 22, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("trade_default_listener", (void (*)(void))__pyx_f_6dxfeed_4core_9listeners_8listener_trade_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("quote_default_listener", (void (*)(void))__pyx_f_6dxfeed_4core_9listeners_8listener_quote_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("summary_default_listener", (void (*)(void))__pyx_f_6dxfeed_4core_9listeners_8listener_summary_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("profile_default_listener", (void (*)(void))__pyx_f_6dxfeed_4core_9listeners_8listener_profile_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("time_and_sale_default_listener", (void (*)(void))__pyx_f_6dxfeed_4core_9listeners_8listener_time_and_sale_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("candle_default_listener", (void (*)(void))__pyx_f_6dxfeed_4core_9listeners_8listener_candle_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("order_default_listener", (void (*)(void))__pyx_f_6dxfeed_4core_9listeners_8listener_order_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("greeks_default_listener", (void (*)(void))__pyx_f_6dxfeed_4core_9listeners_8listener_greeks_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("theo_price_default_listener", (void (*)(void))__pyx_f_6dxfeed_4core_9listeners_8listener_theo_price_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("underlying_default_listener", (void (*)(void))__pyx_f_6dxfeed_4core_9listeners_8listener_underlying_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("series_default_listener", (void (*)(void))__pyx_f_6dxfeed_4core_9listeners_8listener_series_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("configuration_default_listener", (void (*)(void))__pyx_f_6dxfeed_4core_9listeners_8listener_configuration_default_listener, "void (int, dxf_const_string_t, dxf_event_data_t const *, int, void *)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_6dxfeed_4core_9listeners_8listener_FuncWrapper = &__pyx_vtable_6dxfeed_4core_9listeners_8listener_FuncWrapper;
  __pyx_vtable_6dxfeed_4core_9listeners_8listener_FuncWrapper.make_from_ptr = (struct __pyx_obj_6dxfeed_4core_9listeners_8listener_FuncWrapper *(*)(dxf_event_listener_t))__pyx_f_6dxfeed_4core_9listeners_8listener_11FuncWrapper_make_from_ptr;
  if (PyType_Ready(&__pyx_type_6dxfeed_4core_9listeners_8listener_FuncWrapper) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6dxfeed_4core_9listeners_8listener_FuncWrapper.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6dxfeed_4core_9listeners_8listener_FuncWrapper.tp_dictoffset && __pyx_type_6dxfeed_4core_9listeners_8listener_FuncWrapper.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6dxfeed_4core_9listeners_8listener_FuncWrapper.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6dxfeed_4core_9listeners_8listener_FuncWrapper.tp_dict, __pyx_vtabptr_6dxfeed_4core_9listeners_8listener_FuncWrapper) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FuncWrapper, (PyObject *)&__pyx_type_6dxfeed_4core_9listeners_8listener_FuncWrapper) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6dxfeed_4core_9listeners_8listener_FuncWrapper) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __pyx_ptype_6dxfeed_4core_9listeners_8listener_FuncWrapper = &__pyx_type_6dxfeed_4core_9listeners_8listener_FuncWrapper;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("dxfeed.core.utils.helpers"); if (!__pyx_t_1) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "unicode_from_dxf_const_string_t", (void (**)(void))&__pyx_f_6dxfeed_4core_5utils_7helpers_unicode_from_dxf_const_string_t, "PyObject *(dxf_const_string_t)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initlistener(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initlistener(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_listener(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_listener(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_listener(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'listener' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_listener(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(1, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("listener", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_dxfeed__core__listeners__listener) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(1, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "dxfeed.core.listeners.listener")) {
      if (unlikely(PyDict_SetItemString(modules, "dxfeed.core.listeners.listener", __pyx_m) < 0)) __PYX_ERR(1, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) goto __pyx_L1_error;
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) goto __pyx_L1_error;
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() != 0)) goto __pyx_L1_error;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif

  /* "dxfeed/core/listeners/listener.pyx":13
 *         return out
 * 
 * TRADE_COLUMNS = ['Symbol', 'Price', 'ExchangeCode', 'Size', 'Tick', 'Change', 'DayVolume', 'Time', 'IsETH']             # <<<<<<<<<<<<<<
 * cdef void trade_default_listener(int event_type,
 *                                  dxf_const_string_t symbol_name,
 */
  __pyx_t_1 = PyList_New(9); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_Symbol);
  __Pyx_GIVEREF(__pyx_n_u_Symbol);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_Symbol);
  __Pyx_INCREF(__pyx_n_u_Price);
  __Pyx_GIVEREF(__pyx_n_u_Price);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_Price);
  __Pyx_INCREF(__pyx_n_u_ExchangeCode);
  __Pyx_GIVEREF(__pyx_n_u_ExchangeCode);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_ExchangeCode);
  __Pyx_INCREF(__pyx_n_u_Size);
  __Pyx_GIVEREF(__pyx_n_u_Size);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_Size);
  __Pyx_INCREF(__pyx_n_u_Tick);
  __Pyx_GIVEREF(__pyx_n_u_Tick);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_Tick);
  __Pyx_INCREF(__pyx_n_u_Change);
  __Pyx_GIVEREF(__pyx_n_u_Change);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_u_Change);
  __Pyx_INCREF(__pyx_n_u_DayVolume);
  __Pyx_GIVEREF(__pyx_n_u_DayVolume);
  PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_u_DayVolume);
  __Pyx_INCREF(__pyx_n_u_Time);
  __Pyx_GIVEREF(__pyx_n_u_Time);
  PyList_SET_ITEM(__pyx_t_1, 7, __pyx_n_u_Time);
  __Pyx_INCREF(__pyx_n_u_IsETH);
  __Pyx_GIVEREF(__pyx_n_u_IsETH);
  PyList_SET_ITEM(__pyx_t_1, 8, __pyx_n_u_IsETH);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TRADE_COLUMNS, __pyx_t_1) < 0) __PYX_ERR(1, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/listeners/listener.pyx":33
 *                                  trades[i].is_eth])
 * 
 * QUOTE_COLUMNS = ['Symbol', 'BidTime', 'BidExchangeCode', 'BidPrice', 'BidSize', 'AskTime', 'AskExchangeCode',             # <<<<<<<<<<<<<<
 *                  'AskPrice', 'AskSize', 'Scope']
 * cdef void quote_default_listener(int event_type,
 */
  __pyx_t_1 = PyList_New(10); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_Symbol);
  __Pyx_GIVEREF(__pyx_n_u_Symbol);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_Symbol);
  __Pyx_INCREF(__pyx_n_u_BidTime);
  __Pyx_GIVEREF(__pyx_n_u_BidTime);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_BidTime);
  __Pyx_INCREF(__pyx_n_u_BidExchangeCode);
  __Pyx_GIVEREF(__pyx_n_u_BidExchangeCode);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_BidExchangeCode);
  __Pyx_INCREF(__pyx_n_u_BidPrice);
  __Pyx_GIVEREF(__pyx_n_u_BidPrice);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_BidPrice);
  __Pyx_INCREF(__pyx_n_u_BidSize);
  __Pyx_GIVEREF(__pyx_n_u_BidSize);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_BidSize);
  __Pyx_INCREF(__pyx_n_u_AskTime);
  __Pyx_GIVEREF(__pyx_n_u_AskTime);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_u_AskTime);
  __Pyx_INCREF(__pyx_n_u_AskExchangeCode);
  __Pyx_GIVEREF(__pyx_n_u_AskExchangeCode);
  PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_u_AskExchangeCode);
  __Pyx_INCREF(__pyx_n_u_AskPrice);
  __Pyx_GIVEREF(__pyx_n_u_AskPrice);
  PyList_SET_ITEM(__pyx_t_1, 7, __pyx_n_u_AskPrice);
  __Pyx_INCREF(__pyx_n_u_AskSize);
  __Pyx_GIVEREF(__pyx_n_u_AskSize);
  PyList_SET_ITEM(__pyx_t_1, 8, __pyx_n_u_AskSize);
  __Pyx_INCREF(__pyx_n_u_Scope);
  __Pyx_GIVEREF(__pyx_n_u_Scope);
  PyList_SET_ITEM(__pyx_t_1, 9, __pyx_n_u_Scope);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_QUOTE_COLUMNS, __pyx_t_1) < 0) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/listeners/listener.pyx":56
 *                                  <int> quotes[i].scope])
 * 
 * SUMMARY_COLUMNS = ['Symbol', 'DayId', 'DayHighPrice', 'DayLowPrice', 'DayClosePrice', 'PrevDayId', 'PrevDayClosePrice',             # <<<<<<<<<<<<<<
 *                    'PrevDayVolume', 'OpenInterest']
 * cdef void summary_default_listener(int event_type, dxf_const_string_t symbol_name,
 */
  __pyx_t_1 = PyList_New(9); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_Symbol);
  __Pyx_GIVEREF(__pyx_n_u_Symbol);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_Symbol);
  __Pyx_INCREF(__pyx_n_u_DayId);
  __Pyx_GIVEREF(__pyx_n_u_DayId);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_DayId);
  __Pyx_INCREF(__pyx_n_u_DayHighPrice);
  __Pyx_GIVEREF(__pyx_n_u_DayHighPrice);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_DayHighPrice);
  __Pyx_INCREF(__pyx_n_u_DayLowPrice);
  __Pyx_GIVEREF(__pyx_n_u_DayLowPrice);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_DayLowPrice);
  __Pyx_INCREF(__pyx_n_u_DayClosePrice);
  __Pyx_GIVEREF(__pyx_n_u_DayClosePrice);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_DayClosePrice);
  __Pyx_INCREF(__pyx_n_u_PrevDayId);
  __Pyx_GIVEREF(__pyx_n_u_PrevDayId);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_u_PrevDayId);
  __Pyx_INCREF(__pyx_n_u_PrevDayClosePrice);
  __Pyx_GIVEREF(__pyx_n_u_PrevDayClosePrice);
  PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_u_PrevDayClosePrice);
  __Pyx_INCREF(__pyx_n_u_PrevDayVolume);
  __Pyx_GIVEREF(__pyx_n_u_PrevDayVolume);
  PyList_SET_ITEM(__pyx_t_1, 7, __pyx_n_u_PrevDayVolume);
  __Pyx_INCREF(__pyx_n_u_OpenInterest);
  __Pyx_GIVEREF(__pyx_n_u_OpenInterest);
  PyList_SET_ITEM(__pyx_t_1, 8, __pyx_n_u_OpenInterest);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SUMMARY_COLUMNS, __pyx_t_1) < 0) __PYX_ERR(1, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/listeners/listener.pyx":75
 *                                  summary[i].open_interest])
 * 
 * PROFILE_COLUMNS = ['Symbol', 'Beta', 'EPS', 'DivFreq', 'ExdDivAmount', 'ExdDivDate', '52HighPrice', '52LowPrice',             # <<<<<<<<<<<<<<
 *                    'Shares', 'Description', 'RawFlags', 'StatusReason']
 * cdef void profile_default_listener(int event_type,
 */
  __pyx_t_1 = PyList_New(12); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_Symbol);
  __Pyx_GIVEREF(__pyx_n_u_Symbol);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_Symbol);
  __Pyx_INCREF(__pyx_n_u_Beta);
  __Pyx_GIVEREF(__pyx_n_u_Beta);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_Beta);
  __Pyx_INCREF(__pyx_n_u_EPS);
  __Pyx_GIVEREF(__pyx_n_u_EPS);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_EPS);
  __Pyx_INCREF(__pyx_n_u_DivFreq);
  __Pyx_GIVEREF(__pyx_n_u_DivFreq);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_DivFreq);
  __Pyx_INCREF(__pyx_n_u_ExdDivAmount);
  __Pyx_GIVEREF(__pyx_n_u_ExdDivAmount);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_ExdDivAmount);
  __Pyx_INCREF(__pyx_n_u_ExdDivDate);
  __Pyx_GIVEREF(__pyx_n_u_ExdDivDate);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_u_ExdDivDate);
  __Pyx_INCREF(__pyx_kp_u_52HighPrice);
  __Pyx_GIVEREF(__pyx_kp_u_52HighPrice);
  PyList_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u_52HighPrice);
  __Pyx_INCREF(__pyx_kp_u_52LowPrice);
  __Pyx_GIVEREF(__pyx_kp_u_52LowPrice);
  PyList_SET_ITEM(__pyx_t_1, 7, __pyx_kp_u_52LowPrice);
  __Pyx_INCREF(__pyx_n_u_Shares);
  __Pyx_GIVEREF(__pyx_n_u_Shares);
  PyList_SET_ITEM(__pyx_t_1, 8, __pyx_n_u_Shares);
  __Pyx_INCREF(__pyx_n_u_Description);
  __Pyx_GIVEREF(__pyx_n_u_Description);
  PyList_SET_ITEM(__pyx_t_1, 9, __pyx_n_u_Description);
  __Pyx_INCREF(__pyx_n_u_RawFlags);
  __Pyx_GIVEREF(__pyx_n_u_RawFlags);
  PyList_SET_ITEM(__pyx_t_1, 10, __pyx_n_u_RawFlags);
  __Pyx_INCREF(__pyx_n_u_StatusReason);
  __Pyx_GIVEREF(__pyx_n_u_StatusReason);
  PyList_SET_ITEM(__pyx_t_1, 11, __pyx_n_u_StatusReason);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PROFILE_COLUMNS, __pyx_t_1) < 0) __PYX_ERR(1, 75, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/listeners/listener.pyx":99
 *                                  unicode_from_dxf_const_string_t(p[i].status_reason)])
 * 
 * TIME_AND_SALE_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'ExchangeCode', 'Price', 'Size', 'BidPrice',             # <<<<<<<<<<<<<<
 *                          'AskPrice', 'ExchangeSaleConditions', 'RawFlags', 'Buyer', 'Seller', 'Side', 'Type',
 *                          'IsValidTick', 'IsEthTrade', 'TradeThroughExempt', 'IsSpreadLeg']
 */
  __pyx_t_1 = PyList_New(19); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_Symbol);
  __Pyx_GIVEREF(__pyx_n_u_Symbol);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_Symbol);
  __Pyx_INCREF(__pyx_n_u_EventFlags);
  __Pyx_GIVEREF(__pyx_n_u_EventFlags);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_EventFlags);
  __Pyx_INCREF(__pyx_n_u_Index);
  __Pyx_GIVEREF(__pyx_n_u_Index);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_Index);
  __Pyx_INCREF(__pyx_n_u_Time);
  __Pyx_GIVEREF(__pyx_n_u_Time);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_Time);
  __Pyx_INCREF(__pyx_n_u_ExchangeCode);
  __Pyx_GIVEREF(__pyx_n_u_ExchangeCode);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_ExchangeCode);
  __Pyx_INCREF(__pyx_n_u_Price);
  __Pyx_GIVEREF(__pyx_n_u_Price);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_u_Price);
  __Pyx_INCREF(__pyx_n_u_Size);
  __Pyx_GIVEREF(__pyx_n_u_Size);
  PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_u_Size);
  __Pyx_INCREF(__pyx_n_u_BidPrice);
  __Pyx_GIVEREF(__pyx_n_u_BidPrice);
  PyList_SET_ITEM(__pyx_t_1, 7, __pyx_n_u_BidPrice);
  __Pyx_INCREF(__pyx_n_u_AskPrice);
  __Pyx_GIVEREF(__pyx_n_u_AskPrice);
  PyList_SET_ITEM(__pyx_t_1, 8, __pyx_n_u_AskPrice);
  __Pyx_INCREF(__pyx_n_u_ExchangeSaleConditions);
  __Pyx_GIVEREF(__pyx_n_u_ExchangeSaleConditions);
  PyList_SET_ITEM(__pyx_t_1, 9, __pyx_n_u_ExchangeSaleConditions);
  __Pyx_INCREF(__pyx_n_u_RawFlags);
  __Pyx_GIVEREF(__pyx_n_u_RawFlags);
  PyList_SET_ITEM(__pyx_t_1, 10, __pyx_n_u_RawFlags);
  __Pyx_INCREF(__pyx_n_u_Buyer);
  __Pyx_GIVEREF(__pyx_n_u_Buyer);
  PyList_SET_ITEM(__pyx_t_1, 11, __pyx_n_u_Buyer);
  __Pyx_INCREF(__pyx_n_u_Seller);
  __Pyx_GIVEREF(__pyx_n_u_Seller);
  PyList_SET_ITEM(__pyx_t_1, 12, __pyx_n_u_Seller);
  __Pyx_INCREF(__pyx_n_u_Side);
  __Pyx_GIVEREF(__pyx_n_u_Side);
  PyList_SET_ITEM(__pyx_t_1, 13, __pyx_n_u_Side);
  __Pyx_INCREF(__pyx_n_u_Type);
  __Pyx_GIVEREF(__pyx_n_u_Type);
  PyList_SET_ITEM(__pyx_t_1, 14, __pyx_n_u_Type);
  __Pyx_INCREF(__pyx_n_u_IsValidTick);
  __Pyx_GIVEREF(__pyx_n_u_IsValidTick);
  PyList_SET_ITEM(__pyx_t_1, 15, __pyx_n_u_IsValidTick);
  __Pyx_INCREF(__pyx_n_u_IsEthTrade);
  __Pyx_GIVEREF(__pyx_n_u_IsEthTrade);
  PyList_SET_ITEM(__pyx_t_1, 16, __pyx_n_u_IsEthTrade);
  __Pyx_INCREF(__pyx_n_u_TradeThroughExempt);
  __Pyx_GIVEREF(__pyx_n_u_TradeThroughExempt);
  PyList_SET_ITEM(__pyx_t_1, 17, __pyx_n_u_TradeThroughExempt);
  __Pyx_INCREF(__pyx_n_u_IsSpreadLeg);
  __Pyx_GIVEREF(__pyx_n_u_IsSpreadLeg);
  PyList_SET_ITEM(__pyx_t_1, 18, __pyx_n_u_IsSpreadLeg);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TIME_AND_SALE_COLUMNS, __pyx_t_1) < 0) __PYX_ERR(1, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/listeners/listener.pyx":131
 *                                  tns[i].is_spread_leg])
 * 
 * CANDLE_COLUMNS = ['Symbol', 'Index', 'Time', 'Sequence', 'Count', 'Open', 'High', 'Low', 'Close', 'Volume', 'VWap',             # <<<<<<<<<<<<<<
 *                   'BidVolume', 'AskVolume', 'OpenInterest', 'ImpVolatility']
 * cdef void candle_default_listener(int event_type,
 */
  __pyx_t_1 = PyList_New(15); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_Symbol);
  __Pyx_GIVEREF(__pyx_n_u_Symbol);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_Symbol);
  __Pyx_INCREF(__pyx_n_u_Index);
  __Pyx_GIVEREF(__pyx_n_u_Index);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_Index);
  __Pyx_INCREF(__pyx_n_u_Time);
  __Pyx_GIVEREF(__pyx_n_u_Time);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_Time);
  __Pyx_INCREF(__pyx_n_u_Sequence);
  __Pyx_GIVEREF(__pyx_n_u_Sequence);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_Sequence);
  __Pyx_INCREF(__pyx_n_u_Count);
  __Pyx_GIVEREF(__pyx_n_u_Count);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_Count);
  __Pyx_INCREF(__pyx_n_u_Open);
  __Pyx_GIVEREF(__pyx_n_u_Open);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_u_Open);
  __Pyx_INCREF(__pyx_n_u_High);
  __Pyx_GIVEREF(__pyx_n_u_High);
  PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_u_High);
  __Pyx_INCREF(__pyx_n_u_Low);
  __Pyx_GIVEREF(__pyx_n_u_Low);
  PyList_SET_ITEM(__pyx_t_1, 7, __pyx_n_u_Low);
  __Pyx_INCREF(__pyx_n_u_Close);
  __Pyx_GIVEREF(__pyx_n_u_Close);
  PyList_SET_ITEM(__pyx_t_1, 8, __pyx_n_u_Close);
  __Pyx_INCREF(__pyx_n_u_Volume);
  __Pyx_GIVEREF(__pyx_n_u_Volume);
  PyList_SET_ITEM(__pyx_t_1, 9, __pyx_n_u_Volume);
  __Pyx_INCREF(__pyx_n_u_VWap);
  __Pyx_GIVEREF(__pyx_n_u_VWap);
  PyList_SET_ITEM(__pyx_t_1, 10, __pyx_n_u_VWap);
  __Pyx_INCREF(__pyx_n_u_BidVolume);
  __Pyx_GIVEREF(__pyx_n_u_BidVolume);
  PyList_SET_ITEM(__pyx_t_1, 11, __pyx_n_u_BidVolume);
  __Pyx_INCREF(__pyx_n_u_AskVolume);
  __Pyx_GIVEREF(__pyx_n_u_AskVolume);
  PyList_SET_ITEM(__pyx_t_1, 12, __pyx_n_u_AskVolume);
  __Pyx_INCREF(__pyx_n_u_OpenInterest);
  __Pyx_GIVEREF(__pyx_n_u_OpenInterest);
  PyList_SET_ITEM(__pyx_t_1, 13, __pyx_n_u_OpenInterest);
  __Pyx_INCREF(__pyx_n_u_ImpVolatility);
  __Pyx_GIVEREF(__pyx_n_u_ImpVolatility);
  PyList_SET_ITEM(__pyx_t_1, 14, __pyx_n_u_ImpVolatility);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CANDLE_COLUMNS, __pyx_t_1) < 0) __PYX_ERR(1, 131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/listeners/listener.pyx":158
 *                                  candle[i].imp_volatility])
 * 
 * ORDER_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'TimeNanos', 'Sequence', 'Price', 'Size', 'Count', 'Scope',             # <<<<<<<<<<<<<<
 *                  'Side', 'ExchangeCode', 'MarketMaker', 'SpreadSymbol']
 * cdef void order_default_listener(int event_type,
 */
  __pyx_t_1 = PyList_New(14); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_Symbol);
  __Pyx_GIVEREF(__pyx_n_u_Symbol);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_Symbol);
  __Pyx_INCREF(__pyx_n_u_EventFlags);
  __Pyx_GIVEREF(__pyx_n_u_EventFlags);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_EventFlags);
  __Pyx_INCREF(__pyx_n_u_Index);
  __Pyx_GIVEREF(__pyx_n_u_Index);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_Index);
  __Pyx_INCREF(__pyx_n_u_Time);
  __Pyx_GIVEREF(__pyx_n_u_Time);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_Time);
  __Pyx_INCREF(__pyx_n_u_TimeNanos);
  __Pyx_GIVEREF(__pyx_n_u_TimeNanos);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_TimeNanos);
  __Pyx_INCREF(__pyx_n_u_Sequence);
  __Pyx_GIVEREF(__pyx_n_u_Sequence);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_u_Sequence);
  __Pyx_INCREF(__pyx_n_u_Price);
  __Pyx_GIVEREF(__pyx_n_u_Price);
  PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_u_Price);
  __Pyx_INCREF(__pyx_n_u_Size);
  __Pyx_GIVEREF(__pyx_n_u_Size);
  PyList_SET_ITEM(__pyx_t_1, 7, __pyx_n_u_Size);
  __Pyx_INCREF(__pyx_n_u_Count);
  __Pyx_GIVEREF(__pyx_n_u_Count);
  PyList_SET_ITEM(__pyx_t_1, 8, __pyx_n_u_Count);
  __Pyx_INCREF(__pyx_n_u_Scope);
  __Pyx_GIVEREF(__pyx_n_u_Scope);
  PyList_SET_ITEM(__pyx_t_1, 9, __pyx_n_u_Scope);
  __Pyx_INCREF(__pyx_n_u_Side);
  __Pyx_GIVEREF(__pyx_n_u_Side);
  PyList_SET_ITEM(__pyx_t_1, 10, __pyx_n_u_Side);
  __Pyx_INCREF(__pyx_n_u_ExchangeCode);
  __Pyx_GIVEREF(__pyx_n_u_ExchangeCode);
  PyList_SET_ITEM(__pyx_t_1, 11, __pyx_n_u_ExchangeCode);
  __Pyx_INCREF(__pyx_n_u_MarketMaker);
  __Pyx_GIVEREF(__pyx_n_u_MarketMaker);
  PyList_SET_ITEM(__pyx_t_1, 12, __pyx_n_u_MarketMaker);
  __Pyx_INCREF(__pyx_n_u_SpreadSymbol);
  __Pyx_GIVEREF(__pyx_n_u_SpreadSymbol);
  PyList_SET_ITEM(__pyx_t_1, 13, __pyx_n_u_SpreadSymbol);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ORDER_COLUMNS, __pyx_t_1) < 0) __PYX_ERR(1, 158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/listeners/listener.pyx":184
 *                                  unicode_from_dxf_const_string_t(order[i].spread_symbol)])
 * 
 * GREEKS_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'Price', 'Volatility', 'Delta', 'Gamma', 'Theta', 'Rho',             # <<<<<<<<<<<<<<
 *                   'Vega']
 * cdef void greeks_default_listener(int event_type,
 */
  __pyx_t_1 = PyList_New(11); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_Symbol);
  __Pyx_GIVEREF(__pyx_n_u_Symbol);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_Symbol);
  __Pyx_INCREF(__pyx_n_u_EventFlags);
  __Pyx_GIVEREF(__pyx_n_u_EventFlags);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_EventFlags);
  __Pyx_INCREF(__pyx_n_u_Index);
  __Pyx_GIVEREF(__pyx_n_u_Index);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_Index);
  __Pyx_INCREF(__pyx_n_u_Time);
  __Pyx_GIVEREF(__pyx_n_u_Time);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_Time);
  __Pyx_INCREF(__pyx_n_u_Price);
  __Pyx_GIVEREF(__pyx_n_u_Price);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_Price);
  __Pyx_INCREF(__pyx_n_u_Volatility);
  __Pyx_GIVEREF(__pyx_n_u_Volatility);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_u_Volatility);
  __Pyx_INCREF(__pyx_n_u_Delta);
  __Pyx_GIVEREF(__pyx_n_u_Delta);
  PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_u_Delta);
  __Pyx_INCREF(__pyx_n_u_Gamma);
  __Pyx_GIVEREF(__pyx_n_u_Gamma);
  PyList_SET_ITEM(__pyx_t_1, 7, __pyx_n_u_Gamma);
  __Pyx_INCREF(__pyx_n_u_Theta);
  __Pyx_GIVEREF(__pyx_n_u_Theta);
  PyList_SET_ITEM(__pyx_t_1, 8, __pyx_n_u_Theta);
  __Pyx_INCREF(__pyx_n_u_Rho);
  __Pyx_GIVEREF(__pyx_n_u_Rho);
  PyList_SET_ITEM(__pyx_t_1, 9, __pyx_n_u_Rho);
  __Pyx_INCREF(__pyx_n_u_Vega);
  __Pyx_GIVEREF(__pyx_n_u_Vega);
  PyList_SET_ITEM(__pyx_t_1, 10, __pyx_n_u_Vega);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GREEKS_COLUMNS, __pyx_t_1) < 0) __PYX_ERR(1, 184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/listeners/listener.pyx":207
 *                                  greeks[i].vega])
 * 
 * THEO_PRICE_COLUMNS = ['Symbol', 'Time', 'Price', 'UnderlyingPrice', 'Delta', 'Gamma', 'Dividend', 'Interest']             # <<<<<<<<<<<<<<
 * cdef void theo_price_default_listener(int event_type,
 *                                       dxf_const_string_t symbol_name,
 */
  __pyx_t_1 = PyList_New(8); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_Symbol);
  __Pyx_GIVEREF(__pyx_n_u_Symbol);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_Symbol);
  __Pyx_INCREF(__pyx_n_u_Time);
  __Pyx_GIVEREF(__pyx_n_u_Time);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_Time);
  __Pyx_INCREF(__pyx_n_u_Price);
  __Pyx_GIVEREF(__pyx_n_u_Price);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_Price);
  __Pyx_INCREF(__pyx_n_u_UnderlyingPrice);
  __Pyx_GIVEREF(__pyx_n_u_UnderlyingPrice);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_UnderlyingPrice);
  __Pyx_INCREF(__pyx_n_u_Delta);
  __Pyx_GIVEREF(__pyx_n_u_Delta);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_Delta);
  __Pyx_INCREF(__pyx_n_u_Gamma);
  __Pyx_GIVEREF(__pyx_n_u_Gamma);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_u_Gamma);
  __Pyx_INCREF(__pyx_n_u_Dividend);
  __Pyx_GIVEREF(__pyx_n_u_Dividend);
  PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_u_Dividend);
  __Pyx_INCREF(__pyx_n_u_Interest);
  __Pyx_GIVEREF(__pyx_n_u_Interest);
  PyList_SET_ITEM(__pyx_t_1, 7, __pyx_n_u_Interest);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_THEO_PRICE_COLUMNS, __pyx_t_1) < 0) __PYX_ERR(1, 207, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/listeners/listener.pyx":226
 *                                  theo_price[i].interest])
 * 
 * UNDERLYING_COLUMNS = ['Symbol', 'Volatility', 'FrontVolatility', 'BackVolatility', 'PutCallRatio']             # <<<<<<<<<<<<<<
 * cdef void underlying_default_listener(int event_type,
 *                                       dxf_const_string_t symbol_name,
 */
  __pyx_t_1 = PyList_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_Symbol);
  __Pyx_GIVEREF(__pyx_n_u_Symbol);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_Symbol);
  __Pyx_INCREF(__pyx_n_u_Volatility);
  __Pyx_GIVEREF(__pyx_n_u_Volatility);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_Volatility);
  __Pyx_INCREF(__pyx_n_u_FrontVolatility);
  __Pyx_GIVEREF(__pyx_n_u_FrontVolatility);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_FrontVolatility);
  __Pyx_INCREF(__pyx_n_u_BackVolatility);
  __Pyx_GIVEREF(__pyx_n_u_BackVolatility);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_BackVolatility);
  __Pyx_INCREF(__pyx_n_u_PutCallRatio);
  __Pyx_GIVEREF(__pyx_n_u_PutCallRatio);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_PutCallRatio);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_UNDERLYING_COLUMNS, __pyx_t_1) < 0) __PYX_ERR(1, 226, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/listeners/listener.pyx":242
 *                                  underlying[i].put_call_ratio])
 * 
 * SERIES_COLUMNS = ['Symbol', 'EventFlags', 'Index', 'Time', 'Sequence', 'Sequence', 'Expiration', 'Volatility',             # <<<<<<<<<<<<<<
 *                   'PutCallRatio', 'ForwardPrice', 'Dividend', 'Interest']
 * cdef void series_default_listener(int event_type,
 */
  __pyx_t_1 = PyList_New(12); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_Symbol);
  __Pyx_GIVEREF(__pyx_n_u_Symbol);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_Symbol);
  __Pyx_INCREF(__pyx_n_u_EventFlags);
  __Pyx_GIVEREF(__pyx_n_u_EventFlags);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_EventFlags);
  __Pyx_INCREF(__pyx_n_u_Index);
  __Pyx_GIVEREF(__pyx_n_u_Index);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_Index);
  __Pyx_INCREF(__pyx_n_u_Time);
  __Pyx_GIVEREF(__pyx_n_u_Time);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_Time);
  __Pyx_INCREF(__pyx_n_u_Sequence);
  __Pyx_GIVEREF(__pyx_n_u_Sequence);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_Sequence);
  __Pyx_INCREF(__pyx_n_u_Sequence);
  __Pyx_GIVEREF(__pyx_n_u_Sequence);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_u_Sequence);
  __Pyx_INCREF(__pyx_n_u_Expiration);
  __Pyx_GIVEREF(__pyx_n_u_Expiration);
  PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_u_Expiration);
  __Pyx_INCREF(__pyx_n_u_Volatility);
  __Pyx_GIVEREF(__pyx_n_u_Volatility);
  PyList_SET_ITEM(__pyx_t_1, 7, __pyx_n_u_Volatility);
  __Pyx_INCREF(__pyx_n_u_PutCallRatio);
  __Pyx_GIVEREF(__pyx_n_u_PutCallRatio);
  PyList_SET_ITEM(__pyx_t_1, 8, __pyx_n_u_PutCallRatio);
  __Pyx_INCREF(__pyx_n_u_ForwardPrice);
  __Pyx_GIVEREF(__pyx_n_u_ForwardPrice);
  PyList_SET_ITEM(__pyx_t_1, 9, __pyx_n_u_ForwardPrice);
  __Pyx_INCREF(__pyx_n_u_Dividend);
  __Pyx_GIVEREF(__pyx_n_u_Dividend);
  PyList_SET_ITEM(__pyx_t_1, 10, __pyx_n_u_Dividend);
  __Pyx_INCREF(__pyx_n_u_Interest);
  __Pyx_GIVEREF(__pyx_n_u_Interest);
  PyList_SET_ITEM(__pyx_t_1, 11, __pyx_n_u_Interest);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SERIES_COLUMNS, __pyx_t_1) < 0) __PYX_ERR(1, 242, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/listeners/listener.pyx":266
 *                                  series[i].interest])
 * 
 * CONFIGURATION_COLUMNS = ['Symbol', 'Version', 'Object']             # <<<<<<<<<<<<<<
 * cdef void configuration_default_listener(int event_type,
 *                                          dxf_const_string_t symbol_name,
 */
  __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_Symbol);
  __Pyx_GIVEREF(__pyx_n_u_Symbol);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_Symbol);
  __Pyx_INCREF(__pyx_n_u_Version);
  __Pyx_GIVEREF(__pyx_n_u_Version);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_Version);
  __Pyx_INCREF(__pyx_n_u_Object);
  __Pyx_GIVEREF(__pyx_n_u_Object);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_Object);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CONFIGURATION_COLUMNS, __pyx_t_1) < 0) __PYX_ERR(1, 266, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "dxfeed/core/listeners/listener.pyx":1
 * from dxfeed.core.utils.helpers cimport *             # <<<<<<<<<<<<<<
 * 
 * cdef class FuncWrapper:
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init dxfeed.core.listeners.listener", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init dxfeed.core.listeners.listener");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto __PYX_BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto __PYX_BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_wchar_t(wchar_t value) {
    const wchar_t neg_one = (wchar_t) ((wchar_t) 0 - (wchar_t) 1), const_zero = (wchar_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(wchar_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(wchar_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(wchar_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(wchar_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(wchar_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(wchar_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_dxf_int_t(dxf_int_t value) {
    const dxf_int_t neg_one = (dxf_int_t) ((dxf_int_t) 0 - (dxf_int_t) 1), const_zero = (dxf_int_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(dxf_int_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(dxf_int_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(dxf_int_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(dxf_int_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(dxf_int_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(dxf_int_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_dxf_long_t(dxf_long_t value) {
    const dxf_long_t neg_one = (dxf_long_t) ((dxf_long_t) 0 - (dxf_long_t) 1), const_zero = (dxf_long_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(dxf_long_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(dxf_long_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(dxf_long_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(dxf_long_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(dxf_long_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(dxf_long_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_dxf_bool_t(dxf_bool_t value) {
    const dxf_bool_t neg_one = (dxf_bool_t) ((dxf_bool_t) 0 - (dxf_bool_t) 1), const_zero = (dxf_bool_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(dxf_bool_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(dxf_bool_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(dxf_bool_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(dxf_bool_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(dxf_bool_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(dxf_bool_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_dxf_dayid_t(dxf_dayid_t value) {
    const dxf_dayid_t neg_one = (dxf_dayid_t) ((dxf_dayid_t) 0 - (dxf_dayid_t) 1), const_zero = (dxf_dayid_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(dxf_dayid_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(dxf_dayid_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(dxf_dayid_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(dxf_dayid_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(dxf_dayid_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(dxf_dayid_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_dxf_uint_t(dxf_uint_t value) {
    const dxf_uint_t neg_one = (dxf_uint_t) ((dxf_uint_t) 0 - (dxf_uint_t) 1), const_zero = (dxf_uint_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(dxf_uint_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(dxf_uint_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(dxf_uint_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(dxf_uint_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(dxf_uint_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(dxf_uint_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_dxf_order_side_t(dxf_order_side_t value) {
    const dxf_order_side_t neg_one = (dxf_order_side_t) ((dxf_order_side_t) 0 - (dxf_order_side_t) 1), const_zero = (dxf_order_side_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(dxf_order_side_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(dxf_order_side_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(dxf_order_side_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(dxf_order_side_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(dxf_order_side_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(dxf_order_side_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_dxf_tns_type_t(dxf_tns_type_t value) {
    const dxf_tns_type_t neg_one = (dxf_tns_type_t) ((dxf_tns_type_t) 0 - (dxf_tns_type_t) 1), const_zero = (dxf_tns_type_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(dxf_tns_type_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(dxf_tns_type_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(dxf_tns_type_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(dxf_tns_type_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(dxf_tns_type_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(dxf_tns_type_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_dxf_order_scope_t(dxf_order_scope_t value) {
    const dxf_order_scope_t neg_one = (dxf_order_scope_t) ((dxf_order_scope_t) 0 - (dxf_order_scope_t) 1), const_zero = (dxf_order_scope_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(dxf_order_scope_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(dxf_order_scope_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(dxf_order_scope_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(dxf_order_scope_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(dxf_order_scope_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(dxf_order_scope_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
